@@ -1,7 +1,8 @@
 /*
- * Copyright (c) 2015-2016 Cray Inc.  All rights reserved.
- * Copyright (c) 2015-2016 Los Alamos National Security, LLC.
+ * Copyright (c) 2015-2017 Cray Inc.  All rights reserved.
+ * Copyright (c) 2015-2017 Los Alamos National Security, LLC.
  *                         All rights reserved.
+ * Copyright (c) 2015-2016 Cisco Systems, Inc.  All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -35,16 +36,13 @@
 #ifndef _GNIX_H_
 #define _GNIX_H_
 
-#ifdef __cplusplus
-extern "C" {
-#endif
-
 #if HAVE_CONFIG_H
 #include <config.h>
 #endif /* HAVE_CONFIG_H */
 
 #include <stdlib.h>
 #include <stdbool.h>
+#include <inttypes.h>
 
 #include <rdma/fabric.h>
 #include <rdma/fi_atomic.h>
@@ -53,16 +51,22 @@ extern "C" {
 #include <rdma/fi_endpoint.h>
 #include <rdma/fi_eq.h>
 #include <rdma/fi_errno.h>
-#include <rdma/fi_prov.h>
+#include <rdma/providers/fi_prov.h>
 #include <rdma/fi_rma.h>
 #include <rdma/fi_tagged.h>
 #include <rdma/fi_trigger.h>
 
 #include <fi.h>
+#include <ofi_atomic.h>
 #include <fi_enosys.h>
-#include <fi_indexer.h>
 #include <fi_rbuf.h>
 #include <fi_list.h>
+#include <fi_file.h>
+
+#ifdef HAVE_UDREG
+#include <udreg_pub.h>
+#endif
+
 #include "gni_pub.h"
 #include "gnix_util.h"
 #include "gnix_freelist.h"
@@ -71,9 +75,12 @@ extern "C" {
 #include "fi_ext_gni.h"
 #include "gnix_tags.h"
 #include "gnix_mr_cache.h"
+#include "gnix_mr_notifier.h"
+#include "gnix_nic.h"
+#include "gnix_auth_key.h"
 
 #define GNI_MAJOR_VERSION 1
-#define GNI_MINOR_VERSION 0
+#define GNI_MINOR_VERSION 1
 
 /*
  * useful macros
@@ -109,6 +116,11 @@ extern "C" {
 #define compiler_barrier() asm volatile ("" ::: "memory")
 #endif
 
+#define GNIX_MAX_MSG_IOV_LIMIT 8
+#define GNIX_MAX_RMA_IOV_LIMIT 1
+#define GNIX_MAX_ATOMIC_IOV_LIMIT 1
+#define GNIX_ADDR_CACHE_SIZE 5
+
 /*
  * GNI GET alignment
  */
@@ -116,6 +128,17 @@ extern "C" {
 #define GNI_READ_ALIGN		4
 #define GNI_READ_ALIGN_MASK	(GNI_READ_ALIGN - 1)
 
+/*
+ * GNI IOV GET alignment
+ *
+ * We always pull 4byte chucks for unaligned GETs. To prevent stomping on
+ * someone else's head or tail data, each segment must be four bytes
+ * (i.e. GNI_READ_ALIGN bytes).
+ *
+ * Note: "* 2" for head and tail
+ */
+#define GNIX_INT_TX_BUF_SZ (GNIX_MAX_MSG_IOV_LIMIT * GNI_READ_ALIGN * 2)
+
 /*
  * Flags
  * The 64-bit flag field is used as follows:
@@ -150,15 +173,27 @@ extern "C" {
 /*
  * See capabilities section in fi_getinfo.3.
  */
-#define GNIX_EP_RDM_PRIMARY_CAPS                                               \
+
+#define GNIX_DOM_CAPS (FI_REMOTE_COMM)
+
+/* Primary capabilities.  Each must be explicitly requested (unless the full
+ * set is requested by setting input hints->caps to NULL). */
+#define GNIX_EP_PRIMARY_CAPS                                               \
 	(FI_MSG | FI_RMA | FI_TAGGED | FI_ATOMICS |                            \
-	 FI_DIRECTED_RECV | FI_READ |                                          \
+	 FI_DIRECTED_RECV | FI_READ | FI_NAMED_RX_CTX |                        \
 	 FI_WRITE | FI_SEND | FI_RECV | FI_REMOTE_READ | FI_REMOTE_WRITE)
 
-#define GNIX_EP_RDM_SEC_CAPS                                             \
-	(FI_MULTI_RECV | FI_SOURCE | FI_TRIGGER | FI_FENCE)
+/* No overhead secondary capabilities.  These can be silently enabled by the
+ * provider. */
+#define GNIX_EP_SEC_CAPS (FI_MULTI_RECV | FI_TRIGGER | FI_FENCE)
 
-#define GNIX_EP_RDM_CAPS (GNIX_EP_RDM_PRIMARY_CAPS | GNIX_EP_RDM_SEC_CAPS)
+/* Secondary capabilities that introduce overhead.  Must be requested. */
+#define GNIX_EP_SEC_CAPS_OH (FI_SOURCE | FI_RMA_EVENT | FI_SOURCE_ERR)
+
+/* FULL set of capabilities for the provider.  */
+#define GNIX_EP_CAPS_FULL (GNIX_EP_PRIMARY_CAPS | \
+			   GNIX_EP_SEC_CAPS | \
+			   GNIX_EP_SEC_CAPS_OH)
 
 /*
  * see Operations flags in fi_endpoint.3
@@ -172,7 +207,7 @@ extern "C" {
  */
 #define GNIX_SENDMSG_FLAGS	(FI_REMOTE_CQ_DATA | FI_COMPLETION | \
 				 FI_MORE | FI_INJECT | FI_INJECT_COMPLETE | \
-				 FI_TRANSMIT_COMPLETE | FI_FENCE)
+				 FI_TRANSMIT_COMPLETE | FI_FENCE | FI_TRIGGER)
 #define GNIX_RECVMSG_FLAGS	(FI_COMPLETION | FI_MORE | FI_MULTI_RECV)
 #define GNIX_TRECVMSG_FLAGS \
 	(GNIX_RECVMSG_FLAGS | FI_CLAIM | FI_PEEK | FI_DISCARD)
@@ -182,12 +217,15 @@ extern "C" {
  */
 #define GNIX_WRITEMSG_FLAGS	(FI_REMOTE_CQ_DATA | FI_COMPLETION | \
 				 FI_MORE | FI_INJECT | FI_INJECT_COMPLETE | \
-				 FI_TRANSMIT_COMPLETE | FI_FENCE)
-#define GNIX_READMSG_FLAGS	(FI_REMOTE_CQ_DATA | FI_COMPLETION | FI_MORE | \
-				 FI_FENCE)
-#define GNIX_ATOMICMSG_FLAGS	(FI_COMPLETION | FI_MORE | FI_INJECT | FI_FENCE)
-#define GNIX_FATOMICMSG_FLAGS	(FI_COMPLETION | FI_MORE | FI_FENCE)
-#define GNIX_CATOMICMSG_FLAGS	(FI_COMPLETION | FI_MORE | FI_FENCE)
+				 FI_TRANSMIT_COMPLETE | FI_FENCE | FI_TRIGGER)
+#define GNIX_READMSG_FLAGS	(FI_COMPLETION | FI_MORE | \
+				 FI_FENCE | FI_TRIGGER)
+#define GNIX_ATOMICMSG_FLAGS	(FI_COMPLETION | FI_MORE | FI_INJECT | \
+				 FI_FENCE | FI_TRIGGER)
+#define GNIX_FATOMICMSG_FLAGS	(FI_COMPLETION | FI_MORE | FI_FENCE | \
+				 FI_TRIGGER)
+#define GNIX_CATOMICMSG_FLAGS	(FI_COMPLETION | FI_MORE | FI_FENCE | \
+				 FI_TRIGGER)
 
 /*
  * Valid completion event flags.  See fi_cq.3.
@@ -200,13 +238,15 @@ extern "C" {
  */
 #define GNIX_TX_SIZE_DEFAULT	500
 #define GNIX_RX_SIZE_DEFAULT	500
+/*
+ * based on the max number of fma descriptors without fma sharing
+ */
+#define GNIX_RX_CTX_MAX_BITS	8
+#define GNIX_SEP_MAX_CNT	(1 << (GNIX_RX_CTX_MAX_BITS - 1))
 
 /*
  * if this has to be changed, check gnix_getinfo, etc.
  */
-#define GNIX_EP_MSG_CAPS GNIX_EP_RDM_CAPS
-#define GNIX_EP_MSG_SEC_CAPS GNIX_EP_RDM_SEC_CAPS
-
 #define GNIX_MAX_MSG_SIZE ((0x1ULL << 32) - 1)
 #define GNIX_CACHELINE_SIZE (64)
 #define GNIX_INJECT_SIZE GNIX_CACHELINE_SIZE
@@ -222,8 +262,14 @@ extern "C" {
  */
 #define GNIX_FAB_MODES_CLEAR (FI_MSG_PREFIX | FI_ASYNC_IOV)
 
-/*
- * gnix address format - used for fi_send/fi_recv, etc.
+/**
+ * gnix_address struct
+ *
+ * @note - gnix address format - used for fi_send/fi_recv, etc.
+ *         These values are passed to GNI_EpBind
+ *
+ * @var device_addr     physical NIC address of the remote peer
+ * @var cdm_id          user supplied id of the remote instance
  */
 struct gnix_address {
 	uint32_t device_addr;
@@ -243,13 +289,25 @@ struct gnix_address {
 #define GNIX_ADDR_EQUAL(a, b) (((a).device_addr == (b).device_addr) && \
 				((a).cdm_id == (b).cdm_id))
 
+#define GNIX_CREATE_CDM_ID	0
 
-#define GNIX_EPN_TYPE_UNBOUND	0
-#define GNIX_EPN_TYPE_BOUND	1
+#define GNIX_EPN_TYPE_UNBOUND	(1 << 0)
+#define GNIX_EPN_TYPE_BOUND	(1 << 1)
+#define GNIX_EPN_TYPE_SEP	(1 << 2)
 
-/*
- * info returned by fi_getname/fi_getpeer - has enough
- * side band info for RDM ep's to be able to connect, etc.
+/**
+ * gnix_ep_name struct
+ *
+ * @note - info returned by fi_getname/fi_getpeer - has enough
+ *         side band info for RDM ep's to be able to connect, etc.
+ *
+ * @var gnix_addr       address of remote peer
+ * @var name_type       bound, unbound, scalable endpoint name types
+ * @var cm_nic_cdm_id   id of the cm nic associated with this endpoint
+ * @var cookie          communication domain identifier
+ * @var rx_ctx_cnt      number of contexts associated with this endpoint
+ * @var unused1/2       for future use
+ * @var reserved        for future use
  */
 struct gnix_ep_name {
 	struct gnix_address gnix_addr;
@@ -258,17 +316,22 @@ struct gnix_ep_name {
 		uint32_t cm_nic_cdm_id : 24;
 		uint32_t cookie;
 	};
-	uint64_t reserved[4];
+	struct {
+		uint32_t rx_ctx_cnt : 8;
+		uint32_t unused1 : 24;
+		uint32_t unused2;
+	};
+	uint64_t reserved[3];
 };
 
 /* AV address string revision. */
 #define GNIX_AV_STR_ADDR_VERSION  1
 
 /*
- * 49 is the number of characters printed out in gnix_av_straddr.
+ * 52 is the number of characters printed out in gnix_av_straddr.
  *  1 is for the null terminator
  */
-#define GNIX_AV_MAX_STR_ADDR_LEN  (49 + 1)
+#define GNIX_AV_MAX_STR_ADDR_LEN  (52 + 1)
 
 /*
  * 15 is the number of characters for the device addr.
@@ -276,6 +339,12 @@ struct gnix_ep_name {
  */
 #define GNIX_AV_MIN_STR_ADDR_LEN  (15 + 1)
 
+/*
+ * 69 is the number of characters for the printable portion of the address
+ *  1 is for the null terminator
+ */
+#define GNIX_FI_ADDR_STR_LEN (69 + 1)
+
 /*
  * enum for blocking/non-blocking progress
  */
@@ -301,9 +370,14 @@ struct gnix_fid_fabric {
 	 * GNI_PostdataProbeWaitById in gni_pub.h */
 	uint64_t datagram_timeout;
 	struct gnix_reference ref_cnt;
+	struct gnix_mr_notifier mr_notifier;
 };
 
-extern struct fi_ops_cm gnix_cm_ops;
+extern struct fi_ops_cm gnix_ep_msg_ops_cm;
+extern struct fi_ops_cm gnix_ep_ops_cm;
+
+#define GNIX_GET_MR_CACHE_INFO(domain, auth_key) \
+	({ &(domain)->mr_cache_info[(auth_key)->ptag]; })
 
 /*
  * a gnix_fid_domain is associated with one or more gnix_nic's.
@@ -320,24 +394,61 @@ struct gnix_fid_domain {
 	struct gnix_fid_fabric *fabric;
 	struct gnix_cm_nic *cm_nic;
 	fastlock_t cm_nic_lock;
-	uint8_t ptag;
-	uint32_t cookie;
 	uint32_t cdm_id_seed;
+	uint32_t addr_format;
 	/* user tunable parameters accessed via open_ops functions */
 	struct gnix_ops_domain params;
-	/* size of gni tx cqs for this domain */
-	uint32_t gni_tx_cq_size;
-	/* size of gni rx cqs for this domain */
-	uint32_t gni_rx_cq_size;
 	/* additional gni cq modes to use for this domain */
 	gni_cq_mode_t gni_cq_modes;
 	/* additional gni cq modes to use for this domain */
 	enum fi_progress control_progress;
 	enum fi_progress data_progress;
+	enum fi_threading thread_model;
 	struct gnix_reference ref_cnt;
 	gnix_mr_cache_attr_t mr_cache_attr;
-	gnix_mr_cache_t *mr_cache;
+	struct gnix_mr_cache_info *mr_cache_info;
+	struct gnix_mr_ops *mr_ops;
 	fastlock_t mr_cache_lock;
+	int mr_cache_type;
+	/* flag to indicate that memory registration is initialized and should not
+	 * be changed at this point.
+	 */
+	int mr_is_init;
+	int mr_iov_limit;
+	int udreg_reg_limit;
+	struct gnix_auth_key *auth_key;
+#ifdef HAVE_UDREG
+	udreg_cache_handle_t udreg_cache;
+#endif
+	uint32_t num_allocd_stxs;
+};
+
+/**
+ * gnix_fid_pep structure - GNIX passive endpoint
+ *
+ * @var pep_fid		libfabric passive EP fid structure
+ * @var fabric		Fabric associated with this endpoint
+ * @var eq		Event queue bound to this endpoint
+ * @var src_addr	Source address of this endpoint
+ * @var lock		Lock protecting all endpoint fields
+ * @var listen_fd	TCP socket used to listen for connections
+ * @var backlog		Maximum number of pending connetions
+ * @var bound		Flag indicating if the endpoint source address is set
+ * @var cm_data_size	Maximum size of CM data
+ * @var ref_cnt		Endpoint reference count
+ */
+struct gnix_fid_pep {
+	struct fid_pep pep_fid;
+	struct gnix_fid_fabric *fabric;
+	struct fi_info *info;
+	struct gnix_fid_eq *eq;
+	struct gnix_ep_name src_addr;
+	fastlock_t lock;
+	int listen_fd;
+	int backlog;
+	int bound;
+	size_t cm_data_size;
+	struct gnix_reference ref_cnt;
 };
 
 #define GNIX_CQS_PER_EP		8
@@ -353,6 +464,44 @@ struct gnix_fid_ep_ops_en {
 	uint32_t atomic_write_allowed: 1;
 };
 
+#define GNIX_INT_TX_POOL_SIZE 128
+#define GNIX_INT_TX_POOL_COUNT 256
+
+struct gnix_int_tx_buf {
+	struct slist_entry e;
+	uint8_t *buf;
+	struct gnix_fid_mem_desc *md;
+};
+
+struct gnix_int_tx_ptrs {
+	struct slist_entry e;
+	void *sl_ptr;
+	void *buf_ptr;
+	struct gnix_fid_mem_desc *md;
+};
+
+struct gnix_int_tx_pool {
+	bool enabled;
+	int nbufs;
+	fastlock_t lock;
+	struct slist sl;
+	struct slist bl;
+};
+
+struct gnix_addr_cache_entry {
+	fi_addr_t addr;
+	struct gnix_vc *vc;
+};
+
+enum gnix_conn_state {
+	GNIX_EP_UNCONNECTED,
+	GNIX_EP_CONNECTING,
+	GNIX_EP_CONNECTED,
+	GNIX_EP_SHUTDOWN
+};
+
+#define GNIX_EP_CONNECTED(ep)	((ep)->conn_state == GNIX_EP_CONNECTED)
+
 /*
  *   gnix endpoint structure
  *
@@ -371,47 +520,65 @@ struct gnix_fid_ep {
 	struct gnix_fid_cq *recv_cq;
 	struct gnix_fid_cntr *send_cntr;
 	struct gnix_fid_cntr *recv_cntr;
-	struct gnix_fid_cntr *read_cntr;
 	struct gnix_fid_cntr *write_cntr;
+	struct gnix_fid_cntr *read_cntr;
+	struct gnix_fid_cntr *rwrite_cntr;
+	struct gnix_fid_cntr *rread_cntr;
 	struct gnix_fid_av *av;
 	struct gnix_fid_stx *stx_ctx;
-	struct gnix_ep_name my_name;
 	struct gnix_cm_nic *cm_nic;
 	struct gnix_nic *nic;
-	fastlock_t vc_ht_lock;
-	union {
-		struct gnix_hashtable *vc_ht;
-		struct gnix_vc **vc_table;      /* used for FI_AV_TABLE */
-		struct gnix_vc *vc;
-	};
-	/* lock for unexp and posted recv queue */
-	fastlock_t recv_queue_lock;
+	fastlock_t vc_lock;
 	/* used for unexpected receives */
 	struct gnix_tag_storage unexp_recv_queue;
 	/* used for posted receives */
 	struct gnix_tag_storage posted_recv_queue;
 
-	fastlock_t tagged_queue_lock;
 	struct gnix_tag_storage tagged_unexp_recv_queue;
 	struct gnix_tag_storage tagged_posted_recv_queue;
 
-
-	fastlock_t recv_comp_lock;
-	struct slist pending_recv_comp_queue;
 	/* pointer to tag matching engine */
-	void *tag_matcher;
 	int (*progress_fn)(struct gnix_fid_ep *, enum gnix_progress_type);
 	/* RX specific progress fn */
 	int (*rx_progress_fn)(struct gnix_fid_ep *, gni_return_t *rc);
+	struct gnix_xpmem_handle *xpmem_hndl;
 	bool tx_enabled;
 	bool rx_enabled;
+	bool shared_tx;
+	bool requires_lock;
+	struct gnix_auth_key *auth_key;
+	int last_cached;
+	struct gnix_addr_cache_entry addr_cache[GNIX_ADDR_CACHE_SIZE];
 	int send_selective_completion;
 	int recv_selective_completion;
 	int min_multi_recv;
 	/* note this free list will be initialized for thread safe */
-	struct gnix_s_freelist fr_freelist;
+	struct gnix_freelist fr_freelist;
+	struct gnix_int_tx_pool int_tx_pool;
 	struct gnix_reference ref_cnt;
 	struct gnix_fid_ep_ops_en ep_ops;
+
+	struct fi_info *info;
+	struct fi_ep_attr ep_attr;
+	struct gnix_ep_name src_addr;
+
+	/* FI_EP_MSG specific. */
+	struct gnix_vc *vc;
+	int conn_fd;
+	int conn_state;
+	struct gnix_ep_name dest_addr;
+	struct gnix_fid_eq *eq;
+
+	/* Unconnected EP specific. */
+	union {
+		struct gnix_hashtable *vc_ht;	/* FI_AV_MAP */
+		struct gnix_vector *vc_table;	/* FI_AV_TABLE */
+	};
+	struct dlist_entry unmapped_vcs;
+
+	/* FI_MORE specific. */
+	struct slist more_read;
+	struct slist more_write;
 };
 
 #define GNIX_EP_RDM(type)         (type == FI_EP_RDM)
@@ -425,25 +592,104 @@ struct gnix_fid_ep {
 				   (type == FI_EP_DGRAM) || \
 				   (type == FI_EP_MSG))
 
+/**
+ * gnix_fid_sep struct
+ *
+ * @var ep_fid          embedded struct fid_ep field
+ * @var domain          pointer to domain used to create the sep instance
+ * @var info            pointer to dup of info struct supplied to fi_scalable_ep
+ *                      operation
+ * @var op_flags        quick access for op_flags for tx/rx contexts
+ *                      instantiated using this sep
+ * @var caps            quick access for caps for tx/rx contexts instantiated
+ *                      using this sep
+ * @var cdm_id_base     base cdm id to use for tx/rx contexts instantiated
+ *                      using this sep
+ * @var ep_table        array of pointers to EPs used by the rx/tx contexts
+ *                      instantiated using this sep
+ * @var tx_ep_table     array of pointers to tx contexts instantiated using
+ *                      this sep
+ * @var rx_ep_table     array of pointers to rx contexts instantiated using
+ *                      this sep
+ * @var cm_nic          gnix cm nic associated with this SEP.
+ * @var av              address vector bound to this SEP
+ * @var my_name         ep name for this endpoint
+ * @var sep_lock        lock protecting this sep object
+ * @var ref_cnt         ref cnt on this object
+ * @var auth_key		GNIX authorization key
+ */
+struct gnix_fid_sep {
+	struct fid_ep ep_fid;
+	enum fi_ep_type type;
+	struct fid_domain *domain;
+	struct fi_info *info;
+	uint64_t caps;
+	uint32_t cdm_id_base;
+	struct fid_ep **ep_table;
+	struct fid_ep **tx_ep_table;
+	struct fid_ep **rx_ep_table;
+	struct gnix_cm_nic *cm_nic;
+	struct gnix_fid_av *av;
+	struct gnix_ep_name my_name;
+	fastlock_t sep_lock;
+	struct gnix_reference ref_cnt;
+	struct gnix_auth_key *auth_key;
+};
+
+/**
+ * gnix_fid_trx struct
+ *
+ * @var ep_fid          embedded struct fid_ep field
+ * @var ep              pointer to gnix_fid_ep used by this tx/rx context
+ * @var sep             pointer to associated gnix_fid_sep for this context
+ * @var op_flags        op flags for this tx context
+ * @var caps            caps for this tx context
+ * @var ref_cnt         ref cnt on this object
+ */
+struct gnix_fid_trx {
+	struct fid_ep ep_fid;
+	struct gnix_fid_ep *ep;
+	struct gnix_fid_sep *sep;
+	uint64_t op_flags;
+	uint64_t caps;
+	struct gnix_reference ref_cnt;
+};
+
 /**
  * gnix_fid_stx struct
- * @note - this is essentially a dummy structure for the GNI
- *         provider.  Added as a convenience for those consumers
- *         who need to use the fid_stx construct for other
- *         providers.
+ * @note - another way to associated gnix_nic's with an ep
  *
  * @var stx_fid              embedded struct fid_stx field
  * @var domain               pointer to domain used to create the stx instance
+ * @var nic                  pointer to gnix_nic associated with this stx
  * @var ref_cnt              ref cnt on this object
  */
 struct gnix_fid_stx {
 	struct fid_stx stx_fid;
 	struct gnix_fid_domain *domain;
+	struct gnix_nic *nic;
+	struct gnix_auth_key *auth_key;
 	struct gnix_reference ref_cnt;
 };
 
-/*
- * TODO: Support shared named AVs
+/**
+ * gnix_fid_av struct
+ * @TODO - Support shared named AVs
+ *
+ * @var fid_av          embedded struct fid_stx field
+ * @var domain          pointer to domain used to create the av
+ * @var type            the type of the AV, FI_AV_{TABLE,MAP}
+ * @var table
+ * @var valid_entry_vec
+ * @var addrlen
+ * @var capacity        current size of AV
+ * @var count           number of address are currently stored in AV
+ * @var rx_ctx_bits     address bits to identify an rx context
+ * @var mask            mask of the fi_addr to resolve the base address
+ * @var map_ht          Hash table for mapping FI_AV_MAP
+ * @var block_list      linked list of blocks used for allocating entries
+ *                      for FI_AV_MAP
+ * @var ref_cnt         ref cnt on this object
  */
 struct gnix_fid_av {
 	struct fid_av av_fid;
@@ -452,30 +698,37 @@ struct gnix_fid_av {
 	struct gnix_av_addr_entry* table;
 	int *valid_entry_vec;
 	size_t addrlen;
-	/* How many addresses AV can hold before it needs to be resized */
 	size_t capacity;
-	/* How many address are currently stored in AV */
 	size_t count;
-	/* Hash table for mapping FI_AV_MAP */
+	uint64_t rx_ctx_bits;
+	uint64_t mask;
 	struct gnix_hashtable *map_ht;
-	/*
-	 * linked list of blocks used for allocating entries
-	 *  for FI_AV_MAP
-	 */
 	struct slist block_list;
 	struct gnix_reference ref_cnt;
 };
 
 enum gnix_fab_req_type {
 	GNIX_FAB_RQ_SEND,
+	GNIX_FAB_RQ_SENDV,
 	GNIX_FAB_RQ_TSEND,
+	GNIX_FAB_RQ_TSENDV,
 	GNIX_FAB_RQ_RDMA_WRITE,
 	GNIX_FAB_RQ_RDMA_READ,
 	GNIX_FAB_RQ_RECV,
+	GNIX_FAB_RQ_RECVV,
 	GNIX_FAB_RQ_TRECV,
+	GNIX_FAB_RQ_TRECVV,
+	GNIX_FAB_RQ_MRECV,
 	GNIX_FAB_RQ_AMO,
 	GNIX_FAB_RQ_FAMO,
-	GNIX_FAB_RQ_CAMO
+	GNIX_FAB_RQ_CAMO,
+	GNIX_FAB_RQ_END_NON_NATIVE,
+	GNIX_FAB_RQ_START_NATIVE = GNIX_NAMO_AX,
+	GNIX_FAB_RQ_NAMO_AX = GNIX_NAMO_AX,
+	GNIX_FAB_RQ_NAMO_AX_S = GNIX_NAMO_AX_S,
+	GNIX_FAB_RQ_NAMO_FAX = GNIX_NAMO_FAX,
+	GNIX_FAB_RQ_NAMO_FAX_S = GNIX_NAMO_FAX_S,
+	GNIX_FAB_RQ_MAX_TYPES,
 };
 
 struct gnix_fab_req_rma {
@@ -485,28 +738,50 @@ struct gnix_fab_req_rma {
 	uint64_t                 rem_addr;
 	uint64_t                 rem_mr_key;
 	uint64_t                 imm;
-	atomic_t                 outstanding_txds;
+	ofi_atomic32_t           outstanding_txds;
 	gni_return_t             status;
+	struct slist_entry       sle;
 };
 
 struct gnix_fab_req_msg {
 	struct gnix_tag_list_element tle;
-	uint64_t                     send_addr;
-	size_t                       send_len;
-	struct gnix_fid_mem_desc     *send_md;
+
+	struct send_info_t {
+		uint64_t	 send_addr;
+		size_t		 send_len;
+		gni_mem_handle_t mem_hndl;
+		uint32_t	 head;
+		uint32_t	 tail;
+	}			     send_info[GNIX_MAX_MSG_IOV_LIMIT];
+	struct gnix_fid_mem_desc     *send_md[GNIX_MAX_MSG_IOV_LIMIT];
+	size_t                       send_iov_cnt;
 	uint64_t                     send_flags;
-	uint64_t                     recv_addr;
-	size_t                       recv_len;
-	struct gnix_fid_mem_desc     *recv_md;
+	size_t			     cum_send_len;
+	struct gnix_fab_req 	     *parent;
+	size_t                       mrecv_space_left;
+	uint64_t                     mrecv_buf_addr;
+
+	struct recv_info_t {
+		uint64_t	 recv_addr;
+		size_t		 recv_len;
+		gni_mem_handle_t mem_hndl;
+		uint32_t	 tail_len : 2; /* If the send len is > the recv_len, we
+						* need to fetch the unaligned tail into
+						* the txd's int buf
+						*/
+		uint32_t	 head_len : 2;
+	}			     recv_info[GNIX_MAX_MSG_IOV_LIMIT];
+	struct gnix_fid_mem_desc     *recv_md[GNIX_MAX_MSG_IOV_LIMIT];
+	size_t			     recv_iov_cnt;
 	uint64_t                     recv_flags; /* protocol, API info */
+	size_t			     cum_recv_len;
+
 	uint64_t                     tag;
 	uint64_t                     ignore;
 	uint64_t                     imm;
 	gni_mem_handle_t             rma_mdh;
 	uint64_t                     rma_id;
-	uint32_t                     rndzv_head;
-	uint32_t                     rndzv_tail;
-	atomic_t                     outstanding_txds;
+	ofi_atomic32_t               outstanding_txds;
 	gni_return_t                 status;
 };
 
@@ -521,7 +796,6 @@ struct gnix_fab_req_amo {
 	enum fi_op               op;
 	uint64_t                 first_operand;
 	uint64_t                 second_operand;
-	void                     *read_buf;
 };
 
 /*
@@ -714,14 +988,35 @@ static inline int gnix_ops_allowed(struct gnix_fid_ep *ep,
 	return 0;
 }
 
-/*
+/**
  * Fabric request layout, there is a one to one
  * correspondence between an application's invocation of fi_send, fi_recv
  * and a gnix fab_req.
+ *
+ * @var dlist	     a doubly linked list entry used to queue a request in
+ * either the vc's tx_queue or work_queue.
+ * @var addr	     the peer's gnix_address associated with this request.
+ * @var type	     the fabric request type
+ * @var gnix_ep      the gni endpoint associated with this request
+ * @var user_context the user context, typically the receive buffer address for
+ * a send or the send buffer address for a receive.
+ * @var vc	      the virtual channel or connection edge between the sender
+ * and receiver.
+ * @var work_fn	     the function called by the nic progress loop to initiate
+ * the fabric request.
+ * @var flags	      a set of bit patterns that apply to all message types
+ * @cb                optional call back to be invoked when ref cnt on this
+ *                    object drops to zero
+ * @ref_cnt           ref cnt for this object
+ * @var iov_txds      A list of pending Rdma/CtFma GET txds.
+ * @var iov_txd_cnt   The count of outstanding iov txds.
+ * @var tx_failures   tx failure bits.
+ * @var rma	      GNI PostRdma request
+ * @var msg	      GNI SMSG request
+ * @var amo	      GNI Fma request
  */
-
 struct gnix_fab_req {
-	struct slist_entry        slist;
+	struct dlist_entry        dlist;
 	struct gnix_address       addr;
 	enum gnix_fab_req_type    type;
 	struct gnix_fid_ep        *gnix_ep;
@@ -729,18 +1024,48 @@ struct gnix_fab_req {
 	struct gnix_vc            *vc;
 	int                       (*work_fn)(void *);
 	uint64_t                  flags;
-	void                      *txd;
-	uint32_t                  tx_failures;
+	void                      (*cb)(void *);
+	struct gnix_reference     ref_cnt;
+
+	struct slist_entry           *int_tx_buf_e;
+	uint8_t                      *int_tx_buf;
+	gni_mem_handle_t             int_tx_mdh;
+
+	struct gnix_tx_descriptor *iov_txds[GNIX_MAX_MSG_IOV_LIMIT];
+	/*
+	 * special value of UINT_MAX is used to indicate
+	 * an unrecoverable (aka non-transient) error has occurred
+	 * in one of the underlying GNI transactions
+	 */
+	uint32_t		  tx_failures;
 
 	/* common to rma/amo/msg */
 	union {
-		struct gnix_fab_req_rma rma;
-		struct gnix_fab_req_msg msg;
-		struct gnix_fab_req_amo amo;
+		struct gnix_fab_req_rma   rma;
+		struct gnix_fab_req_msg   msg;
+		struct gnix_fab_req_amo   amo;
 	};
 	char inject_buf[GNIX_INJECT_SIZE];
 };
 
+/*
+ * test whether a request is replayable
+ * or not based on the value of the tx_failures field
+ */
+
+static inline bool _gnix_req_replayable(struct gnix_fab_req *req)
+{
+	bool ret = false;
+	uint32_t tx_failures, max_retrans;
+
+	tx_failures = req->tx_failures;
+	max_retrans = req->gnix_ep->domain->params.max_retransmits;
+	if ((req->tx_failures != UINT_MAX) &&
+	    (++tx_failures < max_retrans))
+		ret = true;
+
+	return ret;
+}
 static inline int _gnix_req_inject_err(struct gnix_fab_req *req)
 {
 	int err_cnt = req->gnix_ep->domain->params.err_inject_count;
@@ -768,6 +1093,15 @@ static inline int _gnix_req_inject_smsg_err(struct gnix_fab_req *req)
 	}
 }
 
+extern int gnix_default_user_registration_limit;
+extern int gnix_default_prov_registration_limit;
+extern int gnix_dealloc_aki_on_fabric_close;
+
+/* This is a per-node limitation of the GNI provider. Each process
+   should request only as many registrations as it intends to use
+   and no more than that. */
+#define GNIX_MAX_SCALABLE_REGISTRATIONS 4096
+
 /*
  * work queue struct, used for handling delay ops, etc. in a generic wat
  */
@@ -793,7 +1127,7 @@ struct gnix_work_req {
 extern const char gnix_fab_name[];
 extern const char gnix_dom_name[];
 extern uint32_t gnix_cdm_modes;
-extern atomic_t gnix_id_counter;
+extern ofi_atomic32_t gnix_id_counter;
 
 
 /*
@@ -820,7 +1154,11 @@ int gnix_cq_open(struct fid_domain *domain, struct fi_cq_attr *attr,
 		 struct fid_cq **cq, void *context);
 
 int gnix_ep_open(struct fid_domain *domain, struct fi_info *info,
-		 struct fid_ep **ep, void *context);
+		   struct fid_ep **ep, void *context);
+
+int gnix_pep_open(struct fid_fabric *fabric,
+		  struct fi_info *info, struct fid_pep **pep,
+		  void *context);
 
 int gnix_eq_open(struct fid_fabric *fabric, struct fi_eq_attr *attr,
 		 struct fid_eq **eq, void *context);
@@ -829,9 +1167,28 @@ int gnix_mr_reg(struct fid *fid, const void *buf, size_t len,
 		uint64_t access, uint64_t offset, uint64_t requested_key,
 		uint64_t flags, struct fid_mr **mr_o, void *context);
 
+int gnix_mr_regv(struct fid *fid, const struct iovec *iov,
+                 size_t count, uint64_t access,
+                 uint64_t offset, uint64_t requested_key,
+                 uint64_t flags, struct fid_mr **mr, void *context);
+
+int gnix_mr_regattr(struct fid *fid, const struct fi_mr_attr *attr,
+                    uint64_t flags, struct fid_mr **mr);
+
 int gnix_cntr_open(struct fid_domain *domain, struct fi_cntr_attr *attr,
 		 struct fid_cntr **cntr, void *context);
 
+int gnix_sep_open(struct fid_domain *domain, struct fi_info *info,
+		 struct fid_ep **sep, void *context);
+
+int gnix_ep_bind(fid_t fid, struct fid *bfid, uint64_t flags);
+
+int gnix_ep_close(fid_t fid);
+
+/*
+ * prototype for static data initialization method
+ */
+void _gnix_init(void);
 
 /* Prepend DIRECT_FN to provider specific API functions for global visibility
  * when using fabric direct.  If the API function is static use the STATIC
@@ -845,8 +1202,4 @@ int gnix_cntr_open(struct fid_domain *domain, struct fi_cntr_attr *attr,
 #define STATIC static
 #endif
 
-#ifdef __cplusplus
-} /* extern "C" */
-#endif
-
 #endif /* _GNIX_H_ */
