@@ -1,6 +1,6 @@
 /*
- * Copyright (c) 2015 Cray Inc.  All rights reserved.
- * Copyright (c) 2015 Los Alamos National Security, LLC. All rights reserved.
+ * Copyright (c) 2015-2017 Cray Inc.  All rights reserved.
+ * Copyright (c) 2015-2016 Los Alamos National Security, LLC. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -34,10 +34,6 @@
 #ifndef _GNIX_NIC_H_
 #define _GNIX_NIC_H_
 
-#ifdef __cplusplus
-extern "C" {
-#endif
-
 #if HAVE_CONFIG_H
 #include <config.h>
 #endif /* HAVE_CONFIG_H */
@@ -51,7 +47,14 @@ extern "C" {
 
 #define GNIX_DEF_MAX_NICS_PER_PTAG	4
 
+/*
+ * globals
+ */
+
 extern uint32_t gnix_max_nics_per_ptag;
+extern struct dlist_entry gnix_nic_list_ptag[];
+extern struct dlist_entry gnix_nic_list;
+extern pthread_mutex_t gnix_nic_list_lock;
 
 /*
  * allocation flags for cleaning up GNI resources
@@ -69,6 +72,8 @@ extern uint32_t gnix_max_nics_per_ptag;
  */
 typedef int (*smsg_callback_fn_t)(void  *ptr, void *msg);
 
+extern smsg_callback_fn_t gnix_ep_smsg_callbacks[];
+
 /*
  * typedef for completer functions invoked
  * at initiator when local CQE (tx) is processed
@@ -84,24 +89,30 @@ typedef int (*smsg_completer_fn_t)(void  *desc, gni_return_t);
  * @var gni_nic_hndl         optional previously allocated gni_nic_hndl to
  *                           use for allocating GNI resources (GNI CQs) for
  *                           this nic
+ *
+ * @var gni_cdm_modes	     The mode bits gni_cdm_hndl was created with.
  */
 struct gnix_nic_attr {
 	gni_cdm_handle_t gni_cdm_hndl;
+	uint32_t	 gni_cdm_modes;
 	gni_nic_handle_t gni_nic_hndl;
 	bool use_cdm_id;
 	uint32_t cdm_id;
+	bool must_alloc;
+	struct gnix_auth_key *auth_key;
 };
 
 /**
  * GNIX nic struct
  *
  * @var gnix_nic_list        list element used for global NIC list
- * @var dom_nic_list         list element used for nic linked list associated
- *                           with a given gnix_fid_domain
+ * @var ptag_nic_list        list element used for NIC linked list associated
+ *                           with a given PTAG.
  * @var lock                 lock used for serializing access to
  *                           gni_nic_hndl, rx_cq, and tx_cq
  * @var gni_cdm_hndl         handle for the GNI communication domain (CDM)
  *                           this nic is bound to.
+ * @var gni_cdm_modes	     The mode bits gni_cdm_hndl was created with.
  * @var gni_nic_hndl         handle for the GNI nic to which this GNIX nic is bound
  * @var rx_cq                GNI rx cq (non-blocking) bound to this nic
  * @var rx_cq_blk            GNI rx cq (blocking) bound to this nic
@@ -114,6 +125,8 @@ struct gnix_nic_attr {
  *                           with this nic
  * @var tx_desc_base         base address for the block of memory from which
  *                           tx descriptors were allocated
+ * @var prog_vcs_lock        lock for prog_vcs
+ * @var prog_vcs             list of VCs needing progress
  * @var wq_lock              lock for serializing access to the nic's work queue
  * @var nic_wq               head of linked list of work queue elements
  *                           associated with this nic
@@ -152,10 +165,12 @@ struct gnix_nic_attr {
  */
 struct gnix_nic {
 	struct dlist_entry gnix_nic_list; /* global NIC list */
-	struct dlist_entry dom_nic_list;  /* domain NIC list */
+	struct dlist_entry ptag_nic_list; /* global PTAG NIC list */
+	struct dlist_entry gnix_nic_prog_list; /* temporary list for nic progression */
 	fastlock_t lock;
 	uint32_t allocd_gni_res;
 	gni_cdm_handle_t gni_cdm_hndl;
+	uint32_t	 gni_cdm_modes;
 	gni_nic_handle_t gni_nic_hndl;
 	gni_cq_handle_t rx_cq;
 	gni_cq_handle_t rx_cq_blk;
@@ -166,12 +181,10 @@ struct gnix_nic {
 	struct dlist_entry tx_desc_active_list;
 	struct dlist_entry tx_desc_free_list;
 	struct gnix_tx_descriptor *tx_desc_base;
-	fastlock_t rx_vc_lock;
-	struct dlist_entry rx_vcs;
-	fastlock_t work_vc_lock;
-	struct dlist_entry work_vcs;
-	fastlock_t tx_vc_lock;
-	struct dlist_entry tx_vcs;
+	fastlock_t prog_vcs_lock;
+	struct dlist_entry prog_vcs;
+	/* note this free list will be initialized for thread safe */
+	struct gnix_freelist vc_freelist;
 	uint8_t ptag;
 	uint32_t cookie;
 	uint32_t device_id;
@@ -184,16 +197,16 @@ struct gnix_nic {
 	gnix_bitmap_t vc_id_bitmap;
 	uint32_t mem_per_mbox;
 	struct gnix_mbox_alloc_handle *mbox_hndl;
+	/* TODO: gnix_buddy_alloc_handle_t *alloc_handle */
 	struct gnix_mbox_alloc_handle *s_rdma_buf_hndl;
 	struct gnix_mbox_alloc_handle *r_rdma_buf_hndl;
 	struct gnix_reference ref_cnt;
 	smsg_callback_fn_t const *smsg_callbacks;
 	struct slist err_txds;
-	void *int_bufs;
-	gni_mem_handle_t int_bufs_mdh;
 	gni_mem_handle_t irq_mem_hndl;
 	void *irq_mmap_addr;
 	size_t irq_mmap_len;
+	int requires_lock;
 };
 
 
@@ -240,6 +253,30 @@ struct gnix_smsg_rndzv_start_hdr {
 	uint32_t tail;
 };
 
+/**
+ * gnix_smsg_rndzv_iov_start_hdr
+ *
+ * @var flags	      the sender's flags needed on the receive side.
+ * @var imm	      the immediate data associated with this message.
+ * @var msg_tag       the tag associated with this message.
+ * @var mdh	      the memory handle associated with the iov buffer.
+ * @var iov_cnt       the length of the scatter/gather vector.
+ * @var req_addr      the sender's fabric request address.
+ * @var send_len      the cumulative size (in bytes) of the client's
+ * iov base buffers.
+ *
+ * @note the actual iov base addresses and lengths are placed in the
+ * data section of the start message.
+ */
+struct gnix_smsg_rndzv_iov_start_hdr {
+	uint64_t flags;
+	uint64_t imm;
+	uint64_t msg_tag;
+	uint64_t req_addr;
+	size_t   iov_cnt;
+	uint64_t send_len;
+};
+
 /**
  * gnix_smsg_rndzv_fin_hdr  - first part of a rendezvous send fin SMSG message
  *
@@ -247,17 +284,29 @@ struct gnix_smsg_rndzv_start_hdr {
  */
 struct gnix_smsg_rndzv_fin_hdr {
 	uint64_t req_addr;
+	int status;
 };
 
 /**
  * gnix_smsg_rma_data_hdr  - RMA remote data message
  *
- * @var flags  remote CQ RMA flags
- * @var data   remote CQ immediate data
+ * @var flags       control flags
+ * @var user_flags  remote CQ user flags
+ * @var user_data   remote CQ user immediate data
  */
 struct gnix_smsg_rma_data_hdr {
 	uint64_t flags;
-	uint64_t data;
+	uint64_t user_flags;
+	uint64_t user_data;
+};
+
+/**
+ * gnix_smsg_amo_cntr_hdr  - RMA remote counter message
+ *
+ * @var user_flags  control flags
+ */
+struct gnix_smsg_amo_cntr_hdr {
+	uint64_t flags;
 };
 
 /**
@@ -268,18 +317,22 @@ struct gnix_smsg_rma_data_hdr {
  * @var gni_desc         embedded GNI post descriptor
  * @var gnix_ct_descs    embedded GNI post descriptors for concatenated gets
  *                       used for unaligned gets
+ * @var gni_more_ct_descs embedded GNI post descriptors for concatenated puts
+			  or gets for FI_MORE.
  * @var gnix_smsg_eager_hdr embedded header for SMSG eager protocol
  * @var gnix_smsg_rndzv_start_hdr embedded header for rendezvous protocol
+ * @var gnix_smsg_rndzv_iov_start_hdr embedded header for iovec rndzv protocol
  * @var gnix_smsg_rndzv_fin_hdr embedded header for rendezvous protocol
  * @var gnix_smsg_rndzv_rma_data_hdr embedded header for remote notification for
  *                       rma operations
+ * @var gnix_smsg_amo_cntr_hdr embedded header for AMO remote counter events.
  * @var req              pointer to fab request associated with this descriptor
  * @var completer_fn     call back to invoke when associated GNI CQE's are
  *                       returned.
  * @var id               the id of this descriptor - the value returned
  *                       from GNI_CQ_MSG_ID
  * @var err_list         Error TXD list entry
- * @var int_buf          Intermediate buffer for landing unaligned data, etc.
+ * @var tx_failures	 Number of times this transmission descriptor failed.
  */
 struct gnix_tx_descriptor {
 	struct dlist_entry          list;
@@ -287,25 +340,21 @@ struct gnix_tx_descriptor {
 		struct {
 			gni_post_descriptor_t        gni_desc;
 			gni_ct_get_post_descriptor_t gni_ct_descs[2];
+			void			     *gni_more_ct_descs;
 		};
-		struct gnix_smsg_eager_hdr       eager_hdr;
-		struct gnix_smsg_rndzv_start_hdr rndzv_start_hdr;
-		struct gnix_smsg_rndzv_fin_hdr   rndzv_fin_hdr;
-		struct gnix_smsg_rma_data_hdr    rma_data_hdr;
+		struct gnix_smsg_eager_hdr           eager_hdr;
+		struct gnix_smsg_rndzv_start_hdr     rndzv_start_hdr;
+		struct gnix_smsg_rndzv_iov_start_hdr rndzv_iov_start_hdr;
+		struct gnix_smsg_rndzv_fin_hdr       rndzv_fin_hdr;
+		struct gnix_smsg_rma_data_hdr        rma_data_hdr;
+		struct gnix_smsg_amo_cntr_hdr	     amo_cntr_hdr;
 	};
 	struct gnix_fab_req *req;
 	int  (*completer_fn)(void *, gni_return_t);
 	int id;
 	struct slist_entry err_list;
-	void *int_buf;
 };
 
-/*
- * globals
- */
-
-extern uint32_t gnix_def_max_nics_per_ptag;
-
 /*
  * prototypes
  */
@@ -318,7 +367,26 @@ extern uint32_t gnix_def_max_nics_per_ptag;
  *                     is to be stored
  * @return             FI_SUCCESS on success, -FI_ENOSPC no free tx descriptors
  */
-int _gnix_nic_tx_alloc(struct gnix_nic *nic, struct gnix_tx_descriptor **tdesc);
+static inline int _gnix_nic_tx_alloc(struct gnix_nic *nic,
+               struct gnix_tx_descriptor **desc)
+{
+    struct dlist_entry *entry;
+
+    COND_ACQUIRE(nic->requires_lock, &nic->tx_desc_lock);
+    if (dlist_empty(&nic->tx_desc_free_list)) {
+        COND_RELEASE(nic->requires_lock, &nic->tx_desc_lock);
+        return -FI_ENOSPC;
+    }
+
+    entry = nic->tx_desc_free_list.next;
+    dlist_remove_init(entry);
+    dlist_insert_head(entry, &nic->tx_desc_active_list);
+    *desc = dlist_entry(entry, struct gnix_tx_descriptor, list);
+    COND_RELEASE(nic->requires_lock, &nic->tx_desc_lock);
+
+    return FI_SUCCESS;
+}
+
 
 /**
  * @brief frees a previously allocated tx descriptor
@@ -328,7 +396,17 @@ int _gnix_nic_tx_alloc(struct gnix_nic *nic, struct gnix_tx_descriptor **tdesc);
  * @param[in] tdesc    pointer to previously allocated tx descriptor
  * @return             FI_SUCCESS on success
  */
-int _gnix_nic_tx_free(struct gnix_nic *nic, struct gnix_tx_descriptor *tdesc);
+static inline int _gnix_nic_tx_free(struct gnix_nic *nic,
+                                struct gnix_tx_descriptor *desc)
+{
+    COND_ACQUIRE(nic->requires_lock, &nic->tx_desc_lock);
+    dlist_remove_init(&desc->list);
+    dlist_insert_head(&desc->list, &nic->tx_desc_free_list);
+    COND_RELEASE(nic->requires_lock, &nic->tx_desc_lock);
+
+    return FI_SUCCESS;
+}
+
 
 /**
  * @brief allocate a gnix_nic struct
@@ -362,12 +440,12 @@ int _gnix_nic_free(struct gnix_nic *nic);
 /**
  * @brief progresses control/data operations associated with the nic
  *
- * @param[in] nic      pointer to previously allocated gnix_nic struct
+ * @param[in] arg      pointer to previously allocated gnix_nic struct
  * @return             FI_SUCCESS on success, -FI_EINVAL if an invalid
  *                     nic struct was supplied. TODO: a lot more error
  *                     values can be returned.
  */
-int _gnix_nic_progress(struct gnix_nic *nic);
+int _gnix_nic_progress(void *arg);
 
 /**
  * @brief allocate a remote id for an object, used for looking up an object
@@ -409,17 +487,26 @@ int _gnix_nic_free_rem_id(struct gnix_nic *nic, int remote_id);
  */
 static inline void *__gnix_nic_elem_by_rem_id(struct gnix_nic *nic, int rem_id)
 {
+	void *elem;
+
 	assert(nic);
+
+	COND_ACQUIRE(nic->requires_lock, &nic->vc_id_lock);
+
 	assert(rem_id <= nic->vc_id_table_count);
-	return nic->vc_id_table[rem_id];
-	return 0;
+	elem = nic->vc_id_table[rem_id];
+
+	COND_RELEASE(nic->requires_lock, &nic->vc_id_lock);
+
+	return elem;
 }
 
 void _gnix_nic_txd_err_inject(struct gnix_nic *nic,
 			      struct gnix_tx_descriptor *txd);
 
-#ifdef __cplusplus
-} /* extern "C" */
-#endif
+/**
+ * @brief Initialize global NIC data.
+ */
+void _gnix_nic_init(void);
 
 #endif /* _GNIX_NIC_H_ */
