@@ -1,7 +1,7 @@
 /*
- * Copyright (c) 2015-2016 Los Alamos National Security, LLC.
+ * Copyright (c) 2015-2017 Los Alamos National Security, LLC.
  *                         All rights reserved.
- * Copyright (c) 2015-2016 Cray Inc. All rights reserved.
+ * Copyright (c) 2015-2017 Cray Inc. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -42,9 +42,21 @@
 #include "gnix.h"
 #include "gnix_nic.h"
 #include "gnix_util.h"
+#include "gnix_xpmem.h"
+#include "gnix_hashtable.h"
+#include "gnix_auth_key.h"
+
+#define GNIX_MR_MODE_DEFAULT FI_MR_BASIC
+#define GNIX_NUM_PTAGS 256
 
 gni_cq_mode_t gnix_def_gni_cq_modes = GNI_CQ_PHYS_PAGES;
 
+static char *__gnix_mr_type_to_str[GNIX_MR_MAX_TYPE] = {
+		[GNIX_MR_TYPE_INTERNAL] = "internal",
+		[GNIX_MR_TYPE_UDREG] = "udreg",
+		[GNIX_MR_TYPE_NONE] = "none",
+};
+
 /*******************************************************************************
  * Forward declaration for ops structures.
  ******************************************************************************/
@@ -57,31 +69,29 @@ static struct fi_ops_domain gnix_domain_ops;
 static void __domain_destruct(void *obj)
 {
 	int ret = FI_SUCCESS;
-	struct gnix_nic *p, *next;
 	struct gnix_fid_domain *domain = (struct gnix_fid_domain *) obj;
+	struct gnix_mr_cache_info *info;
+	int i;
 
 	GNIX_TRACE(FI_LOG_DOMAIN, "\n");
 
-	/* if the domain isn't being destructed by close, we need to check the
-	 * cache again. This isn't a likely case. Destroy must succeed since we
-	 * are in the destruct path */
-	if (domain->mr_cache) {
-		ret = _gnix_mr_cache_destroy(domain->mr_cache);
+	for (i = 0; i < GNIX_NUM_PTAGS; i++) {
+		info = &domain->mr_cache_info[i];
+
+		fastlock_acquire(&info->mr_cache_lock);
+		ret = _gnix_close_cache(domain, info);
+		fastlock_release(&info->mr_cache_lock);
+
 		if (ret != FI_SUCCESS)
-			GNIX_WARN(FI_LOG_DOMAIN, "failed to destroy mr cache "
-					"during domain destruct, dom=%p",
-					domain);
-		assert(ret == FI_SUCCESS);
+			GNIX_FATAL(FI_LOG_MR,
+					"failed to close memory "
+					"registration cache\n");
 	}
 
-	/*
-	 * Drop a reference to each NIC used by this domain.
-	 */
-
-	dlist_for_each_safe(&domain->nic_list, p, next, dom_nic_list) {
-		dlist_remove(&p->dom_nic_list);
-		_gnix_ref_put(p);
-	}
+	/* TODO: known issue of multiple consumers of notifier notifications */
+	ret = _gnix_notifier_close(domain->mr_cache_attr.notifier);
+	if (ret != FI_SUCCESS)
+		GNIX_FATAL(FI_LOG_MR, "failed to close MR notifier\n");
 
 	/*
 	 * remove from the list of cdms attached to fabric
@@ -97,10 +107,19 @@ static void __domain_destruct(void *obj)
 
 static void __stx_destruct(void *obj)
 {
+	int ret;
 	struct gnix_fid_stx *stx = (struct gnix_fid_stx *) obj;
 
 	GNIX_TRACE(FI_LOG_DOMAIN, "\n");
 
+	if (stx->nic) {
+		ret = _gnix_nic_free(stx->nic);
+		if (ret != FI_SUCCESS)
+			GNIX_WARN(FI_LOG_EP_CTRL,
+				    "_gnix_nic_free call returned %s\n",
+			     fi_strerror(-ret));
+	}
+
 	memset(stx, 0, sizeof(*stx));
 	free(stx);
 }
@@ -111,7 +130,6 @@ static void __stx_destruct(void *obj)
 
 /**
  * Creates a shared transmit context.
- * @note This is basically a no-op for the GNI provider.
  *
  * @param[in]  val  value to be sign extended
  * @param[in]  len  length to sign extend the value
@@ -142,6 +160,9 @@ DIRECT_FN STATIC int gnix_stx_open(struct fid_domain *dom,
 	}
 
 	stx_priv->domain = domain;
+	stx_priv->auth_key = NULL;
+	stx_priv->nic = NULL;
+
 	_gnix_ref_init(&stx_priv->ref_cnt, 1, __stx_destruct);
 
 	_gnix_ref_get(stx_priv->domain);
@@ -150,6 +171,7 @@ DIRECT_FN STATIC int gnix_stx_open(struct fid_domain *dom,
 	stx_priv->stx_fid.fid.context = context;
 	stx_priv->stx_fid.fid.ops = &gnix_stx_ops;
 	stx_priv->stx_fid.ops = NULL;
+	domain->num_allocd_stxs++;
 
 	*stx = &stx_priv->stx_fid;
 
@@ -189,6 +211,8 @@ static int gnix_domain_close(fid_t fid)
 {
 	int ret = FI_SUCCESS, references_held;
 	struct gnix_fid_domain *domain;
+	int i;
+	struct gnix_mr_cache_info *info;
 
 	GNIX_TRACE(FI_LOG_DOMAIN, "\n");
 
@@ -198,17 +222,38 @@ static int gnix_domain_close(fid_t fid)
 		goto err;
 	}
 
-	/* before checking the refcnt, flush the memory registration cache */
-	if (domain->mr_cache) {
-		fastlock_acquire(&domain->mr_cache_lock);
-		ret = _gnix_mr_cache_flush(domain->mr_cache);
-		if (ret != FI_SUCCESS) {
-			GNIX_WARN(FI_LOG_DOMAIN,
-				  "failed to flush memory cache on domain close\n");
-			fastlock_release(&domain->mr_cache_lock);
-			goto err;
+	for (i = 0; i < GNIX_NUM_PTAGS; i++) {
+		info = &domain->mr_cache_info[i];
+
+		if (!domain->mr_cache_info[i].inuse)
+			continue;
+
+		/* before checking the refcnt,
+		 * flush the memory registration cache
+		 */
+		if (info->mr_cache_ro) {
+			fastlock_acquire(&info->mr_cache_lock);
+			ret = _gnix_mr_cache_flush(info->mr_cache_ro);
+			if (ret != FI_SUCCESS) {
+				GNIX_WARN(FI_LOG_DOMAIN,
+					  "failed to flush memory cache on domain close\n");
+				fastlock_release(&info->mr_cache_lock);
+				goto err;
+			}
+			fastlock_release(&info->mr_cache_lock);
+		}
+
+		if (info->mr_cache_rw) {
+			fastlock_acquire(&info->mr_cache_lock);
+			ret = _gnix_mr_cache_flush(info->mr_cache_rw);
+			if (ret != FI_SUCCESS) {
+				GNIX_WARN(FI_LOG_DOMAIN,
+					  "failed to flush memory cache on domain close\n");
+				fastlock_release(&info->mr_cache_lock);
+				goto err;
+			}
+			fastlock_release(&info->mr_cache_lock);
 		}
-		fastlock_release(&domain->mr_cache_lock);
 	}
 
 	/*
@@ -252,6 +297,35 @@ static const uint32_t default_rx_cq_size = 16384;
 static const uint32_t default_tx_cq_size = 2048;
 static const uint32_t default_max_retransmits = 5;
 static const int32_t default_err_inject_count; /* static var is zeroed */
+static const uint32_t default_dgram_progress_timeout = 100;
+static const uint32_t default_eager_auto_progress = 0;
+
+static int __gnix_string_to_mr_type(const char *name)
+{
+	int i;
+	for (i = 0; i < GNIX_MR_MAX_TYPE; i++)
+		if (strncmp(name, __gnix_mr_type_to_str[i],
+				strlen(__gnix_mr_type_to_str[i])) == 0)
+			return i;
+
+	return -1;
+}
+
+static int
+__gnix_dom_ops_flush_cache(struct fid *fid)
+{
+	struct gnix_fid_domain *domain;
+
+	GNIX_TRACE(FI_LOG_DOMAIN, "\n");
+
+	domain = container_of(fid, struct gnix_fid_domain, domain_fid.fid);
+	if (domain->domain_fid.fid.fclass != FI_CLASS_DOMAIN) {
+		GNIX_WARN(FI_LOG_DOMAIN, ("Invalid domain\n"));
+		return -FI_EINVAL;
+	}
+
+	return _gnix_flush_registration_cache(domain);
+}
 
 static int
 __gnix_dom_ops_get_val(struct fid *fid, dom_ops_val_t t, void *val)
@@ -314,6 +388,34 @@ __gnix_dom_ops_get_val(struct fid *fid, dom_ops_val_t t, void *val)
 	case GNI_MR_CACHE_LAZY_DEREG:
 		*(int32_t *)val = domain->mr_cache_attr.lazy_deregistration;
 		break;
+	case GNI_MR_CACHE:
+		*(char **) val = __gnix_mr_type_to_str[domain->mr_cache_type];
+		break;
+	case GNI_MR_UDREG_REG_LIMIT:
+		*(int32_t *)val = domain->udreg_reg_limit;
+		break;
+	case GNI_MR_HARD_REG_LIMIT:
+		*(int32_t *)val = domain->mr_cache_attr.hard_reg_limit;
+		break;
+	case GNI_MR_SOFT_REG_LIMIT:
+		*(int32_t *)val = domain->mr_cache_attr.soft_reg_limit;
+		break;
+	case GNI_MR_HARD_STALE_REG_LIMIT:
+		*(int32_t *)val = domain->mr_cache_attr.hard_stale_limit;
+		break;
+	case GNI_XPMEM_ENABLE:
+		*(bool *)val = domain->params.xpmem_enabled;
+#if !HAVE_XPMEM
+		GNIX_WARN(FI_LOG_DOMAIN,
+			  "GNI provider XPMEM support not configured\n");
+#endif
+		break;
+	case GNI_DGRAM_PROGRESS_TIMEOUT:
+		*(uint32_t *)val = domain->params.dgram_progress_timeout;
+		break;
+	case GNI_EAGER_AUTO_PROGRESS:
+		*(uint32_t *)val = domain->params.eager_auto_progress;
+		break;
 	default:
 		GNIX_WARN(FI_LOG_DOMAIN, ("Invalid dom_ops_val\n"));
 		return -FI_EINVAL;
@@ -325,9 +427,8 @@ __gnix_dom_ops_get_val(struct fid *fid, dom_ops_val_t t, void *val)
 static int
 __gnix_dom_ops_set_val(struct fid *fid, dom_ops_val_t t, void *val)
 {
-	gni_return_t grc = GNI_RC_SUCCESS;
-	struct gnix_nic *nic;
 	struct gnix_fid_domain *domain;
+	int ret, type;
 
 	GNIX_TRACE(FI_LOG_DOMAIN, "\n");
 
@@ -378,21 +479,6 @@ __gnix_dom_ops_set_val(struct fid *fid, dom_ops_val_t t, void *val)
 		break;
 	case GNI_MAX_RETRANSMITS:
 		domain->params.max_retransmits = *(uint32_t *)val;
-		dlist_for_each(&domain->nic_list, nic, dom_nic_list)
-		{
-			fastlock_acquire(&nic->lock);
-			grc = GNI_SmsgSetMaxRetrans(nic->gni_nic_hndl,
-						    *(uint16_t *)val);
-			fastlock_release(&nic->lock);
-			if (grc != GNI_RC_SUCCESS)
-				break;
-		}
-
-		if (unlikely(grc != GNI_RC_SUCCESS)) {
-			GNIX_WARN(FI_LOG_DOMAIN, "failed to set smsg max "
-				  "retrans, ret=%s\n", gni_err_str[grc]);
-			return gnixu_to_fi_errno(grc);
-		}
 		break;
 	case GNI_ERR_INJECT_COUNT:
 		domain->params.err_inject_count = *(int32_t *)val;
@@ -400,6 +486,51 @@ __gnix_dom_ops_set_val(struct fid *fid, dom_ops_val_t t, void *val)
 	case GNI_MR_CACHE_LAZY_DEREG:
 		domain->mr_cache_attr.lazy_deregistration = *(int32_t *)val;
 		break;
+	case GNI_MR_CACHE:
+		if (val != NULL) {
+			GNIX_DEBUG(FI_LOG_DOMAIN, "user provided value=%s\n",
+					*(char **) val);
+
+			type = __gnix_string_to_mr_type(*(const char **) val);
+			if (type < 0 || type >= GNIX_MR_MAX_TYPE)
+				return -FI_EINVAL;
+
+			GNIX_DEBUG(FI_LOG_DOMAIN, "setting domain mr type to %s\n",
+					__gnix_mr_type_to_str[type]);
+
+			ret = _gnix_open_cache(domain, type);
+			if (ret != FI_SUCCESS)
+				return -FI_EINVAL;
+		}
+		break;
+	case GNI_MR_HARD_REG_LIMIT:
+		domain->mr_cache_attr.hard_reg_limit = *(int32_t *) val;
+		break;
+	case GNI_MR_SOFT_REG_LIMIT:
+		domain->mr_cache_attr.soft_reg_limit = *(int32_t *) val;
+		break;
+	case GNI_MR_HARD_STALE_REG_LIMIT:
+		domain->mr_cache_attr.hard_stale_limit = *(int32_t *) val;
+		break;
+	case GNI_MR_UDREG_REG_LIMIT:
+		if (*(int32_t *) val < 0)
+			return -FI_EINVAL;
+		domain->udreg_reg_limit = *(int32_t *) val;
+		break;
+	case GNI_XPMEM_ENABLE:
+#if HAVE_XPMEM
+		domain->params.xpmem_enabled = *(bool *)val;
+#else
+		GNIX_WARN(FI_LOG_DOMAIN,
+			  "GNI provider XPMEM support not configured\n");
+#endif
+		break;
+	case GNI_DGRAM_PROGRESS_TIMEOUT:
+		domain->params.dgram_progress_timeout = *(uint32_t *)val;
+		break;
+	case GNI_EAGER_AUTO_PROGRESS:
+		domain->params.eager_auto_progress = *(uint32_t *)val;
+		break;
 	default:
 		GNIX_WARN(FI_LOG_DOMAIN, ("Invalid dom_ops_val\n"));
 		return -FI_EINVAL;
@@ -408,10 +539,10 @@ __gnix_dom_ops_set_val(struct fid *fid, dom_ops_val_t t, void *val)
 	return FI_SUCCESS;
 }
 
-
 static struct fi_gni_ops_domain gnix_ops_domain = {
 	.set_val = __gnix_dom_ops_set_val,
-	.get_val = __gnix_dom_ops_get_val
+	.get_val = __gnix_dom_ops_get_val,
+	.flush_cache = __gnix_dom_ops_flush_cache,
 };
 
 DIRECT_FN int gnix_domain_bind(struct fid_domain *domain, struct fid *fid,
@@ -439,42 +570,26 @@ DIRECT_FN int gnix_domain_open(struct fid_fabric *fabric, struct fi_info *info,
 {
 	struct gnix_fid_domain *domain = NULL;
 	int ret = FI_SUCCESS;
-	uint8_t ptag;
-	uint32_t cookie;
 	struct gnix_fid_fabric *fabric_priv;
+	struct gnix_auth_key *auth_key = NULL;
+	int i;
 
 	GNIX_TRACE(FI_LOG_DOMAIN, "\n");
 
 	fabric_priv = container_of(fabric, struct gnix_fid_fabric, fab_fid);
-	if (!info->domain_attr->name ||
-	    strncmp(info->domain_attr->name, gnix_dom_name,
-		    strlen(gnix_dom_name))) {
-		ret = -FI_EINVAL;
-		goto err;
-	}
 
-	/*
-	 * check cookie/ptag credentials - for FI_EP_MSG we may be creating a
-	 * domain
-	 * using a cookie supplied being used by the server.  Otherwise, we use
-	 * use the cookie/ptag supplied by the job launch system.
-	 */
-	if (info->dest_addr) {
-		ret =
-		    gnixu_get_rdma_credentials(info->dest_addr, &ptag, &cookie);
-		if (ret) {
-			GNIX_WARN(FI_LOG_DOMAIN,
-				   "gnixu_get_rdma_credentials returned ptag %u cookie 0x%x\n",
-				   ptag, cookie);
-			goto err;
-		}
-	} else {
-		ret = gnixu_get_rdma_credentials(NULL, &ptag, &cookie);
-	}
+	if (FI_VERSION_LT(fabric->api_version, FI_VERSION(1, 5)) &&
+		(info->domain_attr->auth_key_size || info->domain_attr->auth_key))
+			return -FI_EINVAL;
+
+	auth_key = GNIX_GET_AUTH_KEY(info->domain_attr->auth_key,
+			info->domain_attr->auth_key_size);
+	if (!auth_key)
+		return -FI_EINVAL;
 
 	GNIX_INFO(FI_LOG_DOMAIN,
-		  "gnix rdma credentials returned ptag %u cookie 0x%x\n",
-		  ptag, cookie);
+		  "authorization key=%p ptag %u cookie 0x%x\n",
+		  auth_key, auth_key->ptag, auth_key->cookie);
 
 	domain = calloc(1, sizeof *domain);
 	if (domain == NULL) {
@@ -482,12 +597,32 @@ DIRECT_FN int gnix_domain_open(struct fid_fabric *fabric, struct fi_info *info,
 		goto err;
 	}
 
+	domain->mr_cache_info = calloc(sizeof(*domain->mr_cache_info),
+		GNIX_NUM_PTAGS);
+	if (!domain->mr_cache_info) {
+		ret = -FI_ENOMEM;
+		goto err;
+	}
+
+	domain->auth_key = auth_key;
+
 	domain->mr_cache_attr = _gnix_default_mr_cache_attr;
 	domain->mr_cache_attr.reg_context = (void *) domain;
 	domain->mr_cache_attr.dereg_context = NULL;
 	domain->mr_cache_attr.destruct_context = NULL;
-	domain->mr_cache = NULL;
+
+	ret = _gnix_notifier_open(&domain->mr_cache_attr.notifier);
+	if (ret != FI_SUCCESS)
+		goto err;
+
 	fastlock_init(&domain->mr_cache_lock);
+	for (i = 0; i < GNIX_NUM_PTAGS; i++) {
+		domain->mr_cache_info[i].inuse = 0;
+		domain->mr_cache_info[i].domain = domain;
+		fastlock_init(&domain->mr_cache_info[i].mr_cache_lock);
+	}
+
+	domain->udreg_reg_limit = 4096;
 
 	dlist_init(&domain->nic_list);
 	dlist_init(&domain->list);
@@ -497,9 +632,8 @@ DIRECT_FN int gnix_domain_open(struct fid_fabric *fabric, struct fi_info *info,
 	domain->fabric = fabric_priv;
 	_gnix_ref_get(domain->fabric);
 
-	domain->ptag = ptag;
-	domain->cookie = cookie;
 	domain->cdm_id_seed = getpid();  /* TODO: direct syscall better */
+	domain->addr_format = info->addr_format;
 
 	/* user tunables */
 	domain->params.msg_rendezvous_thresh = default_msg_rendezvous_thresh;
@@ -512,11 +646,18 @@ DIRECT_FN int gnix_domain_open(struct fid_fabric *fabric, struct fi_info *info,
 	domain->params.mbox_num_per_slab = default_mbox_num_per_slab;
 	domain->params.mbox_maxcredit = default_mbox_maxcredit;
 	domain->params.mbox_msg_maxsize = default_mbox_msg_maxsize;
+	domain->params.rx_cq_size = default_rx_cq_size;
+	domain->params.tx_cq_size = default_tx_cq_size;
 	domain->params.max_retransmits = default_max_retransmits;
 	domain->params.err_inject_count = default_err_inject_count;
+#if HAVE_XPMEM
+	domain->params.xpmem_enabled = true;
+#else
+	domain->params.xpmem_enabled = false;
+#endif
+	domain->params.dgram_progress_timeout = default_dgram_progress_timeout;
+	domain->params.eager_auto_progress = default_eager_auto_progress;
 
-	domain->gni_tx_cq_size = default_tx_cq_size;
-	domain->gni_rx_cq_size = default_rx_cq_size;
 	domain->gni_cq_modes = gnix_def_gni_cq_modes;
 	_gnix_ref_init(&domain->ref_cnt, 1, __domain_destruct);
 
@@ -528,12 +669,21 @@ DIRECT_FN int gnix_domain_open(struct fid_fabric *fabric, struct fi_info *info,
 
 	domain->control_progress = info->domain_attr->control_progress;
 	domain->data_progress = info->domain_attr->data_progress;
+	domain->thread_model = info->domain_attr->threading;
+	domain->mr_is_init = 0;
+	domain->mr_iov_limit = info->domain_attr->mr_iov_limit;
+
 	fastlock_init(&domain->cm_nic_lock);
 
+	_gnix_open_cache(domain, GNIX_DEFAULT_CACHE_TYPE);
+
 	*dom = &domain->domain_fid;
 	return FI_SUCCESS;
 
 err:
+	if (domain && domain->mr_cache_info)
+		free(domain->mr_cache_info);
+
 	if (domain != NULL) {
 		free(domain);
 	}
@@ -547,18 +697,6 @@ DIRECT_FN int gnix_srx_context(struct fid_domain *domain,
 	return -FI_ENOSYS;
 }
 
-DIRECT_FN int gnix_scalable_ep_open(struct fid_domain *domain,
-				    struct fi_info *info,
-				    struct fid_ep **sep, void *context)
-{
-	return -FI_ENOSYS;
-}
-
-DIRECT_FN int gnix_scalable_ep_bind(fid_t fid, struct fid *bfid, uint64_t flags)
-{
-	return -FI_ENOSYS;
-}
-
 /*******************************************************************************
  * FI_OPS_* data structures.
  ******************************************************************************/
@@ -581,8 +719,8 @@ static struct fi_ops gnix_domain_fi_ops = {
 static struct fi_ops_mr gnix_domain_mr_ops = {
 	.size = sizeof(struct fi_ops_mr),
 	.reg = gnix_mr_reg,
-	.regv = fi_no_mr_regv,
-	.regattr = fi_no_mr_regattr
+	.regv = gnix_mr_regv,
+	.regattr = gnix_mr_regattr,
 };
 
 static struct fi_ops_domain gnix_domain_ops = {
@@ -590,7 +728,7 @@ static struct fi_ops_domain gnix_domain_ops = {
 	.av_open = gnix_av_open,
 	.cq_open = gnix_cq_open,
 	.endpoint = gnix_ep_open,
-	.scalable_ep = fi_no_scalable_ep,
+	.scalable_ep = gnix_sep_open,
 	.cntr_open = gnix_cntr_open,
 	.poll_open = fi_no_poll_open,
 	.stx_ctx = gnix_stx_open,
