@@ -31,17 +31,24 @@
  */
 
 #include <gnix_mr_cache.h>
+#include <gnix_mr_notifier.h>
 #include <gnix.h>
 
-typedef enum cache_entry_flags {
-	GNIX_CE_RETIRED = 1 << 0,
-} cache_entry_flags_e;
+typedef unsigned long long int cache_entry_state_t;
+/* These are used for entry state and should be unique */
+#define GNIX_CES_INUSE       (1ULL << 8)    /* in use */
+#define GNIX_CES_STALE       (2ULL << 8)    /* cached for possible reuse */
+#define GNIX_CES_STATE_MASK  (0xFULL << 8)
 
-typedef enum cache_entry_state {
-	GNIX_CES_DEAD = 0,
-	GNIX_CES_INUSE,
-	GNIX_CES_STALE,
-} cache_entry_state_e;
+typedef unsigned long long int cache_entry_flag_t;
+/* One or more of these can be combined with the above */
+#define GNIX_CE_RETIRED     (1ULL << 61)   /* in use, but not to be reused */
+#define GNIX_CE_MERGED      (1ULL << 62)   /* merged entry, i.e., not
+					    * an original request from
+					    * fi_mr_reg */
+#define GNIX_CE_UNMAPPED    (1ULL << 63)   /* at least 1 page of the
+					    * entry has been unmapped
+					    * by the OS */
 
 /**
  * @brief structure for containing the fields relevant to the memory cache key
@@ -67,24 +74,18 @@ typedef struct gnix_mr_cache_key {
  * @var   lru_entry  lru list entry
  * @var   siblings   list of sibling entries
  * @var   children   list of subsumed child entries
- * @var   flags      cache entry flags @see cache_entry_flags_e
  */
 typedef struct gnix_mr_cache_entry {
-	cache_entry_state_e state;
+	cache_entry_state_t state;
 	gnix_mr_cache_key_t key;
-	atomic_t ref_cnt;
+	ofi_atomic32_t ref_cnt;
 	struct dlist_entry lru_entry;
 	struct dlist_entry siblings;
 	struct dlist_entry children;
-	cache_entry_flags_e flags;
 	uint64_t mr[0];
 } gnix_mr_cache_entry_t;
 
 /* forward declarations */
-int _gnix_mr_cache_init(
-		gnix_mr_cache_t      **cache,
-		gnix_mr_cache_attr_t *attr);
-
 int _gnix_mr_cache_register(
 		gnix_mr_cache_t             *cache,
 		uint64_t                    address,
@@ -104,6 +105,9 @@ static inline int __mr_cache_entry_get(
 		gnix_mr_cache_t       *cache,
 		gnix_mr_cache_entry_t *entry);
 
+static inline int __mr_cache_entry_destroy(gnix_mr_cache_t *cache,
+					   gnix_mr_cache_entry_t *entry);
+
 static int __mr_cache_create_registration(
 		gnix_mr_cache_t             *cache,
 		uint64_t                    address,
@@ -120,9 +124,73 @@ gnix_mr_cache_attr_t __default_mr_cache_attr = {
 		.soft_reg_limit      = 4096,
 		.hard_reg_limit      = -1,
 		.hard_stale_limit    = 128,
+#if HAVE_KDREG
 		.lazy_deregistration = 1,
+#else
+		.lazy_deregistration = 0,
+#endif
 };
 
+/* Functions for using and manipulating cache entry state */
+static inline cache_entry_state_t __entry_get_state(gnix_mr_cache_entry_t *e)
+{
+	return e->state & GNIX_CES_STATE_MASK;
+}
+
+static inline void __entry_set_state(gnix_mr_cache_entry_t *e,
+				     cache_entry_state_t s)
+{
+	e->state = (e->state & ~GNIX_CES_STATE_MASK) |
+		(s & GNIX_CES_STATE_MASK);
+}
+
+static inline void __entry_reset_state(gnix_mr_cache_entry_t *e)
+{
+	e->state = 0ULL;
+}
+
+static inline bool __entry_is_flag(gnix_mr_cache_entry_t *e,
+				   cache_entry_flag_t f)
+{
+	return (e->state & f) != 0;
+}
+
+static inline void __entry_set_flag(gnix_mr_cache_entry_t *e,
+				    cache_entry_flag_t f)
+{
+	e->state = e->state | f;
+}
+
+static inline bool __entry_is_retired(gnix_mr_cache_entry_t *e)
+{
+	return __entry_is_flag(e, GNIX_CE_RETIRED);
+}
+
+static inline bool __entry_is_merged(gnix_mr_cache_entry_t *e)
+{
+	return __entry_is_flag(e, GNIX_CE_MERGED);
+}
+
+static inline bool __entry_is_unmapped(gnix_mr_cache_entry_t *e)
+{
+	return __entry_is_flag(e, GNIX_CE_UNMAPPED);
+}
+
+static inline void __entry_set_retired(gnix_mr_cache_entry_t *e)
+{
+	__entry_set_flag(e, GNIX_CE_RETIRED);
+}
+
+static inline void __entry_set_merged(gnix_mr_cache_entry_t *e)
+{
+	__entry_set_flag(e, GNIX_CE_MERGED);
+}
+
+static inline void __entry_set_unmapped(gnix_mr_cache_entry_t *e)
+{
+	__entry_set_flag(e, GNIX_CE_UNMAPPED);
+}
+
 /**
  * Key comparison function for finding overlapping gnix memory
  * registration cache entries
@@ -130,7 +198,7 @@ gnix_mr_cache_attr_t __default_mr_cache_attr = {
  * @param[in] x key to be inserted or found
  * @param[in] y key to be compared against
  *
- * @return    -1 if it should be positioned at the left, 0 if the same,
+ * @return    -1 if it should be positioned at the left, 0 if it overlaps,
  *             1 otherwise
  */
 static int __find_overlapping_addr(
@@ -139,8 +207,8 @@ static int __find_overlapping_addr(
 {
 	gnix_mr_cache_key_t *to_find  = (gnix_mr_cache_key_t *) x;
 	gnix_mr_cache_key_t *to_compare = (gnix_mr_cache_key_t *) y;
-	uint64_t to_find_end = to_find->address + to_find->length;
-	uint64_t to_compare_end = to_compare->address + to_compare->length;
+	uint64_t to_find_end = to_find->address + to_find->length - 1;
+	uint64_t to_compare_end = to_compare->address + to_compare->length - 1;
 
 	/* format: (x_addr,  x_len) - (y_addr,  y_len) truth_value
 	 *
@@ -259,7 +327,11 @@ static inline void __attach_retired_entries_to_registration(
 			__mr_cache_entry_put(cache, entry);
 		}
 	}
-	assert(dlist_empty(retired_entries));
+
+	if (!dlist_empty(retired_entries)) {
+		GNIX_FATAL(FI_LOG_MR, "retired_entries not empty\n");
+	}
+
 	__mr_cache_entry_get(cache, parent);
 }
 
@@ -274,18 +346,19 @@ static inline void __remove_sibling_entries_from_tree(
 
 	dlist_for_each(list, entry, siblings)
 	{
-		GNIX_INFO(FI_LOG_MR,
-				"removing key from tree, key=%llx:%llx\n",
-				entry->key.address,
-				entry->key.length);
+		GNIX_DEBUG(FI_LOG_MR,
+			   "removing key from tree, key=%llx:%llx\n",
+			   entry->key.address, entry->key.length);
 		iter = rbtFind(tree, &entry->key);
-		assert(iter);
+		if (unlikely(!iter)) {
+			GNIX_FATAL(FI_LOG_MR, "key not found\n");
+		}
 
 		rc = rbtErase(tree, iter);
-		if (unlikely(rc != RBT_STATUS_OK))
-			GNIX_INFO(FI_LOG_MR,
-					"could not remove entry from tree\n");
-		assert(rc == RBT_STATUS_OK);
+		if (unlikely(rc != RBT_STATUS_OK)) {
+			GNIX_FATAL(FI_LOG_MR,
+				   "could not remove entry from tree\n");
+		}
 	}
 }
 
@@ -336,6 +409,250 @@ static inline int __mr_cache_lru_dequeue(
 	return FI_SUCCESS;
 }
 
+/**
+ * Removes a particular registration cache entry from the lru cache.
+ *
+ * @param[in] cache  a memory registration cache
+ * @param[in] entry  a memory registration cache entry
+ *
+ * @return           FI_SUCCESS, on success
+ * @return           -FI_ENOENT, on empty LRU
+ */
+static inline int __mr_cache_lru_remove(
+		gnix_mr_cache_t       *cache,
+		gnix_mr_cache_entry_t *entry)
+{
+	/* Could do some error checking to see if in cache */
+
+	dlist_remove(&entry->lru_entry);
+
+	return FI_SUCCESS;
+}
+
+/**
+ * Remove entries that have been unmapped as indicated by the notifer
+ *
+ * @param[in] cache  a memory registration cache
+ *
+ * @return           nothing
+ */
+static bool __notifier_warned = false;
+static void
+__clear_notifier_events(gnix_mr_cache_t *cache)
+{
+	int ret;
+	gnix_mr_cache_entry_t *entry;
+	RbtIterator iter;
+	uint64_t cookie;
+
+	if (!cache->attr.notifier) {
+		return;
+	}
+
+	if (!cache->attr.lazy_deregistration) {
+		return;
+	}
+
+	ret = _gnix_notifier_get_event(cache->attr.notifier,
+				       &cookie, sizeof(cookie));
+	while (ret > 0) {
+		if (ret == sizeof(cookie)) {
+			entry = (gnix_mr_cache_entry_t *) cookie;
+			switch (__entry_get_state(entry)) {
+			case GNIX_CES_INUSE:
+				/* First, warn that this might be a
+				 * problem.*/
+				if ((__notifier_warned == false) &&
+				    !__entry_is_merged(entry)) {
+					GNIX_WARN(FI_LOG_MR,
+						  "Registered memory region"
+						  " includes unmapped pages."
+						  "  Have you freed memory"
+						  " without closing the memory"
+						  " region?\n");
+					__notifier_warned = true;
+				}
+
+				GNIX_DEBUG(FI_LOG_MR,
+					   "Marking unmapped entry (%p)"
+					   " as retired %llx:%llx\n", entry,
+					   entry->key.address,
+					   entry->key.length);
+
+				__entry_set_unmapped(entry);
+
+				if (__entry_is_retired(entry)) {
+					/* Nothing to do */
+					break;
+				}
+
+				/* Retire this entry (remove from
+				 * inuse tree) */
+
+				__entry_set_retired(entry);
+				iter = rbtFind(cache->inuse.rb_tree,
+					       &entry->key);
+				if (likely(iter != NULL)) {
+					ret = rbtErase(cache->inuse.rb_tree,
+						       iter);
+					if (ret != RBT_STATUS_OK) {
+						GNIX_FATAL(FI_LOG_MR,
+							   "Unmapped entry"
+							   " could not be"
+							   " removed from"
+							   " in usetree.\n");
+					}
+				} else {
+					/*  The only way we should get
+					 *  here is if we're in the
+					 *  middle of retiring this
+					 *  entry.  Not sure if this
+					 *  is worth a separate
+					 *  warning from the one
+					 *  above. */
+				}
+
+				break;
+			case GNIX_CES_STALE:
+				__entry_set_unmapped(entry);
+				iter = rbtFind(cache->stale.rb_tree,
+					       &entry->key);
+				if (!iter) {
+					break;
+				}
+
+				ret = rbtErase(cache->stale.rb_tree, iter);
+				if (ret != RBT_STATUS_OK) {
+					GNIX_FATAL(FI_LOG_MR,
+						   "Unmapped entry could"
+						   " not be removed "
+						   " from stale tree.\n");
+				}
+
+				GNIX_DEBUG(FI_LOG_MR, "Removed unmapped entry"
+					   " (%p) from stale tree %llx:%llx\n",
+					   entry, entry->key.address,
+					   entry->key.length);
+
+				if (__mr_cache_lru_remove(cache, entry)
+				    == FI_SUCCESS) {
+					GNIX_DEBUG(FI_LOG_MR, "Removed"
+						   " unmapped entry (%p)"
+						   " from lru list %llx:%llx\n",
+						   entry, entry->key.address,
+						   entry->key.length);
+
+					ofi_atomic_dec32(&cache->stale.elements);
+
+				} else {
+					GNIX_WARN(FI_LOG_MR, "Failed to remove"
+						  " unmapped entry"
+						  " from lru list (%p) %p\n",
+						  entry, iter);
+				}
+				
+				__mr_cache_entry_destroy(cache, entry);
+
+				break;
+			default:
+				GNIX_FATAL(FI_LOG_MR,
+					   "Unmapped entry (%p) in incorrect"
+					   " state: %d\n",
+					   entry, entry->state);
+			}
+
+		} else {
+			/* Should we do something else here? */
+			GNIX_FATAL(FI_LOG_MR,
+				   "_gnix_notifier_get_event returned incomplete event\n");
+		}
+		ret = _gnix_notifier_get_event(cache->attr.notifier,
+					       &cookie, sizeof(cookie));
+	}
+	if (ret != -FI_EAGAIN) {
+		/* Should we do something else here? */
+		GNIX_WARN(FI_LOG_MR,
+			  "_gnix_notifier_get_event returned error: %s\n",
+			  fi_strerror(-ret));
+	}
+
+	return;
+}
+
+/**
+ * Start monitoring a memory region
+ *
+ * @param[in] cache  a memory registration cache
+ * @param[in] entry  a memory registration entry
+ *
+ * @return           return code from _gnix_notifier_monitor
+ */
+static int
+__notifier_monitor(gnix_mr_cache_t *cache,
+		   gnix_mr_cache_entry_t *entry)
+{
+
+	if (!cache->attr.lazy_deregistration) {
+		return FI_SUCCESS;
+	}
+
+	if (cache->attr.notifier == NULL) {
+		return FI_SUCCESS;
+	}
+
+	GNIX_DEBUG(FI_LOG_MR, "monitoring entry=%p %llx:%llx\n", entry,
+		   entry->key.address, entry->key.length);
+
+	return  _gnix_notifier_monitor(cache->attr.notifier,
+					      (void *) entry->key.address,
+					      entry->key.length,
+					      (uint64_t) entry);
+}
+
+/**
+ * Stop monitoring a memory region
+ *
+ * @param[in] cache  a memory registration cache
+ * @param[in] entry  a memory registration entry
+ *
+ * @return           nothing
+ */
+static void
+__notifier_unmonitor(gnix_mr_cache_t *cache,
+		     gnix_mr_cache_entry_t *entry)
+{
+	int rc;
+
+	if (!cache->attr.lazy_deregistration) {
+		return;
+	}
+
+	if (cache->attr.notifier == NULL) {
+		return;
+	}
+
+	__clear_notifier_events(cache);
+
+	if (!__entry_is_unmapped(entry)) {
+		GNIX_DEBUG(FI_LOG_MR, "unmonitoring entry=%p (state=%d)\n",
+			   entry, entry->state);
+		rc = _gnix_notifier_unmonitor(cache->attr.notifier,
+					      (uint64_t) entry);
+		if (rc != FI_SUCCESS) {
+			/* This probably is okay (ESRCH), because the
+			 * memory could have been unmapped in the
+			 * interim, so clear the notifier events
+			 * again */
+			GNIX_DEBUG(FI_LOG_MR,
+				   "failed to unmonitor memory (entry=%p),"
+				   " so clear notifier events again\n",
+				   entry, fi_strerror(-rc));
+
+			__clear_notifier_events(cache);
+		}
+	}
+}
+
 /**
  * Destroys the memory registration cache entry and deregisters the memory
  *   region with uGNI
@@ -352,15 +669,21 @@ static inline int __mr_cache_entry_destroy(gnix_mr_cache_t *cache,
 	rc = cache->attr.dereg_callback(entry->mr,
 			cache->attr.destruct_context);
 	if (!rc) {
-		entry->state = GNIX_CES_DEAD;
+		/* Should we bother with this check?  If we don't, the
+		 * only difference it that __clear_notifier_events
+		 * will be called one additional time. */
+		if (!__entry_is_unmapped(entry)) {
+			__notifier_unmonitor(cache, entry);
+		}
+		__entry_reset_state(entry);
 
 		rc = cache->attr.destruct_callback(cache->attr.dereg_context);
 		if (!rc)
 			free(entry);
 	} else {
-		GNIX_WARN(FI_LOG_MR, "failed to deregister memory"
-				" region with callback, "
-				"cache_entry=%p ret=%i\n", entry, rc);
+		GNIX_INFO(FI_LOG_MR, "failed to deregister memory"
+			  " region with callback, "
+			  "cache_entry=%p ret=%i\n", entry, rc);
 	}
 
 	return rc;
@@ -373,6 +696,14 @@ static inline int __insert_entry_into_stale(
 	RbtStatus rc;
 	int ret = 0;
 
+	if (__entry_is_unmapped(entry)) {
+		GNIX_DEBUG(FI_LOG_MR, "entry (%p) unmapped, not inserting"
+			   " into stale %llx:%llx", entry,
+			   entry->key.address, entry->key.length);
+		/* Should we return some other value? */
+		return ret;
+	}
+
 	rc = rbtInsert(cache->stale.rb_tree,
 			&entry->key,
 			entry);
@@ -387,14 +718,23 @@ static inline int __insert_entry_into_stale(
 
 		ret = __mr_cache_entry_destroy(cache, entry);
 	} else {
-		GNIX_INFO(FI_LOG_MR,
-				"inserted key=%llx:%llx into stale\n",
-				entry->key.address,
-				entry->key.length);
+		GNIX_DEBUG(FI_LOG_MR,
+			   "inserted key=%llx:%llx into stale\n",
+			   entry->key.address, entry->key.length);
 
 		__mr_cache_lru_enqueue(cache, entry);
-		atomic_inc(&cache->stale.elements);
-		entry->state = GNIX_CES_STALE;
+		ofi_atomic_inc32(&cache->stale.elements);
+		switch (__entry_get_state(entry)) {
+		case  GNIX_CES_INUSE:
+			__entry_set_state(entry, GNIX_CES_STALE);
+			break;
+		default:
+			GNIX_FATAL(FI_LOG_MR,
+				   "stale entry (%p) %llx:%llx in bad"
+				   " state (%d)\n", entry,
+				   entry->key.address, entry->key.length,
+				   entry->state);
+		}
 	}
 
 	return ret;
@@ -413,7 +753,11 @@ static inline void __resolve_stale_entry_collision(
 	RbtIterator iter = found;
 	int add_new_entry = 1, cmp;
 
-	GNIX_INFO(FI_LOG_MR, "resolving collisions\n");
+	GNIX_TRACE(FI_LOG_MR, "\n");
+
+	GNIX_DEBUG(FI_LOG_MR,
+		   "resolving collisions with entry (%p) %llx:%llx\n",
+		   entry, entry->key.address, entry->key.length);
 
 	while (iter) {
 		rbtKeyValue(cache->stale.rb_tree, iter, (void **) &c_key,
@@ -425,9 +769,10 @@ static inline void __resolve_stale_entry_collision(
 
 		if (__can_subsume(&entry->key, c_key) ||
 				(entry->key.length > c_key->length)) {
-			GNIX_INFO(FI_LOG_MR,
-					"adding stale entry to destroy list, key=%llx:%llx",
-					c_key->address, c_key->length);
+			GNIX_DEBUG(FI_LOG_MR,
+				   "adding stale entry (%p) to destroy list,"
+				   "  key=%llx:%llx\n", c_entry,
+				   c_key->address, c_key->length);
 			dlist_insert_tail(&c_entry->siblings, &to_destroy);
 		} else {
 			add_new_entry = 0;
@@ -441,34 +786,50 @@ static inline void __resolve_stale_entry_collision(
 	 */
 	dlist_for_each_safe(&to_destroy, c_entry, tmp, siblings)
 	{
-		GNIX_INFO(FI_LOG_MR, "removing key from tree, key=%llx:%ll\n",
-				c_entry->key.address, c_entry->key.length);
+		GNIX_DEBUG(FI_LOG_MR, "removing key from tree, entry %p"
+			   " key=%llx:%llx\n", c_entry,
+			   c_entry->key.address, c_entry->key.length);
 		iter = rbtFind(cache->stale.rb_tree, &c_entry->key);
-		assert(iter);
+		if (unlikely(!iter)) {
+			GNIX_FATAL(FI_LOG_MR, "key not found\n");
+		}
 
 		rc = rbtErase(cache->stale.rb_tree,
 						iter);
-		assert(rc == RBT_STATUS_OK);
+		if (unlikely(rc != RBT_STATUS_OK)) {
+			GNIX_FATAL(FI_LOG_MR,
+				   "could not remove entry from tree\n");
+		}
 
-		dlist_remove(&c_entry->lru_entry);
+		if (__mr_cache_lru_remove(cache, c_entry) != FI_SUCCESS) {
+			GNIX_WARN(FI_LOG_MR, "Failed to remove entry"
+				  " from lru list (%p)\n",
+				  c_entry);
+		}
+		ofi_atomic_dec32(&cache->stale.elements);
 		dlist_remove(&c_entry->siblings);
-		atomic_dec(&cache->stale.elements);
 		__mr_cache_entry_destroy(cache, c_entry);
 	}
-	assert(dlist_empty(&to_destroy));
+	if (unlikely(!dlist_empty(&to_destroy))) {
+		GNIX_FATAL(FI_LOG_MR, "to_destroy not empty\n");
+	}
 
 	if (add_new_entry) {
 		ret = __insert_entry_into_stale(cache, entry);
-		assert(!ret);
+		if (ret) {
+			GNIX_FATAL(FI_LOG_MR,
+				   "Failed to insert subsumed MR "
+				   " entry (%p) into stale list\n",
+				   entry);
+		}
 	} else {
 		/* stale entry is larger than this one
 		 * so lets just toss this entry out
 		 */
-		GNIX_INFO(FI_LOG_MR,
-				"larger entry already exists, "
-				"to_destroy=%llx:%llx\n",
-				entry->key.address,
-				entry->key.length);
+		GNIX_DEBUG(FI_LOG_MR,
+			   "larger entry already exists, "
+			   "to_destroy=%llx:%llx\n",
+			   entry->key.address, entry->key.length);
 
 		ret = __mr_cache_entry_destroy(cache, entry);
 		if (ret) {
@@ -492,7 +853,11 @@ static inline int __mr_cache_entry_get(
 		gnix_mr_cache_t       *cache,
 		gnix_mr_cache_entry_t *entry)
 {
-	return atomic_inc(&entry->ref_cnt);
+	GNIX_TRACE(FI_LOG_MR, "\n");
+
+	GNIX_DEBUG(FI_LOG_MR,
+		   "Up ref cnt on entry %p\n", entry);
+	return ofi_atomic_inc32(&entry->ref_cnt);
 }
 
 /**
@@ -508,12 +873,21 @@ static inline int __mr_cache_entry_put(
 		gnix_mr_cache_entry_t *entry)
 {
 	RbtIterator iter;
+	int rc;
 	gni_return_t grc = GNI_RC_SUCCESS;
 	RbtIterator found;
 	gnix_mr_cache_entry_t *parent = NULL;
 	struct dlist_entry *next;
 
-	if (atomic_dec(&entry->ref_cnt) == 0) {
+	GNIX_TRACE(FI_LOG_MR, "\n");
+
+	if (cache->attr.lazy_deregistration) {
+		__clear_notifier_events(cache);
+	}
+
+	GNIX_DEBUG(FI_LOG_MR,
+		   "Decrease ref cnt on entry %p\n", entry);
+	if (ofi_atomic_dec32(&entry->ref_cnt) == 0) {
 		next = entry->siblings.next;
 		dlist_remove(&entry->children);
 		dlist_remove(&entry->siblings);
@@ -524,26 +898,30 @@ static inline int __mr_cache_entry_put(
 		if (next != &entry->siblings && dlist_empty(next)) {
 			parent = container_of(next, gnix_mr_cache_entry_t,
 					children);
-
 			grc = __mr_cache_entry_put(cache, parent);
 			if (unlikely(grc != GNI_RC_SUCCESS)) {
 				GNIX_ERR(FI_LOG_MR,
 						"failed to release reference to parent, "
 						"parent=%p refs=%d\n",
 						parent,
-						atomic_get(&parent->ref_cnt));
+						ofi_atomic_get32(&parent->ref_cnt));
 			}
 		}
 
-		atomic_dec(&cache->inuse.elements);
+		ofi_atomic_dec32(&cache->inuse.elements);
 
-		if (!(entry->flags & GNIX_CE_RETIRED)) {
+		if (!__entry_is_retired(entry)) {
 			iter = rbtFind(cache->inuse.rb_tree, &entry->key);
 			if (unlikely(!iter)) {
 				GNIX_ERR(FI_LOG_MR,
 						"failed to find entry in the inuse cache\n");
 			} else {
-				rbtErase(cache->inuse.rb_tree, iter);
+				rc = rbtErase(cache->inuse.rb_tree, iter);
+				if (unlikely(rc != RBT_STATUS_OK)) {
+					GNIX_ERR(FI_LOG_MR,
+						 "failed to erase lru entry"
+						 " from stale tree\n");
+				}
 			}
 		}
 
@@ -551,9 +929,10 @@ static inline int __mr_cache_entry_put(
 		 * isn't retired, put it in the stale cache
 		 */
 		if (cache->attr.lazy_deregistration &&
-				!(entry->flags & GNIX_CE_RETIRED)) {
-			GNIX_INFO(FI_LOG_MR, "moving key %llx:%llx to stale\n",
-					entry->key.address, entry->key.length);
+		    !(__entry_is_retired(entry))) {
+			GNIX_DEBUG(FI_LOG_MR,
+				   "moving key %llx:%llx to stale\n",
+				   entry->key.address, entry->key.length);
 
 			found = rbtFindLeftmost(cache->stale.rb_tree,
 					&entry->key, __find_overlapping_addr);
@@ -569,18 +948,16 @@ static inline int __mr_cache_entry_put(
 			}
 		} else {
 			/* if retired or not using lazy registration */
-			GNIX_INFO(FI_LOG_MR,
-					"destroying entry, key=%llx:%llx\n",
-					entry->key.address,
-					entry->key.length);
+			GNIX_DEBUG(FI_LOG_MR,
+				   "destroying entry, key=%llx:%llx\n",
+				   entry->key.address, entry->key.length);
 
 			grc = __mr_cache_entry_destroy(cache, entry);
 		}
 
 		if (unlikely(grc != GNI_RC_SUCCESS)) {
-			GNIX_WARN(FI_LOG_MR,
-					"dereg callback returned '%s'\n",
-					gni_err_str[grc]);
+			GNIX_INFO(FI_LOG_MR, "dereg callback returned '%s'\n",
+				  gni_err_str[grc]);
 		}
 	}
 
@@ -612,8 +989,8 @@ static inline int __check_mr_cache_attr_sanity(gnix_mr_cache_attr_t *attr)
 }
 
 int _gnix_mr_cache_init(
-		gnix_mr_cache_t      **cache,
-		gnix_mr_cache_attr_t *attr)
+		gnix_mr_cache_t         **cache,
+		gnix_mr_cache_attr_t    *attr)
 {
 	gnix_mr_cache_attr_t *cache_attr = &__default_mr_cache_attr;
 	gnix_mr_cache_t *cache_p;
@@ -621,9 +998,6 @@ int _gnix_mr_cache_init(
 
 	GNIX_TRACE(FI_LOG_MR, "\n");
 
-	if (!cache)
-		return -FI_EINVAL;
-
 	/* if the provider asks us to use their attributes, are they sane? */
 	if (attr) {
 		if (__check_mr_cache_attr_sanity(attr) != FI_SUCCESS)
@@ -664,14 +1038,15 @@ int _gnix_mr_cache_init(
 	 *   destroy will have already set the element counts
 	 */
 	if (cache_p->state == GNIX_MRC_STATE_UNINITIALIZED) {
-		atomic_initialize(&cache_p->inuse.elements, 0);
-		atomic_initialize(&cache_p->stale.elements, 0);
+		ofi_atomic_initialize32(&cache_p->inuse.elements, 0);
+		ofi_atomic_initialize32(&cache_p->stale.elements, 0);
 	}
 
 	cache_p->hits = 0;
 	cache_p->misses = 0;
 
 	cache_p->state = GNIX_MRC_STATE_READY;
+
 	*cache = cache_p;
 
 	return FI_SUCCESS;
@@ -702,7 +1077,7 @@ int _gnix_mr_cache_destroy(gnix_mr_cache_t *cache)
 	 *   then someone forgot to deregister memory.
 	 *   We probably shouldn't destroy the cache at this point.
 	 */
-	if (atomic_get(&cache->inuse.elements) != 0) {
+	if (ofi_atomic_get32(&cache->inuse.elements) != 0) {
 		return -FI_EAGAIN;
 	}
 
@@ -725,13 +1100,12 @@ int _gnix_mr_cache_destroy(gnix_mr_cache_t *cache)
 int __mr_cache_flush(gnix_mr_cache_t *cache, int flush_count) {
 	int rc;
 	RbtIterator iter;
-	gnix_mr_cache_key_t *e_key;
-	gnix_mr_cache_entry_t *entry, *e_entry;
+	gnix_mr_cache_entry_t *entry;
 	int destroyed = 0;
 
 	GNIX_TRACE(FI_LOG_MR, "\n");
 
-	GNIX_INFO(FI_LOG_MR, "starting flush on memory registration cache\n");
+	GNIX_DEBUG(FI_LOG_MR, "starting flush on memory registration cache\n");
 
 	/* flushes are unnecessary for caches without lazy deregistration */
 	if (!cache->attr.lazy_deregistration)
@@ -749,30 +1123,18 @@ int __mr_cache_flush(gnix_mr_cache_t *cache, int flush_count) {
 			break;
 		}
 
-		GNIX_INFO(FI_LOG_MR, "attempting to flush key %llx:%llx\n",
-				entry->key.address, entry->key.length);
+		GNIX_DEBUG(FI_LOG_MR, "attempting to flush key %llx:%llx\n",
+			   entry->key.address, entry->key.length);
 		iter = rbtFind(cache->stale.rb_tree, &entry->key);
 		if (unlikely(!iter)) {
 			GNIX_ERR(FI_LOG_MR,
-					"lru entries MUST be present in the cache,"
-					" could not find key in stale tree\n");
+				 "lru entries MUST be present in the cache,"
+				 " could not find entry (%p) in stale tree"
+				 " %llx:%llx\n",
+				 entry, entry->key.address, entry->key.length);
 			break;
 		}
 
-		rbtKeyValue(cache->stale.rb_tree, iter, (void **) &e_key,
-			    (void **) &e_entry);
-		if (e_entry != entry) {
-			/* If not an exact match, remove the found entry,
-			 and then put the original entry back on the LRU list */
-			GNIX_INFO(FI_LOG_MR,
-				  "Flushing non-lru entry %llx:%llx\n",
-				  e_entry->key.address, e_entry->key.length);
-			dlist_remove(&e_entry->lru_entry);
-			dlist_insert_tail(&entry->lru_entry, &cache->lru_head);
-			/* Destroy the actual entry below */
-			entry = e_entry;
-		}
-
 		rc = rbtErase(cache->stale.rb_tree, iter);
 		if (unlikely(rc != RBT_STATUS_OK)) {
 			GNIX_ERR(FI_LOG_MR,
@@ -785,12 +1147,12 @@ int __mr_cache_flush(gnix_mr_cache_t *cache, int flush_count) {
 		++destroyed;
 	}
 
-	GNIX_INFO(FI_LOG_MR, "flushed %i of %i entries from memory "
-				"registration cache\n", destroyed,
-				atomic_get(&cache->stale.elements));
+	GNIX_DEBUG(FI_LOG_MR, "flushed %i of %i entries from memory "
+		   "registration cache\n", destroyed,
+		   ofi_atomic_get32(&cache->stale.elements));
 
 	if (destroyed > 0) {
-		atomic_sub(&cache->stale.elements, destroyed);
+		ofi_atomic_sub32(&cache->stale.elements, destroyed);
 	}
 
 	return FI_SUCCESS;
@@ -822,6 +1184,10 @@ static int __mr_cache_search_inuse(
 	uint64_t new_end, found_end;
 	DLIST_HEAD(retired_entries);
 
+	if (cache->attr.lazy_deregistration) {
+		__clear_notifier_events(cache);
+	}
+
 	/* first we need to find an entry that overlaps with this one.
 	 * care should be taken to find the left most entry that overlaps
 	 * with this entry since the entry we are searching for might overlap
@@ -831,31 +1197,31 @@ static int __mr_cache_search_inuse(
 	iter = rbtFindLeftmost(cache->inuse.rb_tree, (void *) key,
 			__find_overlapping_addr);
 	if (!iter) {
-		GNIX_INFO(FI_LOG_MR,
-				"could not find key in inuse, key=%llx:%llx\n",
-				key->address, key->length);
+		GNIX_DEBUG(FI_LOG_MR,
+			   "could not find key in inuse, key=%llx:%llx\n",
+			   key->address, key->length);
 		return -FI_ENOENT;
 	}
 
 	rbtKeyValue(cache->inuse.rb_tree, iter, (void **) &found_key,
 			(void **) &found_entry);
 
-	GNIX_INFO(FI_LOG_MR,
-			"found a key that matches the search criteria, "
-			"found=%llx:%llx key=%llx:%llx\n",
-			found_key->address, found_key->length,
-			key->address, key->length);
+	GNIX_DEBUG(FI_LOG_MR,
+		   "found a key that matches the search criteria, "
+		   "found=%llx:%llx key=%llx:%llx\n",
+		   found_key->address, found_key->length,
+		   key->address, key->length);
 
 	/* if the entry that we've found completely subsumes
 	 * the requested entry, just return a reference to
 	 * that existing registration
 	 */
 	if (__can_subsume(found_key, key)) {
-		GNIX_INFO(FI_LOG_MR,
-				"found an entry that subsumes the request, "
-				"existing=%llx:%llx key=%llx:%llx\n",
-				found_key->address, found_key->length,
-				key->address, key->length);
+		GNIX_DEBUG(FI_LOG_MR,
+			   "found an entry that subsumes the request, "
+			   "existing=%llx:%llx key=%llx:%llx entry %p\n",
+			   found_key->address, found_key->length,
+			   key->address, key->length, found_entry);
 		*entry = found_entry;
 		__mr_cache_entry_get(cache, found_entry);
 
@@ -875,10 +1241,10 @@ static int __mr_cache_search_inuse(
 
 
 		cmp = __find_overlapping_addr(found_key, key);
-		GNIX_INFO(FI_LOG_MR,
-				"candidate: key=%llx:%llx result=%d\n",
-				found_key->address,
-				found_key->length, cmp);
+		GNIX_DEBUG(FI_LOG_MR,
+			   "candidate: key=%llx:%llx result=%d\n",
+			   found_key->address,
+			   found_key->length, cmp);
 		if (cmp != 0)
 			break;
 
@@ -886,9 +1252,9 @@ static int __mr_cache_search_inuse(
 		found_end = found_key->address + found_key->length;
 
 		/* mark the entry as retired */
-		GNIX_INFO(FI_LOG_MR, "retiring entry, key=%llx:%llx\n",
-				found_key->address, found_key->length);
-		found_entry->flags |= GNIX_CE_RETIRED;
+		GNIX_DEBUG(FI_LOG_MR, "retiring entry, key=%llx:%llx entry %p\n",
+			   found_key->address, found_key->length, found_entry);
+		__entry_set_retired(found_entry);
 		dlist_insert_tail(&found_entry->siblings, &retired_entries);
 
 		iter = rbtNext(cache->inuse.rb_tree, iter);
@@ -900,23 +1266,38 @@ static int __mr_cache_search_inuse(
 
 
 	/* remove retired entries from tree */
-	GNIX_INFO(FI_LOG_MR, "removing retired entries from inuse tree\n");
+	GNIX_DEBUG(FI_LOG_MR, "removing retired entries from inuse tree\n");
 	__remove_sibling_entries_from_tree(cache,
 			&retired_entries, cache->inuse.rb_tree);
 
 	/* create new registration */
-	GNIX_INFO(FI_LOG_MR,
-			"creating a new merged registration, key=%llx:%llx\n",
-			new_key.address, new_key.length);
+	GNIX_DEBUG(FI_LOG_MR,
+		   "creating a new merged registration, key=%llx:%llx\n",
+		   new_key.address, new_key.length);
 	ret = __mr_cache_create_registration(cache,
 			new_key.address, new_key.length,
 			entry, &new_key, fi_reg_context);
 	if (ret) {
-		GNIX_ERR(FI_LOG_MR, "failure in registration, can't really "
-				"recover at this point since we've already retired "
-				"the existing entries\n");
+		/* If we get here, one of two things have happened.
+		 * Either some part of the new merged registration was
+		 * unmapped (i.e., freed by user) or the merged
+		 * registration failed for some other reason (probably
+		 * GNI_RC_ERROR_RESOURCE).  The first case is a user
+		 * error (which they should have been warned about by
+		 * the notifier), and the second case is always
+		 * possible.  Neither case is a problem.  The entries
+		 * above have been retired, and here we return the
+		 * error */
+		GNIX_DEBUG(FI_LOG_MR,
+			   "failed to create merged registration, key=",
+			   new_key.address, new_key.length);
 		return ret;
 	}
+	GNIX_DEBUG(FI_LOG_MR,
+		   "created a new merged registration, key=%llx:%llx entry %p\n",
+		   new_key.address, new_key.length, *entry);
+
+	__entry_set_merged(*entry);
 
 	/* move retired entries to the head of the new entry's child list */
 	if (!dlist_empty(&retired_entries)) {
@@ -943,8 +1324,12 @@ static int __mr_cache_search_stale(
 	gnix_mr_cache_key_t *mr_key;
 	gnix_mr_cache_entry_t *mr_entry, *tmp;
 
-	GNIX_INFO(FI_LOG_MR, "searching for stale entry, key=%llx:%llx\n",
-			key->address, key->length);
+	if (cache->attr.lazy_deregistration) {
+		__clear_notifier_events(cache);
+	}
+
+	GNIX_DEBUG(FI_LOG_MR, "searching for stale entry, key=%llx:%llx\n",
+		   key->address, key->length);
 
 	iter = rbtFindLeftmost(cache->stale.rb_tree, (void *) key,
 			__find_overlapping_addr);
@@ -954,10 +1339,10 @@ static int __mr_cache_search_stale(
 	rbtKeyValue(cache->stale.rb_tree, iter, (void **) &mr_key,
 			(void **) &mr_entry);
 
-	GNIX_INFO(FI_LOG_MR,
-			"found a matching entry, found=%llx:%llx key=%llx:%llx\n",
-			mr_key->address, mr_key->length,
-			key->address, key->length);
+	GNIX_DEBUG(FI_LOG_MR,
+		   "found a matching entry, found=%llx:%llx key=%llx:%llx\n",
+		   mr_key->address, mr_key->length,
+		   key->address, key->length);
 
 
 	/* if the entry that we've found completely subsumes
@@ -976,37 +1361,47 @@ static int __mr_cache_search_stale(
 			 * The old entry (mr_entry) should be destroyed
 			 * now as it is no longer needed.
 			 */
-			GNIX_INFO(FI_LOG_MR,
-					"removing entry from stale key=%llx:%llx\n",
-					mr_key->address, mr_key->length);
+			GNIX_DEBUG(FI_LOG_MR,
+				   "removing entry from stale key=%llx:%llx\n",
+				   mr_key->address, mr_key->length);
 
 			rc = rbtErase(cache->stale.rb_tree, iter);
 			if (unlikely(rc != RBT_STATUS_OK)) {
 				GNIX_ERR(FI_LOG_MR,
 						"failed to erase entry from stale tree\n");
 			} else {
-				dlist_remove(&mr_entry->lru_entry);
-
-				atomic_dec(&cache->stale.elements);
-
+				if (__mr_cache_lru_remove(cache, mr_entry)
+				    == FI_SUCCESS) {
+					ofi_atomic_dec32(&cache->stale.elements);
+				} else {
+					GNIX_WARN(FI_LOG_MR, "Failed to remove"
+						  " entry (%p) from lru list\n",
+						  mr_entry);
+				}
 				__mr_cache_entry_destroy(cache, mr_entry);
 			}
 
 			*entry = tmp;
 		} else {
-			GNIX_INFO(FI_LOG_MR,
-					"removing entry from stale and migrating to inuse, "
-					"key=%llx:%llx\n",
-					mr_key->address, mr_key->length);
+			GNIX_DEBUG(FI_LOG_MR,
+				   "removing entry (%p) from stale and"
+				   " migrating to inuse, key=%llx:%llx\n",
+				   mr_entry, mr_key->address, mr_key->length);
 			rc = rbtErase(cache->stale.rb_tree, iter);
-			if (unlikely(rc != RBT_STATUS_OK))
-				GNIX_WARN(FI_LOG_MR,
-						"failed to erase entry from stale tree\n");
-			assert(rc == RBT_STATUS_OK);
+			if (unlikely(rc != RBT_STATUS_OK)) {
+				GNIX_FATAL(FI_LOG_MR,
+					   "failed to erase entry (%p) from "
+					   " stale tree\n", mr_entry);
+			}
 
-			dlist_remove(&mr_entry->lru_entry);
+			if (__mr_cache_lru_remove(cache, mr_entry)
+			    != FI_SUCCESS) {
+				GNIX_WARN(FI_LOG_MR, "Failed to remove"
+					  " entry (%p) from lru list\n",
+					  mr_entry);
+			}
 
-			atomic_dec(&cache->stale.elements);
+			ofi_atomic_dec32(&cache->stale.elements);
 
 			/* if we made it to this point, there weren't
 			 * any entries in the inuse tree that would
@@ -1014,13 +1409,14 @@ static int __mr_cache_search_stale(
 			 */
 			rc = rbtInsert(cache->inuse.rb_tree,
 					&mr_entry->key, mr_entry);
-			if (unlikely(rc != RBT_STATUS_OK))
-				GNIX_WARN(FI_LOG_MR,
-						"failed to insert entry into inuse tree\n");
-			assert(rc == RBT_STATUS_OK);
+			if (unlikely(rc != RBT_STATUS_OK)) {
+				GNIX_FATAL(FI_LOG_MR,
+					   "failed to insert entry into"
+					   "inuse tree\n");
+			}
 
-			atomic_set(&mr_entry->ref_cnt, 1);
-			atomic_inc(&cache->inuse.elements);
+			ofi_atomic_set32(&mr_entry->ref_cnt, 1);
+			ofi_atomic_inc32(&cache->inuse.elements);
 
 			*entry = mr_entry;
 		}
@@ -1028,10 +1424,10 @@ static int __mr_cache_search_stale(
 		return FI_SUCCESS;
 	}
 
-	GNIX_INFO(FI_LOG_MR,
-			"could not use matching entry, "
-			"found=%llx:%llx\n",
-			mr_key->address, mr_key->length);
+	GNIX_DEBUG(FI_LOG_MR,
+		   "could not use matching entry, "
+		   "found=%llx:%llx\n",
+		   mr_key->address, mr_key->length);
 
 	return -FI_ENOENT;
 }
@@ -1062,15 +1458,23 @@ static int __mr_cache_create_registration(
 	handle = cache->attr.reg_callback(handle, (void *) address, length,
 			fi_reg_context, cache->attr.reg_context);
 	if (unlikely(!handle)) {
-		free(current_entry);
-		GNIX_WARN(FI_LOG_MR, "failed to register memory with callback");
-		return -FI_ENOMEM;
+		GNIX_INFO(FI_LOG_MR,
+			  "failed to register memory with callback\n");
+		goto err;
 	}
 
+	__entry_reset_state(current_entry);
+
 	/* set up the entry's key */
 	current_entry->key.address = address;
 	current_entry->key.length = length;
-	current_entry->flags = 0;
+
+	rc = __notifier_monitor(cache, current_entry);
+	if (unlikely(rc != FI_SUCCESS)) {
+		GNIX_INFO(FI_LOG_MR,
+			  "failed to monitor memory with notifier\n");
+		goto err_dereg;
+	}
 
 	rc = rbtInsert(cache->inuse.rb_tree, &current_entry->key,
 			current_entry);
@@ -1078,30 +1482,31 @@ static int __mr_cache_create_registration(
 		GNIX_ERR(FI_LOG_MR, "failed to insert registration "
 				"into cache, ret=%i\n", rc);
 
-		rc = cache->attr.dereg_callback(current_entry->mr,
-				cache->attr.dereg_context);
-		if (unlikely(rc)) {
-			GNIX_WARN(FI_LOG_MR,
-					"failed to deregister memory with "
-					"callback, ret=%d\n",
-					rc);
-		}
-
-		free(current_entry);
-
-		return -FI_ENOMEM;
+		goto err_dereg;
 	}
 
-	GNIX_INFO(FI_LOG_MR, "inserted key %llx:%llx into inuse\n",
-			current_entry->key.address, current_entry->key.length);
+	GNIX_DEBUG(FI_LOG_MR, "inserted key %llx:%llx into inuse %p\n",
+		   current_entry->key.address, current_entry->key.length, current_entry);
 
 
-	atomic_inc(&cache->inuse.elements);
-	atomic_initialize(&current_entry->ref_cnt, 1);
+	ofi_atomic_inc32(&cache->inuse.elements);
+	ofi_atomic_initialize32(&current_entry->ref_cnt, 1);
 
 	*entry = current_entry;
 
 	return FI_SUCCESS;
+
+err_dereg:
+	rc = cache->attr.dereg_callback(current_entry->mr,
+					cache->attr.dereg_context);
+	if (unlikely(rc)) {
+		GNIX_INFO(FI_LOG_MR,
+			  "failed to deregister memory with "
+			  "callback, ret=%d\n", rc);
+	}
+err:
+	free(current_entry);
+	return -FI_ENOMEM;
 }
 
 /**
@@ -1140,11 +1545,15 @@ int _gnix_mr_cache_register(
 
 	/* if we shouldn't introduce any new elements, return -FI_ENOSPC */
 	if (unlikely(cache->attr.hard_reg_limit > 0 &&
-			(atomic_get(&cache->inuse.elements) >=
-					cache->attr.hard_reg_limit)))
-		return -FI_ENOSPC;
+			(ofi_atomic_get32(&cache->inuse.elements) >=
+			 cache->attr.hard_reg_limit))) {
+		ret = -FI_ENOSPC;
+		goto err;
+	}
 
 	if (cache->attr.lazy_deregistration) {
+		__clear_notifier_events(cache);
+
 		/* if lazy deregistration is in use, we can check the
 		 *   stale tree
 		 */
@@ -1160,28 +1569,31 @@ int _gnix_mr_cache_register(
 	 *   room for the new entry. This works because we check above to see if
 	 *   the number of inuse entries exceeds the hard reg limit
 	 */
-	if ((atomic_get(&cache->inuse.elements) +
-			atomic_get(&cache->stale.elements)) == cache->attr.hard_reg_limit)
+	if ((ofi_atomic_get32(&cache->inuse.elements) +
+			ofi_atomic_get32(&cache->stale.elements)) == cache->attr.hard_reg_limit)
 		__mr_cache_flush(cache, 1);
 
 	ret = __mr_cache_create_registration(cache, address, length,
 			&entry, &key, fi_reg_context);
-	if (ret)
-		return ret;
+	if (ret) {
+		goto err;
+	}
 
 	cache->misses++;
 
 success:
-	entry->state = GNIX_CES_INUSE;
+	__entry_set_state(entry, GNIX_CES_INUSE);
 	*handle = (void *) entry->mr;
 
 	return FI_SUCCESS;
+
+err:
+	return ret;
 }
 
 /**
  * Function to deregister memory in the cache
  *
- * @param[in]  cache  gnix memory registration cache pointer
  * @param[in]  mr     gnix memory registration descriptor pointer
  *
  * @return     FI_SUCCESS on success
@@ -1203,18 +1615,21 @@ int _gnix_mr_cache_deregister(
 	 */
 
 	entry = container_of(handle, gnix_mr_cache_entry_t, mr);
-	if (entry->state != GNIX_CES_INUSE)
+	if (__entry_get_state(entry) != GNIX_CES_INUSE) {
+		GNIX_WARN(FI_LOG_MR, "entry (%p) in incorrect state (%d)\n",
+			  entry, entry->state);
 		return -FI_EINVAL;
+	}
 
-	GNIX_INFO(FI_LOG_MR, "entry found, entry=%p refs=%d\n",
-			entry, atomic_get(&entry->ref_cnt));
+	GNIX_DEBUG(FI_LOG_MR, "entry found, entry=%p refs=%d\n",
+		   entry, ofi_atomic_get32(&entry->ref_cnt));
 
 	grc = __mr_cache_entry_put(cache, entry);
 
 	/* Since we check this on each deregistration, the amount of elements
 	 * over the limit should always be 1
 	 */
-	if (atomic_get(&cache->stale.elements) > cache->attr.hard_stale_limit)
+	if (ofi_atomic_get32(&cache->stale.elements) > cache->attr.hard_stale_limit)
 		__mr_cache_flush(cache, 1);
 
 	return gnixu_to_fi_errno(grc);
