@@ -1,6 +1,7 @@
 /*
- * Copyright (c) 2015-2016 Cray Inc. All rights reserved.
- * Copyright (c) 2015 Los Alamos National Security, LLC. All rights reserved.
+ * Copyright (c) 2015-2017 Cray Inc. All rights reserved.
+ * Copyright (c) 2015-2016 Los Alamos National Security, LLC.
+ *                         All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -43,9 +44,16 @@
 #include "gnix_vc.h"
 #include "gnix_mbox_allocator.h"
 
-static int gnix_nics_per_ptag[GNI_PTAG_USER_END];
-static DLIST_HEAD(gnix_nic_list);
-static pthread_mutex_t gnix_nic_list_lock = PTHREAD_MUTEX_INITIALIZER;
+/*
+ * TODO: make this a domain parameter
+ */
+#define GNIX_VC_FL_MIN_SIZE 128
+#define GNIX_VC_FL_INIT_REFILL_SIZE 10
+
+static int gnix_nics_per_ptag[GNI_PTAG_MAX];
+struct dlist_entry gnix_nic_list_ptag[GNI_PTAG_MAX];
+DLIST_HEAD(gnix_nic_list);
+pthread_mutex_t gnix_nic_list_lock = PTHREAD_MUTEX_INITIALIZER;
 
 /*
  * globals
@@ -142,13 +150,15 @@ try_again:
 		/*
 		 * first dequeue RX CQEs
 		 */
-		if (which == 1) {
+		if (nic->rx_cq_blk != nic->rx_cq && which == 1) {
 			do {
 				status = GNI_CqGetEvent(nic->rx_cq_blk,
 							&cqe);
 			} while (status == GNI_RC_SUCCESS);
 		}
+		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &prev_state);
 		_gnix_nic_progress(nic);
+		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &prev_state);
 		retry = 1;
 		break;
 	case GNI_RC_TIMEOUT:
@@ -251,13 +261,16 @@ static int __nic_teardown_irq_cq(struct gnix_nic *nic)
 	if (nic->irq_mmap_addr == NULL)
 		return ret;
 
-	status = GNI_MemDeregister(nic->gni_nic_hndl,
-				  &nic->irq_mem_hndl);
-	if (status != GNI_RC_SUCCESS) {
-		ret = gnixu_to_fi_errno(status);
-		GNIX_WARN(FI_LOG_EP_CTRL,
-			  "GNI_MemDeregister returned %s\n",
-			  gni_err_str[status]);
+	if ((nic->irq_mem_hndl.qword1) ||
+		(nic->irq_mem_hndl.qword2)) {
+		status = GNI_MemDeregister(nic->gni_nic_hndl,
+					  &nic->irq_mem_hndl);
+		if (status != GNI_RC_SUCCESS) {
+			ret = gnixu_to_fi_errno(status);
+			GNIX_WARN(FI_LOG_EP_CTRL,
+				  "GNI_MemDeregister returned %s\n",
+				  gni_err_str[status]);
+		}
 	}
 
 	munmap(nic->irq_mmap_addr,
@@ -294,12 +307,16 @@ static int __nic_rx_overrun(struct gnix_nic *nic)
 	GNIX_WARN(FI_LOG_EP_DATA, "\n");
 
 	/* clear out the CQ */
+	/*
+	 * TODO:  really need to process CQEs better for error reporting,
+	 * etc.
+	 */
 	while ((status = GNI_CqGetEvent(nic->rx_cq, &cqe)) == GNI_RC_SUCCESS);
 	assert(status == GNI_RC_NOT_DONE);
 
-	fastlock_acquire(&nic->vc_id_lock);
+	COND_ACQUIRE(nic->requires_lock, &nic->vc_id_lock);
 	max_id = nic->vc_id_table_count;
-	fastlock_release(&nic->vc_id_lock);
+	COND_RELEASE(nic->requires_lock, &nic->vc_id_lock);
 	/*
 	 * TODO: optimization would
 	 * be to keep track of last time
@@ -310,12 +327,8 @@ static int __nic_rx_overrun(struct gnix_nic *nic)
 		ret = _gnix_test_bit(&nic->vc_id_bitmap, i);
 		if (ret) {
 			vc = __gnix_nic_elem_by_rem_id(nic, i);
-			ret = _gnix_vc_dequeue_smsg(vc);
-			if (ret != FI_SUCCESS) {
-				GNIX_WARN(FI_LOG_EP_DATA,
-					  "_gnix_vc_dqueue_smsg returned %d\n",
-					  ret);
-			}
+			ret = _gnix_vc_rx_schedule(vc);
+			assert(ret == FI_SUCCESS);
 		}
 	}
 
@@ -328,30 +341,34 @@ static int __process_rx_cqe(struct gnix_nic *nic, gni_cq_entry_t cqe)
 	struct gnix_vc *vc;
 
 	vc_id =  GNI_CQ_GET_INST_ID(cqe);
-	vc = __gnix_nic_elem_by_rem_id(nic, vc_id);
 
-#if 1 /* Process RX inline with arrival of an RX CQE. */
-	if (unlikely(vc->conn_state != GNIX_VC_CONNECTED)) {
-		GNIX_INFO(FI_LOG_EP_DATA,
-			  "Scheduling VC for RX processing (%p)\n",
-			  vc);
-		ret = _gnix_vc_rx_schedule(vc);
-		assert(ret == FI_SUCCESS);
-	} else {
-		GNIX_INFO(FI_LOG_EP_DATA,
-			  "Processing VC RX (%p)\n",
-			  vc);
-		ret = _gnix_vc_dequeue_smsg(vc);
-		if (ret != FI_SUCCESS) {
-			GNIX_WARN(FI_LOG_EP_DATA,
-					"_gnix_vc_dqueue_smsg returned %d\n",
-					ret);
+	/*
+	 * its possible this vc has been destroyed, so may get NULL
+	 * back.
+	 */
+
+	vc = __gnix_nic_elem_by_rem_id(nic, vc_id);
+	if (vc != NULL) {
+		switch (vc->conn_state) {
+		case GNIX_VC_CONNECTING:
+			GNIX_DEBUG(FI_LOG_EP_DATA,
+				  "Scheduling VC for RX processing (%p)\n",
+				  vc);
+			ret = _gnix_vc_rx_schedule(vc);
+			assert(ret == FI_SUCCESS);
+			break;
+		case GNIX_VC_CONNECTED:
+			GNIX_DEBUG(FI_LOG_EP_DATA,
+				  "Processing VC RX (%p)\n",
+				  vc);
+			ret = _gnix_vc_rx_schedule(vc);
+			assert(ret == FI_SUCCESS);
+			break;
+		default:
+			break;  /* VC not in a state for scheduling or
+				   SMSG processing */
 		}
 	}
-#else /* Defer RX processing until after the RX CQ is cleared. */
-	ret = _gnix_vc_rx_schedule(vc);
-	assert(ret == FI_SUCCESS);
-#endif
 
 	return ret;
 }
@@ -366,7 +383,7 @@ static int __nic_rx_progress(struct gnix_nic *nic)
 	if (status == GNI_RC_NOT_DONE)
 		return FI_SUCCESS;
 
-	fastlock_acquire(&nic->lock);
+	COND_ACQUIRE(nic->requires_lock, &nic->lock);
 
 	do {
 		status = GNI_CqGetEvent(nic->rx_cq, &cqe);
@@ -398,7 +415,7 @@ static int __nic_rx_progress(struct gnix_nic *nic)
 		}
 	} while (1);
 
-	fastlock_release(&nic->lock);
+	COND_RELEASE(nic->requires_lock, &nic->lock);
 
 	return ret;
 }
@@ -434,10 +451,11 @@ static void __nic_get_completed_txd(struct gnix_nic *nic,
 {
 	gni_post_descriptor_t *gni_desc;
 	struct gnix_tx_descriptor *txd_p = NULL;
+	struct gnix_fab_req *req;
 	gni_return_t status;
 	int msg_id;
 	gni_cq_entry_t cqe;
-	uint32_t recov;
+	uint32_t recov = 1;
 
 	if (__gnix_nic_txd_err_get(nic, &txd_p)) {
 		*txd = txd_p;
@@ -455,13 +473,22 @@ static void __nic_get_completed_txd(struct gnix_nic *nic,
 	assert(status == GNI_RC_SUCCESS ||
 	       status == GNI_RC_TRANSACTION_ERROR);
 
-	if (status == GNI_RC_TRANSACTION_ERROR) {
+	if (unlikely(status == GNI_RC_TRANSACTION_ERROR)) {
 		status = GNI_CqErrorRecoverable(cqe, &recov);
-		if (status != GNI_RC_SUCCESS || !recov) {
-			char ebuf[512];
-			GNI_CqErrorStr(cqe, ebuf, sizeof(ebuf));
-			GNIX_FATAL(FI_LOG_EP_DATA, "CQ error status: %s\n",
-				   ebuf);
+		if (status == GNI_RC_SUCCESS) {
+			if (!recov) {
+				char ebuf[512];
+
+				GNI_CqErrorStr(cqe, ebuf, sizeof(ebuf));
+				GNIX_WARN(FI_LOG_EP_DATA,
+					  "CQ error status: %s\n",
+					   ebuf);
+			}
+		} else {
+			GNIX_WARN(FI_LOG_EP_DATA,
+				  "GNI_CqErrorRecover returned: %s\n",
+				   gni_err_str[status]);
+			recov = 0;  /* assume something bad has happened */
 		}
 	}
 
@@ -479,8 +506,18 @@ static void __nic_get_completed_txd(struct gnix_nic *nic,
 		txd_p = __desc_lkup_by_id(nic, msg_id);
 	}
 
-	if (!txd_p) {
+	if (unlikely(txd_p == NULL))
 		GNIX_FATAL(FI_LOG_EP_DATA, "Unexpected CQE: 0x%lx", cqe);
+
+	/*
+	 * set retry count on the request to max to force
+	 * delivering error'd CQ event to application
+	 */
+	if (!recov) {
+		status = GNI_RC_TRANSACTION_ERROR;
+		req = txd_p->req;
+		if (req)
+			req->tx_failures = UINT_MAX;
 	}
 
 	*tx_status = status;
@@ -497,10 +534,10 @@ static int __nic_tx_progress(struct gnix_nic *nic, gni_cq_handle_t cq)
 	do {
 		txd = NULL;
 
-		fastlock_acquire(&nic->lock);
+		COND_ACQUIRE(nic->requires_lock, &nic->lock);
 		__nic_get_completed_txd(nic, cq, &txd,
 					&tx_status);
-		fastlock_release(&nic->lock);
+		COND_RELEASE(nic->requires_lock, &nic->lock);
 
 		if (txd && txd->completer_fn) {
 			ret = txd->completer_fn(txd, tx_status);
@@ -520,26 +557,27 @@ static int __nic_tx_progress(struct gnix_nic *nic, gni_cq_handle_t cq)
 	return ret;
 }
 
-int _gnix_nic_progress(struct gnix_nic *nic)
+int _gnix_nic_progress(void *arg)
 {
+	struct gnix_nic *nic = (struct gnix_nic *)arg;
 	int ret = FI_SUCCESS;
 
 	ret =  __nic_tx_progress(nic, nic->tx_cq);
 	if (unlikely(ret != FI_SUCCESS))
 		return ret;
 
-	if (nic->tx_cq_blk) {
+	if (nic->tx_cq_blk && nic->tx_cq_blk != nic->tx_cq) {
 		ret =  __nic_tx_progress(nic, nic->tx_cq_blk);
 		if (unlikely(ret != FI_SUCCESS))
 			return ret;
 	}
 
 	ret = __nic_rx_progress(nic);
-	if (unlikely(ret != FI_SUCCESS))
+	if (ret != FI_SUCCESS)
 		return ret;
 
 	ret = _gnix_vc_nic_progress(nic);
-	if (unlikely(ret != FI_SUCCESS))
+	if (ret != FI_SUCCESS)
 		return ret;
 
 	return ret;
@@ -576,7 +614,7 @@ int _gnix_nic_get_rem_id(struct gnix_nic *nic, int *remote_id, void *entry)
 	 * bit before resizing the table
 	 */
 
-	fastlock_acquire(&nic->vc_id_lock);
+	COND_ACQUIRE(nic->requires_lock, &nic->vc_id_lock);
 	if (nic->vc_id_table_capacity == nic->vc_id_table_count) {
 		table_base = realloc(nic->vc_id_table,
 				     2 * nic->vc_id_table_capacity *
@@ -587,6 +625,13 @@ int _gnix_nic_get_rem_id(struct gnix_nic *nic, int *remote_id, void *entry)
 		}
 		nic->vc_id_table_capacity *= 2;
 		nic->vc_id_table = table_base;
+
+		ret = _gnix_realloc_bitmap(&nic->vc_id_bitmap,
+					   nic->vc_id_table_capacity);
+		if (ret != FI_SUCCESS) {
+			assert(ret == -FI_ENOMEM);
+			goto err;
+		}
 	}
 
 	nic->vc_id_table[nic->vc_id_table_count] = entry;
@@ -600,50 +645,10 @@ int _gnix_nic_get_rem_id(struct gnix_nic *nic, int *remote_id, void *entry)
 
 	++(nic->vc_id_table_count);
 err:
-	fastlock_release(&nic->vc_id_lock);
+	COND_RELEASE(nic->requires_lock, &nic->vc_id_lock);
 	return ret;
 }
 
-/*
- * allocate a tx desc for this nic
- */
-
-int _gnix_nic_tx_alloc(struct gnix_nic *nic,
-		       struct gnix_tx_descriptor **desc)
-{
-	struct dlist_entry *entry;
-
-	fastlock_acquire(&nic->tx_desc_lock);
-	if (dlist_empty(&nic->tx_desc_free_list)) {
-		fastlock_release(&nic->tx_desc_lock);
-		return -FI_ENOSPC;
-	}
-
-	entry = nic->tx_desc_free_list.next;
-	dlist_remove_init(entry);
-	dlist_insert_head(entry, &nic->tx_desc_active_list);
-	*desc = dlist_entry(entry, struct gnix_tx_descriptor, list);
-	fastlock_release(&nic->tx_desc_lock);
-
-	return FI_SUCCESS;
-}
-
-/*
- * free a tx desc for this nic - the nic is not embedded in the
- * descriptor to help keep it small
- */
-
-int _gnix_nic_tx_free(struct gnix_nic *nic,
-		      struct gnix_tx_descriptor *desc)
-{
-	fastlock_acquire(&nic->tx_desc_lock);
-	dlist_remove_init(&desc->list);
-	dlist_insert_head(&desc->list, &nic->tx_desc_free_list);
-	fastlock_release(&nic->tx_desc_lock);
-
-	return FI_SUCCESS;
-}
-
 /*
  * allocate a free list of tx descs for a gnix_nic struct.
  */
@@ -652,8 +657,6 @@ static int __gnix_nic_tx_freelist_init(struct gnix_nic *nic, int n_descs)
 {
 	int i, ret = FI_SUCCESS;
 	struct gnix_tx_descriptor *desc_base, *desc_ptr;
-	void *int_bufs;
-	gni_return_t status;
 
 	GNIX_TRACE(FI_LOG_EP_CTRL, "\n");
 
@@ -667,28 +670,11 @@ static int __gnix_nic_tx_freelist_init(struct gnix_nic *nic, int n_descs)
 		goto err;
 	}
 
-	int_bufs = calloc(n_descs, GNIX_CACHELINE_SIZE);
-	if (int_bufs == NULL) {
-		ret = -FI_ENOMEM;
-		goto err_buf_alloc;
-	}
-
-	/* We don't have a domain here to use for caching the registration. */
-	status = GNI_MemRegister(nic->gni_nic_hndl, (uint64_t)int_bufs,
-				 n_descs * GNIX_CACHELINE_SIZE, NULL,
-				 GNI_MEM_READWRITE, 0, &nic->int_bufs_mdh);
-	if (status != GNI_RC_SUCCESS) {
-		ret = gnixu_to_fi_errno(status);
-		goto err_buf_reg;
-	}
-	nic->int_bufs = int_bufs;
-
 	dlist_init(&nic->tx_desc_free_list);
 	dlist_init(&nic->tx_desc_active_list);
 
 	for (i = 0, desc_ptr = desc_base; i < n_descs; i++, desc_ptr++) {
 		desc_ptr->id = i;
-		desc_ptr->int_buf = int_bufs + (i * GNIX_CACHELINE_SIZE);
 		dlist_insert_tail(&desc_ptr->list,
 				  &nic->tx_desc_free_list);
 	}
@@ -700,10 +686,6 @@ static int __gnix_nic_tx_freelist_init(struct gnix_nic *nic, int n_descs)
 
 	return ret;
 
-err_buf_reg:
-	free(int_bufs);
-err_buf_alloc:
-	free(desc_base);
 err:
 	return ret;
 
@@ -714,16 +696,8 @@ err:
  */
 static void __gnix_nic_tx_freelist_destroy(struct gnix_nic *nic)
 {
-	gni_return_t status;
-
 	GNIX_TRACE(FI_LOG_EP_CTRL, "\n");
 
-	status = GNI_MemDeregister(nic->gni_nic_hndl, &nic->int_bufs_mdh);
-	if (status != GNI_RC_SUCCESS)
-		GNIX_WARN(FI_LOG_DOMAIN, "GNI_MemDeregister() failed: %s\n",
-			  gni_err_str[status]);
-	free(nic->int_bufs);
-
 	free(nic->tx_desc_base);
 	fastlock_destroy(&nic->tx_desc_lock);
 }
@@ -740,6 +714,17 @@ static void __nic_destruct(void *obj)
 
 	GNIX_TRACE(FI_LOG_EP_CTRL, "\n");
 
+	/* Get us out of the progression tables we are destroying the nic
+	 * and we don't want the wait progression thread to progress us
+	 * after our structures are destroyed.
+	 */
+	pthread_mutex_lock(&gnix_nic_list_lock);
+
+	dlist_remove(&nic->gnix_nic_list);
+	--gnix_nics_per_ptag[nic->ptag];
+	dlist_remove(&nic->ptag_nic_list);
+
+	pthread_mutex_unlock(&gnix_nic_list_lock);
 	__gnix_nic_tx_freelist_destroy(nic);
 
 	/*
@@ -785,6 +770,11 @@ static void __nic_destruct(void *obj)
 			  "_gnix_mbox_allocator_destroy returned %s\n",
 			  fi_strerror(-ret));
 
+	/*
+	 * see comments in the nic constructor about why
+	 * the following code section is currently stubbed out.
+	 */
+#if 0
 	ret = _gnix_mbox_allocator_destroy(nic->s_rdma_buf_hndl);
 	if (ret != FI_SUCCESS)
 		GNIX_WARN(FI_LOG_EP_CTRL,
@@ -796,6 +786,7 @@ static void __nic_destruct(void *obj)
 		GNIX_WARN(FI_LOG_EP_CTRL,
 			  "_gnix_mbox_allocator_destroy returned %s\n",
 			  fi_strerror(-ret));
+#endif
 
 	if (!nic->gni_cdm_hndl) {
 		GNIX_WARN(FI_LOG_EP_CTRL, "No CDM attached to nic, nic=%p");
@@ -803,6 +794,17 @@ static void __nic_destruct(void *obj)
 
 	assert(nic->gni_cdm_hndl != NULL);
 
+	if (nic->rx_cq != NULL && nic->rx_cq != nic->rx_cq_blk) {
+		status = GNI_CqDestroy(nic->rx_cq);
+		if (status != GNI_RC_SUCCESS) {
+			GNIX_WARN(FI_LOG_EP_CTRL,
+				  "GNI_CqDestroy returned %s\n",
+				 gni_err_str[status]);
+			ret = gnixu_to_fi_errno(status);
+			goto err;
+		}
+	}
+
 	if (nic->rx_cq_blk != NULL) {
 		status = GNI_CqDestroy(nic->rx_cq_blk);
 		if (status != GNI_RC_SUCCESS) {
@@ -814,8 +816,8 @@ static void __nic_destruct(void *obj)
 		}
 	}
 
-	if (nic->rx_cq != NULL) {
-		status = GNI_CqDestroy(nic->rx_cq);
+	if (nic->tx_cq != NULL && nic->tx_cq != nic->tx_cq_blk) {
+		status = GNI_CqDestroy(nic->tx_cq);
 		if (status != GNI_RC_SUCCESS) {
 			GNIX_WARN(FI_LOG_EP_CTRL,
 				  "GNI_CqDestroy returned %s\n",
@@ -836,17 +838,6 @@ static void __nic_destruct(void *obj)
 		}
 	}
 
-	if (nic->tx_cq != NULL) {
-		status = GNI_CqDestroy(nic->tx_cq);
-		if (status != GNI_RC_SUCCESS) {
-			GNIX_WARN(FI_LOG_EP_CTRL,
-				  "GNI_CqDestroy returned %s\n",
-				 gni_err_str[status]);
-			ret = gnixu_to_fi_errno(status);
-			goto err;
-		}
-	}
-
 	if (nic->allocd_gni_res & GNIX_NIC_CDM_ALLOCD) {
 		status = GNI_CdmDestroy(nic->gni_cdm_hndl);
 		if (status != GNI_RC_SUCCESS) {
@@ -864,6 +855,12 @@ static void __nic_destruct(void *obj)
 		GNIX_WARN(FI_LOG_EP_CTRL, "vc_id_table was NULL\n");
 	}
 
+	/*
+	 * destroy VC free list associated with this nic
+	 */
+
+	_gnix_fl_destroy(&nic->vc_freelist);
+
 	/*
 	 * remove the nic from the linked lists
 	 * for the domain and the global nic list
@@ -872,14 +869,6 @@ static void __nic_destruct(void *obj)
 err:
 	_gnix_free_bitmap(&nic->vc_id_bitmap);
 
-	pthread_mutex_lock(&gnix_nic_list_lock);
-
-	dlist_remove(&nic->gnix_nic_list);
-	--gnix_nics_per_ptag[nic->ptag];
-	dlist_remove(&nic->dom_nic_list);
-
-	pthread_mutex_unlock(&gnix_nic_list_lock);
-
 	free(nic);
 }
 
@@ -907,24 +896,29 @@ int gnix_nic_alloc(struct gnix_fid_domain *domain,
 	struct gnix_nic *nic = NULL;
 	uint32_t device_addr;
 	gni_return_t status;
-	uint32_t fake_cdm_id, seed;
+	uint32_t fake_cdm_id = GNIX_CREATE_CDM_ID;
 	gni_smsg_attr_t smsg_mbox_attr;
 	struct gnix_nic_attr *nic_attr = &default_attr;
+	uint32_t num_corespec_cpus = 0;
 	bool must_alloc_nic = false;
+	bool free_list_inited = false;
+	struct gnix_auth_key *auth_key;
 
 	GNIX_TRACE(FI_LOG_EP_CTRL, "\n");
 
 	*nic_ptr = NULL;
+	nic_attr->gni_cdm_modes = gnix_cdm_modes;
 
 	if (attr) {
 		ret = __gnix_nic_check_attr_sanity(attr);
 		if (ret != FI_SUCCESS)
 			return ret;
 		nic_attr = attr;
-		if (nic_attr->use_cdm_id == true)
-			must_alloc_nic = true;
+		must_alloc_nic = nic_attr->must_alloc;
 	}
 
+	auth_key = nic_attr->auth_key;
+
 	/*
 	 * If we've maxed out the number of nics for this domain/ptag,
 	 * search the list of existing nics.  Take the gnix_nic_list_lock
@@ -945,17 +939,18 @@ int gnix_nic_alloc(struct gnix_fid_domain *domain,
 
 	/*
 	 * we can reuse previously allocated nics as long as a
-	 * cdm_id is not specified in the nic_attr arg.
+	 * must_alloc is not specified in the nic_attr arg.
 	 */
 
-	if ((must_alloc_nic == false) && (gnix_nics_per_ptag[domain->ptag] >=
-					 gnix_max_nics_per_ptag)) {
-		assert(!dlist_empty(&domain->nic_list));
+	if ((must_alloc_nic == false) &&
+	    (gnix_nics_per_ptag[auth_key->ptag] >= gnix_max_nics_per_ptag)) {
+		assert(!dlist_empty(&gnix_nic_list_ptag[auth_key->ptag]));
 
-		nic = dlist_first_entry(&domain->nic_list, struct gnix_nic,
-					dom_nic_list);
-		dlist_remove(&nic->dom_nic_list);
-		dlist_insert_tail(&nic->dom_nic_list, &domain->nic_list);
+		nic = dlist_first_entry(&gnix_nic_list_ptag[auth_key->ptag],
+					struct gnix_nic, ptag_nic_list);
+		dlist_remove(&nic->ptag_nic_list);
+		dlist_insert_tail(&nic->ptag_nic_list,
+				  &gnix_nic_list_ptag[auth_key->ptag]);
 		_gnix_ref_get(nic);
 
 		GNIX_INFO(FI_LOG_EP_CTRL, "Reusing NIC:%p\n", nic);
@@ -974,15 +969,7 @@ int gnix_nic_alloc(struct gnix_fid_domain *domain,
 		}
 
 		if (nic_attr->use_cdm_id == false) {
-			ret = _gnix_cm_nic_get_cdm_seed_set(domain, 1, &seed);
-			if (ret != FI_SUCCESS) {
-				GNIX_WARN(FI_LOG_EP_CTRL,
-					  "_gnix_cm_nic_get_cdm_seed_set returned %s\n",
-					  fi_strerror(-ret));
-				goto err;
-			}
-			ret = _gnix_cm_nic_create_cdm_id(domain, seed,
-							 &fake_cdm_id);
+			ret = _gnix_cm_nic_create_cdm_id(domain, &fake_cdm_id);
 			if (ret != FI_SUCCESS) {
 				GNIX_WARN(FI_LOG_EP_CTRL,
 					  "_gnix_cm_nic_create_cdm_id returned %s\n",
@@ -994,8 +981,8 @@ int gnix_nic_alloc(struct gnix_fid_domain *domain,
 
 		if (nic_attr->gni_cdm_hndl == NULL) {
 			status = GNI_CdmCreate(fake_cdm_id,
-						domain->ptag,
-						domain->cookie,
+						auth_key->ptag,
+						auth_key->cookie,
 						gnix_cdm_modes,
 						&nic->gni_cdm_hndl);
 			if (status != GNI_RC_SUCCESS) {
@@ -1005,8 +992,9 @@ int gnix_nic_alloc(struct gnix_fid_domain *domain,
 				goto err1;
 			}
 			nic->allocd_gni_res |= GNIX_NIC_CDM_ALLOCD;
-		} else
+		} else {
 			nic->gni_cdm_hndl = nic_attr->gni_cdm_hndl;
+		}
 
 		/*
 		 * Okay, now go for the attach
@@ -1020,6 +1008,7 @@ int gnix_nic_alloc(struct gnix_fid_domain *domain,
 			if (status != GNI_RC_SUCCESS) {
 				GNIX_WARN(FI_LOG_EP_CTRL, "GNI_CdmAttach returned %s\n",
 					 gni_err_str[status]);
+				_gnix_dump_gni_res(auth_key->ptag);
 				ret = gnixu_to_fi_errno(status);
 				goto err1;
 			}
@@ -1031,63 +1020,55 @@ int gnix_nic_alloc(struct gnix_fid_domain *domain,
 		 */
 
 		status = GNI_CqCreate(nic->gni_nic_hndl,
-					domain->gni_tx_cq_size,
+					domain->params.tx_cq_size,
 					0,                  /* no delay count */
-					GNI_CQ_NOBLOCK |
-					domain->gni_cq_modes,
+					GNI_CQ_BLOCKING |
+						domain->gni_cq_modes,
 					NULL,              /* useless handler */
 					NULL,               /* useless handler
 								context */
-					&nic->tx_cq);
-		if (status != GNI_RC_SUCCESS) {
-			GNIX_WARN(FI_LOG_EP_CTRL,
-				  "GNI_CqCreate returned %s\n",
-				  gni_err_str[status]);
-			ret = gnixu_to_fi_errno(status);
-			goto err1;
-		}
-
-		status = GNI_CqCreate(nic->gni_nic_hndl,
-					domain->gni_tx_cq_size,
-					0,
-					GNI_CQ_BLOCKING |
-						domain->gni_cq_modes,
-					NULL,
-					NULL,
 					&nic->tx_cq_blk);
 		if (status != GNI_RC_SUCCESS) {
 			GNIX_WARN(FI_LOG_EP_CTRL,
 				  "GNI_CqCreate returned %s\n",
 				  gni_err_str[status]);
+			_gnix_dump_gni_res(auth_key->ptag);
 			ret = gnixu_to_fi_errno(status);
 			goto err1;
 		}
 
+		/* Use blocking CQs for all operations if eager_auto_progress
+		 * is used.  */
+		if (domain->params.eager_auto_progress) {
+			nic->tx_cq = nic->tx_cq_blk;
+		} else {
+			status = GNI_CqCreate(nic->gni_nic_hndl,
+						domain->params.tx_cq_size,
+						0, /* no delay count */
+						domain->gni_cq_modes,
+						NULL, /* useless handler */
+						NULL, /* useless handler ctx */
+						&nic->tx_cq);
+			if (status != GNI_RC_SUCCESS) {
+				GNIX_WARN(FI_LOG_EP_CTRL,
+					  "GNI_CqCreate returned %s\n",
+					  gni_err_str[status]);
+				_gnix_dump_gni_res(auth_key->ptag);
+				ret = gnixu_to_fi_errno(status);
+				goto err1;
+			}
+		}
+
+
 		/*
 		 * create RX CQs - first polling, then blocking
 		 */
 
 		status = GNI_CqCreate(nic->gni_nic_hndl,
-					domain->gni_rx_cq_size,
-					0,
-					GNI_CQ_NOBLOCK |
-						domain->gni_cq_modes,
-					NULL,
-					NULL,
-					&nic->rx_cq);
-		if (status != GNI_RC_SUCCESS) {
-			GNIX_WARN(FI_LOG_EP_CTRL,
-				  "GNI_CqCreate returned %s\n",
-				  gni_err_str[status]);
-			ret = gnixu_to_fi_errno(status);
-			goto err1;
-		}
-
-		status = GNI_CqCreate(nic->gni_nic_hndl,
-					domain->gni_rx_cq_size,
+					domain->params.rx_cq_size,
 					0,
 					GNI_CQ_BLOCKING |
-					domain->gni_cq_modes,
+						domain->gni_cq_modes,
 					NULL,
 					NULL,
 					&nic->rx_cq_blk);
@@ -1095,13 +1076,36 @@ int gnix_nic_alloc(struct gnix_fid_domain *domain,
 			GNIX_WARN(FI_LOG_EP_CTRL,
 				  "GNI_CqCreate returned %s\n",
 				  gni_err_str[status]);
+			_gnix_dump_gni_res(auth_key->ptag);
 			ret = gnixu_to_fi_errno(status);
 			goto err1;
 		}
 
+		/* Use blocking CQs for all operations if eager_auto_progress
+		 * is used.  */
+		if (domain->params.eager_auto_progress) {
+			nic->rx_cq = nic->rx_cq_blk;
+		} else {
+			status = GNI_CqCreate(nic->gni_nic_hndl,
+						domain->params.rx_cq_size,
+						0,
+						domain->gni_cq_modes,
+						NULL,
+						NULL,
+						&nic->rx_cq);
+			if (status != GNI_RC_SUCCESS) {
+				GNIX_WARN(FI_LOG_EP_CTRL,
+					  "GNI_CqCreate returned %s\n",
+					  gni_err_str[status]);
+				_gnix_dump_gni_res(auth_key->ptag);
+				ret = gnixu_to_fi_errno(status);
+				goto err1;
+			}
+		}
+
 		nic->device_addr = device_addr;
-		nic->ptag = domain->ptag;
-		nic->cookie = domain->cookie;
+		nic->ptag = auth_key->ptag;
+		nic->cookie = auth_key->cookie;
 
 		nic->vc_id_table_capacity = domain->params.vc_id_table_capacity;
 		nic->vc_id_table = malloc(sizeof(void *) *
@@ -1113,21 +1117,8 @@ int gnix_nic_alloc(struct gnix_fid_domain *domain,
 			goto err1;
 		}
 
-		fastlock_init(&nic->lock);
-
-		ret = __gnix_nic_tx_freelist_init(nic, domain->gni_tx_cq_size);
-		if (ret != FI_SUCCESS)
-			goto err1;
-
-		fastlock_init(&nic->rx_vc_lock);
-		dlist_init(&nic->rx_vcs);
-		fastlock_init(&nic->work_vc_lock);
-		dlist_init(&nic->work_vcs);
-		fastlock_init(&nic->tx_vc_lock);
-		dlist_init(&nic->tx_vcs);
-
-		_gnix_ref_init(&nic->ref_cnt, 1, __nic_destruct);
-		ret = _gnix_alloc_bitmap(&nic->vc_id_bitmap, 1000);
+		ret = _gnix_alloc_bitmap(&nic->vc_id_bitmap,
+					 nic->vc_id_table_capacity);
 		if (ret != FI_SUCCESS) {
 			GNIX_WARN(FI_LOG_EP_CTRL,
 				  "alloc_bitmap returned %d\n", ret);
@@ -1135,6 +1126,50 @@ int gnix_nic_alloc(struct gnix_fid_domain *domain,
 		}
 		fastlock_init(&nic->vc_id_lock);
 
+		/*
+		 * initialize free list for VC's
+		 * In addition to hopefully allowing for a more compact
+		 * allocation of VC structs, the free list is also import
+		 * because there is a window of time when using auto progress
+		 * that a thread may be going through the progress engine
+		 * while one of the application threads is actively tearing
+		 * down an endpoint (and hence its associated VCs) before the
+		 * rem_id for the vc is removed from the vector.
+		 * As a consequence, it is important that
+		 * the memory allocated within the freelist allocator not be
+		 * returned to the system prior to the freelist being destroyed
+		 * as part of the nic destructor procedure.  The freelist is
+		 * destroyed in that procedure after the progress thread
+		 * has been joined.
+		 */
+
+		ret = _gnix_fl_init_ts(sizeof(struct gnix_vc),
+				       offsetof(struct gnix_vc, fr_list),
+				       GNIX_VC_FL_MIN_SIZE,
+				       GNIX_VC_FL_INIT_REFILL_SIZE,
+				       0,
+				       0,
+				       &nic->vc_freelist);
+		if (ret == FI_SUCCESS) {
+			free_list_inited = true;
+		} else {
+			GNIX_DEBUG(FI_LOG_EP_DATA, "_gnix_fl_init returned: %s\n",
+				   fi_strerror(-ret));
+			goto err1;
+		}
+
+		fastlock_init(&nic->lock);
+
+		ret = __gnix_nic_tx_freelist_init(nic,
+						  domain->params.tx_cq_size);
+		if (ret != FI_SUCCESS)
+			goto err1;
+
+		fastlock_init(&nic->prog_vcs_lock);
+		dlist_init(&nic->prog_vcs);
+
+		_gnix_ref_init(&nic->ref_cnt, 1, __nic_destruct);
+
 		smsg_mbox_attr.msg_type = GNI_SMSG_TYPE_MBOX_AUTO_RETRANSMIT;
 		smsg_mbox_attr.mbox_maxcredit = domain->params.mbox_maxcredit;
 		smsg_mbox_attr.msg_maxsize =  domain->params.mbox_msg_maxsize;
@@ -1159,6 +1194,7 @@ int gnix_nic_alloc(struct gnix_fid_domain *domain,
 					  (size_t)nic->mem_per_mbox,
 					  domain->params.mbox_num_per_slab,
 					  &nic->mbox_hndl);
+
 		if (ret != FI_SUCCESS) {
 			GNIX_WARN(FI_LOG_EP_CTRL,
 				  "_gnix_mbox_alloc returned %s\n",
@@ -1180,8 +1216,11 @@ int gnix_nic_alloc(struct gnix_fid_domain *domain,
 		 * TODO: hardwired constants, uff
 		 * TODO: better to use a buddy allocator or some other
 		 * allocator
+		 * Disable these for now as we're not using and they
+		 * chew up a lot of IOMMU space per nic.
 		 */
 
+#if 0
 		ret = _gnix_mbox_allocator_create(nic,
 						  NULL,
 						  GNIX_PAGE_2MB,
@@ -1192,6 +1231,7 @@ int gnix_nic_alloc(struct gnix_fid_domain *domain,
 			GNIX_WARN(FI_LOG_EP_CTRL,
 				  "_gnix_mbox_alloc returned %s\n",
 				  fi_strerror(-ret));
+			_gnix_dump_gni_res(domain->ptag);
 			goto err1;
 		}
 
@@ -1205,14 +1245,17 @@ int gnix_nic_alloc(struct gnix_fid_domain *domain,
 			GNIX_WARN(FI_LOG_EP_CTRL,
 				  "_gnix_mbox_alloc returned %s\n",
 				  fi_strerror(-ret));
+			_gnix_dump_gni_res(domain->ptag);
 			goto err1;
 		}
+#endif
 
 		ret =  __nic_setup_irq_cq(nic);
 		if (ret != FI_SUCCESS) {
 			GNIX_WARN(FI_LOG_EP_CTRL,
 				  "__nic_setup_irq_cq returned %s\n",
 				  fi_strerror(-ret));
+			_gnix_dump_gni_res(auth_key->ptag);
 			goto err1;
 		}
 
@@ -1230,10 +1273,20 @@ int gnix_nic_alloc(struct gnix_fid_domain *domain,
 			 * though...
 			 */
 
-			ret = _gnix_job_disable_affinity_apply();
+			ret = _gnix_get_num_corespec_cpus(&num_corespec_cpus);
+			if (ret != FI_SUCCESS) {
+				GNIX_WARN(FI_LOG_EP_CTRL,
+				  "failed to get num corespec cpus\n");
+			}
+			if (num_corespec_cpus > 0) {
+				ret = _gnix_job_disable_affinity_apply();
+			} else {
+				ret = _gnix_job_enable_unassigned_cpus();
+			}
 			if (ret != 0)
 				GNIX_WARN(FI_LOG_EP_CTRL,
-				"_gnix_job_disable call returned %d\n", ret);
+				"job_disable/unassigned cpus returned %d\n",
+					 ret);
 
 			ret = pthread_create(&nic->progress_thread,
 					     NULL,
@@ -1245,18 +1298,25 @@ int gnix_nic_alloc(struct gnix_fid_domain *domain,
 		}
 
 		dlist_insert_tail(&nic->gnix_nic_list, &gnix_nic_list);
-		dlist_insert_tail(&nic->dom_nic_list, &domain->nic_list);
+		dlist_insert_tail(&nic->ptag_nic_list,
+				  &gnix_nic_list_ptag[auth_key->ptag]);
 
-		++gnix_nics_per_ptag[domain->ptag];
+		nic->smsg_callbacks = gnix_ep_smsg_callbacks;
+
+		++gnix_nics_per_ptag[auth_key->ptag];
 
 		GNIX_INFO(FI_LOG_EP_CTRL, "Allocated NIC:%p\n", nic);
 	}
 
+	if (nic) {
+		nic->requires_lock = domain->thread_model != FI_THREAD_COMPLETION;
+	}
+
 	*nic_ptr = nic;
 	goto out;
 
 err1:
-	atomic_dec(&gnix_id_counter);
+	ofi_atomic_dec32(&gnix_id_counter);
 err:
 	if (nic != NULL) {
 		__nic_teardown_irq_cq(nic);
@@ -1266,17 +1326,19 @@ err:
 			_gnix_mbox_allocator_destroy(nic->s_rdma_buf_hndl);
 		if (nic->mbox_hndl != NULL)
 			_gnix_mbox_allocator_destroy(nic->mbox_hndl);
+		if (nic->rx_cq != NULL && nic->rx_cq != nic->rx_cq_blk)
+			GNI_CqDestroy(nic->rx_cq);
 		if (nic->rx_cq_blk != NULL)
 			GNI_CqDestroy(nic->rx_cq_blk);
-		if (nic->rx_cq != NULL)
-			GNI_CqDestroy(nic->rx_cq);
+		if (nic->tx_cq != NULL && nic->tx_cq != nic->tx_cq_blk)
+			GNI_CqDestroy(nic->tx_cq);
 		if (nic->tx_cq_blk != NULL)
 			GNI_CqDestroy(nic->tx_cq_blk);
-		if (nic->tx_cq != NULL)
-			GNI_CqDestroy(nic->tx_cq);
 		if ((nic->gni_cdm_hndl != NULL) && (nic->allocd_gni_res &
 		    GNIX_NIC_CDM_ALLOCD))
 			GNI_CdmDestroy(nic->gni_cdm_hndl);
+		if (free_list_inited == true)
+			_gnix_fl_destroy(&nic->vc_freelist);
 		free(nic);
 	}
 
@@ -1284,3 +1346,35 @@ out:
 	pthread_mutex_unlock(&gnix_nic_list_lock);
 	return ret;
 }
+
+void _gnix_nic_init(void)
+{
+	int i, rc;
+
+	for (i = 0; i < GNI_PTAG_MAX; i++) {
+		dlist_init(&gnix_nic_list_ptag[i]);
+	}
+
+	rc = _gnix_nics_per_rank(&gnix_max_nics_per_ptag);
+	if (rc == FI_SUCCESS) {
+		GNIX_DEBUG(FI_LOG_FABRIC, "gnix_max_nics_per_ptag: %u\n",
+			   gnix_max_nics_per_ptag);
+	} else {
+		GNIX_WARN(FI_LOG_FABRIC, "_gnix_nics_per_rank failed: %d\n",
+			  rc);
+	}
+
+	if (getenv("GNIX_MAX_NICS") != NULL)
+		gnix_max_nics_per_ptag = atoi(getenv("GNIX_MAX_NICS"));
+
+	/*
+	 * Well if we didn't get 1 nic, that means we must really be doing
+	 * FMA sharing.
+	 */
+
+	if (gnix_max_nics_per_ptag == 0) {
+		gnix_max_nics_per_ptag = 1;
+		GNIX_WARN(FI_LOG_FABRIC, "Using inter-procss FMA sharing\n");
+	}
+}
+
