@@ -1,6 +1,7 @@
 /*
- * Copyright (c) 2015 Los Alamos National Security, LLC. All rights reserved.
- * Copyright (c) 2015-2016 Cray Inc. All rights reserved.
+ * Copyright (c) 2015-2017 Los Alamos National Security, LLC.
+ *                         All rights reserved.
+ * Copyright (c) 2015-2017 Cray Inc. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -43,20 +44,15 @@
 #include <limits.h>
 #include <assert.h>
 
-#include <rdma/fabric.h>
-#include <rdma/fi_domain.h>
-#include <rdma/fi_errno.h>
-#include <rdma/fi_endpoint.h>
-#include <rdma/fi_cm.h>
-#include <rdma/fi_atomic.h>
-#include <rdma/fi_tagged.h>
-#include "fi_ext_gni.h"
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <inttypes.h>
 
 #include <criterion/criterion.h>
+#include "gnix_rdma_headers.h"
+#include "fi_ext_gni.h"
+#include "gnix_util.h"
+#include "common.h"
 
 #if 1
 #define dbg_printf(...)
@@ -70,6 +66,8 @@
 
 #define NUMEPS 2
 
+/* Note: Set to ~FI_NOTIFY_FLAGS_ONLY since this was written before api 1.5 */
+static uint64_t mode_bits = ~FI_NOTIFY_FLAGS_ONLY;
 static struct fid_fabric *fab;
 static struct fid_domain *dom[NUMEPS];
 struct fi_gni_ops_domain *gni_domain_ops[NUMEPS];
@@ -107,11 +105,13 @@ void rdm_api_setup_ep(void)
 
 	/* Get info about fabric services with the provided hints */
 	for (i = 0; i < NUMEPS; i++) {
-		ret = fi_getinfo(FI_VERSION(1, 0), NULL, 0, 0, hints[i],
+		hints[i]->domain_attr->mr_mode = GNIX_DEFAULT_MR_MODE;
+		ret = fi_getinfo(fi_version(), NULL, 0, 0, hints[i],
 				 &fi[i]);
 		cr_assert(!ret, "fi_getinfo");
 	}
 
+	memset(&attr, 0, sizeof(attr));
 	attr.type = FI_AV_MAP;
 	attr.count = NUMEPS;
 
@@ -174,9 +174,6 @@ void rdm_api_setup_ep(void)
 		ret = fi_ep_bind(ep[i], &av[i]->fid, 0);
 		cr_assert(!ret, "fi_ep_bind");
 
-		ret = fi_enable(ep[i]);
-		cr_assert(!ret, "fi_ep_enable");
-
 		ret = fi_cntr_open(dom[i], &cntr_attr, send_cntr + i, 0);
 		cr_assert(!ret, "fi_cntr_open");
 
@@ -188,6 +185,14 @@ void rdm_api_setup_ep(void)
 
 		ret = fi_ep_bind(ep[i], &recv_cntr[i]->fid, FI_RECV);
 		cr_assert(!ret, "fi_ep_bind");
+
+		ret = fi_enable(ep[i]);
+		cr_assert(!ret, "fi_ep_enable");
+
+		ret = fi_enable(ep[i]);
+		cr_assert_eq(ret, -FI_EOPBADSTATE);
+
+
 	}
 
 	for (i = 0; i < NUMEPS; i++) {
@@ -213,21 +218,24 @@ void rdm_api_setup(void)
 
 		hints[i]->domain_attr->cq_data_size = NUMEPS * 2;
 		hints[i]->domain_attr->data_progress = FI_PROGRESS_AUTO;
-		hints[i]->mode = ~0;
-		hints[i]->domain_attr->mr_mode = FI_MR_BASIC;
-		hints[i]->fabric_attr->name = strdup("gni");
+		hints[i]->mode = mode_bits;
+		hints[i]->domain_attr->mr_mode = GNIX_DEFAULT_MR_MODE;
+		hints[i]->fabric_attr->prov_name = strdup("gni");
 	}
 }
 
+void api_setup(void)
+{
+}
+
+void api_teardown(void)
+{
+}
+
 static void rdm_api_teardown_common(bool unreg)
 {
 	int ret = 0, i = 0;
 
-	free(uc_source);
-	free(uc_target);
-	free(target);
-	free(source);
-
 	for (; i < NUMEPS; i++) {
 		fi_close(&recv_cntr[i]->fid);
 		fi_close(&send_cntr[i]->fid);
@@ -254,6 +262,11 @@ static void rdm_api_teardown_common(bool unreg)
 		fi_freeinfo(hints[i]);
 	}
 
+	free(uc_source);
+	free(uc_target);
+	free(target);
+	free(source);
+
 	ret = fi_close(&fab->fid);
 	cr_assert(!ret, "failure in closing fabric.");
 }
@@ -288,8 +301,10 @@ int rdm_api_check_data(char *buf1, char *buf2, int len)
 
 void rdm_api_check_cqe(struct fi_cq_tagged_entry *cqe, void *ctx,
 		      uint64_t flags, void *addr, size_t len,
-		      uint64_t data)
+		      uint64_t data, struct fid_ep *fid_ep)
 {
+	struct gnix_fid_ep *gnix_ep = get_gnix_ep(fid_ep);
+
 	cr_assert(cqe->op_context == ctx, "CQE Context mismatch");
 	cr_assert(cqe->flags == flags, "CQE flags mismatch");
 
@@ -297,7 +312,8 @@ void rdm_api_check_cqe(struct fi_cq_tagged_entry *cqe, void *ctx,
 		cr_assert(cqe->len == len, "CQE length mismatch");
 		cr_assert(cqe->buf == addr, "CQE address mismatch");
 
-		if (flags & FI_REMOTE_CQ_DATA)
+	/* TODO: Remove GNIX_ALLOW_FI_REMOTE_CQ_DATA and only check flags for FI_RMA_EVENT */
+	if (GNIX_ALLOW_FI_REMOTE_CQ_DATA(flags, gnix_ep->caps))
 			cr_assert(cqe->data == data, "CQE data mismatch");
 	} else {
 		cr_assert(cqe->len == 0, "Invalid CQE length");
@@ -430,6 +446,36 @@ void api_send_recv(int len)
 	}
 }
 
+Test(api, dom_caps)
+{
+	int ret;
+
+	hints[0] = fi_allocinfo();
+	cr_assert(hints[0], "fi_allocinfo");
+
+	hints[0]->mode = mode_bits;
+	hints[0]->fabric_attr->prov_name = strdup("gni");
+	hints[0]->domain_attr->mr_mode = GNIX_DEFAULT_MR_MODE;
+
+	/* we only support REMOTE_COMM */
+	hints[0]->domain_attr->caps = FI_LOCAL_COMM;
+	ret = fi_getinfo(fi_version(), NULL, 0, 0, hints[0], &fi[0]);
+	cr_assert_eq(ret, -FI_ENODATA, "fi_getinfo");
+
+	hints[0]->domain_attr->caps = FI_REMOTE_COMM;
+	ret = fi_getinfo(fi_version(), NULL, 0, 0, hints[0], &fi[0]);
+	cr_assert_eq(ret, 0, "fi_getinfo");
+
+	fi_freeinfo(fi[0]);
+
+	hints[0]->domain_attr->mr_mode = FI_MR_UNSPEC;
+	ret = fi_getinfo(FI_VERSION(1, 0), NULL, 0, 0, hints[0], &fi[0]);
+	cr_assert_eq(ret, 0, "fi_getinfo");
+
+	fi_freeinfo(hints[0]);
+	fi_freeinfo(fi[0]);
+}
+
 Test(rdm_api, msg_no_caps)
 {
 	hints[0]->caps = 0;
@@ -815,3 +861,13 @@ Test(rdm_api, amo_write_read_w_msg)
 	api_write_read(BUF_SZ);
 }
 
+TestSuite(api, .init = api_setup, .fini = api_teardown, .disabled = false);
+
+Test(api, getinfo_w_null_hints)
+{
+	int ret;
+
+	ret = fi_getinfo(fi_version(), NULL, 0, 0, NULL, &fi[0]);
+	cr_assert(ret == FI_SUCCESS, "fi_getinfo returned: %s",
+		  fi_strerror(-ret));
+}
