@@ -1,6 +1,6 @@
 /*
- * Copyright (c) 2015 Los Alamos National Security, LLC. All rights reserved.
- * Copyright (c) 2015-2016 Cray Inc.  All rights reserved.
+ * Copyright (c) 2015-2017 Los Alamos National Security, LLC. All rights reserved.
+ * Copyright (c) 2015-2017 Cray Inc.  All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -40,11 +40,6 @@
 #include <string.h>
 #include <pthread.h>
 
-#include <rdma/fabric.h>
-#include <rdma/fi_domain.h>
-#include <rdma/fi_errno.h>
-#include <rdma/fi_endpoint.h>
-#include <rdma/fi_cm.h>
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -56,6 +51,8 @@
 #include "gnix_rma.h"
 
 #include <criterion/criterion.h>
+#include "gnix_rdma_headers.h"
+#include "common.h"
 
 #if 1
 #define dbg_printf(...)
@@ -68,29 +65,37 @@ do {				\
 #endif
 
 #define NUM_EPS 5
+/* NUM_THREADS must be <= NUM_EPS */
+#define NUM_THREADS NUM_EPS
 
+/* Note: Set to ~FI_NOTIFY_FLAGS_ONLY since this was written before api 1.5 */
+static uint64_t mode_bits = ~FI_NOTIFY_FLAGS_ONLY;
 static struct fid_fabric *fab;
-static struct fid_domain *dom;
+static struct fid_domain *dom[NUM_EPS];
 static struct fid_ep *ep[NUM_EPS];
-static struct fid_av *av;
+static struct fid_av *av[NUM_EPS];
 static struct fi_info *hints;
 static struct fi_info *fi;
 static void *ep_name[NUM_EPS];
-static size_t gni_addr[NUM_EPS];
+static size_t gni_addr[NUM_EPS][NUM_EPS];
 static struct fid_cq *send_cq;
 static struct fid_cq *recv_cq;
 static struct fi_cq_attr cq_attr;
-static struct fid_cntr *write_cntr, *read_cntr, *rcv_cntr;
+static struct fid_cntr *write_cntrs[NUM_EPS];
+static struct fid_cntr *read_cntrs[NUM_EPS];
+static struct fid_cntr *recv_cntrs[NUM_EPS];
+
 static struct fi_cntr_attr cntr_attr = {.events = FI_CNTR_EVENTS_COMP,
+					.wait_obj = FI_WAIT_UNSPEC,
 					.flags = 0};
 
 #define BUF_SZ (64*1024)
 char *target;
 char *source;
-struct fid_mr *rem_mr, *loc_mr;
-uint64_t mr_key;
+struct fid_mr *rem_mr[NUM_EPS], *loc_mr[NUM_EPS];
+uint64_t mr_key[NUM_EPS];
 
-static inline void cntr_setup_eps(void)
+static inline void cntr_setup_eps(const uint64_t caps)
 {
 	int i, ret;
 	struct fi_av_attr attr;
@@ -98,32 +103,37 @@ static inline void cntr_setup_eps(void)
 	hints = fi_allocinfo();
 	cr_assert(hints, "fi_allocinfo");
 
+	hints->domain_attr->mr_mode = GNIX_DEFAULT_MR_MODE;
 	hints->domain_attr->cq_data_size = 4;
-	hints->mode = ~0;
+	hints->mode = mode_bits;
 
-	hints->fabric_attr->name = strdup("gni");
+	hints->fabric_attr->prov_name = strdup("gni");
 
-	ret = fi_getinfo(FI_VERSION(1, 0), NULL, 0, 0, hints, &fi);
+	hints->caps = caps;
+
+	ret = fi_getinfo(fi_version(), NULL, 0, 0, hints, &fi);
 	cr_assert(!ret, "fi_getinfo");
 
 	ret = fi_fabric(fi->fabric_attr, &fab, NULL);
 	cr_assert(!ret, "fi_fabric");
 
-	ret = fi_domain(fab, fi, &dom, NULL);
-	cr_assert(!ret, "fi_domain");
-
+	memset(&attr, 0, sizeof(attr));
 	attr.type = FI_AV_MAP;
 	attr.count = 16;
 
-	ret = fi_av_open(dom, &attr, &av, NULL);
-	cr_assert(!ret, "fi_av_open");
-
 	for (i = 0; i < NUM_EPS; i++) {
-		ret = fi_endpoint(dom, fi, &ep[i], NULL);
+		ret = fi_domain(fab, fi, &dom[i], NULL);
+		cr_assert(!ret, "fi_domain");
+
+		ret = fi_av_open(dom[i], &attr, &av[i], NULL);
+		cr_assert(!ret, "fi_av_open");
+
+		ret = fi_endpoint(dom[i], fi, &ep[i], NULL);
 		cr_assert(!ret, "fi_endpoint");
 	}
 }
 
+/* Only used by cntr suite for now */
 static inline void cntr_setup_cqs(void)
 {
 	int ret;
@@ -132,32 +142,64 @@ static inline void cntr_setup_cqs(void)
 	cq_attr.size = 1024;
 	cq_attr.wait_obj = FI_WAIT_NONE;
 
-	ret = fi_cq_open(dom, &cq_attr, &send_cq, 0);
+	ret = fi_cq_open(dom[0], &cq_attr, &send_cq, 0);
 	cr_assert(!ret, "fi_cq_open");
+	ret = fi_ep_bind(ep[0], &send_cq->fid, FI_SEND);
+	cr_assert(!ret, "fi_ep_bind cq");
 
-	ret = fi_cq_open(dom, &cq_attr, &recv_cq, 0);
+	ret = fi_cq_open(dom[1], &cq_attr, &recv_cq, 0);
 	cr_assert(!ret, "fi_cq_open");
+	ret = fi_ep_bind(ep[1], &recv_cq->fid, FI_RECV);
+	cr_assert(!ret, "fi_ep_bind cq");
 
 }
 
-static inline void cntr_setup_cntrs(void)
+static inline void cntr_setup_cntrs(uint64_t write_flags,
+				    uint64_t read_flags,
+				    uint64_t recv_flags)
 {
-	int ret;
+	int i, ret;
 
-	ret = fi_cntr_open(dom, &cntr_attr, &write_cntr, 0);
-	cr_assert(!ret, "fi_cntr_open");
+	for (i = 0; i < NUM_EPS; i++) {
+		if (write_flags) {
+			ret = fi_cntr_open(dom[i], &cntr_attr,
+					   &write_cntrs[i], 0);
+			cr_assert(!ret, "fi_cntr_open");
+			ret = fi_ep_bind(ep[i], &write_cntrs[i]->fid,
+					 write_flags);
+			cr_assert(!ret, "fi_ep_bind cntr");
+		} else {
+			write_cntrs[i] = NULL;
+		}
 
-	ret = fi_cntr_open(dom, &cntr_attr, &read_cntr, 0);
-	cr_assert(!ret, "fi_cntr_open");
+		if (read_flags) {
+			ret = fi_cntr_open(dom[i], &cntr_attr,
+					   &read_cntrs[i], 0);
+			cr_assert(!ret, "fi_cntr_open");
+			ret = fi_ep_bind(ep[i], &read_cntrs[i]->fid,
+					 read_flags);
+			cr_assert(!ret, "fi_ep_bind cntr");
+		} else {
+			read_cntrs[i] = NULL;
+		}
 
-	ret = fi_cntr_open(dom, &cntr_attr, &rcv_cntr, 0);
-	cr_assert(!ret, "fi_cntr_open");
+		if (recv_flags) {
+			ret = fi_cntr_open(dom[i], &cntr_attr,
+					   &recv_cntrs[i], 0);
+			cr_assert(!ret, "fi_cntr_open");
+			ret = fi_ep_bind(ep[i], &recv_cntrs[i]->fid,
+					 recv_flags);
+			cr_assert(!ret, "fi_ep_bind cntr");
+		} else {
+			recv_cntrs[i] = NULL;
+		}
+	}
 
 }
 
 static inline void cntr_setup_av(void)
 {
-	int i, ret;
+	int i, j, ret;
 	size_t addrlen = 0;
 
 	ret = fi_getname(&ep[0]->fid, NULL, &addrlen);
@@ -170,46 +212,19 @@ static inline void cntr_setup_av(void)
 		ret = fi_getname(&ep[i]->fid, ep_name[i], &addrlen);
 		cr_assert(ret == FI_SUCCESS);
 
-		ret = fi_av_insert(av, ep_name[i], 1, &gni_addr[i], 0,
-				   NULL);
-		cr_assert(ret == 1);
+		for (j = 0; j < NUM_EPS; j++) {
+			ret = fi_av_insert(av[j], ep_name[i], 1,
+					   &gni_addr[j][i], 0, NULL);
+			cr_assert(ret == 1);
+		}
 	}
 
 	for (i = 0; i < NUM_EPS; i++) {
-		ret = fi_ep_bind(ep[i], &av->fid, 0);
+		ret = fi_ep_bind(ep[i], &av[i]->fid, 0);
 		cr_assert(!ret, "fi_ep_bind av");
 	}
 }
 
-static inline void cntr_setup_bind_cqs(void)
-{
-	int i, ret;
-
-	for (i = 0; i < NUM_EPS; i++) {
-		ret = fi_ep_bind(ep[i], &send_cq->fid, FI_SEND);
-		cr_assert(!ret, "fi_ep_bind cq");
-
-		ret = fi_ep_bind(ep[i], &recv_cq->fid, FI_RECV);
-		cr_assert(!ret, "fi_ep_bind cq");
-	}
-}
-
-static inline void cntr_setup_bind_cntrs(void)
-{
-	int i, ret;
-
-	for (i = 0; i < NUM_EPS; i++) {
-		ret = fi_ep_bind(ep[i], &write_cntr->fid, FI_WRITE | FI_SEND);
-		cr_assert(!ret, "fi_ep_bind cntr");
-
-		ret = fi_ep_bind(ep[i], &read_cntr->fid, FI_READ);
-		cr_assert(!ret, "fi_ep_bind cntr");
-
-		ret = fi_ep_bind(ep[i], &rcv_cntr->fid, FI_RECV);
-		cr_assert(!ret, "fi_ep_bind cntr");
-	}
-}
-
 static inline void cntr_setup_enable_ep(void)
 {
 	int i, ret;
@@ -222,7 +237,7 @@ static inline void cntr_setup_enable_ep(void)
 
 static inline void cntr_setup_mr(void)
 {
-	int ret;
+	int i, ret;
 
 	target = malloc(BUF_SZ);
 	assert(target);
@@ -230,33 +245,38 @@ static inline void cntr_setup_mr(void)
 	source = malloc(BUF_SZ);
 	assert(source);
 
-	ret = fi_mr_reg(dom, target, BUF_SZ,
-			FI_REMOTE_WRITE, 0, 0, 0, &rem_mr, &target);
-	cr_assert_eq(ret, 0);
+	for (i = 0; i < NUM_EPS; i++) {
+		ret = fi_mr_reg(dom[i], target, BUF_SZ,
+				FI_REMOTE_READ | FI_REMOTE_WRITE,
+				0, 0, 0, &rem_mr[i], &target);
+		cr_assert_eq(ret, 0);
 
-	ret = fi_mr_reg(dom, source, BUF_SZ,
-			FI_REMOTE_WRITE, 0, 0, 0, &loc_mr, &source);
-	cr_assert_eq(ret, 0);
+		ret = fi_mr_reg(dom[i], source, BUF_SZ,	FI_READ | FI_WRITE,
+				0, 0, 0, &loc_mr[i], &source);
+		cr_assert_eq(ret, 0);
 
-	mr_key = fi_mr_key(rem_mr);
+		mr_key[i] = fi_mr_key(rem_mr[i]);
+	}
 }
 
 static void cntr_setup(void)
 {
-	cntr_setup_eps();
+	cntr_setup_eps(GNIX_EP_PRIMARY_CAPS);
 	cntr_setup_av();
 	cntr_setup_cqs();
-	cntr_setup_bind_cqs();
-	cntr_setup_cntrs();
-	cntr_setup_bind_cntrs();
+	cntr_setup_cntrs(FI_WRITE | FI_SEND, FI_READ, FI_RECV);
 	cntr_setup_enable_ep();
 	cntr_setup_mr();
 }
 
 static inline void cntr_teardown_mr(void)
 {
-	fi_close(&loc_mr->fid);
-	fi_close(&rem_mr->fid);
+	int i;
+
+	for (i = 0; i < NUM_EPS; i++) {
+		fi_close(&loc_mr[i]->fid);
+		fi_close(&rem_mr[i]->fid);
+	}
 
 	free(target);
 	free(source);
@@ -286,16 +306,24 @@ static inline void cntr_teardown_cqs(void)
 
 static inline void cntr_teardown_cntrs(void)
 {
-	int ret;
+	int i, ret;
 
-	ret = fi_close(&write_cntr->fid);
-	cr_assert(!ret, "failure in closing write_cntr.");
+	for (i = 0; i < NUM_EPS; i++) {
+		if (write_cntrs[i]) {
+			ret = fi_close(&write_cntrs[i]->fid);
+			cr_assert(!ret, "failure in closing write_cntr.");
+		}
 
-	ret = fi_close(&read_cntr->fid);
-	cr_assert(!ret, "failure in closing read_cntr.");
+		if (read_cntrs[i]) {
+			ret = fi_close(&read_cntrs[i]->fid);
+			cr_assert(!ret, "failure in closing read_cntr.");
+		}
 
-	ret = fi_close(&rcv_cntr->fid);
-	cr_assert(!ret, "failure in closing read_cntr.");
+		if (recv_cntrs[i]) {
+			ret = fi_close(&recv_cntrs[i]->fid);
+			cr_assert(!ret, "failure in closing read_cntr.");
+		}
+	}
 
 }
 
@@ -303,21 +331,22 @@ static inline void cntr_teardown_fini(void)
 {
 	int i, ret;
 
-	ret = fi_close(&av->fid);
-	cr_assert(!ret, "failure in closing av.");
+	for (i = 0; i < NUM_EPS; i++) {
+		ret = fi_close(&av[i]->fid);
+		cr_assert(!ret, "failure in closing av.");
 
-	ret = fi_close(&dom->fid);
-	cr_assert(!ret, "failure in closing domain.");
+		ret = fi_close(&dom[i]->fid);
+		cr_assert(!ret, "failure in closing domain.");
+
+		free(ep_name[i]);
+	}
 
 	ret = fi_close(&fab->fid);
 	cr_assert(!ret, "failure in closing fabric.");
 
+	/* This can be moved up to set up */
 	fi_freeinfo(fi);
 	fi_freeinfo(hints);
-
-	for (i = 0; i < NUM_EPS; i++) {
-		free(ep_name[i]);
-	}
 }
 
 static void cntr_teardown(void)
@@ -377,16 +406,15 @@ static void do_write(int len)
 	init_data(target, len, 0);
 
 
-	old_w_cnt = fi_cntr_read(write_cntr);
-	old_r_cnt = fi_cntr_read(read_cntr);
+	old_w_cnt = fi_cntr_read(write_cntrs[0]);
+	old_r_cnt = fi_cntr_read(read_cntrs[0]);
 
-	sz = fi_write(ep[0], source, len,
-			 loc_mr, gni_addr[1], (uint64_t)target, mr_key,
-			 target);
+	sz = fi_write(ep[0], source, len, loc_mr[0], gni_addr[0][1],
+		      (uint64_t)target, mr_key[1], target);
 	cr_assert_eq(sz, 0);
 
 	do {
-		new_w_cnt = fi_cntr_read(write_cntr);
+		new_w_cnt = fi_cntr_read(write_cntrs[0]);
 		if (new_w_cnt == (old_w_cnt + 1))
 			break;
 		pthread_yield();
@@ -394,7 +422,7 @@ static void do_write(int len)
 
 	cr_assert(check_data(source, target, len), "Data mismatch");
 
-	new_r_cnt = fi_cntr_read(read_cntr);
+	new_r_cnt = fi_cntr_read(read_cntrs[0]);
 
 	/*
 	 * no fi_read called so old and new read cnts should be equal
@@ -418,23 +446,23 @@ static void do_write_wait(int len)
 	init_data(source, len, 0xab);
 	init_data(target, len, 0);
 
-	old_w_cnt = fi_cntr_read(write_cntr);
-	old_r_cnt = fi_cntr_read(read_cntr);
+	old_w_cnt = fi_cntr_read(write_cntrs[0]);
+	old_r_cnt = fi_cntr_read(read_cntrs[0]);
 
 	for (i = 0; i < iters; i++) {
-		sz = fi_write(ep[0], source, len,
-			      loc_mr, gni_addr[1], (uint64_t)target, mr_key,
+		sz = fi_write(ep[0], source, len, loc_mr[0],
+			      gni_addr[0][1], (uint64_t)target, mr_key[1],
 			      target);
 		cr_assert_eq(sz, 0);
 	}
 
-	fi_cntr_wait(write_cntr, old_w_cnt+iters, -1);
-	new_w_cnt = fi_cntr_read(write_cntr);
+	fi_cntr_wait(write_cntrs[0], old_w_cnt+iters, -1);
+	new_w_cnt = fi_cntr_read(write_cntrs[0]);
 	cr_assert(old_w_cnt + iters == new_w_cnt);
 
 	cr_assert(check_data(source, target, len), "Data mismatch");
 
-	new_r_cnt = fi_cntr_read(read_cntr);
+	new_r_cnt = fi_cntr_read(read_cntrs[0]);
 
 	/*
 	 * no fi_read called so old and new read cnts should be equal
@@ -457,16 +485,16 @@ static void do_read(int len)
 	init_data(source, len, 0);
 	init_data(target, len, 0xad);
 
-	old_w_cnt = fi_cntr_read(write_cntr);
-	old_r_cnt = fi_cntr_read(read_cntr);
+	old_w_cnt = fi_cntr_read(write_cntrs[0]);
+	old_r_cnt = fi_cntr_read(read_cntrs[0]);
 
-	sz = fi_read(ep[0], source, len,
-			loc_mr, gni_addr[1], (uint64_t)target, mr_key,
+	sz = fi_read(ep[0], source, len, loc_mr[0],
+		     gni_addr[0][1], (uint64_t)target, mr_key[1],
 			(void *)READ_CTX);
 	cr_assert_eq(sz, 0);
 
 	do {
-		new_r_cnt = fi_cntr_read(read_cntr);
+		new_r_cnt = fi_cntr_read(read_cntrs[0]);
 		if (new_r_cnt == (old_r_cnt + 1))
 			break;
 		pthread_yield();
@@ -474,7 +502,7 @@ static void do_read(int len)
 
 	cr_assert(check_data(source, target, len), "Data mismatch");
 
-	new_w_cnt = fi_cntr_read(write_cntr);
+	new_w_cnt = fi_cntr_read(write_cntrs[0]);
 
 	/*
 	 * no fi_read called so old and new read cnts should be equal
@@ -493,21 +521,21 @@ static void do_read_wait(int len)
 	init_data(source, len, 0);
 	init_data(target, len, 0xad);
 
-	old_w_cnt = fi_cntr_read(write_cntr);
-	old_r_cnt = fi_cntr_read(read_cntr);
+	old_w_cnt = fi_cntr_read(write_cntrs[0]);
+	old_r_cnt = fi_cntr_read(read_cntrs[0]);
 
 	for (i = 0; i < iters; i++) {
-		sz = fi_read(ep[0], source, len,
-				loc_mr, gni_addr[1], (uint64_t)target,
-				mr_key, (void *)READ_CTX);
+		sz = fi_read(ep[0], source, len, loc_mr[0],
+			     gni_addr[0][1], (uint64_t)target,
+			     mr_key[1], (void *)READ_CTX);
 		cr_assert_eq(sz, 0);
 	}
 
-	fi_cntr_wait(read_cntr, old_r_cnt + iters, -1);
+	fi_cntr_wait(read_cntrs[0], old_r_cnt + iters, -1);
 
 	cr_assert(check_data(source, target, len), "Data mismatch");
 
-	new_w_cnt = fi_cntr_read(write_cntr);
+	new_w_cnt = fi_cntr_read(write_cntrs[0]);
 
 	/*
 	 * no fi_read called so old and new read cnts should be equal
@@ -535,8 +563,8 @@ Test(cntr, send_recv)
 	uint64_t old_r_cnt, new_r_cnt;
 	char s_buffer[128], r_buffer[128];
 
-	old_s_cnt = fi_cntr_read(write_cntr);
-	old_r_cnt = fi_cntr_read(rcv_cntr);
+	old_s_cnt = fi_cntr_read(write_cntrs[0]);
+	old_r_cnt = fi_cntr_read(recv_cntrs[1]);
 
 	for (i = 0; i < 16; i++) {
 		sprintf(s_buffer, "Hello there iter=%d", i);
@@ -545,14 +573,14 @@ Test(cntr, send_recv)
 			      r_buffer,
 			      sizeof(r_buffer),
 			      NULL,
-			      gni_addr[0],
+			      gni_addr[1][0],
 			      &r_context);
 		cr_assert_eq(ret, FI_SUCCESS, "fi_recv");
 		ret = fi_send(ep[0],
 			      s_buffer,
 			      strlen(s_buffer),
 			      NULL,
-			      gni_addr[1],
+			      gni_addr[0][1],
 			      &s_context);
 		cr_assert_eq(ret, FI_SUCCESS, "fi_send");
 
@@ -564,11 +592,11 @@ Test(cntr, send_recv)
 		got_r = (cqe.op_context == &r_context) ? 1 : 0;
 
 		if (got_r) {
-			new_r_cnt = fi_cntr_read(rcv_cntr);
+			new_r_cnt = fi_cntr_read(recv_cntrs[1]);
 			old_r_cnt++;
 			cr_assert(new_r_cnt == old_r_cnt);
 		} else {
-			new_s_cnt = fi_cntr_read(write_cntr);
+			new_s_cnt = fi_cntr_read(write_cntrs[0]);
 			old_s_cnt++;
 			cr_assert(new_s_cnt == old_s_cnt);
 		}
@@ -581,11 +609,11 @@ Test(cntr, send_recv)
 			cr_assert((cqe.op_context == &r_context), "fi_cq_read");
 
 		if (got_r) {
-			new_s_cnt = fi_cntr_read(write_cntr);
+			new_s_cnt = fi_cntr_read(write_cntrs[0]);
 			old_s_cnt++;
 			cr_assert(new_s_cnt == old_s_cnt);
 		} else {
-			new_r_cnt = fi_cntr_read(rcv_cntr);
+			new_r_cnt = fi_cntr_read(recv_cntrs[1]);
 			old_r_cnt++;
 			cr_assert(new_r_cnt == old_r_cnt);
 		}
@@ -608,63 +636,32 @@ struct tinfo {
 
 #define get_mark(i) ((char) (((i)%255)+0x31))
 
-static atomic_t cntr_test_next_tid;
+static ofi_atomic32_t cntr_test_next_tid;
 static __thread uint32_t cntr_test_tid = ~(uint32_t) 0;
 #define cntr_test_get_tid()						\
 	((cntr_test_tid  == ~(uint32_t) 0) ?				\
-	 atomic_inc(&cntr_test_next_tid) :				\
+	 ofi_atomic_inc32(&cntr_test_next_tid) :				\
 	 cntr_test_tid)
 
 
-static struct fid_cntr *ep_write_cntrs[NUM_EPS];
-static struct fid_cntr *ep_read_cntrs[NUM_EPS];
-
 static void cntr_setup_mt(void)
 {
-	int i, ret;
+	cr_assert(NUM_EPS >= NUM_THREADS);
 
-	cntr_setup_eps();
+	cntr_setup_eps(GNIX_EP_PRIMARY_CAPS);
 	cntr_setup_av();
-	cntr_setup_cqs();
-	cntr_setup_bind_cqs();
-
-	for (i = 0; i < NUM_EPS; i++) {
-		ret = fi_cntr_open(dom, &cntr_attr, &ep_write_cntrs[i], 0);
-		cr_assert(!ret, "fi_cntr_open");
-
-		ret = fi_ep_bind(ep[i], &ep_write_cntrs[i]->fid,
-				 FI_WRITE | FI_SEND);
-		cr_assert(!ret, "fi_ep_bind cntr");
-
-		ret = fi_cntr_open(dom, &cntr_attr, &ep_read_cntrs[i], 0);
-		cr_assert(!ret, "fi_cntr_open");
-
-		ret = fi_ep_bind(ep[i], &ep_read_cntrs[i]->fid, FI_READ);
-		cr_assert(!ret, "fi_ep_bind cntr");
-	}
-
+	cntr_setup_cntrs(FI_WRITE | FI_SEND, FI_READ, 0x0);
 	cntr_setup_enable_ep();
 	cntr_setup_mr();
 
-	atomic_initialize(&cntr_test_next_tid, 0);
+	ofi_atomic_initialize32(&cntr_test_next_tid, 0);
 }
 
 static void cntr_teardown_mt(void)
 {
-	int i, ret;
-
 	cntr_teardown_mr();
 	cntr_teardown_eps();
-	cntr_teardown_cqs();
-
-	for (i = 0; i < NUM_EPS; i++) {
-		ret = fi_close(&ep_write_cntrs[i]->fid);
-		cr_assert(!ret, "failure in closing write_cntr.");
-
-		ret = fi_close(&ep_read_cntrs[i]->fid);
-		cr_assert(!ret, "failure in closing read_cntr.");
-	}
-
+	cntr_teardown_cntrs();
 	cntr_teardown_fini();
 }
 
@@ -683,14 +680,15 @@ static void *do_thread_read_wait(void *data)
 
 	dbg_printf("%d: reading\n", tid);
 	for (i = 0; i < iters; i++) {
-		sz = fi_read(ep[tid], &source[tid*msg_size], msg_size, loc_mr,
-			     gni_addr[0], (uint64_t)&target[tid*msg_size],
-			     mr_key, (void *)(READ_CTX+i));
+		sz = fi_read(ep[tid], &source[tid*msg_size], msg_size,
+			     loc_mr[tid], gni_addr[tid][0],
+			     (uint64_t)&target[tid*msg_size],
+			     mr_key[0], (void *)(READ_CTX+i));
 		cr_assert_eq(sz, 0);
 	}
 
 	dbg_printf("%d: waiting\n", tid);
-	ret = fi_cntr_wait(ep_read_cntrs[tid], iters, -1);
+	ret = fi_cntr_wait(read_cntrs[tid], iters, -1);
 	cr_assert(ret == FI_SUCCESS);
 
 	dbg_printf("%d: done\n", tid);
@@ -700,31 +698,31 @@ static void *do_thread_read_wait(void *data)
 Test(cntr_mt, read_wait)
 {
 	int i, j;
-	pthread_t threads[NUM_EPS];
+	pthread_t threads[NUM_THREADS];
 	const int msg_size = 128;
 	struct tinfo info = { msg_size, 500 /* iters */};
 
-	cr_assert(NUM_EPS*msg_size <= BUF_SZ);
+	cr_assert(NUM_THREADS*msg_size <= BUF_SZ);
 
-	memset(source, 0, NUM_EPS*msg_size);
-	for (i = 0; i < NUM_EPS; i++) {
+	memset(source, 0, NUM_THREADS*msg_size);
+	for (i = 0; i < NUM_THREADS; i++) {
 		memset(&target[i*msg_size], get_mark(i), msg_size);
 	}
 
 	dbg_printf("creating threads\n");
-	for (i = 1; i < NUM_EPS; i++) {
+	for (i = 1; i < NUM_THREADS; i++) {
 		pthread_create(&threads[i], NULL, do_thread_read_wait, &info);
 	}
 
 	dbg_printf("joining\n");
 
-	for (i = 1; i < NUM_EPS; i++) {
+	for (i = 1; i < NUM_THREADS; i++) {
 		pthread_join(threads[i], NULL);
 	}
 
 	/* Must wait until all threads are done, since we don't know
 	 * which thread got which id */
-	for (i = 1; i < NUM_EPS; i++) {
+	for (i = 1; i < NUM_THREADS; i++) {
 		for (j = 0; j < msg_size; j++) {
 			cr_assert(source[i*msg_size+j] == get_mark(i));
 		}
@@ -746,14 +744,15 @@ static void *do_thread_write_wait(void *data)
 
 	dbg_printf("%d: writing\n", tid);
 	for (i = 0; i < iters; i++) {
-		sz = fi_write(ep[tid], &source[tid*msg_size], msg_size, loc_mr,
-			      gni_addr[0], (uint64_t)&target[tid*msg_size],
-			      mr_key, (void *)(READ_CTX+i));
+		sz = fi_write(ep[tid], &source[tid*msg_size], msg_size,
+			      loc_mr[tid], gni_addr[tid][0],
+			      (uint64_t)&target[tid*msg_size],
+			      mr_key[0], (void *)(READ_CTX+i));
 		cr_assert_eq(sz, 0);
 	}
 
 	dbg_printf("%d: waiting\n", tid);
-	ret = fi_cntr_wait(ep_write_cntrs[tid], iters, -1);
+	ret = fi_cntr_wait(write_cntrs[tid], iters, -1);
 	cr_assert(ret == FI_SUCCESS);
 
 	dbg_printf("%d: done\n", tid);
@@ -763,31 +762,31 @@ static void *do_thread_write_wait(void *data)
 Test(cntr_mt, write_wait)
 {
 	int i, j;
-	pthread_t threads[NUM_EPS];
+	pthread_t threads[NUM_THREADS];
 	const int msg_size = 128;
 	struct tinfo info = { msg_size, 500 /* iters */};
 
-	cr_assert(NUM_EPS*msg_size <= BUF_SZ);
+	cr_assert(NUM_THREADS*msg_size <= BUF_SZ);
 
-	memset(target, 0, NUM_EPS*msg_size);
-	for (i = 0; i < NUM_EPS; i++) {
+	memset(target, 0, NUM_THREADS*msg_size);
+	for (i = 0; i < NUM_THREADS; i++) {
 		memset(&source[i*msg_size], get_mark(i), msg_size);
 	}
 
 	dbg_printf("creating threads\n");
-	for (i = 1; i < NUM_EPS; i++) {
+	for (i = 1; i < NUM_THREADS; i++) {
 		pthread_create(&threads[i], NULL, do_thread_write_wait, &info);
 	}
 
 	dbg_printf("joining\n");
 
-	for (i = 1; i < NUM_EPS; i++) {
+	for (i = 1; i < NUM_THREADS; i++) {
 		pthread_join(threads[i], NULL);
 	}
 
 	/* Must wait until all threads are done, since we don't know
 	 * which thread got which id */
-	for (i = 1; i < NUM_EPS; i++) {
+	for (i = 1; i < NUM_THREADS; i++) {
 		for (j = 0; j < msg_size; j++) {
 			cr_assert(target[i*msg_size+j] == get_mark(i));
 		}
@@ -795,3 +794,192 @@ Test(cntr_mt, write_wait)
 
 	dbg_printf("done\n");
 }
+
+void *do_add_cntr_mt(void *arg)
+{
+	int i = 0, ret, iters = ((int *)arg)[0];
+
+	for (; i < iters; i++) {
+		ret = fi_cntr_add(write_cntrs[0], 1);
+		cr_assert(ret == FI_SUCCESS);
+	}
+
+	return NULL;
+}
+
+void *do_add_err_cntr_mt(void *arg)
+{
+	int i = 0, ret, iters = ((int *)arg)[0];
+
+	for (; i < iters; i++) {
+		ret = fi_cntr_adderr(write_cntrs[0], 1);
+		cr_assert(ret == FI_SUCCESS);
+	}
+
+	return NULL;
+}
+
+Test(cntr_mt, set_add_read_cntr)
+{
+	int iters = 128, nthreads = 4, i, ret;
+	uint64_t cntr_val;
+	pthread_t threads[nthreads];
+	void *pt_ret;
+
+	ret = fi_cntr_set(write_cntrs[0], 0);
+	cr_assert(ret == FI_SUCCESS);
+
+	cntr_val = fi_cntr_read(write_cntrs[0]);
+	cr_assert_eq(cntr_val, 0, "write error counter is incorrect.");
+
+	/* Create & Spawn threads */
+	for (i = 0; i < nthreads; i++) {
+		cr_assert(!pthread_create(threads + i, NULL, do_add_cntr_mt,
+					  (void *) &iters));
+	}
+
+	/* Wait until all threads are done */
+	for (i = 0; i < nthreads; i++) {
+		cr_assert(!pthread_join(threads[i], &pt_ret));
+		cr_assert(!pt_ret);
+	}
+
+	cntr_val = fi_cntr_read(write_cntrs[0]);
+	cr_assert_eq(cntr_val, (iters * nthreads), "write error counter "
+		"is incorrect.");
+}
+
+Test(cntr_mt, set_add_read_err_cntr)
+{
+	int iters = 128, nthreads = 4, i, ret;
+	uint64_t err_cntr_val;
+	pthread_t threads[nthreads];
+	void *pt_ret;
+
+	ret = fi_cntr_seterr(write_cntrs[0], 0);
+	cr_assert(ret == FI_SUCCESS);
+
+	err_cntr_val = fi_cntr_readerr(write_cntrs[0]);
+	cr_assert_eq(err_cntr_val, 0, "write error counter is incorrect.");
+
+	/* Create & Spawn threads */
+	for (i = 0; i < nthreads; i++) {
+		cr_assert(!pthread_create(threads + i, NULL, do_add_err_cntr_mt,
+					  (void *) &iters));
+	}
+
+	/* Wait until all threads are done */
+	for (i = 0; i < nthreads; i++) {
+		cr_assert(!pthread_join(threads[i], &pt_ret));
+		cr_assert(!pt_ret);
+	}
+
+	err_cntr_val = fi_cntr_readerr(write_cntrs[0]);
+	cr_assert_eq(err_cntr_val, (iters * nthreads), "write error counter "
+		"is incorrect.");
+}
+
+static void *do_thread_adderr_wait(void *data)
+{
+	int i, ret;
+	i = *((int *) data);
+
+	dbg_printf("%d: waiting\n", i);
+	ret = fi_cntr_wait(write_cntrs[i], ~0, -1);
+	cr_assert(ret != FI_SUCCESS, "Bad return value from fi_cntr_wait");
+
+	dbg_printf("%d: done\n", i);
+	return NULL;
+}
+
+Test(cntr_mt, adderr_wait)
+{
+	int i, ret;
+	pthread_t threads[NUM_THREADS];
+	int thread_args[NUM_THREADS];
+	void *pt_ret;
+
+	/* Each thread waits for an err cntr change on the i'th ep_write_cntr */
+	dbg_printf("creating threads\n");
+	for (i = 0; i < NUM_THREADS; i++) {
+		thread_args[i] = i;
+		cr_assert(!pthread_create(&threads[i], NULL,
+					  do_thread_adderr_wait, (void *) &thread_args[i]));
+	}
+
+	dbg_printf("Adding errors\n");
+	for (i = 0; i < NUM_THREADS; i++) {
+		ret = fi_cntr_adderr(write_cntrs[i], 1);
+		cr_assert(ret == FI_SUCCESS, "Bad return value from "
+			"fi_cntr_adderr");
+	}
+
+	for (i = 0; i < NUM_THREADS; i++) {
+		cr_assert(!pthread_join(threads[i], &pt_ret));
+		cr_assert(!pt_ret);
+	}
+
+	dbg_printf("done\n");
+}
+
+Test(cntr, adderr_wait)
+{
+	int ret;
+
+	ret = fi_cntr_adderr(write_cntrs[0], 1);
+	cr_assert(ret == FI_SUCCESS, "Bad return value from fi_cntr_adderr");
+
+	ret = fi_cntr_wait(write_cntrs[0], ~0, -1);
+	cr_assert(ret != FI_SUCCESS, "Bad return value from fi_cntr_wait");
+}
+
+Test(cntr, set_add_read_cntr)
+{
+	int iters = 128, ret, i = 0, init_val = 0xabcdefab;
+	uint64_t cntr_val, prev_cntr_val;
+
+	ret = fi_cntr_set(write_cntrs[0], init_val);
+	cr_assert(ret == FI_SUCCESS);
+
+	cntr_val = fi_cntr_read(write_cntrs[0]);
+	cr_assert_eq(cntr_val, init_val, "write or counter is incorrect.");
+
+	ret = fi_cntr_set(write_cntrs[0], 0);
+	cr_assert(ret == FI_SUCCESS);
+
+	for (; i < iters; i++) {
+		prev_cntr_val = fi_cntr_read(write_cntrs[0]);
+		ret = fi_cntr_add(write_cntrs[0], 1);
+		cr_assert(ret == FI_SUCCESS);
+
+		cntr_val = fi_cntr_read(write_cntrs[0]);
+		cr_assert_eq(cntr_val, prev_cntr_val + 1, "counter is "
+			"incorrect");
+	}
+}
+
+Test(cntr, set_add_read_err_cntr)
+{
+	int iters = 128, ret, i = 0,  init_val = 0xabcdefab;
+	uint64_t cntr_val, prev_cntr_val;
+
+	ret = fi_cntr_seterr(write_cntrs[0], init_val);
+	cr_assert(ret == FI_SUCCESS);
+
+	cntr_val = fi_cntr_readerr(write_cntrs[0]);
+	cr_assert_eq(cntr_val, init_val, "write or counter is incorrect.");
+
+	ret = fi_cntr_seterr(write_cntrs[0], 0);
+	cr_assert(ret == FI_SUCCESS);
+
+	for (; i < iters; i++) {
+		prev_cntr_val = fi_cntr_readerr(write_cntrs[0]);
+		ret = fi_cntr_adderr(write_cntrs[0], 1);
+		cr_assert(ret == FI_SUCCESS);
+
+		cntr_val = fi_cntr_readerr(write_cntrs[0]);
+		cr_assert_eq(cntr_val, prev_cntr_val + 1, "error counter is "
+			"incorrect");
+	}
+}
+
