@@ -1,6 +1,6 @@
 /*
- * Copyright (c) 2015 Los Alamos National Security, LLC. All rights reserved.
- * Copyright (c) 2015 Cray Inc.  All rights reserved.
+ * Copyright (c) 2015-2017 Los Alamos National Security, LLC. All rights reserved.
+ * Copyright (c) 2015-2017 Cray Inc.  All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -41,12 +41,6 @@
 #include <pthread.h>
 #include <sys/time.h>
 
-#include <rdma/fabric.h>
-#include <rdma/fi_domain.h>
-#include <rdma/fi_errno.h>
-#include <rdma/fi_endpoint.h>
-#include <rdma/fi_cm.h>
-#include <rdma/fi_tagged.h>
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -58,6 +52,8 @@
 #include "gnix_rma.h"
 
 #include <criterion/criterion.h>
+#include "gnix_rdma_headers.h"
+#include "common.h"
 
 /* Both the send and recv paths use independent state machines within
  * each test to simulate the behavior you would expect in a client/server
@@ -252,6 +248,8 @@ static char *recv_state_strings[R_STATE_INVALID + 1] = {
 struct timeval begin, end;
 struct timeval loop_start, loop_end;
 
+/* Note: Set to ~FI_NOTIFY_FLAGS_ONLY since this was written before api 1.5 */
+static uint64_t mode_bits = ~FI_NOTIFY_FLAGS_ONLY;
 static struct fid_fabric *fab;
 static struct fid_domain *dom;
 static struct fid_ep *ep[2];
@@ -268,7 +266,7 @@ static char *target;
 static char *source;
 static struct fid_mr *rem_mr, *loc_mr;
 static uint64_t mr_key;
-static const int max_test_time = 5;
+static const int max_test_time = 10;
 
 /* test variables */
 static int elapsed;
@@ -342,12 +340,13 @@ static void rdm_fi_pdc_setup(void)
 	hints = fi_allocinfo();
 	cr_assert(hints, "fi_allocinfo");
 
+	hints->domain_attr->mr_mode = GNIX_DEFAULT_MR_MODE;
 	hints->domain_attr->cq_data_size = 4;
-	hints->mode = ~0;
+	hints->mode = mode_bits;
 
-	hints->fabric_attr->name = strdup("gni");
+	hints->fabric_attr->prov_name = strdup("gni");
 
-	ret = fi_getinfo(FI_VERSION(1, 0), NULL, 0, 0, hints, &fi);
+	ret = fi_getinfo(fi_version(), NULL, 0, 0, hints, &fi);
 	cr_assert(!ret, "fi_getinfo");
 
 	ret = fi_fabric(fi->fabric_attr, &fab, NULL);
@@ -356,6 +355,7 @@ static void rdm_fi_pdc_setup(void)
 	ret = fi_domain(fab, fi, &dom, NULL);
 	cr_assert(!ret, "fi_domain");
 
+	memset(&attr, 0, sizeof(attr));
 	attr.type = FI_AV_MAP;
 	attr.count = 16;
 
@@ -618,8 +618,9 @@ static void validate_cqe_contents(
 	cr_assert_eq(entry->data, 0);
 
 	if (flags & FI_RECV) {
-		if (!(flags & FI_DISCARD))
+		if (!(flags & FI_DISCARD)) {
 			cr_assert_eq(entry->len, len);
+		}
 		else {
 			cr_assert_eq(entry->len, 0);
 			cr_assert_eq(entry->buf, NULL);
@@ -1769,11 +1770,12 @@ static void pdc_peek_event_present_small_buffer_provided(int len)
 {
 	/* Like pdc_peek_event_present_buffer_provided except uses an
 	 * undersized receive buffer with the FI_PEEK request. */
-	int ret;
+	int ret, i;
 	struct fi_msg_tagged msg;
 	struct iovec iov;
 	struct fi_cq_tagged_entry s_cqe;
 	struct fi_cq_tagged_entry d_cqe;
+	size_t cum_recv_len = 0;
 
 	rdm_fi_pdc_init_data(source, len, 0xab);
 	rdm_fi_pdc_init_data(target, len, 0);
@@ -1790,6 +1792,7 @@ static void pdc_peek_event_present_small_buffer_provided(int len)
 	build_peek_message(&peek_msg, &msg);
 	peek_iov.iov_len /= 2;
 
+
 	start_test_timer();
 	do {
 		PROGRESS_CQS(msg_cq);
@@ -1858,10 +1861,14 @@ static void pdc_peek_event_present_small_buffer_provided(int len)
 
 	ASSERT_SEND_RECV_DONE;
 
+	for (i = 0; i < peek_msg.iov_count; i++) {
+		cum_recv_len += peek_msg.msg_iov[i].iov_len;
+	}
+
 	/* validate the expected results */
-	validate_cqe_contents(&s_cqe, TSEND_FLAGS, source, len, len, target);
+	validate_cqe_contents(&s_cqe, TSEND_FLAGS, source, cum_recv_len, len, target);
 	validate_cqe_contents(&d_peek_cqe, TRECV_FLAGS | FI_PEEK,
-			peek_msg.msg_iov[0].iov_base, len, peek_msg.tag,
+			peek_msg.msg_iov[0].iov_base, cum_recv_len, peek_msg.tag,
 			peek_msg.context);
 
 	/* if CQE provided a buffer back, the data was copied.
@@ -1873,6 +1880,7 @@ static void pdc_peek_event_present_small_buffer_provided(int len)
 	}
 
 	validate_cqe_with_message(&d_cqe, &msg, TRECV_FLAGS);
+
 	cr_assert(rdm_fi_pdc_check_data(source, target, len), "Data mismatch");
 }
 
