@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2014 Intel Corporation. All rights reserved.
+ * Copyright (c) 2013-2017 Intel Corporation. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -32,8 +32,12 @@
 
 #include "psmx.h"
 #include "prov.h"
+#include <glob.h>
 
 static int psmx_init_count = 0;
+static int psmx_lib_initialized = 0;
+static pthread_mutex_t psmx_lib_mutex; 
+static int psmx_compat_lib = 0;
 
 struct psmx_env psmx_env = {
 	.name_server	= 1,
@@ -63,6 +67,57 @@ static void psmx_init_env(void)
 	fi_param_get_str(&psmx_prov, "prog_affinity", &psmx_env.prog_affinity);
 }
 
+static int psmx_init_lib(void)
+{
+	int major, minor;
+	int ret = 0, err;
+
+	if (psmx_lib_initialized)
+		return 0;
+
+	pthread_mutex_lock(&psmx_lib_mutex);
+
+	if (psmx_lib_initialized)
+		goto out;
+
+	psm_error_register_handler(NULL, PSM_ERRHANDLER_NO_HANDLER);
+
+	major = PSM_VERNO_MAJOR;
+	minor = PSM_VERNO_MINOR;
+
+	err = psm_init(&major, &minor);
+	if (err != PSM_OK) {
+		FI_WARN(&psmx_prov, FI_LOG_CORE,
+			"psm_init failed: %s\n", psm_error_get_string(err));
+		ret = err;
+		goto out;
+	}
+
+	FI_INFO(&psmx_prov, FI_LOG_CORE,
+		"PSM header version = (%d, %d)\n", PSM_VERNO_MAJOR, PSM_VERNO_MINOR);
+	FI_INFO(&psmx_prov, FI_LOG_CORE,
+		"PSM library version = (%d, %d)\n", major, minor);
+
+	if (major != PSM_VERNO_MAJOR) {
+		psmx_am_compat_mode = 1;
+		FI_INFO(&psmx_prov, FI_LOG_CORE,
+			"PSM AM compat mode enabled: appliation %d.%d, library %d.%d.\n",
+			PSM_VERNO_MAJOR, PSM_VERNO_MINOR, major, minor);
+	}
+
+	if (major > 1) {
+		psmx_compat_lib = 1;
+		FI_INFO(&psmx_prov, FI_LOG_CORE,
+			"PSM is supported via the psm2-compat library over PSM2.\n");
+	}
+
+	psmx_lib_initialized = 1;
+
+out:
+	pthread_mutex_unlock(&psmx_lib_mutex);
+	return ret;
+}
+
 static int psmx_reserve_tag_bits(int *caps, uint64_t *max_tag_value)
 {
 	uint64_t reserved_bits = 0;
@@ -125,7 +180,8 @@ static int psmx_getinfo(uint32_t version, const char *node, const char *service,
 {
 	struct fi_info *psmx_info;
 	uint32_t cnt = 0;
-	void *dest_addr = NULL;
+	psm_epid_t *dest_addr = NULL;
+	struct psmx_src_name *src_addr;
 	int ep_type = FI_EP_RDM;
 	int av_type = FI_AV_UNSPEC;
 	uint64_t mode = FI_CONTEXT;
@@ -136,11 +192,44 @@ static int psmx_getinfo(uint32_t version, const char *node, const char *service,
 	int caps = 0;
 	uint64_t max_tag_value = 0;
 	int err = -FI_ENODATA;
+	int svc0, svc = PSMX_ANY_SERVICE;
 
 	FI_INFO(&psmx_prov, FI_LOG_CORE,"\n");
 
 	*info = NULL;
 
+	if (FI_VERSION_GE(version, FI_VERSION(1,5)))
+		mr_mode = 0;
+
+	if (psmx_init_lib())
+		return -FI_ENODATA;
+
+	if (psmx_compat_lib) {
+		/*
+		 * native PSM running over TrueScale doesn't have the issue handled
+		 * here. it's only present when PSM is supported via the psm2-compat
+		 * library, where the PSM functions are just wrappers around the PSM2
+		 * counterparts.
+		 *
+		 * psm2_ep_num_devunits() may wait for 15 seconds before return
+		 * when /dev/hfi1_0 is not present. Check the existence of any hfi1
+		 * device interface first to avoid this delay. Note that the devices
+		 * don't necessarily appear consecutively so we need to check all
+		 * possible device names before returning "no device found" error.
+		 * This also means if "/dev/hfi1_0" doesn't exist but other devices
+		 * exist, we are still going to see the delay; but that's a rare case.
+		 */
+		glob_t glob_buf;
+
+		if ((glob("/dev/hfi1_[0-9]", 0, NULL, &glob_buf) != 0) &&
+		    (glob("/dev/hfi1_[0-9][0-9]", GLOB_APPEND, NULL, &glob_buf) != 0)) {
+			FI_INFO(&psmx_prov, FI_LOG_CORE,
+				"no hfi1 device is found.\n");
+			return -FI_ENODATA;
+		}
+		globfree(&glob_buf);
+	}
+
 	if (psm_ep_num_devunits(&cnt) || !cnt) {
 		FI_INFO(&psmx_prov, FI_LOG_CORE,
 			"no PSM device is found.\n");
@@ -149,8 +238,59 @@ static int psmx_getinfo(uint32_t version, const char *node, const char *service,
 
 	psmx_init_env();
 
-	if (node && !(flags & FI_SOURCE))
-		dest_addr = psmx_resolve_name(node, 0);
+	src_addr = calloc(1, sizeof(*src_addr));
+	if (!src_addr) {
+		FI_INFO(&psmx_prov, FI_LOG_CORE,
+			"failed to allocate src addr.\n");
+		return -FI_ENODATA;
+	}
+	src_addr->signature = 0xFFFF;
+	src_addr->unit = PSMX_DEFAULT_UNIT;
+	src_addr->port = PSMX_DEFAULT_PORT;
+	src_addr->service = PSMX_ANY_SERVICE;
+
+	if (flags & FI_SOURCE) {
+		if (node)
+			sscanf(node, "%*[^:]:%" SCNi8 ":%" SCNu8, &src_addr->unit, &src_addr->port);
+		if (service)
+			sscanf(service, "%" SCNu32, &src_addr->service);
+		FI_INFO(&psmx_prov, FI_LOG_CORE,
+			"node '%s' service '%s' converted to <unit=%d, port=%d, service=%d>\n",
+			node, service, src_addr->unit, src_addr->port, src_addr->service);
+	} else if (node) {
+		struct util_ns ns = (const struct util_ns){ 0 };
+		struct util_ns_attr ns_attr = (const struct util_ns_attr){ 0 };
+		psm_uuid_t uuid;
+
+		psmx_get_uuid(uuid);
+		ns_attr.ns_port = psmx_uuid_to_port(uuid);
+		ns_attr.name_len = sizeof(*dest_addr);
+		ns_attr.service_len = sizeof(svc);
+		ns_attr.service_cmp = psmx_ns_service_cmp;
+		ns_attr.is_service_wildcard = psmx_ns_is_service_wildcard;
+		err = ofi_ns_init(&ns_attr, &ns);
+		if (err) {
+			FI_INFO(&psmx_prov, FI_LOG_CORE,
+				"ofi_ns_init returns %d\n", err);
+			err = -FI_ENODATA;
+			goto err_out;
+		}
+
+		if (service)
+			svc = atoi(service);
+		svc0 = svc;
+		dest_addr = (psm_epid_t *)ofi_ns_resolve_name(&ns, node, &svc);
+		if (dest_addr) {
+			FI_INFO(&psmx_prov, FI_LOG_CORE,
+				"'%s:%u' resolved to <epid=%"PRIu64">:%u\n",
+				node, svc0, *dest_addr, svc);
+		} else {
+			FI_INFO(&psmx_prov, FI_LOG_CORE,
+				"failed to resolve '%s:%u'.\n", node, svc);
+			err = -FI_ENODATA;
+			goto err_out;
+		}
+	}
 
 	if (hints) {
 		switch (hints->addr_format) {
@@ -190,16 +330,17 @@ static int psmx_getinfo(uint32_t version, const char *node, const char *service,
 				goto err_out;
 			}
 
-			if (hints->ep_attr->tx_ctx_cnt > 1) {
+			if (hints->ep_attr->tx_ctx_cnt > 1 &&
+			    hints->ep_attr->tx_ctx_cnt != FI_SHARED_CONTEXT) {
 				FI_INFO(&psmx_prov, FI_LOG_CORE,
-					"hints->ep_attr->tx_ctx_cnt=%d, supported=0,1\n",
+					"hints->ep_attr->tx_ctx_cnt=%"PRIu64", supported=0,1\n",
 					hints->ep_attr->tx_ctx_cnt);
 				goto err_out;
 			}
 
 			if (hints->ep_attr->rx_ctx_cnt > 1) {
 				FI_INFO(&psmx_prov, FI_LOG_CORE,
-					"hints->ep_attr->rx_ctx_cnt=%d, supported=0,1\n",
+					"hints->ep_attr->rx_ctx_cnt=%"PRIu64", supported=0,1\n",
 					hints->ep_attr->rx_ctx_cnt);
 				goto err_out;
 			}
@@ -207,26 +348,32 @@ static int psmx_getinfo(uint32_t version, const char *node, const char *service,
 
 		if ((hints->caps & PSMX_CAPS) != hints->caps &&
 		    (hints->caps & PSMX_CAPS2) != hints->caps) {
-			FI_INFO(&psmx_prov, FI_LOG_CORE,
-				"hints->caps=0x%llx, supported=0x%llx,0x%llx\n",
-				hints->caps, PSMX_CAPS, PSMX_CAPS2);
+			uint64_t psmx_caps = PSMX_CAPS;
+			uint64_t psmx_caps2 = PSMX_CAPS2;
+			FI_INFO(&psmx_prov, FI_LOG_CORE, "hints->caps: %s,\n"
+				"supported = %s and %s.\n",
+				fi_tostr(&hints->caps, FI_TYPE_CAPS),
+				fi_tostr(&psmx_caps, FI_TYPE_CAPS),
+				fi_tostr(&psmx_caps2, FI_TYPE_CAPS));
 			goto err_out;
 		}
 
 		if (hints->tx_attr) {
 			if ((hints->tx_attr->op_flags & PSMX_OP_FLAGS) !=
 			    hints->tx_attr->op_flags) {
-				FI_INFO(&psmx_prov, FI_LOG_CORE,
-					"hints->tx->flags=0x%llx, "
-					"supported=0x%llx\n",
-					hints->tx_attr->op_flags,
-					PSMX_OP_FLAGS);
+				uint64_t psmx_op_flags = PSMX_OP_FLAGS;
+				FI_INFO(&psmx_prov, FI_LOG_CORE, "hints->tx->flags=%s,\n"
+					"supported=%s.\n",
+					fi_tostr(&hints->tx_attr->op_flags,
+						 FI_TYPE_OP_FLAGS),
+					fi_tostr(&psmx_op_flags,
+						 FI_TYPE_OP_FLAGS));
 				goto err_out;
 			}
 			if (hints->tx_attr->inject_size > PSMX_INJECT_SIZE) {
 				FI_INFO(&psmx_prov, FI_LOG_CORE,
-					"hints->tx_attr->inject_size=%ld,"
-					"supported=%ld.\n",
+					"hints->tx_attr->inject_size=%"PRIu64","
+					"supported=%d.\n",
 					hints->tx_attr->inject_size,
 					PSMX_INJECT_SIZE);
 				goto err_out;
@@ -236,18 +383,24 @@ static int psmx_getinfo(uint32_t version, const char *node, const char *service,
 		if (hints->rx_attr &&
 		    (hints->rx_attr->op_flags & PSMX_OP_FLAGS) !=
 		     hints->rx_attr->op_flags) {
-			FI_INFO(&psmx_prov, FI_LOG_CORE,
-				"hints->rx->flags=0x%llx, supported=0x%llx\n",
-				hints->rx_attr->op_flags, PSMX_OP_FLAGS);
+			uint64_t psmx_op_flags = PSMX_OP_FLAGS;
+			FI_INFO(&psmx_prov, FI_LOG_CORE, "hints->rx->flags=%s,\n"
+				"supported=%s.\n",
+				fi_tostr(&hints->rx_attr->op_flags,
+					 FI_TYPE_OP_FLAGS),
+				fi_tostr(&psmx_op_flags,
+					 FI_TYPE_OP_FLAGS));
 			goto err_out;
 		}
 
 		if ((hints->caps & FI_TAGGED) ||
 		    ((hints->caps & FI_MSG) && !psmx_env.am_msg)) {
 			if ((hints->mode & FI_CONTEXT) != FI_CONTEXT) {
+				uint64_t psmx_mode = FI_CONTEXT;
 				FI_INFO(&psmx_prov, FI_LOG_CORE,
-					"hints->mode=0x%llx, required=0x%llx\n",
-					hints->mode, FI_CONTEXT);
+					"hints->mode=%s,\n required=%s.\n",
+					fi_tostr(&hints->mode, FI_TYPE_MODE),
+					fi_tostr(&psmx_mode, FI_TYPE_MODE));
 				goto err_out;
 			}
 		} else {
@@ -285,18 +438,14 @@ static int psmx_getinfo(uint32_t version, const char *node, const char *service,
 				goto err_out;
 			}
 
-			switch (hints->domain_attr->mr_mode) {
-			case FI_MR_UNSPEC:
-				break;
-			case FI_MR_BASIC:
-			case FI_MR_SCALABLE:
-				mr_mode = hints->domain_attr->mr_mode;
-				break;
-			default:
+			if (hints->domain_attr->mr_mode == FI_MR_BASIC) {
+				mr_mode = FI_MR_BASIC;
+			} else if (hints->domain_attr->mr_mode == FI_MR_SCALABLE) {
+				mr_mode = FI_MR_SCALABLE;
+			} else if (hints->domain_attr->mr_mode & (FI_MR_BASIC | FI_MR_SCALABLE)) {
 				FI_INFO(&psmx_prov, FI_LOG_CORE,
-					"hints->domain_attr->mr_mode=%d, supported=%d %d %d\n",
-					hints->domain_attr->mr_mode, FI_MR_UNSPEC, FI_MR_BASIC,
-					FI_MR_SCALABLE);
+					"hints->domain_attr->mr_mode has FI_MR_BASIC or FI_MR_SCALABLE "
+					"combined with other bits\n");
 				goto err_out;
 			}
 
@@ -347,13 +496,20 @@ static int psmx_getinfo(uint32_t version, const char *node, const char *service,
 					FI_PROGRESS_MANUAL, FI_PROGRESS_AUTO);
 				goto err_out;
 			}
+
+			if (hints->domain_attr->caps & FI_SHARED_AV) {
+				FI_INFO(&psmx_prov, FI_LOG_CORE,
+					"hints->domain_attr->caps=%lx, shared AV is unsupported\n",
+					hints->domain_attr->caps);
+				goto err_out;
+			}
 		}
 
 		if (hints->ep_attr) {
 			if (hints->ep_attr->max_msg_size > PSMX_MAX_MSG_SIZE) {
 				FI_INFO(&psmx_prov, FI_LOG_CORE,
-					"hints->ep_attr->max_msg_size=%ld,"
-					"supported=%ld.\n",
+					"hints->ep_attr->max_msg_size=%"PRIu64","
+					"supported=%llu.\n",
 					hints->ep_attr->max_msg_size,
 					PSMX_MAX_MSG_SIZE);
 				goto err_out;
@@ -364,20 +520,20 @@ static int psmx_getinfo(uint32_t version, const char *node, const char *service,
 		if (hints->tx_attr) {
 			if ((hints->tx_attr->msg_order & PSMX_MSG_ORDER) !=
 			    hints->tx_attr->msg_order) {
+				uint64_t psmx_msg_order = PSMX_MSG_ORDER;
 				FI_INFO(&psmx_prov, FI_LOG_CORE,
-					"hints->tx_attr->msg_order=%lx,"
-					"supported=%lx.\n",
-					hints->tx_attr->msg_order,
-					PSMX_MSG_ORDER);
+					"hints->tx_attr->msg_order=%s,\n supported=%s.\n",
+					fi_tostr(&hints->tx_attr->msg_order, FI_TYPE_MSG_ORDER),
+					fi_tostr(&psmx_msg_order, FI_TYPE_MSG_ORDER));
 				goto err_out;
 			}
 			if ((hints->tx_attr->comp_order & PSMX_COMP_ORDER) !=
 			    hints->tx_attr->comp_order) {
+				uint64_t psmx_comp_order = PSMX_COMP_ORDER;
 				FI_INFO(&psmx_prov, FI_LOG_CORE,
-					"hints->tx_attr->msg_order=%lx,"
-					"supported=%lx.\n",
-					hints->tx_attr->comp_order,
-					PSMX_COMP_ORDER);
+					"hints->tx_attr->comp_order=%s,\n supported=%s.\n",
+					fi_tostr(&hints->tx_attr->comp_order, FI_TYPE_MSG_ORDER),
+					fi_tostr(&psmx_comp_order, FI_TYPE_MSG_ORDER));
 				goto err_out;
 			}
 			if (hints->tx_attr->inject_size > PSMX_INJECT_SIZE) {
@@ -407,20 +563,20 @@ static int psmx_getinfo(uint32_t version, const char *node, const char *service,
 		if (hints->rx_attr) {
 			if ((hints->rx_attr->msg_order & PSMX_MSG_ORDER) !=
 			    hints->rx_attr->msg_order) {
+				uint64_t psmx_msg_order = PSMX_MSG_ORDER;
 				FI_INFO(&psmx_prov, FI_LOG_CORE,
-					"hints->rx_attr->msg_order=%lx,"
-					"supported=%lx.\n",
-					hints->rx_attr->msg_order,
-					PSMX_MSG_ORDER);
+					"hints->rx_attr->msg_order=%s,\n supported=%s.\n",
+					fi_tostr(&hints->rx_attr->msg_order, FI_TYPE_MSG_ORDER),
+					fi_tostr(&psmx_msg_order, FI_TYPE_MSG_ORDER));
 				goto err_out;
 			}
 			if ((hints->rx_attr->comp_order & PSMX_COMP_ORDER) !=
 			    hints->rx_attr->comp_order) {
+				uint64_t psmx_comp_order = PSMX_COMP_ORDER;
 				FI_INFO(&psmx_prov, FI_LOG_CORE,
-					"hints->rx_attr->msg_order=%lx,"
-					"supported=%lx.\n",
-					hints->rx_attr->comp_order,
-					PSMX_COMP_ORDER);
+					"hints->rx_attr->comp_order=%s,\n supported=%s.\n",
+					fi_tostr(&hints->rx_attr->comp_order, FI_TYPE_MSG_ORDER),
+					fi_tostr(&psmx_comp_order, FI_TYPE_MSG_ORDER));
 				goto err_out;
 			}
 			if (hints->rx_attr->iov_limit > 1) {
@@ -467,21 +623,27 @@ static int psmx_getinfo(uint32_t version, const char *node, const char *service,
 	psmx_info->domain_attr->ep_cnt = 65535;
 	psmx_info->domain_attr->tx_ctx_cnt = 1;
 	psmx_info->domain_attr->rx_ctx_cnt = 1;
-	psmx_info->domain_attr->max_ep_tx_ctx = 65535;
+	psmx_info->domain_attr->max_ep_tx_ctx = 1;
 	psmx_info->domain_attr->max_ep_rx_ctx = 1;
 	psmx_info->domain_attr->max_ep_stx_ctx = 65535;
 	psmx_info->domain_attr->max_ep_srx_ctx = 0;
+	psmx_info->domain_attr->cntr_cnt = 65535;
+	psmx_info->domain_attr->mr_iov_limit = 65535;
+	psmx_info->domain_attr->caps = PSMX_DOM_CAPS;
+	psmx_info->domain_attr->mode = 0;
+	psmx_info->domain_attr->mr_cnt = 65535;
 
 	psmx_info->next = NULL;
 	psmx_info->caps = (hints && hints->caps) ? hints->caps : caps;
 	psmx_info->mode = mode;
 	psmx_info->addr_format = FI_ADDR_PSMX;
-	psmx_info->src_addrlen = 0;
-	psmx_info->dest_addrlen = sizeof(psm_epid_t);
-	psmx_info->src_addr = NULL;
+	psmx_info->src_addr = src_addr;
+	psmx_info->src_addrlen = sizeof(*src_addr);
 	psmx_info->dest_addr = dest_addr;
+	psmx_info->dest_addrlen = sizeof(*dest_addr);
 	psmx_info->fabric_attr->name = strdup(PSMX_FABRIC_NAME);
 	psmx_info->fabric_attr->prov_name = NULL;
+	psmx_info->fabric_attr->prov_version = PSMX_VERSION;
 
 	psmx_info->tx_attr->caps = psmx_info->caps;
 	psmx_info->tx_attr->mode = psmx_info->mode;
@@ -508,8 +670,8 @@ static int psmx_getinfo(uint32_t version, const char *node, const char *service,
 	return 0;
 
 err_out:
-	if (dest_addr)
-		free(dest_addr);
+	free(dest_addr);
+	free(src_addr);
 
 	return err;
 }
@@ -518,7 +680,7 @@ static void psmx_fini(void)
 {
 	FI_INFO(&psmx_prov, FI_LOG_CORE, "\n");
 
-	if (! --psmx_init_count) {
+	if (! --psmx_init_count && psmx_lib_initialized) {
 		/* This function is called from a library destructor, which is called
 		 * automatically when exit() is called. The call to psm_finalize()
 		 * might cause deadlock if the applicaiton is terminated with Ctrl-C
@@ -526,18 +688,20 @@ static void psmx_fini(void)
 		 * psm_finalize() tries to acquire. This can be avoided by only
 		 * calling psm_finalize() when PSM is guaranteed to be unused.
 		 */
-		if (psmx_active_fabric)
+		if (psmx_active_fabric) {
 			FI_INFO(&psmx_prov, FI_LOG_CORE,
 				"psmx_active_fabric != NULL, skip psm_finalize\n");
-		else
+		} else {
 			psm_finalize();
+			psmx_lib_initialized = 0;
+		}
 	}
 }
 
 struct fi_provider psmx_prov = {
 	.name = PSMX_PROV_NAME,
-	.version = FI_VERSION(0, 9),
-	.fi_version = FI_VERSION(1, 3),
+	.version = PSMX_VERSION,
+	.fi_version = PSMX_VERSION,
 	.getinfo = psmx_getinfo,
 	.fabric = psmx_fabric,
 	.cleanup = psmx_fini
@@ -545,9 +709,6 @@ struct fi_provider psmx_prov = {
 
 PROVIDER_INI
 {
-	int major, minor;
-	int err;
-
 	FI_INFO(&psmx_prov, FI_LOG_CORE, "\n");
 
 	fi_param_define(&psmx_prov, "name_server", FI_PARAM_BOOL,
@@ -588,30 +749,7 @@ PROVIDER_INI
 			"(when >=0) or core_id - num_cores (when <0). "
 			"(default: affinity not set)");
 
-        psm_error_register_handler(NULL, PSM_ERRHANDLER_NO_HANDLER);
-
-	major = PSM_VERNO_MAJOR;
-	minor = PSM_VERNO_MINOR;
-
-        err = psm_init(&major, &minor);
-	if (err != PSM_OK) {
-		FI_WARN(&psmx_prov, FI_LOG_CORE,
-			"psm_init failed: %s\n", psm_error_get_string(err));
-		return NULL;
-	}
-
-	FI_INFO(&psmx_prov, FI_LOG_CORE,
-		"PSM header version = (%d, %d)\n", PSM_VERNO_MAJOR, PSM_VERNO_MINOR);
-	FI_INFO(&psmx_prov, FI_LOG_CORE,
-		"PSM library version = (%d, %d)\n", major, minor);
-
-	if (major != PSM_VERNO_MAJOR) {
-		psmx_am_compat_mode = 1;
-		FI_INFO(&psmx_prov, FI_LOG_CORE,
-			"PSM AM compat mode enabled: appliation %d.%d, library %d.%d.\n",
-			PSM_VERNO_MAJOR, PSM_VERNO_MINOR, major, minor);
-	}
-
+	pthread_mutex_init(&psmx_lib_mutex, NULL);
 	psmx_init_count++;
 	return (&psmx_prov);
 }
