@@ -1,5 +1,6 @@
 /*
- * Copyright (c) 2013-2014 Intel Corporation. All rights reserved.
+ * Copyright (c) 2013-2017 Intel Corporation. All rights reserved.
+ * Copyright (c) 2016 Cisco Systems, Inc. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -47,13 +48,14 @@ extern "C" {
 #include <errno.h>
 #include <unistd.h>
 #include <fcntl.h>
+#include <inttypes.h>
 #include <pthread.h>
 #include <sys/types.h>
+#include <sys/stat.h>
 #include <sys/socket.h>
 #include <netdb.h>
 #include <complex.h>
 #include <rdma/fabric.h>
-#include <rdma/fi_prov.h>
 #include <rdma/fi_domain.h>
 #include <rdma/fi_endpoint.h>
 #include <rdma/fi_tagged.h>
@@ -62,8 +64,8 @@ extern "C" {
 #include <rdma/fi_trigger.h>
 #include <rdma/fi_cm.h>
 #include <rdma/fi_errno.h>
-#include <rdma/fi_log.h>
 #include "fi.h"
+#include "ofi_atomic.h"
 #include "fi_enosys.h"
 #include "fi_list.h"
 #include "fi_util.h"
@@ -72,7 +74,7 @@ extern "C" {
 
 extern struct fi_provider psmx2_prov;
 
-#define PSMX2_VERSION	(FI_VERSION(1,3))
+#define PSMX2_VERSION	(FI_VERSION(1,5))
 
 #define PSMX2_OP_FLAGS	(FI_INJECT | FI_MULTI_RECV | FI_COMPLETION | \
 			 FI_TRIGGER | FI_INJECT_COMPLETE | \
@@ -82,12 +84,17 @@ extern struct fi_provider psmx2_prov;
 			 FI_RMA | FI_MULTI_RECV | \
                          FI_READ | FI_WRITE | FI_SEND | FI_RECV | \
                          FI_REMOTE_READ | FI_REMOTE_WRITE | \
-			 FI_TRIGGER | FI_RMA_EVENT | \
-			 FI_REMOTE_CQ_DATA | FI_SOURCE | FI_DIRECTED_RECV)
+			 FI_TRIGGER | FI_RMA_EVENT | FI_REMOTE_CQ_DATA | \
+			 FI_SOURCE | FI_SOURCE_ERR | FI_DIRECTED_RECV | \
+			 FI_NAMED_RX_CTX)
 
 #define PSMX2_SUB_CAPS	(FI_READ | FI_WRITE | FI_REMOTE_READ | FI_REMOTE_WRITE | \
 			 FI_SEND | FI_RECV)
 
+#define PSMX2_DOM_CAPS	(FI_LOCAL_COMM | FI_REMOTE_COMM)
+
+#define PSMX2_MAX_TRX_CTXT	(80)
+#define PSMX2_ALL_TRX_CTXT	((void *)-1)
 #define PSMX2_MAX_MSG_SIZE	((0x1ULL << 32) - 1)
 #define PSMX2_INJECT_SIZE	(64)
 #define PSMX2_MSG_ORDER		FI_ORDER_SAS
@@ -96,6 +103,7 @@ extern struct fi_provider psmx2_prov;
 #define PSMX2_MSG_BIT	(0x80000000)
 #define PSMX2_RMA_BIT	(0x40000000)
 #define PSMX2_IOV_BIT	(0x20000000)
+#define PSMX2_IMM_BIT	(0x10000000)
 #define PSMX2_SEQ_BITS	(0x0FFF0000)
 #define PSMX2_SRC_BITS	(0x0000FF00)
 #define PSMX2_DST_BITS	(0x000000FF)
@@ -113,18 +121,44 @@ extern struct fi_provider psmx2_prov;
 						tag96.tag1 = (uint32_t)(tag64>>32); \
 						tag96.tag2 = tag32; \
 					} while (0)
+#define PSMX2_SET_TAG_FIRST64(tag96,tag64) do { \
+						memcpy(&(tag96).tag0, &(tag64), sizeof(uint64_t)); \
+					} while (0)
+#define PSMX2_SET_TAG_LAST32(tag96,tag32) do { \
+						tag96.tag2 = tag32; \
+					} while (0)
 
-#define PSMX2_GET_TAG64(tag96)		(tag96.tag0 | ((uint64_t)tag96.tag1<<32))
+#define PSMX2_GET_TAG64(tag96)		(psmx2_get_tag64(&(tag96)))
 
-/* Canonical virtual address on X86_64 only uses 48 bits and the higher 16 bits
- * are sign extensions. We can put vlane into part of these 16 bits of an epaddr.
+static inline uint64_t psmx2_get_tag64(const psm2_mq_tag_t *tag96)
+{
+	uint64_t tag64;
+
+	memcpy(&tag64, &tag96->tag0, sizeof(tag64));
+	return tag64;
+}
+
+/* When using the long RMA protocol, set a bit in the unused SEQ bits to
+ * indicate whether or not the operation is a read or a write. This prevents tag
+ * collisions. */
+#define PSMX2_TAG32_LONG_WRITE(tag32) PSMX2_TAG32_SET_SEQ(tag32, 0x1)
+#define PSMX2_TAG32_LONG_READ(tag32)  PSMX2_TAG32_SET_SEQ(tag32, 0x2)
+
+/*
+ * Canonical virtual address on X86_64 only uses 48 bits and the higher 16 bits
+ * are sign extensions. We can put some extra information into the 16 bits.
+ *
+ * Here is the layout:  AA-B-C-DDDDDDDDDDDD
+ *
+ * C == 0xE: scalable endpoint, AAB is context index, DDDDDDDDDDDD is the address
+ * C != 0xE: regular endpoint, AA is vlane, BCDDDDDDDDDDDD is epaddr
  */
 #define PSMX2_MAX_VL			(0xFF)
 #define PSMX2_EP_MASK			(0x00FFFFFFFFFFFFFFUL)
 #define PSMX2_SIGN_MASK  		(0x0080000000000000UL)
 #define PSMX2_SIGN_EXT			(0xFF00000000000000UL)
 #define PSMX2_VL_MASK			(0xFF00000000000000UL)
- 
+
 #define PSMX2_EP_TO_ADDR(ep,vl)		((((uint64_t)vl) << 56) | \
 						((uint64_t)ep & PSMX2_EP_MASK))
 #define PSMX2_ADDR_TO_VL(addr)		((uint8_t)((addr & PSMX2_VL_MASK) >> 56))
@@ -133,10 +167,20 @@ extern struct fi_provider psmx2_prov;
                                                  (addr | PSMX2_SIGN_EXT) : \
                                                  (addr & PSMX2_EP_MASK)))
 
+#define PSMX2_MAX_RX_CTX_BITS		(12)
+#define PSMX2_SEP_ADDR_FLAG		(0x000E000000000000UL)
+#define PSMX2_SEP_ADDR_MASK		(0x000F000000000000UL)
+#define PSMX2_SEP_CTXT_MASK		(0xFFF0000000000000UL)
+#define PSMX2_SEP_IDX_MASK		(0x0000FFFFFFFFFFFFUL)
+#define PSMX2_SEP_ADDR_TEST(addr)	(((addr) & PSMX2_SEP_ADDR_MASK) == PSMX2_SEP_ADDR_FLAG)
+#define PSMX2_SEP_ADDR_CTXT(addr, ctxt_bits) \
+					(((addr) & PSMX2_SEP_CTXT_MASK) >> (64-(ctxt_bits)))
+#define PSMX2_SEP_ADDR_IDX(addr)	((addr) & PSMX2_SEP_IDX_MASK)
+
 /* Bits 60 .. 63 of the flag are provider specific */
 #define PSMX2_NO_COMPLETION	(1ULL << 60)
 
-#define PSMX2_CTXT_ALLOC_FLAG		0x80000000
+
 enum psmx2_context_type {
 	PSMX2_NOCOMP_SEND_CONTEXT = 1,
 	PSMX2_NOCOMP_RECV_CONTEXT,
@@ -154,7 +198,7 @@ enum psmx2_context_type {
 	PSMX2_SENDV_CONTEXT,
 	PSMX2_IOV_SEND_CONTEXT,
 	PSMX2_IOV_RECV_CONTEXT,
-	PSMX2_NOCOMP_RECV_CONTEXT_ALLOC = PSMX2_NOCOMP_RECV_CONTEXT | PSMX2_CTXT_ALLOC_FLAG,
+	PSMX2_NOCOMP_RECV_CONTEXT_ALLOC
 };
 
 struct psmx2_context {
@@ -175,6 +219,7 @@ union psmx2_pi {
 
 #define PSMX2_AM_RMA_HANDLER	0
 #define PSMX2_AM_ATOMIC_HANDLER	1
+#define PSMX2_AM_SEP_HANDLER	2
 
 #define PSMX2_AM_OP_MASK	0x000000FF
 #define PSMX2_AM_DST_MASK	0x0000FF00
@@ -207,13 +252,16 @@ enum {
 	PSMX2_AM_REQ_ATOMIC_COMPWRITE,
 	PSMX2_AM_REP_ATOMIC_COMPWRITE,
 	PSMX2_AM_REQ_WRITEV,
+	PSMX2_AM_REQ_READV,
+	PSMX2_AM_REQ_SEP_QUERY,
+	PSMX2_AM_REP_SEP_QUERY,
 };
 
 struct psmx2_am_request {
 	int op;
 	union {
 		struct {
-			void	*buf;
+			uint8_t	*buf;
 			size_t	len;
 			uint64_t addr;
 			uint64_t key;
@@ -224,7 +272,10 @@ struct psmx2_am_request {
 			uint64_t data;
 		} write;
 		struct {
-			void	*buf;
+			union {
+				uint8_t	*buf;	   /* for read */
+				size_t	iov_count; /* for readv */
+			};
 			size_t	len;
 			uint64_t addr;
 			uint64_t key;
@@ -235,12 +286,16 @@ struct psmx2_am_request {
 			size_t	len_read;
 		} read;
 		struct {
-			void	*buf;
+			union {
+				uint8_t	*buf;	   /* for result_count == 1 */
+				size_t	iov_count; /* for result_count > 1 */
+			};
 			size_t	len;
 			uint64_t addr;
 			uint64_t key;
 			void	*context;
-			void 	*result;
+			uint8_t *result;
+			int	datatype;
 		} atomic;
 	};
 	uint64_t cq_flags;
@@ -249,6 +304,10 @@ struct psmx2_am_request {
 	int no_event;
 	int error;
 	struct slist_entry list_entry;
+	union {
+		struct iovec iov[0];	/* for readv, must be the last field */
+		struct fi_ioc ioc[0];	/* for atomic read, must be the last field */
+	};
 };
 
 #define PSMX2_IOV_PROTO_PACK	0
@@ -282,7 +341,7 @@ struct psmx2_sendv_reply {
 	struct fi_context fi_context;
 	int no_completion;
 	int multi_recv;
-	void *buf;
+	uint8_t *buf;
 	void *user_context;
 	size_t iov_done;
 	size_t bytes_received;
@@ -301,7 +360,7 @@ struct psmx2_multi_recv {
 	psm2_epaddr_t	src_addr;
 	psm2_mq_tag_t	tag;
 	psm2_mq_tag_t	tagsel;
-	void		*buf;
+	uint8_t		*buf;
 	size_t		len;
 	size_t		offset;
 	int		min_buf_size;
@@ -313,29 +372,19 @@ struct psmx2_fid_fabric {
 	struct util_fabric	util_fabric;
 	struct psmx2_fid_domain	*active_domain;
 	psm2_uuid_t		uuid;
-	pthread_t		name_server_thread;
+	struct util_ns		name_server;
 };
 
-struct psmx2_fid_domain {
-	struct util_domain	util_domain;
-	struct psmx2_fid_fabric	*fabric;
+struct psmx2_trx_ctxt {
 	psm2_ep_t		psm2_ep;
 	psm2_epid_t		psm2_epid;
 	psm2_mq_t		psm2_mq;
-	uint64_t		mode;
-	uint64_t		caps;
-
-	enum fi_mr_mode		mr_mode;
-	fastlock_t		mr_lock;
-	uint64_t		mr_reserved_key;
-	RbtHandle		mr_map;
-
-	fastlock_t		vl_lock;
-	uint64_t		vl_map[(PSMX2_MAX_VL+1)/sizeof(uint64_t)];
-	int			vl_alloc;
-	struct psmx2_fid_ep	*eps[PSMX2_MAX_VL+1];
-	
 	int			am_initialized;
+	int			id;
+	struct psm2_am_parameters psm2_am_param;
+
+	/* ep bound to this tx/rx context, NULL if multiplexed */
+	struct psmx2_fid_ep	*ep;
 
 	/* incoming req queue for AM based RMA request. */
 	struct psmx2_req_queue	rma_queue;
@@ -348,15 +397,71 @@ struct psmx2_fid_domain {
 	 */
 	fastlock_t		poll_lock;
 
+	struct dlist_entry	entry;
+};
+
+struct psmx2_fid_domain {
+	struct util_domain	util_domain;
+	struct psmx2_fid_fabric	*fabric;
+	uint64_t		mode;
+	uint64_t		caps;
+
+	enum fi_mr_mode		mr_mode;
+	fastlock_t		mr_lock;
+	uint64_t		mr_reserved_key;
+	RbtHandle		mr_map;
+
+	/*
+	 * A list of all opened hw contexts, including the base hw context.
+	 * The list is used for making progress.
+	 */
+	fastlock_t		trx_ctxt_lock;
+	struct dlist_entry	trx_ctxt_list;
+
+	/*
+	 * The base hw context is multiplexed for all regular endpoints via
+	 * logical "virtual lanes".
+	 */
+	struct psmx2_trx_ctxt	*base_trx_ctxt;
+	fastlock_t		vl_lock;
+	uint64_t		vl_map[(PSMX2_MAX_VL+1)/sizeof(uint64_t)];
+	int			vl_alloc;
+	struct psmx2_fid_ep	*eps[PSMX2_MAX_VL+1];
+
+	ofi_atomic32_t		sep_cnt;
+	fastlock_t		sep_lock;
+	struct dlist_entry	sep_list;
+
 	int			progress_thread_enabled;
 	pthread_t		progress_thread;
+
+	int			addr_format;
 };
 
+#define PSMX2_EP_REGULAR	0
+#define PSMX2_EP_SCALABLE	1
+#define PSMX2_EP_SRC_ADDR	2
+
+#define PSMX2_RESERVED_EPID	(0xFFFFULL)
+#define PSMX2_DEFAULT_UNIT	(-1)
+#define PSMX2_DEFAULT_PORT	0
+#define PSMX2_ANY_SERVICE	0
+
 struct psmx2_ep_name {
 	psm2_epid_t		epid;
-	uint8_t			vlane;
+	uint8_t			type;
+	union {
+		uint8_t		vlane;		/* for regular ep */
+		uint8_t		sep_id;		/* for scalable ep */
+		int8_t		unit;		/* for src addr. start from 0. -1 means any */
+	};
+	uint8_t			port;		/* for src addr. start from 1, 0 means any */
+	uint8_t			padding;
+	uint32_t		service;	/* for src addr. 0 means any */
 };
 
+#define PSMX2_MAX_STRING_NAME_LEN	64	/* "fi_addr_psmx2://<uint64_t>:<uint64_t>"  */
+
 struct psmx2_cq_event {
 	union {
 		struct fi_cq_entry		context;
@@ -366,13 +471,18 @@ struct psmx2_cq_event {
 		struct fi_cq_err_entry		err;
 	} cqe;
 	int error;
+	int source_is_valid;
 	fi_addr_t source;
+	struct psmx2_fid_av *source_av;
 	struct slist_entry list_entry;
 };
 
+#define PSMX2_ERR_DATA_SIZE		64	/* large enough to hold a string address */
+
 struct psmx2_fid_cq {
 	struct fid_cq			cq;
 	struct psmx2_fid_domain		*domain;
+	struct psmx2_trx_ctxt		*trx_ctxt;
 	int 				format;
 	int				entry_size;
 	size_t				event_count;
@@ -383,6 +493,7 @@ struct psmx2_fid_cq {
 	struct util_wait		*wait;
 	int				wait_cond;
 	int				wait_is_local;
+	uint8_t				error_data[PSMX2_ERR_DATA_SIZE];
 };
 
 enum psmx2_triggered_op {
@@ -395,9 +506,13 @@ enum psmx2_triggered_op {
 	PSMX2_TRIGGERED_WRITE,
 	PSMX2_TRIGGERED_WRITEV,
 	PSMX2_TRIGGERED_READ,
+	PSMX2_TRIGGERED_READV,
 	PSMX2_TRIGGERED_ATOMIC_WRITE,
+	PSMX2_TRIGGERED_ATOMIC_WRITEV,
 	PSMX2_TRIGGERED_ATOMIC_READWRITE,
+	PSMX2_TRIGGERED_ATOMIC_READWRITEV,
 	PSMX2_TRIGGERED_ATOMIC_COMPWRITE,
+	PSMX2_TRIGGERED_ATOMIC_COMPWRITEV,
 };
 
 struct psmx2_trigger {
@@ -502,6 +617,17 @@ struct psmx2_trigger {
 			void		*context;
 			uint64_t	flags;
 		} read;
+		struct {
+			struct fid_ep	*ep;
+			const struct iovec *iov;
+			size_t		count;
+			void		*desc;
+			fi_addr_t	src_addr;
+			uint64_t	addr;
+			uint64_t	key;
+			void		*context;
+			uint64_t	flags;
+		} readv;
 		struct {
 			struct fid_ep	*ep;
 			const void	*buf;
@@ -515,6 +641,19 @@ struct psmx2_trigger {
 			void		*context;
 			uint64_t	flags;
 		} atomic_write;
+		struct {
+			struct fid_ep	*ep;
+			const struct fi_ioc *iov;
+			size_t		count;
+			void		*desc;
+			fi_addr_t	dest_addr;
+			uint64_t	addr;
+			uint64_t	key;
+			enum fi_datatype datatype;
+			enum fi_op	atomic_op;
+			void		*context;
+			uint64_t	flags;
+		} atomic_writev;
 		struct {
 			struct fid_ep	*ep;
 			const void	*buf;
@@ -530,6 +669,22 @@ struct psmx2_trigger {
 			void		*context;
 			uint64_t	flags;
 		} atomic_readwrite;
+		struct {
+			struct fid_ep	*ep;
+			const struct fi_ioc *iov;
+			size_t		count;
+			void		**desc;
+			struct fi_ioc	*resultv;
+			void		**result_desc;
+			size_t		result_count;
+			fi_addr_t	dest_addr;
+			uint64_t	addr;
+			uint64_t	key;
+			enum fi_datatype datatype;
+			enum fi_op	atomic_op;
+			void		*context;
+			uint64_t	flags;
+		} atomic_readwritev;
 		struct {
 			struct fid_ep	*ep;
 			const void	*buf;
@@ -547,24 +702,55 @@ struct psmx2_trigger {
 			void		*context;
 			uint64_t	flags;
 		} atomic_compwrite;
+		struct {
+			struct fid_ep	*ep;
+			const struct fi_ioc *iov;
+			size_t		count;
+			void		**desc;
+			const struct fi_ioc *comparev;
+			void		**compare_desc;
+			size_t		compare_count;
+			struct fi_ioc	*resultv;
+			void		**result_desc;
+			size_t		result_count;
+			fi_addr_t	dest_addr;
+			uint64_t	addr;
+			uint64_t	key;
+			enum fi_datatype datatype;
+			enum fi_op	atomic_op;
+			void		*context;
+			uint64_t	flags;
+		} atomic_compwritev;
 	};
 	struct psmx2_trigger *next;	/* used for randomly accessed trigger list */
 	struct slist_entry list_entry;	/* used for ready-to-fire trigger queue */
 };
 
 struct psmx2_fid_cntr {
-	struct fid_cntr		cntr;
+	union {
+		struct fid_cntr		cntr;
+		struct util_cntr	util_cntr; /* for util_poll_run */
+	};
 	struct psmx2_fid_domain	*domain;
+	struct psmx2_trx_ctxt	*trx_ctxt;
 	int			events;
 	uint64_t		flags;
-	volatile uint64_t	counter;
-	volatile uint64_t	error_counter;
-	uint64_t		counter_last_read;
-	uint64_t		error_counter_last_read;
+	ofi_atomic64_t		counter;
+	ofi_atomic64_t		error_counter;
 	struct util_wait	*wait;
 	int			wait_is_local;
 	struct psmx2_trigger	*trigger;
-	pthread_mutex_t		trigger_lock;
+	fastlock_t		trigger_lock;
+};
+
+struct psmx2_ctxt_addr {
+	psm2_epid_t		epid;
+	psm2_epaddr_t		*epaddrs;
+};
+
+struct psmx2_sep_addr {
+	int			ctxt_cnt;
+	struct psmx2_ctxt_addr	ctxt_addrs[];
 };
 
 struct psmx2_fid_av {
@@ -572,6 +758,8 @@ struct psmx2_fid_av {
 	struct psmx2_fid_domain	*domain;
 	struct fid_eq		*eq;
 	int			type;
+	int			addr_format;
+	int			rx_ctx_bits;
 	uint64_t		flags;
 	size_t			addrlen;
 	size_t			count;
@@ -579,11 +767,18 @@ struct psmx2_fid_av {
 	psm2_epid_t		*epids;
 	psm2_epaddr_t		*epaddrs;
 	uint8_t			*vlanes;
+	uint8_t			*types;
+	struct psmx2_sep_addr	**sepaddrs;
 };
 
 struct psmx2_fid_ep {
 	struct fid_ep		ep;
+	int			type;
 	struct psmx2_fid_domain	*domain;
+	/* above fields are common with sep */
+
+	struct psmx2_trx_ctxt	*trx_ctxt;
+	struct psmx2_fid_ep	*base_ep;
 	struct psmx2_fid_av	*av;
 	struct psmx2_fid_cq	*send_cq;
 	struct psmx2_fid_cq	*recv_cq;
@@ -596,14 +791,39 @@ struct psmx2_fid_ep {
 	uint8_t			vlane;
 	unsigned		send_selective_completion:1;
 	unsigned		recv_selective_completion:1;
-	uint64_t		flags;
+	unsigned		enabled:1;
+	uint64_t		tx_flags;
+	uint64_t		rx_flags;
 	uint64_t		caps;
+	ofi_atomic32_t		ref;
 	struct fi_context	nocomp_send_context;
 	struct fi_context	nocomp_recv_context;
 	struct slist		free_context_list;
 	fastlock_t		context_lock;
 	size_t			min_multi_recv;
 	uint32_t		iov_seq_num;
+	int			service;
+};
+
+struct psmx2_sep_ctxt {
+	struct psmx2_trx_ctxt	*trx_ctxt;
+	struct psmx2_fid_ep	*ep;
+};
+
+struct psmx2_fid_sep {
+	struct fid_ep		ep;
+	int			type;
+	struct psmx2_fid_domain	*domain;
+	/* above fields are common with regular ep */
+
+	struct dlist_entry	entry;
+
+	ofi_atomic32_t		ref;
+	int			service;
+	uint8_t			id;
+	uint8_t			enabled;
+	size_t			ctxt_cnt;
+	struct psmx2_sep_ctxt	ctxts[]; /* must be last element */
 };
 
 struct psmx2_fid_stx {
@@ -619,11 +839,11 @@ struct psmx2_fid_mr {
 	uint64_t		flags;
 	uint64_t		offset;
 	size_t			iov_count;
-	struct iovec		iov[0];	/* must be the last field */
+	struct iovec		iov[];	/* must be the last field */
 };
 
 struct psmx2_epaddr_context {
-	struct psmx2_fid_domain	*domain;
+	struct psmx2_trx_ctxt	*trx_ctxt;
 	psm2_epid_t		epid;
 };
 
@@ -635,6 +855,12 @@ struct psmx2_env {
 	int timeout;
 	int prog_interval;
 	char *prog_affinity;
+	int sep;
+	int max_trx_ctxt;
+	int sep_trx_ctxt;
+	int num_devunits;
+	int inject_size;
+	int lock_level;
 };
 
 extern struct fi_ops_mr		psmx2_mr_ops;
@@ -652,16 +878,91 @@ extern struct fi_ops_msg	psmx2_msg_ops;
 extern struct fi_ops_msg	psmx2_msg2_ops;
 extern struct fi_ops_rma	psmx2_rma_ops;
 extern struct fi_ops_atomic	psmx2_atomic_ops;
-extern struct psm2_am_parameters psmx2_am_param;
 extern struct psmx2_env		psmx2_env;
 extern struct psmx2_fid_fabric	*psmx2_active_fabric;
 
+/*
+ * Lock levels:
+ *     0 -- always lock
+ *     1 -- lock needed if there is more than one thread (including internal threads)
+ *     2 -- lock needed if more then one thread accesses the same psm2 ep
+ */
+static inline void psmx2_lock(fastlock_t *lock, int lock_level)
+{
+	if (psmx2_env.lock_level >= lock_level)
+		fastlock_acquire(lock);
+}
+
+static inline int psmx2_trylock(fastlock_t *lock, int lock_level)
+{
+	if (psmx2_env.lock_level >= lock_level)
+		return fastlock_tryacquire(lock);
+	else
+		return 0;
+}
+
+static inline void psmx2_unlock(fastlock_t *lock, int lock_level)
+{
+	if (psmx2_env.lock_level >= lock_level)
+		fastlock_release(lock);
+}
+
+#ifdef PSM2_MULTI_EP_CAP
+
+static inline int psmx2_sep_ok(void)
+{
+	uint64_t caps = PSM2_MULTI_EP_CAP;
+	return (psm2_get_capability_mask(caps) == caps);
+}
+
+static inline psm2_error_t psmx2_ep_epid_lookup(psm2_ep_t ep, psm2_epid_t epid,
+						psm2_epconn_t *epconn)
+{
+	return psm2_ep_epid_lookup2(ep, epid, epconn);
+}
+
+static inline psm2_epid_t psmx2_epaddr_to_epid(psm2_epaddr_t epaddr)
+{
+	psm2_epid_t epid;
+
+	/* Caller ensures that epaddr is not NULL */
+	psm2_epaddr_to_epid(epaddr, &epid);
+	return epid;
+}
+
+#else
+
+static inline int psmx2_sep_ok(void)
+{
+	return 0;
+}
+
+static inline psm2_error_t psmx2_ep_epid_lookup(psm2_ep_t ep, psm2_epid_t epid,
+						psm2_epconn_t *epconn)
+{
+	return psm2_ep_epid_lookup(epid, epconn);
+}
+
+static inline psm2_epid_t psmx2_epaddr_to_epid(psm2_epaddr_t epaddr)
+{
+	/*
+	 * This is a hack based on the fact that the internal representation of
+	 * epaddr has epid as the first field. This is a workaround before a PSM2
+	 * function is availale to retrieve this information.
+	 */
+	return *(psm2_epid_t *)epaddr;
+}
+
+#endif /* PSM2_MULTI_EP_CAP */
+
 int	psmx2_fabric(struct fi_fabric_attr *attr,
 		    struct fid_fabric **fabric, void *context);
 int	psmx2_domain_open(struct fid_fabric *fabric, struct fi_info *info,
 			 struct fid_domain **domain, void *context);
 int	psmx2_ep_open(struct fid_domain *domain, struct fi_info *info,
 		     struct fid_ep **ep, void *context);
+int	psmx2_sep_open(struct fid_domain *domain, struct fi_info *info,
+		       struct fid_ep **sep, void *context);
 int	psmx2_stx_ctx(struct fid_domain *domain, struct fi_tx_attr *attr,
 		     struct fid_stx **stx, void *context);
 int	psmx2_cq_open(struct fid_domain *domain, struct fi_cq_attr *attr,
@@ -672,37 +973,62 @@ int	psmx2_cntr_open(struct fid_domain *domain, struct fi_cntr_attr *attr,
 		       struct fid_cntr **cntr, void *context);
 int	psmx2_wait_open(struct fid_fabric *fabric, struct fi_wait_attr *attr,
 			struct fid_wait **waitset);
+int	psmx2_wait_trywait(struct fid_fabric *fabric, struct fid **fids,
+			   int count);
+int	psmx2_query_atomic(struct fid_domain *doamin, enum fi_datatype datatype,
+			   enum fi_op op, struct fi_atomic_attr *attr,
+			   uint64_t flags);
 
 static inline void psmx2_fabric_acquire(struct psmx2_fid_fabric *fabric)
 {
-	atomic_inc(&fabric->util_fabric.ref);
+	ofi_atomic_inc32(&fabric->util_fabric.ref);
 }
 
 static inline void psmx2_fabric_release(struct psmx2_fid_fabric *fabric)
 {
-	atomic_dec(&fabric->util_fabric.ref);
+	ofi_atomic_dec32(&fabric->util_fabric.ref);
 }
 
 static inline void psmx2_domain_acquire(struct psmx2_fid_domain *domain)
 {
-	atomic_inc(&domain->util_domain.ref);
+	ofi_atomic_inc32(&domain->util_domain.ref);
 }
 
 static inline void psmx2_domain_release(struct psmx2_fid_domain *domain)
 {
-	atomic_dec(&domain->util_domain.ref);
+	ofi_atomic_dec32(&domain->util_domain.ref);
 }
 
 int	psmx2_domain_check_features(struct psmx2_fid_domain *domain, int ep_cap);
 int	psmx2_domain_enable_ep(struct psmx2_fid_domain *domain, struct psmx2_fid_ep *ep);
-void 	*psmx2_name_server(void *args);
-void	*psmx2_resolve_name(const char *servername, int port);
+
+void	psmx2_trx_ctxt_free(struct psmx2_trx_ctxt *trx_ctxt);
+struct	psmx2_trx_ctxt *psmx2_trx_ctxt_alloc(struct psmx2_fid_domain *domain,
+					     struct psmx2_ep_name *src_addr,
+					     int sep_ctxt_idx);
+
+
+static inline
+int	psmx2_ns_service_cmp(void *svc1, void *svc2)
+{
+	int service1 = *(int *)svc1, service2 = *(int *)svc2;
+	if (service1 == PSMX2_ANY_SERVICE ||
+	    service2 == PSMX2_ANY_SERVICE)
+		return 0;
+	return (service1 < service2) ?
+		-1 : (service1 > service2);
+}
+static inline
+int	psmx2_ns_is_service_wildcard(void *svc)
+{
+	return (*(int *)svc == PSMX2_ANY_SERVICE);
+}
 void	psmx2_get_uuid(psm2_uuid_t uuid);
 int	psmx2_uuid_to_port(psm2_uuid_t uuid);
 char	*psmx2_uuid_to_string(psm2_uuid_t uuid);
+void	*psmx2_ep_name_to_string(const struct psmx2_ep_name *name, size_t *len);
+struct	psmx2_ep_name *psmx2_string_to_ep_name(const void *s);
 int	psmx2_errno(int err);
-int	psmx2_epid_to_epaddr(struct psmx2_fid_domain *domain,
-			    psm2_epid_t epid, psm2_epaddr_t *epaddr);
 void	psmx2_query_mpi(void);
 
 struct	fi_context *psmx2_ep_get_op_context(struct psmx2_fid_ep *ep);
@@ -713,24 +1039,33 @@ struct	psmx2_cq_event *psmx2_cq_create_event(struct psmx2_fid_cq *cq,
 					uint64_t flags, size_t len,
 					uint64_t data, uint64_t tag,
 					size_t olen, int err);
-int	psmx2_cq_poll_mq(struct psmx2_fid_cq *cq, struct psmx2_fid_domain *domain,
+int	psmx2_cq_poll_mq(struct psmx2_fid_cq *cq, struct psmx2_trx_ctxt *trx_ctxt,
 			struct psmx2_cq_event *event, int count, fi_addr_t *src_addr);
 
-int	psmx2_am_init(struct psmx2_fid_domain *domain);
-int	psmx2_am_fini(struct psmx2_fid_domain *domain);
-int	psmx2_am_progress(struct psmx2_fid_domain *domain);
-int	psmx2_am_process_send(struct psmx2_fid_domain *domain,
+psm2_epaddr_t psmx2_av_translate_sep(struct psmx2_fid_av *av,
+				     struct psmx2_trx_ctxt *trx_ctxt, fi_addr_t addr);
+
+void	psmx2_am_global_init(void);
+void	psmx2_am_global_fini(void);
+int	psmx2_am_init(struct psmx2_trx_ctxt *trx_ctxt);
+void	psmx2_am_fini(struct psmx2_trx_ctxt *trx_ctxt);
+int	psmx2_am_progress(struct psmx2_trx_ctxt *trx_ctxt);
+int	psmx2_am_process_send(struct psmx2_trx_ctxt *trx_ctxt,
 				struct psmx2_am_request *req);
-int	psmx2_am_process_rma(struct psmx2_fid_domain *domain,
+int	psmx2_am_process_rma(struct psmx2_trx_ctxt *trx_ctxt,
 				struct psmx2_am_request *req);
-int	psmx2_process_trigger(struct psmx2_fid_domain *domain,
+int	psmx2_process_trigger(struct psmx2_trx_ctxt *trx_ctxt,
 				struct psmx2_trigger *trigger);
-int	psmx2_am_rma_handler(psm2_am_token_t token,
-				psm2_amarg_t *args, int nargs, void *src, uint32_t len);
-int	psmx2_am_atomic_handler(psm2_am_token_t token,
-				psm2_amarg_t *args, int nargs, void *src, uint32_t len);
-void	psmx2_atomic_init(void);
-void	psmx2_atomic_fini(void);
+int	psmx2_am_rma_handler_ext(psm2_am_token_t token,
+				 psm2_amarg_t *args, int nargs, void *src, uint32_t len,
+				 struct psmx2_trx_ctxt *trx_ctxt);
+int	psmx2_am_atomic_handler_ext(psm2_am_token_t token,
+				    psm2_amarg_t *args, int nargs, void *src, uint32_t len,
+				    struct psmx2_trx_ctxt *trx_ctxt);
+int	psmx2_am_sep_handler(psm2_am_token_t token, psm2_amarg_t *args, int nargs,
+			     void *src, uint32_t len);
+void	psmx2_atomic_global_init(void);
+void	psmx2_atomic_global_fini(void);
 
 void	psmx2_am_ack_rma(struct psmx2_am_request *req);
 
@@ -744,21 +1079,59 @@ int	psmx2_handle_sendv_req(struct psmx2_fid_ep *ep, psm2_mq_status2_t *psm2_stat
 
 static inline void psmx2_cntr_inc(struct psmx2_fid_cntr *cntr)
 {
-	cntr->counter++;
+	ofi_atomic_inc64(&cntr->counter);
 	psmx2_cntr_check_trigger(cntr);
 	if (cntr->wait)
 		cntr->wait->signal(cntr->wait);
 }
 
-static inline void psmx2_progress(struct psmx2_fid_domain *domain)
+fi_addr_t psmx2_av_translate_source(struct psmx2_fid_av *av, fi_addr_t source);
+
+static inline void psmx2_get_source_name(fi_addr_t source, struct psmx2_ep_name *name)
 {
-	if (domain) {
-		psmx2_cq_poll_mq(NULL, domain, NULL, 0, NULL);
-		if (domain->am_initialized)
-			psmx2_am_progress(domain);
+	psm2_epaddr_t epaddr = PSMX2_ADDR_TO_EP(source);
+
+	memset(name, 0, sizeof(*name));
+	name->epid = psmx2_epaddr_to_epid(epaddr);
+	name->vlane = PSMX2_ADDR_TO_VL(source);
+	name->type = PSMX2_EP_REGULAR;
+}
+
+static inline void psmx2_get_source_string_name(fi_addr_t source, char *name, size_t *len)
+{
+	struct psmx2_ep_name ep_name;
+	psm2_epaddr_t epaddr = PSMX2_ADDR_TO_EP(source);
+
+	memset(&ep_name, 0, sizeof(ep_name));
+	ep_name.epid = psmx2_epaddr_to_epid(epaddr);
+	ep_name.vlane = PSMX2_ADDR_TO_VL(source);
+	ep_name.type = PSMX2_EP_REGULAR;
+
+	ofi_straddr(name, len, FI_ADDR_PSMX2, &ep_name);
+}
+
+static inline void psmx2_progress(struct psmx2_trx_ctxt *trx_ctxt)
+{
+	if (trx_ctxt) {
+		psmx2_cq_poll_mq(NULL, trx_ctxt, NULL, 0, NULL);
+		if (trx_ctxt->am_initialized)
+			psmx2_am_progress(trx_ctxt);
 	}
 }
 
+static inline void psmx2_progress_all(struct psmx2_fid_domain *domain)
+{
+	struct dlist_entry *item;
+	struct psmx2_trx_ctxt *trx_ctxt;
+
+	psmx2_lock(&domain->trx_ctxt_lock, 1);
+	dlist_foreach(&domain->trx_ctxt_list, item) {
+		trx_ctxt = container_of(item, struct psmx2_trx_ctxt, entry);
+		psmx2_progress(trx_ctxt);
+	}
+	psmx2_unlock(&domain->trx_ctxt_lock, 1);
+}
+
 /* The following functions are used by triggered operations */
 
 ssize_t psmx2_send_generic(
@@ -825,6 +1198,13 @@ ssize_t psmx2_read_generic(
 			uint64_t addr, uint64_t key,
 			void *context, uint64_t flags);
 
+ssize_t psmx2_readv_generic(
+			struct fid_ep *ep,
+			const struct iovec *iov, void *desc,
+			size_t count, fi_addr_t src_addr,
+			uint64_t addr, uint64_t key,
+			void *context, uint64_t flags);
+
 ssize_t psmx2_atomic_write_generic(
 			struct fid_ep *ep,
 			const void *buf,
