@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2014 Intel Corporation. All rights reserved.
+ * Copyright (c) 2013-2017 Intel Corporation. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -37,38 +37,21 @@ struct psmx2_fid_fabric *psmx2_active_fabric = NULL;
 static int psmx2_fabric_close(fid_t fid)
 {
 	struct psmx2_fid_fabric *fabric;
-	void *exit_code;
-	int ret;
 
 	fabric = container_of(fid, struct psmx2_fid_fabric,
 			      util_fabric.fabric_fid.fid);
 
 	FI_INFO(&psmx2_prov, FI_LOG_CORE, "refcnt=%d\n",
-		atomic_get(&fabric->util_fabric.ref));
+		ofi_atomic_get32(&fabric->util_fabric.ref));
+
+	if (psmx2_env.name_server)
+		ofi_ns_stop_server(&fabric->name_server);
 
 	psmx2_fabric_release(fabric);
 
-	if (util_fabric_close(&fabric->util_fabric))
+	if (ofi_fabric_close(&fabric->util_fabric))
 		return 0;
 
-	if (psmx2_env.name_server &&
-	    !pthread_equal(fabric->name_server_thread, pthread_self())) {
-		ret = pthread_cancel(fabric->name_server_thread);
-		if (ret) {
-			FI_INFO(&psmx2_prov, FI_LOG_CORE,
-				"pthread_cancel returns %d\n", ret);
-		}
-		ret = pthread_join(fabric->name_server_thread, &exit_code);
-		if (ret) {
-			FI_INFO(&psmx2_prov, FI_LOG_CORE,
-				"pthread_join returns %d\n", ret);
-		} else {
-			FI_INFO(&psmx2_prov, FI_LOG_CORE,
-				"name server thread exited with code %ld (%s)\n",
-				(uintptr_t)exit_code,
-				(exit_code == PTHREAD_CANCELED) ? "PTHREAD_CANCELED" : "?");
-		}
-	}
 	if (fabric->active_domain) {
 		FI_WARN(&psmx2_prov, FI_LOG_CORE, "forced closing of active_domain\n");
 		fi_close(&fabric->active_domain->util_domain.domain_fid.fid);
@@ -89,9 +72,9 @@ static struct fi_ops_fabric psmx2_fabric_ops = {
 	.size = sizeof(struct fi_ops_fabric),
 	.domain = psmx2_domain_open,
 	.passive_ep = fi_no_passive_ep,
-	.eq_open = fi_eq_create,
+	.eq_open = ofi_eq_create,
 	.wait_open = psmx2_wait_open,
-	.trywait = util_trywait
+	.trywait = psmx2_wait_trywait
 };
 
 static struct fi_fabric_attr psmx2_fabric_attr = {
@@ -120,30 +103,41 @@ int psmx2_fabric(struct fi_fabric_attr *attr,
 	if (!fabric_priv)
 		return -FI_ENOMEM;
 
-	ret = fi_fabric_init(&psmx2_prov, &psmx2_fabric_attr, attr,
+	psmx2_get_uuid(fabric_priv->uuid);
+	if (psmx2_env.name_server) {
+		struct util_ns_attr ns_attr = {
+			.ns_port = psmx2_uuid_to_port(fabric_priv->uuid),
+			.name_len = sizeof(struct psmx2_ep_name),
+			.service_len = sizeof(int),
+			.service_cmp = psmx2_ns_service_cmp,
+			.is_service_wildcard = psmx2_ns_is_service_wildcard,
+		};
+		ret = ofi_ns_init(&ns_attr,
+				  &fabric_priv->name_server);
+		if (ret) {
+			FI_INFO(&psmx2_prov, FI_LOG_CORE,
+			        "ofi_ns_init returns %d\n", ret);
+			free(fabric_priv);
+			return ret;
+		}
+
+		ofi_ns_start_server(&fabric_priv->name_server);
+	}
+
+	ret = ofi_fabric_init(&psmx2_prov, &psmx2_fabric_attr, attr,
 			     &fabric_priv->util_fabric, context);
 	if (ret) {
-		FI_INFO(&psmx2_prov, FI_LOG_CORE, "fi_fabric_init returns %d\n", ret);
+		FI_INFO(&psmx2_prov, FI_LOG_CORE, "ofi_fabric_init returns %d\n", ret);
+		if (psmx2_env.name_server)
+			ofi_ns_stop_server(&fabric_priv->name_server);
 		free(fabric_priv);
 		return ret;
 	}
 
-	/* fclass & context initialized in fi_fabric_init */
+	/* fclass & context initialized in ofi_fabric_init */
 	fabric_priv->util_fabric.fabric_fid.fid.ops = &psmx2_fabric_fi_ops;
 	fabric_priv->util_fabric.fabric_fid.ops = &psmx2_fabric_ops;
 
-	psmx2_get_uuid(fabric_priv->uuid);
-
-	if (psmx2_env.name_server) {
-		ret = pthread_create(&fabric_priv->name_server_thread, NULL,
-				     psmx2_name_server, (void *)fabric_priv);
-		if (ret) {
-			FI_INFO(&psmx2_prov, FI_LOG_CORE, "pthread_create returns %d\n", ret);
-			/* use the main thread's ID as invalid value for the new thread */
-			fabric_priv->name_server_thread = pthread_self();
-		}
-	}
-
 	psmx2_query_mpi();
 
 	/* take the reference to count for multiple fabric open calls */
