@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2014 Intel Corporation. All rights reserved.
+ * Copyright (c) 2013-2017 Intel Corporation. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -32,12 +32,40 @@
 
 #include "psmx2.h"
 
-static inline void psmx2_am_enqueue_rma(struct psmx2_fid_domain *domain,
+static inline void psmx2_am_enqueue_rma(struct psmx2_trx_ctxt *trx_ctxt,
 				        struct psmx2_am_request *req)
 {
-	fastlock_acquire(&domain->rma_queue.lock);
-	slist_insert_tail(&req->list_entry, &domain->rma_queue.list);
-	fastlock_release(&domain->rma_queue.lock);
+	psmx2_lock(&trx_ctxt->rma_queue.lock, 2);
+	slist_insert_tail(&req->list_entry, &trx_ctxt->rma_queue.list);
+	psmx2_unlock(&trx_ctxt->rma_queue.lock, 2);
+}
+
+static inline void psmx2_iov_copy(struct iovec *iov, size_t count,
+				  size_t offset, const void *src,
+				  size_t len)
+{
+	int i;
+	size_t copy_len;
+
+	for (i=0; i<count && len; i++) {
+		if (offset >= iov[i].iov_len) {
+			offset -= iov[i].iov_len;
+			continue;
+		}
+
+		copy_len = iov[i].iov_len - offset;
+		
+		if (copy_len > len)
+			copy_len = len;
+
+		memcpy((uint8_t *)iov[i].iov_base + offset, src, copy_len);
+
+		src = (const uint8_t *)src + copy_len;
+		len -= copy_len;
+
+		if (offset)
+			offset = 0;
+	}
 }
 
 /* RMA protocol:
@@ -70,11 +98,12 @@ static inline void psmx2_am_enqueue_rma(struct psmx2_fid_domain *domain,
  *	args[2].u64	offset
  */
 
-int psmx2_am_rma_handler(psm2_am_token_t token, psm2_amarg_t *args,
-			 int nargs, void *src, uint32_t len)
+int psmx2_am_rma_handler_ext(psm2_am_token_t token, psm2_amarg_t *args,
+			     int nargs, void *src, uint32_t len,
+			     struct psmx2_trx_ctxt *trx_ctxt)
 {
 	psm2_amarg_t rep_args[8];
-	void *rma_addr;
+	uint8_t *rma_addr;
 	ssize_t rma_len;
 	uint64_t key;
 	int err = 0;
@@ -92,10 +121,15 @@ int psmx2_am_rma_handler(psm2_am_token_t token, psm2_amarg_t *args,
 	psm2_am_get_source(token, &epaddr);
 
 	cmd = PSMX2_AM_GET_OP(args[0].u32w0);
-	dst_vl = PSMX2_AM_GET_DST(args[0].u32w0);
-
 	domain = psmx2_active_fabric->active_domain;
-	ep = domain->eps[dst_vl];
+
+	if (trx_ctxt->ep) {
+		dst_vl = 0;
+		ep = trx_ctxt->ep;
+	} else {
+		dst_vl = PSMX2_AM_GET_DST(args[0].u32w0);
+		ep = domain->eps[dst_vl];
+	}
 
 	eom = args[0].u32w0 & PSMX2_AM_EOM;
 	has_data = args[0].u32w0 & PSMX2_AM_DATA;
@@ -103,7 +137,7 @@ int psmx2_am_rma_handler(psm2_am_token_t token, psm2_amarg_t *args,
 	switch (cmd) {
 	case PSMX2_AM_REQ_WRITE:
 		rma_len = args[0].u32w1;
-		rma_addr = (void *)(uintptr_t)args[2].u64;
+		rma_addr = (uint8_t *)(uintptr_t)args[2].u64;
 		key = args[3].u64;
 		mr = psmx2_mr_get(domain, key);
 		op_error = mr ?
@@ -152,7 +186,7 @@ int psmx2_am_rma_handler(psm2_am_token_t token, psm2_amarg_t *args,
 	case PSMX2_AM_REQ_WRITE_LONG:
 		src_vl = PSMX2_AM_GET_SRC(args[0].u32w0);
 		rma_len = args[0].u32w1;
-		rma_addr = (void *)(uintptr_t)args[2].u64;
+		rma_addr = (uint8_t *)(uintptr_t)args[2].u64;
 		key = args[3].u64;
 		mr = psmx2_mr_get(domain, key);
 		op_error = mr ?
@@ -188,13 +222,13 @@ int psmx2_am_rma_handler(psm2_am_token_t token, psm2_amarg_t *args,
 					(has_data ? FI_REMOTE_CQ_DATA : 0),
 			PSMX2_CTXT_TYPE(&req->fi_context) = PSMX2_REMOTE_WRITE_CONTEXT;
 			PSMX2_CTXT_USER(&req->fi_context) = mr;
-			psmx2_am_enqueue_rma(mr->domain, req);
+			psmx2_am_enqueue_rma(trx_ctxt, req);
 		}
 		break;
 
 	case PSMX2_AM_REQ_READ:
 		rma_len = args[0].u32w1;
-		rma_addr = (void *)(uintptr_t)args[2].u64;
+		rma_addr = (uint8_t *)(uintptr_t)args[2].u64;
 		key = args[3].u64;
 		offset = args[4].u64;
 		mr = psmx2_mr_get(domain, key);
@@ -225,7 +259,7 @@ int psmx2_am_rma_handler(psm2_am_token_t token, psm2_amarg_t *args,
 	case PSMX2_AM_REQ_READ_LONG:
 		src_vl = PSMX2_AM_GET_SRC(args[0].u32w0);
 		rma_len = args[0].u32w1;
-		rma_addr = (void *)(uintptr_t)args[2].u64;
+		rma_addr = (uint8_t *)(uintptr_t)args[2].u64;
 		key = args[3].u64;
 		mr = psmx2_mr_get(domain, key);
 		op_error = mr ?
@@ -259,7 +293,7 @@ int psmx2_am_rma_handler(psm2_am_token_t token, psm2_amarg_t *args,
 			req->read.peer_vl = src_vl;
 			PSMX2_CTXT_TYPE(&req->fi_context) = PSMX2_REMOTE_READ_CONTEXT;
 			PSMX2_CTXT_USER(&req->fi_context) = mr;
-			psmx2_am_enqueue_rma(mr->domain, req);
+			psmx2_am_enqueue_rma(trx_ctxt, req);
 		}
 		break;
 
@@ -270,7 +304,7 @@ int psmx2_am_rma_handler(psm2_am_token_t token, psm2_amarg_t *args,
 		if (!req->error)
 			req->error = op_error;
 		if (eom) {
-			if (req->ep->send_cq && !req->no_event) {
+			if (req->ep->send_cq && (!req->no_event || req->error)) {
 				event = psmx2_cq_create_event(
 						req->ep->send_cq,
 						req->write.context,
@@ -296,17 +330,24 @@ int psmx2_am_rma_handler(psm2_am_token_t token, psm2_amarg_t *args,
 
 	case PSMX2_AM_REP_READ:
 		req = (struct psmx2_am_request *)(uintptr_t)args[1].u64;
-		assert(req->op == PSMX2_AM_REQ_READ);
+		assert(req->op == PSMX2_AM_REQ_READ || req->op == PSMX2_AM_REQ_READV);
 		op_error = (int)args[0].u32w1;
 		offset = args[2].u64;
 		if (!req->error)
 			req->error = op_error;
 		if (!op_error) {
-			memcpy(req->read.buf + offset, src, len);
+			if (req->op == PSMX2_AM_REQ_READ)
+				memcpy(req->read.buf + offset, src, len);
+			else 
+				psmx2_iov_copy(req->iov, req->read.iov_count, offset, src, len);
+
 			req->read.len_read += len;
 		}
-		if (eom) {
-			if (req->ep->send_cq && !req->no_event) {
+		if (eom || req->read.len == req->read.len_read) {
+			if (!eom)
+				FI_INFO(&psmx2_prov, FI_LOG_EP_DATA,
+					"readv: short protocol finishes after long protocol.\n");
+			if (req->ep->send_cq && (!req->no_event || req->error)) {
 				event = psmx2_cq_create_event(
 						req->ep->send_cq,
 						req->read.context,
@@ -352,7 +393,7 @@ static ssize_t psmx2_rma_self(int am_cmd,
 	int err = 0;
 	int op_error = 0;
 	int access;
-	void *dst;
+	uint8_t *dst, *src;
 	uint64_t cq_flags;
 	struct iovec *iov = buf;
 	size_t iov_count = len;
@@ -374,6 +415,13 @@ static ssize_t psmx2_rma_self(int am_cmd,
 		access = FI_REMOTE_READ;
 		cq_flags = FI_READ | FI_RMA;
 		break;
+	case PSMX2_AM_REQ_READV:
+		access = FI_REMOTE_READ;
+		cq_flags = FI_READ | FI_RMA;
+		len = 0;
+		for (i=0; i<iov_count; i++)
+			len += iov[i].iov_len;
+		break;
 	default:
 		return -FI_EINVAL;
 	}
@@ -411,6 +459,16 @@ static ssize_t psmx2_rma_self(int am_cmd,
 			cntr = dst_ep->remote_read_cntr;
 			memcpy(buf, (void *)addr, len);
 			break;
+
+		case PSMX2_AM_REQ_READV:
+			cntr = dst_ep->remote_read_cntr;
+			src = (void *)addr;
+			for (i=0; i<iov_count; i++)
+				if (iov[i].iov_len) {
+					memcpy(iov[i].iov_base, src, iov[i].iov_len);
+					src += iov[i].iov_len;
+				}
+			break;
 		}
 
 		if (cq) {
@@ -441,7 +499,7 @@ static ssize_t psmx2_rma_self(int am_cmd,
 	no_event = (flags & PSMX2_NO_COMPLETION) ||
 		   (ep->send_selective_completion && !(flags & FI_COMPLETION));
 
-	if (ep->send_cq && !no_event) {
+	if (ep->send_cq && (!no_event || op_error)) {
 		event = psmx2_cq_create_event(
 				ep->send_cq,
 				context,
@@ -460,11 +518,13 @@ static ssize_t psmx2_rma_self(int am_cmd,
 
 	switch (am_cmd) {
 	case PSMX2_AM_REQ_WRITE:
+	case PSMX2_AM_REQ_WRITEV:
 		if (ep->write_cntr)
 			psmx2_cntr_inc(ep->write_cntr);
 		break;
 
 	case PSMX2_AM_REQ_READ:
+	case PSMX2_AM_REQ_READV:
 		if (ep->read_cntr)
 			psmx2_cntr_inc(ep->read_cntr);
 		break;
@@ -489,7 +549,7 @@ void psmx2_am_ack_rma(struct psmx2_am_request *req)
 			      PSM2_AM_FLAG_NOREPLY, NULL, NULL);
 }
 
-int psmx2_am_process_rma(struct psmx2_fid_domain *domain,
+int psmx2_am_process_rma(struct psmx2_trx_ctxt *trx_ctxt,
 			 struct psmx2_am_request *req)
 {
 	int err;
@@ -499,17 +559,19 @@ int psmx2_am_process_rma(struct psmx2_fid_domain *domain,
 
 	if ((req->op & PSMX2_AM_OP_MASK) == PSMX2_AM_REQ_WRITE_LONG) {
 		tag32 = PSMX2_TAG32(PSMX2_RMA_BIT, req->write.peer_vl, req->write.vl);
+		PSMX2_TAG32_LONG_WRITE(tag32);
 		PSMX2_SET_TAG(psm2_tag, (uint64_t)req->write.context, tag32);
 		PSMX2_SET_TAG(psm2_tagsel, -1ULL, -1);
-		err = psm2_mq_irecv2(domain->psm2_mq,
+		err = psm2_mq_irecv2(trx_ctxt->psm2_mq,
 				     (psm2_epaddr_t)req->write.peer_addr,
 				     &psm2_tag, &psm2_tagsel, 0,
 				     (void *)req->write.addr, req->write.len,
 				     (void *)&req->fi_context, &psm2_req);
 	} else {
 		tag32 = PSMX2_TAG32(PSMX2_RMA_BIT, req->read.vl, req->read.peer_vl);
+		PSMX2_TAG32_LONG_READ(tag32);
 		PSMX2_SET_TAG(psm2_tag, (uint64_t)req->read.context, tag32);
-		err = psm2_mq_isend2(domain->psm2_mq,
+		err = psm2_mq_isend2(trx_ctxt->psm2_mq,
 				     (psm2_epaddr_t)req->read.peer_addr,
 				     0, &psm2_tag,
 				     (void *)req->read.addr, req->read.len,
@@ -537,6 +599,7 @@ ssize_t psmx2_read_generic(struct fid_ep *ep, void *buf, size_t len,
 	psm2_mq_tag_t psm2_tag, psm2_tagsel;
 	uint32_t tag32;
 	size_t idx;
+	int sep_target = 0;
 
 	ep_priv = container_of(ep, struct psmx2_fid_ep, ep);
 
@@ -570,7 +633,11 @@ ssize_t psmx2_read_generic(struct fid_ep *ep, void *buf, size_t len,
 		return -FI_EINVAL;
 
 	av = ep_priv->av;
-	if (av && av->type == FI_AV_TABLE) {
+	if (av && PSMX2_SEP_ADDR_TEST(src_addr)) {
+		psm2_epaddr = psmx2_av_translate_sep(av, ep_priv->trx_ctxt, src_addr);
+		vlane = 0;
+		sep_target = 1;
+	} else if (av && av->type == FI_AV_TABLE) {
 		idx = src_addr;
 		if (idx >= av->last)
 			return -FI_EINVAL;
@@ -586,9 +653,10 @@ ssize_t psmx2_read_generic(struct fid_ep *ep, void *buf, size_t len,
 	}
 
 	epaddr_context = psm2_epaddr_getctxt((void *)psm2_epaddr);
-	if (epaddr_context->epid == ep_priv->domain->psm2_epid)
+	if (epaddr_context->epid == ep_priv->trx_ctxt->psm2_epid)
 		return psmx2_rma_self(PSMX2_AM_REQ_READ, ep_priv,
-				      ep_priv->domain->eps[vlane],
+				      (sep_target ? ep_priv :
+					ep_priv->domain->eps[vlane]),
 				      buf, len, desc, addr, key,
 				      context, flags, 0);
 
@@ -613,7 +681,7 @@ ssize_t psmx2_read_generic(struct fid_ep *ep, void *buf, size_t len,
 		req->no_event = 1;
 	}
 
-	chunk_size = psmx2_am_param.max_reply_short;
+	chunk_size = ep_priv->trx_ctxt->psm2_am_param.max_reply_short;
 
 	args[0].u32w0 = 0;
 	PSMX2_AM_SET_SRC(args[0].u32w0, ep_priv->vlane);
@@ -621,9 +689,10 @@ ssize_t psmx2_read_generic(struct fid_ep *ep, void *buf, size_t len,
 
 	if (psmx2_env.tagged_rma && len > chunk_size) {
 		tag32 = PSMX2_TAG32(PSMX2_RMA_BIT, vlane, ep_priv->vlane);
+		PSMX2_TAG32_LONG_READ(tag32);
 		PSMX2_SET_TAG(psm2_tag, (uint64_t)req, tag32);
 		PSMX2_SET_TAG(psm2_tagsel, -1ULL, -1);
-		psm2_mq_irecv2(ep_priv->domain->psm2_mq, psm2_epaddr,
+		psm2_mq_irecv2(ep_priv->trx_ctxt->psm2_mq, psm2_epaddr,
 			       &psm2_tag, &psm2_tagsel, 0, buf, len,
 			       (void *)&req->fi_context, &psm2_req);
 
@@ -662,6 +731,177 @@ ssize_t psmx2_read_generic(struct fid_ep *ep, void *buf, size_t len,
 	return 0;
 }
 
+ssize_t psmx2_readv_generic(struct fid_ep *ep, const struct iovec *iov,
+			    void *desc, size_t count, fi_addr_t src_addr,
+			    uint64_t addr, uint64_t key, void *context,
+			    uint64_t flags)
+{
+	struct psmx2_fid_ep *ep_priv;
+	struct psmx2_fid_av *av;
+	struct psmx2_epaddr_context *epaddr_context;
+	struct psmx2_am_request *req;
+	psm2_amarg_t args[8];
+	int chunk_size;
+	size_t offset = 0;
+	psm2_epaddr_t psm2_epaddr;
+	uint8_t vlane;
+	psm2_mq_req_t psm2_req;
+	psm2_mq_tag_t psm2_tag, psm2_tagsel;
+	uint32_t tag32;
+	size_t idx;
+	size_t total_len, long_len = 0, short_len;
+	void *long_buf = NULL;
+	int i;
+	int sep_target = 0;
+
+	ep_priv = container_of(ep, struct psmx2_fid_ep, ep);
+
+	if (flags & FI_TRIGGER) {
+		struct psmx2_trigger *trigger;
+		struct fi_triggered_context *ctxt = context;
+
+		trigger = calloc(1, sizeof(*trigger));
+		if (!trigger)
+			return -FI_ENOMEM;
+
+		trigger->op = PSMX2_TRIGGERED_READV;
+		trigger->cntr = container_of(ctxt->trigger.threshold.cntr,
+					     struct psmx2_fid_cntr, cntr);
+		trigger->threshold = ctxt->trigger.threshold.threshold;
+		trigger->readv.ep = ep;
+		trigger->readv.iov = iov;
+		trigger->readv.count = count;
+		trigger->readv.desc = desc;
+		trigger->readv.src_addr = src_addr;
+		trigger->readv.addr = addr;
+		trigger->readv.key = key;
+		trigger->readv.context = context;
+		trigger->readv.flags = flags & ~FI_TRIGGER;
+
+		psmx2_cntr_add_trigger(trigger->cntr, trigger);
+		return 0;
+	}
+
+	av = ep_priv->av;
+	if (av && PSMX2_SEP_ADDR_TEST(src_addr)) {
+		psm2_epaddr = psmx2_av_translate_sep(av, ep_priv->trx_ctxt, src_addr);
+		vlane = 0;
+		sep_target = 1;
+	} else if (av && av->type == FI_AV_TABLE) {
+		idx = src_addr;
+		if (idx >= av->last)
+			return -FI_EINVAL;
+
+		psm2_epaddr = av->epaddrs[idx];
+		vlane = av->vlanes[idx];
+	} else {
+		if (!src_addr)
+			return -FI_EINVAL;
+
+		psm2_epaddr = PSMX2_ADDR_TO_EP(src_addr);
+		vlane = PSMX2_ADDR_TO_VL(src_addr);
+	}
+
+	epaddr_context = psm2_epaddr_getctxt((void *)psm2_epaddr);
+	if (epaddr_context->epid == ep_priv->trx_ctxt->psm2_epid)
+		return psmx2_rma_self(PSMX2_AM_REQ_READV, ep_priv,
+				      (sep_target ? ep_priv :
+					ep_priv->domain->eps[vlane]),
+				      (void *)iov, count, desc, addr,
+				      key, context, flags, 0);
+
+	total_len = 0;
+	for (i=0; i<count; i++)
+		total_len += iov[i].iov_len;
+
+	req = calloc(1, sizeof(*req) + count * sizeof(struct iovec));
+	if (!req)
+		return -FI_ENOMEM;
+
+	memcpy(req->iov, iov, count * sizeof(struct iovec));
+
+	req->op = PSMX2_AM_REQ_READV;
+	req->read.iov_count = count;
+	req->read.len = total_len;
+	req->read.addr = addr;	/* needed? */
+	req->read.key = key; 	/* needed? */
+	req->read.context = context;
+	req->ep = ep_priv;
+	req->cq_flags = FI_READ | FI_RMA;
+	PSMX2_CTXT_TYPE(&req->fi_context) = PSMX2_READ_CONTEXT;
+	PSMX2_CTXT_USER(&req->fi_context) = context;
+	PSMX2_CTXT_EP(&req->fi_context) = ep_priv;
+
+	if (ep_priv->send_selective_completion && !(flags & FI_COMPLETION)) {
+		PSMX2_CTXT_TYPE(&req->fi_context) = PSMX2_NOCOMP_READ_CONTEXT;
+		req->no_event = 1;
+	}
+
+	chunk_size = ep_priv->trx_ctxt->psm2_am_param.max_reply_short;
+
+	if (psmx2_env.tagged_rma) {
+		for (i=count-1; i>=0; i--) {
+			if (iov[i].iov_len > chunk_size) {
+				long_buf = iov[i].iov_base;
+				long_len = iov[i].iov_len;
+				break;
+			} else if (iov[i].iov_len) {
+				break;
+			}
+		}
+	}
+
+	short_len = total_len - long_len;
+
+	/* Use short protocol for all but the last segment (long_len) */
+	args[0].u32w0 = 0;
+	PSMX2_AM_SET_SRC(args[0].u32w0, ep_priv->vlane);
+	PSMX2_AM_SET_DST(args[0].u32w0, vlane);
+	PSMX2_AM_SET_OP(args[0].u32w0, PSMX2_AM_REQ_READ);
+	args[1].u64 = (uint64_t)(uintptr_t)req;
+	args[3].u64 = key;
+	while (short_len > chunk_size) {
+		args[0].u32w1 = chunk_size;
+		args[2].u64 = addr;
+		args[4].u64 = offset;
+		psm2_am_request_short(psm2_epaddr, PSMX2_AM_RMA_HANDLER,
+				      args, 5, NULL, 0, 0, NULL, NULL);
+		addr += chunk_size;
+		short_len -= chunk_size;
+		offset += chunk_size;
+	}
+
+	if (!long_len)
+		PSMX2_AM_SET_FLAG(args[0].u32w0, PSMX2_AM_EOM);
+	args[0].u32w1 = short_len;
+	args[2].u64 = addr;
+	args[4].u64 = offset;
+	psm2_am_request_short(psm2_epaddr, PSMX2_AM_RMA_HANDLER,
+			      args, 5, NULL, 0, 0, NULL, NULL);
+
+	/* Use the long protocol for the last segment */
+	if (long_len) {
+		tag32 = PSMX2_TAG32(PSMX2_RMA_BIT, vlane, ep_priv->vlane);
+		PSMX2_TAG32_LONG_READ(tag32);
+		PSMX2_SET_TAG(psm2_tag, (uint64_t)req, tag32);
+		PSMX2_SET_TAG(psm2_tagsel, -1ULL, -1);
+		psm2_mq_irecv2(ep_priv->trx_ctxt->psm2_mq, psm2_epaddr,
+			       &psm2_tag, &psm2_tagsel, 0,
+			       long_buf, long_len,
+			       (void *)&req->fi_context, &psm2_req);
+
+		PSMX2_AM_SET_OP(args[0].u32w0, PSMX2_AM_REQ_READ_LONG);
+		args[0].u32w1 = long_len;
+		args[1].u64 = (uint64_t)req;
+		args[2].u64 = addr + short_len;
+		args[3].u64 = key;
+		psm2_am_request_short(psm2_epaddr, PSMX2_AM_RMA_HANDLER,
+				      args, 4, NULL, 0, 0, NULL, NULL);
+	}
+
+	return 0;
+}
+
 static ssize_t psmx2_read(struct fid_ep *ep, void *buf, size_t len,
 			  void *desc, fi_addr_t src_addr,
 			  uint64_t addr, uint64_t key, void *context)
@@ -671,16 +911,25 @@ static ssize_t psmx2_read(struct fid_ep *ep, void *buf, size_t len,
 	ep_priv = container_of(ep, struct psmx2_fid_ep, ep);
 
 	return psmx2_read_generic(ep, buf, len, desc, src_addr, addr,
-				  key, context, ep_priv->flags);
+				  key, context, ep_priv->tx_flags);
 }
 
 static ssize_t psmx2_readmsg(struct fid_ep *ep,
 			     const struct fi_msg_rma *msg,
 			     uint64_t flags)
 {
-	if (!msg || msg->iov_count != 1 || !msg->msg_iov || !msg->rma_iov)
+	if (!msg || !msg->iov_count || !msg->msg_iov ||
+	    !msg->rma_iov || msg->rma_iov_count != 1)
 		return -FI_EINVAL;
 
+	if (msg->iov_count > 1)
+		return psmx2_readv_generic(ep, msg->msg_iov,
+					   msg->desc ? msg->desc[0] : NULL,
+					   msg->iov_count, msg->addr,
+					   msg->rma_iov[0].addr,
+					   msg->rma_iov[0].key,
+					   msg->context, flags);
+
 	return psmx2_read_generic(ep, msg->msg_iov[0].iov_base,
 				  msg->msg_iov[0].iov_len,
 				  msg->desc ? msg->desc[0] : NULL,
@@ -693,9 +942,18 @@ static ssize_t psmx2_readv(struct fid_ep *ep, const struct iovec *iov,
 		           void **desc, size_t count, fi_addr_t src_addr,
 		           uint64_t addr, uint64_t key, void *context)
 {
-	if (!iov || count != 1)
+	struct psmx2_fid_ep *ep_priv;
+
+	ep_priv = container_of(ep, struct psmx2_fid_ep, ep);
+
+	if (!iov || !count)
 		return -FI_EINVAL;
 
+	if (count > 1)
+		return psmx2_readv_generic(ep, iov, desc ? desc[0] : NULL,
+					   count, src_addr, addr, key,
+					   context, ep_priv->tx_flags);
+
 	return psmx2_read(ep, iov->iov_base, iov->iov_len,
 			  desc ? desc[0] : NULL, src_addr, addr, key, context);
 }
@@ -721,6 +979,7 @@ ssize_t psmx2_write_generic(struct fid_ep *ep, const void *buf, size_t len,
 	size_t idx;
 	void *psm2_context;
 	int no_event;
+	int sep_target = 0;
 
 	ep_priv = container_of(ep, struct psmx2_fid_ep, ep);
 
@@ -755,7 +1014,11 @@ ssize_t psmx2_write_generic(struct fid_ep *ep, const void *buf, size_t len,
 		return -FI_EINVAL;
 
 	av = ep_priv->av;
-	if (av && av->type == FI_AV_TABLE) {
+	if (av && PSMX2_SEP_ADDR_TEST(dest_addr)) {
+		psm2_epaddr = psmx2_av_translate_sep(av, ep_priv->trx_ctxt, dest_addr);
+		vlane = 0;
+		sep_target = 1;
+	} else if (av && av->type == FI_AV_TABLE) {
 		idx = dest_addr;
 		if (idx >= av->last)
 			return -FI_EINVAL;
@@ -771,9 +1034,10 @@ ssize_t psmx2_write_generic(struct fid_ep *ep, const void *buf, size_t len,
 	}
 
 	epaddr_context = psm2_epaddr_getctxt((void *)psm2_epaddr);
-	if (epaddr_context->epid == ep_priv->domain->psm2_epid)
+	if (epaddr_context->epid == ep_priv->trx_ctxt->psm2_epid)
 		return psmx2_rma_self(PSMX2_AM_REQ_WRITE, ep_priv,
-				      ep_priv->domain->eps[vlane],
+				      (sep_target ? ep_priv :
+					ep_priv->domain->eps[vlane]),
 				      (void *)buf, len, desc, addr,
 				      key, context, flags, data);
 
@@ -788,9 +1052,9 @@ ssize_t psmx2_write_generic(struct fid_ep *ep, const void *buf, size_t len,
 		if (!req)
 			return -FI_ENOMEM;
 
-		memset((void *)req, 0, sizeof(*req));
-		memcpy((void *)req + sizeof(*req), (void *)buf, len);
-		buf = (void *)req + sizeof(*req);
+		memset(req, 0, sizeof(*req));
+		memcpy((uint8_t *)req + sizeof(*req), (void *)buf, len);
+		buf = (uint8_t *)req + sizeof(*req);
 	} else {
 		req = calloc(1, sizeof(*req));
 		if (!req)
@@ -813,7 +1077,7 @@ ssize_t psmx2_write_generic(struct fid_ep *ep, const void *buf, size_t len,
 	PSMX2_CTXT_USER(&req->fi_context) = context;
 	PSMX2_CTXT_EP(&req->fi_context) = ep_priv;
 
-	chunk_size = psmx2_am_param.max_request_short;
+	chunk_size = ep_priv->trx_ctxt->psm2_am_param.max_request_short;
 
 	args[0].u32w0 = 0;
 	PSMX2_AM_SET_SRC(args[0].u32w0, ep_priv->vlane);
@@ -821,6 +1085,7 @@ ssize_t psmx2_write_generic(struct fid_ep *ep, const void *buf, size_t len,
 
 	if (psmx2_env.tagged_rma && len > chunk_size) {
 		tag32 = PSMX2_TAG32(PSMX2_RMA_BIT, ep_priv->vlane, vlane);
+		PSMX2_TAG32_LONG_WRITE(tag32);
 		PSMX2_SET_TAG(psm2_tag, (uint64_t)req, tag32);
 		PSMX2_AM_SET_OP(args[0].u32w0, PSMX2_AM_REQ_WRITE_LONG);
 		args[0].u32w1 = len;
@@ -844,7 +1109,7 @@ ssize_t psmx2_write_generic(struct fid_ep *ep, const void *buf, size_t len,
 		psm2_am_request_short(psm2_epaddr, PSMX2_AM_RMA_HANDLER, args,
 				      nargs, NULL, 0, am_flags, NULL, NULL);
 
-		psm2_mq_isend2(ep_priv->domain->psm2_mq, psm2_epaddr, 0,
+		psm2_mq_isend2(ep_priv->trx_ctxt->psm2_mq, psm2_epaddr, 0,
 			       &psm2_tag, buf, len, psm2_context, &psm2_req);
 
 		return 0;
@@ -860,7 +1125,7 @@ ssize_t psmx2_write_generic(struct fid_ep *ep, const void *buf, size_t len,
 		psm2_am_request_short(psm2_epaddr, PSMX2_AM_RMA_HANDLER, args,
 				      nargs, (void *)buf, chunk_size, am_flags,
 				      NULL, NULL);
-		buf += chunk_size;
+		buf = (const uint8_t *)buf + chunk_size;
 		addr += chunk_size;
 		len -= chunk_size;
 	}
@@ -904,8 +1169,9 @@ ssize_t psmx2_writev_generic(struct fid_ep *ep, const struct iovec *iov,
 	void *psm2_context;
 	int no_event;
 	size_t total_len, len, len_sent;
-	void *buf, *p;
+	uint8_t *buf, *p;
 	int i;
+	int sep_target = 0;
 
 	ep_priv = container_of(ep, struct psmx2_fid_ep, ep);
 
@@ -937,7 +1203,11 @@ ssize_t psmx2_writev_generic(struct fid_ep *ep, const struct iovec *iov,
 	}
 
 	av = ep_priv->av;
-	if (av && av->type == FI_AV_TABLE) {
+	if (av && PSMX2_SEP_ADDR_TEST(dest_addr)) {
+		psm2_epaddr = psmx2_av_translate_sep(av, ep_priv->trx_ctxt, dest_addr);
+		vlane = 0;
+		sep_target = 1;
+	} else if (av && av->type == FI_AV_TABLE) {
 		idx = dest_addr;
 		if (idx >= av->last)
 			return -FI_EINVAL;
@@ -953,9 +1223,10 @@ ssize_t psmx2_writev_generic(struct fid_ep *ep, const struct iovec *iov,
 	}
 
 	epaddr_context = psm2_epaddr_getctxt((void *)psm2_epaddr);
-	if (epaddr_context->epid == ep_priv->domain->psm2_epid)
+	if (epaddr_context->epid == ep_priv->trx_ctxt->psm2_epid)
 		return psmx2_rma_self(PSMX2_AM_REQ_WRITEV, ep_priv,
-				      ep_priv->domain->eps[vlane],
+				      (sep_target ? ep_priv :
+					ep_priv->domain->eps[vlane]),
 				      (void *)iov, count, desc, addr,
 				      key, context, flags, data);
 
@@ -966,7 +1237,7 @@ ssize_t psmx2_writev_generic(struct fid_ep *ep, const struct iovec *iov,
 	for (i=0; i<count; i++)
 		total_len += iov[i].iov_len;
 
-	chunk_size = psmx2_am_param.max_request_short;
+	chunk_size = ep_priv->trx_ctxt->psm2_am_param.max_request_short;
 
 	/* Case 1: fit into a AM message, then pack and send */
 	if (total_len <= chunk_size) {
@@ -974,15 +1245,15 @@ ssize_t psmx2_writev_generic(struct fid_ep *ep, const struct iovec *iov,
 		if (!req)
 			return -FI_ENOMEM;
 
-		memset((void *)req, 0, sizeof(*req));
-		p = (void *)req + sizeof(*req);
+		memset(req, 0, sizeof(*req));
+		p = (uint8_t *)req + sizeof(*req);
 		for (i=0; i<count; i++) {
 			if (iov[i].iov_len) {
 				memcpy(p, iov[i].iov_base, iov[i].iov_len);
 				p += iov[i].iov_len;
 			}
 		}
-		buf = (void *)req + sizeof(*req);
+		buf = (uint8_t *)req + sizeof(*req);
 		len = total_len;
 
 		req->no_event = no_event;
@@ -1056,6 +1327,7 @@ ssize_t psmx2_writev_generic(struct fid_ep *ep, const struct iovec *iov,
 		if (psmx2_env.tagged_rma && iov[i].iov_len > chunk_size &&
 		    len_sent + iov[i].iov_len == total_len) {
 			tag32 = PSMX2_TAG32(PSMX2_RMA_BIT, ep_priv->vlane, vlane);
+			PSMX2_TAG32_LONG_WRITE(tag32);
 			PSMX2_SET_TAG(psm2_tag, (uint64_t)req, tag32);
 			PSMX2_AM_SET_OP(args[0].u32w0, PSMX2_AM_REQ_WRITE_LONG);
 			args[0].u32w1 = iov[i].iov_len;
@@ -1079,7 +1351,7 @@ ssize_t psmx2_writev_generic(struct fid_ep *ep, const struct iovec *iov,
 			psm2_am_request_short(psm2_epaddr, PSMX2_AM_RMA_HANDLER, args,
 					      nargs, NULL, 0, am_flags, NULL, NULL);
 
-			psm2_mq_isend2(ep_priv->domain->psm2_mq, psm2_epaddr, 0,
+			psm2_mq_isend2(ep_priv->trx_ctxt->psm2_mq, psm2_epaddr, 0,
 				       &psm2_tag, iov[i].iov_base, iov[i].iov_len,
 				       psm2_context, &psm2_req);
 
@@ -1137,7 +1409,7 @@ static ssize_t psmx2_write(struct fid_ep *ep, const void *buf, size_t len,
 	ep_priv = container_of(ep, struct psmx2_fid_ep, ep);
 
 	return psmx2_write_generic(ep, buf, len, desc, dest_addr, addr,
-				   key, context, ep_priv->flags, 0);
+				   key, context, ep_priv->tx_flags, 0);
 }
 
 static ssize_t psmx2_writemsg(struct fid_ep *ep,
@@ -1175,11 +1447,11 @@ static ssize_t psmx2_writev(struct fid_ep *ep, const struct iovec *iov,
 
 	if (count > 1)
 		return psmx2_writev_generic(ep, iov, desc, count, dest_addr,
-					    addr, key, context, ep_priv->flags, 0);
+					    addr, key, context, ep_priv->tx_flags, 0);
 
 	return psmx2_write_generic(ep, iov->iov_base, iov->iov_len,
 				   desc ? desc[0] : NULL, dest_addr, addr, key,
-				   context, ep_priv->flags, 0);
+				   context, ep_priv->tx_flags, 0);
 }
 
 static ssize_t psmx2_inject(struct fid_ep *ep, const void *buf, size_t len,
@@ -1190,7 +1462,7 @@ static ssize_t psmx2_inject(struct fid_ep *ep, const void *buf, size_t len,
 	ep_priv = container_of(ep, struct psmx2_fid_ep, ep);
 
 	return psmx2_write_generic(ep, buf, len, NULL, dest_addr, addr, key, NULL,
-				   ep_priv->flags | FI_INJECT | PSMX2_NO_COMPLETION,
+				   ep_priv->tx_flags | FI_INJECT | PSMX2_NO_COMPLETION,
 				   0);
 }
 
@@ -1203,7 +1475,7 @@ static ssize_t psmx2_writedata(struct fid_ep *ep, const void *buf, size_t len,
 	ep_priv = container_of(ep, struct psmx2_fid_ep, ep);
 
 	return psmx2_write_generic(ep, buf, len, desc, dest_addr, addr, key,
-				   context, ep_priv->flags | FI_REMOTE_CQ_DATA,
+				   context, ep_priv->tx_flags | FI_REMOTE_CQ_DATA,
 				   data);
 }
 
@@ -1216,7 +1488,7 @@ static ssize_t psmx2_injectdata(struct fid_ep *ep, const void *buf, size_t len,
 	ep_priv = container_of(ep, struct psmx2_fid_ep, ep);
 
 	return psmx2_write_generic(ep, buf, len, NULL, dest_addr, addr, key, NULL,
-				   ep_priv->flags | FI_INJECT | PSMX2_NO_COMPLETION,
+				   ep_priv->tx_flags | FI_INJECT | PSMX2_NO_COMPLETION,
 				   data);
 }
 
