@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2014 Intel Corporation. All rights reserved.
+ * Copyright (c) 2013-2017 Intel Corporation. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -52,9 +52,12 @@ static ssize_t psmx2_tagged_peek_generic(struct fid_ep *ep,
 
 	ep_priv = container_of(ep, struct psmx2_fid_ep, ep);
 
-	if (src_addr != FI_ADDR_UNSPEC) {
+	if ((ep_priv->caps & FI_DIRECTED_RECV) && src_addr != FI_ADDR_UNSPEC) {
 		av = ep_priv->av;
-		if (av && av->type == FI_AV_TABLE) {
+		if (av && PSMX2_SEP_ADDR_TEST(src_addr)) {
+			psm2_epaddr = psmx2_av_translate_sep(av, ep_priv->trx_ctxt, src_addr);
+			vlane = 0;
+		} else if (av && av->type == FI_AV_TABLE) {
 			idx = (size_t)src_addr;
 			if (idx >= av->last)
 				return -FI_EINVAL;
@@ -77,18 +80,24 @@ static ssize_t psmx2_tagged_peek_generic(struct fid_ep *ep,
 	PSMX2_SET_TAG(psm2_tagsel, ~ignore, tagsel32);
 
 	if (flags & (FI_CLAIM | FI_DISCARD))
-		err = psm2_mq_improbe2(ep_priv->domain->psm2_mq,
+		err = psm2_mq_improbe2(ep_priv->trx_ctxt->psm2_mq,
 				       psm2_epaddr, &psm2_tag,
 				       &psm2_tagsel, &req, &psm2_status);
 	else
-		err = psm2_mq_iprobe2(ep_priv->domain->psm2_mq,
+		err = psm2_mq_iprobe2(ep_priv->trx_ctxt->psm2_mq,
 				      psm2_epaddr, &psm2_tag, &psm2_tagsel,
 				      &psm2_status);
 	switch (err) {
 	case PSM2_OK:
 		if (ep_priv->recv_cq) {
-			if ((flags & FI_CLAIM) && context)
-				PSMX2_CTXT_REQ((struct fi_context *)context) = req;
+			if (flags & FI_CLAIM) {
+				if (context)
+					PSMX2_CTXT_REQ((struct fi_context *)context) = req;
+			} else if (flags & FI_DISCARD) {
+				if (!psm2_mq_imrecv(ep_priv->trx_ctxt->psm2_mq, 0,
+						    NULL, 0, req, &req))
+					psm2_mq_wait2(&req, NULL);
+			}
 
 			tag = PSMX2_GET_TAG64(psm2_status.msg_tag);
 			len = psm2_status.msg_length;
@@ -107,7 +116,9 @@ static ssize_t psmx2_tagged_peek_generic(struct fid_ep *ep,
 				return -FI_ENOMEM;
 
 			vlane = PSMX2_TAG32_GET_SRC(psm2_status.msg_tag.tag2);
+			event->source_is_valid = 1;
 			event->source = PSMX2_EP_TO_ADDR(psm2_status.msg_peer, vlane);
+			event->source_av = ep_priv->av;
 			psmx2_cq_enqueue_event(ep_priv->recv_cq, event);
 		}
 		return 0;
@@ -192,7 +203,48 @@ ssize_t psmx2_tagged_recv_generic(struct fid_ep *ep, void *buf,
 		if (!context)
 			return -FI_EINVAL;
 
-		/* TODO: handle FI_DISCARD */
+		if (flags & FI_DISCARD) {
+			psm2_mq_status2_t psm2_status;
+			struct psmx2_cq_event *event;
+
+			fi_context = context;
+			psm2_req = PSMX2_CTXT_REQ(fi_context);
+			err = psm2_mq_imrecv(ep_priv->trx_ctxt->psm2_mq, 0,
+					     NULL, 0, context, &psm2_req);
+			if (err != PSM2_OK)
+				return psmx2_errno(err);
+
+			psm2_mq_wait2(&psm2_req, &psm2_status);
+
+			if (ep_priv->recv_cq &&
+			    (!ep_priv->recv_selective_completion || (flags & FI_COMPLETION))) {
+				tag = PSMX2_GET_TAG64(psm2_status.msg_tag);
+				event = psmx2_cq_create_event(
+						ep_priv->recv_cq,
+						context,		/* op_context */
+						NULL,			/* buf */
+						flags|FI_RECV|FI_TAGGED,/* flags */
+						0,			/* len */
+						0,			/* data */
+						tag,			/* tag */
+						0,			/* olen */
+						0);			/* err */
+
+				if (!event)
+					return -FI_ENOMEM;
+
+				vlane = PSMX2_TAG32_GET_SRC(psm2_status.msg_tag.tag2);
+				event->source_is_valid = 1;
+				event->source = PSMX2_EP_TO_ADDR(psm2_status.msg_peer, vlane);
+				event->source_av = ep_priv->av;
+				psmx2_cq_enqueue_event(ep_priv->recv_cq, event);
+			}
+
+			if (ep_priv->recv_cntr)
+				psmx2_cntr_inc(ep_priv->recv_cntr);
+
+			return 0;
+		}
 
 		fi_context = context;
 		psm2_req = PSMX2_CTXT_REQ(fi_context);
@@ -200,7 +252,7 @@ ssize_t psmx2_tagged_recv_generic(struct fid_ep *ep, void *buf,
 		PSMX2_CTXT_USER(fi_context) = buf;
 		PSMX2_CTXT_EP(fi_context) = ep_priv;
 
-		err = psm2_mq_imrecv(ep_priv->domain->psm2_mq, 0,
+		err = psm2_mq_imrecv(ep_priv->trx_ctxt->psm2_mq, 0,
 				     buf, len, context, &psm2_req);
 		if (err != PSM2_OK)
 			return psmx2_errno(err);
@@ -228,7 +280,10 @@ ssize_t psmx2_tagged_recv_generic(struct fid_ep *ep, void *buf,
 
 	if ((ep_priv->caps & FI_DIRECTED_RECV) && src_addr != FI_ADDR_UNSPEC) {
 		av = ep_priv->av;
-		if (av && av->type == FI_AV_TABLE) {
+		if (av && PSMX2_SEP_ADDR_TEST(src_addr)) {
+			psm2_epaddr = psmx2_av_translate_sep(av, ep_priv->trx_ctxt, src_addr);
+			vlane = 0;
+		} else if (av && av->type == FI_AV_TABLE) {
 			idx = (size_t)src_addr;
 			if (idx >= av->last)
 				return -FI_EINVAL;
@@ -250,7 +305,7 @@ ssize_t psmx2_tagged_recv_generic(struct fid_ep *ep, void *buf,
 	PSMX2_SET_TAG(psm2_tag, tag, tag32);
 	PSMX2_SET_TAG(psm2_tagsel, ~ignore, tagsel32);
 
-	err = psm2_mq_irecv2(ep_priv->domain->psm2_mq, psm2_epaddr,
+	err = psm2_mq_irecv2(ep_priv->trx_ctxt->psm2_mq, psm2_epaddr,
 			     &psm2_tag, &psm2_tagsel, 0, buf, len,
 			     (void *)fi_context, &psm2_req);
 
@@ -272,6 +327,7 @@ psmx2_tagged_recv_no_flag_av_map(struct fid_ep *ep, void *buf, size_t len,
 				 void *context)
 {
 	struct psmx2_fid_ep *ep_priv;
+	struct psmx2_fid_av *av;
 	psm2_epaddr_t psm2_epaddr;
 	uint8_t vlane;
 	psm2_mq_req_t psm2_req;
@@ -289,8 +345,14 @@ psmx2_tagged_recv_no_flag_av_map(struct fid_ep *ep, void *buf, size_t len,
 	PSMX2_CTXT_SIZE(fi_context) = len;
 
 	if ((ep_priv->caps & FI_DIRECTED_RECV) && src_addr != FI_ADDR_UNSPEC) {
-		psm2_epaddr = PSMX2_ADDR_TO_EP(src_addr);
-		vlane = PSMX2_ADDR_TO_VL(src_addr);
+		av = ep_priv->av;
+		if (av && PSMX2_SEP_ADDR_TEST(src_addr)) {
+			psm2_epaddr = psmx2_av_translate_sep(av, ep_priv->trx_ctxt, src_addr);
+			vlane = 0;
+		} else {
+			psm2_epaddr = PSMX2_ADDR_TO_EP(src_addr);
+			vlane = PSMX2_ADDR_TO_VL(src_addr);
+		}
 		tag32 = PSMX2_TAG32(0, vlane, ep_priv->vlane);
 		tagsel32 = ~PSMX2_IOV_BIT;
 	} else {
@@ -302,7 +364,7 @@ psmx2_tagged_recv_no_flag_av_map(struct fid_ep *ep, void *buf, size_t len,
 	PSMX2_SET_TAG(psm2_tag, tag, tag32);
 	PSMX2_SET_TAG(psm2_tagsel, ~ignore, tagsel32);
 
-	err = psm2_mq_irecv2(ep_priv->domain->psm2_mq, psm2_epaddr,
+	err = psm2_mq_irecv2(ep_priv->trx_ctxt->psm2_mq, psm2_epaddr,
 			     &psm2_tag, &psm2_tagsel, 0, buf, len,
 			     (void *)fi_context, &psm2_req);
 	if (err != PSM2_OK)
@@ -340,12 +402,18 @@ psmx2_tagged_recv_no_flag_av_table(struct fid_ep *ep, void *buf, size_t len,
 
 	if ((ep_priv->caps & FI_DIRECTED_RECV) && src_addr != FI_ADDR_UNSPEC) {
 		av = ep_priv->av;
-		idx = (size_t)src_addr;
-		if (idx >= av->last)
-			return -FI_EINVAL;
+		assert(av != NULL);
+		if (PSMX2_SEP_ADDR_TEST(src_addr)) {
+			psm2_epaddr = psmx2_av_translate_sep(av, ep_priv->trx_ctxt, src_addr);
+			vlane = 0;
+		} else {
+			idx = (size_t)src_addr;
+			if (idx >= av->last)
+				return -FI_EINVAL;
 
-		psm2_epaddr = av->epaddrs[idx];
-		vlane = av->vlanes[idx];
+			psm2_epaddr = av->epaddrs[idx];
+			vlane = av->vlanes[idx];
+		}
 		tag32 = PSMX2_TAG32(0, vlane, ep_priv->vlane);
 		tagsel32 = ~PSMX2_IOV_BIT;
 	} else {
@@ -357,7 +425,7 @@ psmx2_tagged_recv_no_flag_av_table(struct fid_ep *ep, void *buf, size_t len,
 	PSMX2_SET_TAG(psm2_tag, tag, tag32);
 	PSMX2_SET_TAG(psm2_tagsel, ~ignore, tagsel32);
 
-	err = psm2_mq_irecv2(ep_priv->domain->psm2_mq, psm2_epaddr,
+	err = psm2_mq_irecv2(ep_priv->trx_ctxt->psm2_mq, psm2_epaddr,
 			     &psm2_tag, &psm2_tagsel, 0, buf, len,
 			     (void *)fi_context, &psm2_req);
 	if (err != PSM2_OK)
@@ -375,6 +443,7 @@ psmx2_tagged_recv_no_event_av_map(struct fid_ep *ep, void *buf, size_t len,
 				  void *context)
 {
 	struct psmx2_fid_ep *ep_priv;
+	struct psmx2_fid_av *av;
 	psm2_epaddr_t psm2_epaddr;
 	uint8_t vlane;
 	psm2_mq_req_t psm2_req;
@@ -392,8 +461,14 @@ psmx2_tagged_recv_no_event_av_map(struct fid_ep *ep, void *buf, size_t len,
 	PSMX2_CTXT_SIZE(fi_context) = len;
 
 	if ((ep_priv->caps & FI_DIRECTED_RECV) && src_addr != FI_ADDR_UNSPEC) {
-		psm2_epaddr = PSMX2_ADDR_TO_EP(src_addr);
-		vlane = PSMX2_ADDR_TO_VL(src_addr);
+		av = ep_priv->av;
+		if (av && PSMX2_SEP_ADDR_TEST(src_addr)) {
+			psm2_epaddr = psmx2_av_translate_sep(av, ep_priv->trx_ctxt, src_addr);
+			vlane = 0;
+		} else {
+			psm2_epaddr = PSMX2_ADDR_TO_EP(src_addr);
+			vlane = PSMX2_ADDR_TO_VL(src_addr);
+		}
 		tag32 = PSMX2_TAG32(0, vlane, ep_priv->vlane);
 		tagsel32 = ~PSMX2_IOV_BIT;
 	} else {
@@ -405,7 +480,7 @@ psmx2_tagged_recv_no_event_av_map(struct fid_ep *ep, void *buf, size_t len,
 	PSMX2_SET_TAG(psm2_tag, tag, tag32);
 	PSMX2_SET_TAG(psm2_tagsel, ~ignore, tagsel32);
 
-	err = psm2_mq_irecv2(ep_priv->domain->psm2_mq, psm2_epaddr,
+	err = psm2_mq_irecv2(ep_priv->trx_ctxt->psm2_mq, psm2_epaddr,
 			     &psm2_tag, &psm2_tagsel, 0, buf, len,
 			     (void *)fi_context, &psm2_req);
 
@@ -440,12 +515,18 @@ psmx2_tagged_recv_no_event_av_table(struct fid_ep *ep, void *buf, size_t len,
 
 	if ((ep_priv->caps & FI_DIRECTED_RECV) && src_addr != FI_ADDR_UNSPEC) {
 		av = ep_priv->av;
-		idx = (size_t)src_addr;
-		if (idx >= av->last)
-			return -FI_EINVAL;
+		assert(av != NULL);
+		if (PSMX2_SEP_ADDR_TEST(src_addr)) {
+			psm2_epaddr = psmx2_av_translate_sep(av, ep_priv->trx_ctxt, src_addr);
+			vlane = 0;
+		} else {
+			idx = (size_t)src_addr;
+			if (idx >= av->last)
+				return -FI_EINVAL;
 
-		psm2_epaddr = av->epaddrs[idx];
-		vlane = av->vlanes[idx];
+			psm2_epaddr = av->epaddrs[idx];
+			vlane = av->vlanes[idx];
+		}
 		tag32 = PSMX2_TAG32(0, vlane, ep_priv->vlane);
 		tagsel32 = ~PSMX2_IOV_BIT;
 	} else {
@@ -457,7 +538,7 @@ psmx2_tagged_recv_no_event_av_table(struct fid_ep *ep, void *buf, size_t len,
 	PSMX2_SET_TAG(psm2_tag, tag, tag32);
 	PSMX2_SET_TAG(psm2_tagsel, ~ignore, tagsel32);
 
-	err = psm2_mq_irecv2(ep_priv->domain->psm2_mq, psm2_epaddr,
+	err = psm2_mq_irecv2(ep_priv->trx_ctxt->psm2_mq, psm2_epaddr,
 			     &psm2_tag, &psm2_tagsel, 0, buf, len,
 			     (void *)fi_context, &psm2_req);
 
@@ -473,7 +554,7 @@ static ssize_t psmx2_tagged_recv(struct fid_ep *ep, void *buf, size_t len,
 	ep_priv = container_of(ep, struct psmx2_fid_ep, ep);
 
 	return psmx2_tagged_recv_generic(ep, buf, len, desc, src_addr, tag,
-					 ignore, context, ep_priv->flags);
+					 ignore, context, ep_priv->rx_flags);
 }
 
 static ssize_t psmx2_tagged_recvmsg(struct fid_ep *ep,
@@ -487,7 +568,7 @@ static ssize_t psmx2_tagged_recvmsg(struct fid_ep *ep,
 		return -FI_EINVAL;
 
 	if (msg->iov_count > 1) {
-		return -FI_EINVAL;
+		return -FI_ENOSYS;
 	} else if (msg->iov_count) {
 		buf = msg->msg_iov[0].iov_base;
 		len = msg->msg_iov[0].iov_len;
@@ -511,8 +592,10 @@ psmx2_tagged_recvv##suffix(struct fid_ep *ep, const struct iovec *iov,	\
 {									\
 	void *buf;							\
 	size_t len;							\
-	if ((count && !iov) || (count > 1))				\
+	if (count && !iov)						\
 		return -FI_EINVAL;					\
+	if (count > 1)							\
+		return -FI_ENOSYS;					\
 	if (count) {							\
 		buf = iov[0].iov_base;					\
 		len = iov[0].iov_len;					\
@@ -581,7 +664,10 @@ ssize_t psmx2_tagged_send_generic(struct fid_ep *ep,
 	}
 
 	av = ep_priv->av;
-	if (av && av->type == FI_AV_TABLE) {
+	if (av && PSMX2_SEP_ADDR_TEST(dest_addr)) {
+		psm2_epaddr = psmx2_av_translate_sep(av, ep_priv->trx_ctxt, dest_addr);
+		vlane = 0;
+	} else  if (av && av->type == FI_AV_TABLE) {
 		idx = (size_t)dest_addr;
 		if (idx >= av->last)
 			return -FI_EINVAL;
@@ -601,10 +687,10 @@ ssize_t psmx2_tagged_send_generic(struct fid_ep *ep,
 		no_completion = 1;
 
 	if (flags & FI_INJECT) {
-		if (len > PSMX2_INJECT_SIZE)
+		if (len > psmx2_env.inject_size)
 			return -FI_EMSGSIZE;
 
-		err = psm2_mq_send2(ep_priv->domain->psm2_mq, psm2_epaddr,
+		err = psm2_mq_send2(ep_priv->trx_ctxt->psm2_mq, psm2_epaddr,
 				    0, &psm2_tag, buf, len);
 
 		if (err != PSM2_OK)
@@ -630,7 +716,7 @@ ssize_t psmx2_tagged_send_generic(struct fid_ep *ep,
 		return 0;
 	}
 
-	if (no_completion && !context) {
+	if (no_completion) {
 		fi_context = &ep_priv->nocomp_send_context;
 	} else {
 		if (!context)
@@ -642,7 +728,7 @@ ssize_t psmx2_tagged_send_generic(struct fid_ep *ep,
 		PSMX2_CTXT_EP(fi_context) = ep_priv;
 	}
 
-	err = psm2_mq_isend2(ep_priv->domain->psm2_mq, psm2_epaddr, 0,
+	err = psm2_mq_isend2(ep_priv->trx_ctxt->psm2_mq, psm2_epaddr, 0,
 			     &psm2_tag, buf, len, (void*)fi_context,
 			     &psm2_req);
 
@@ -663,6 +749,7 @@ psmx2_tagged_send_no_flag_av_map(struct fid_ep *ep, const void *buf,
 				 void *context)
 {
 	struct psmx2_fid_ep *ep_priv;
+	struct psmx2_fid_av *av;
 	psm2_epaddr_t psm2_epaddr;
 	uint8_t vlane;
 	psm2_mq_req_t psm2_req;
@@ -673,8 +760,14 @@ psmx2_tagged_send_no_flag_av_map(struct fid_ep *ep, const void *buf,
 
 	ep_priv = container_of(ep, struct psmx2_fid_ep, ep);
 
-	psm2_epaddr = PSMX2_ADDR_TO_EP(dest_addr);
-	vlane = PSMX2_ADDR_TO_VL(dest_addr);
+	av = ep_priv->av;
+	if (av && PSMX2_SEP_ADDR_TEST(dest_addr)) {
+		psm2_epaddr = psmx2_av_translate_sep(av, ep_priv->trx_ctxt, dest_addr);
+		vlane = 0;
+	} else {
+		psm2_epaddr = PSMX2_ADDR_TO_EP(dest_addr);
+		vlane = PSMX2_ADDR_TO_VL(dest_addr);
+	}
 	tag32 = PSMX2_TAG32(0, ep_priv->vlane, vlane);
 	PSMX2_SET_TAG(psm2_tag, tag, tag32);
 
@@ -683,7 +776,7 @@ psmx2_tagged_send_no_flag_av_map(struct fid_ep *ep, const void *buf,
 	PSMX2_CTXT_USER(fi_context) = (void *)buf;
 	PSMX2_CTXT_EP(fi_context) = ep_priv;
 
-	err = psm2_mq_isend2(ep_priv->domain->psm2_mq, psm2_epaddr, 0,
+	err = psm2_mq_isend2(ep_priv->trx_ctxt->psm2_mq, psm2_epaddr, 0,
 			     &psm2_tag, buf, len, (void*)fi_context,
 			     &psm2_req);
 
@@ -715,12 +808,18 @@ psmx2_tagged_send_no_flag_av_table(struct fid_ep *ep, const void *buf,
 	ep_priv = container_of(ep, struct psmx2_fid_ep, ep);
 
 	av = ep_priv->av;
-	idx = (size_t)dest_addr;
-	if (idx >= av->last)
-		return -FI_EINVAL;
+	assert(av != NULL);
+	if (PSMX2_SEP_ADDR_TEST(dest_addr)) {
+		psm2_epaddr = psmx2_av_translate_sep(av, ep_priv->trx_ctxt, dest_addr);
+		vlane = 0;
+	} else {
+		idx = (size_t)dest_addr;
+		if (idx >= av->last)
+			return -FI_EINVAL;
 
-	psm2_epaddr = av->epaddrs[idx];
-	vlane = av->vlanes[idx];
+		psm2_epaddr = av->epaddrs[idx];
+		vlane = av->vlanes[idx];
+	}
 	tag32 = PSMX2_TAG32(0, ep_priv->vlane, vlane);
 	PSMX2_SET_TAG(psm2_tag, tag, tag32);
 
@@ -729,7 +828,7 @@ psmx2_tagged_send_no_flag_av_table(struct fid_ep *ep, const void *buf,
 	PSMX2_CTXT_USER(fi_context) = (void *)buf;
 	PSMX2_CTXT_EP(fi_context) = ep_priv;
 
-	err = psm2_mq_isend2(ep_priv->domain->psm2_mq, psm2_epaddr, 0,
+	err = psm2_mq_isend2(ep_priv->trx_ctxt->psm2_mq, psm2_epaddr, 0,
 			     &psm2_tag, buf, len, (void*)fi_context,
 			     &psm2_req);
 
@@ -748,6 +847,7 @@ psmx2_tagged_send_no_event_av_map(struct fid_ep *ep, const void *buf,
 			          void *context)
 {
 	struct psmx2_fid_ep *ep_priv;
+	struct psmx2_fid_av *av;
 	psm2_epaddr_t psm2_epaddr;
 	uint8_t vlane;
 	psm2_mq_req_t psm2_req;
@@ -758,14 +858,20 @@ psmx2_tagged_send_no_event_av_map(struct fid_ep *ep, const void *buf,
 
 	ep_priv = container_of(ep, struct psmx2_fid_ep, ep);
 
-	psm2_epaddr = PSMX2_ADDR_TO_EP(dest_addr);
-	vlane = PSMX2_ADDR_TO_VL(dest_addr);
+	av = ep_priv->av;
+	if (av && PSMX2_SEP_ADDR_TEST(dest_addr)) {
+		psm2_epaddr = psmx2_av_translate_sep(av, ep_priv->trx_ctxt, dest_addr);
+		vlane = 0;
+	} else {
+		psm2_epaddr = PSMX2_ADDR_TO_EP(dest_addr);
+		vlane = PSMX2_ADDR_TO_VL(dest_addr);
+	}
 	tag32 = PSMX2_TAG32(0, ep_priv->vlane, vlane);
 	PSMX2_SET_TAG(psm2_tag, tag, tag32);
 
 	fi_context = &ep_priv->nocomp_send_context;
 
-	err = psm2_mq_isend2(ep_priv->domain->psm2_mq, psm2_epaddr, 0,
+	err = psm2_mq_isend2(ep_priv->trx_ctxt->psm2_mq, psm2_epaddr, 0,
 			     &psm2_tag, buf, len, (void*)fi_context,
 			     &psm2_req);
 
@@ -796,18 +902,24 @@ psmx2_tagged_send_no_event_av_table(struct fid_ep *ep, const void *buf,
 	ep_priv = container_of(ep, struct psmx2_fid_ep, ep);
 
 	av = ep_priv->av;
-	idx = (size_t)dest_addr;
-	if (idx >= av->last)
-		return -FI_EINVAL;
+	assert(av != NULL);
+	if (PSMX2_SEP_ADDR_TEST(dest_addr)) {
+		psm2_epaddr = psmx2_av_translate_sep(av, ep_priv->trx_ctxt, dest_addr);
+		vlane = 0;
+	} else {
+		idx = (size_t)dest_addr;
+		if (idx >= av->last)
+			return -FI_EINVAL;
 
-	psm2_epaddr = av->epaddrs[idx];
-	vlane = av->vlanes[idx];
+		psm2_epaddr = av->epaddrs[idx];
+		vlane = av->vlanes[idx];
+	}
 	tag32 = PSMX2_TAG32(0, ep_priv->vlane, vlane);
 	PSMX2_SET_TAG(psm2_tag, tag, tag32);
 
 	fi_context = &ep_priv->nocomp_send_context;
 
-	err = psm2_mq_isend2(ep_priv->domain->psm2_mq, psm2_epaddr, 0,
+	err = psm2_mq_isend2(ep_priv->trx_ctxt->psm2_mq, psm2_epaddr, 0,
 			     &psm2_tag, buf, len, (void*)fi_context,
 			     &psm2_req);
 
@@ -824,23 +936,32 @@ psmx2_tagged_inject_no_flag_av_map(struct fid_ep *ep, const void *buf,
 				   uint64_t tag)
 {
 	struct psmx2_fid_ep *ep_priv;
+	struct psmx2_fid_av *av;
 	psm2_epaddr_t psm2_epaddr;
 	uint8_t vlane;
 	psm2_mq_tag_t psm2_tag;
 	uint32_t tag32;
 	int err;
 
-	if (len > PSMX2_INJECT_SIZE)
+	if (OFI_UNLIKELY(len > psmx2_env.inject_size))
 		return -FI_EMSGSIZE;
 
+	PSMX2_SET_TAG_FIRST64(psm2_tag, tag);
+
 	ep_priv = container_of(ep, struct psmx2_fid_ep, ep);
 
-	psm2_epaddr = PSMX2_ADDR_TO_EP(dest_addr);
-	vlane = PSMX2_ADDR_TO_VL(dest_addr);
+	av = ep_priv->av;
+	if (av && PSMX2_SEP_ADDR_TEST(dest_addr)) {
+		psm2_epaddr = psmx2_av_translate_sep(av, ep_priv->trx_ctxt, dest_addr);
+		vlane = 0;
+	} else {
+		psm2_epaddr = PSMX2_ADDR_TO_EP(dest_addr);
+		vlane = PSMX2_ADDR_TO_VL(dest_addr);
+	}
 	tag32 = PSMX2_TAG32(0, ep_priv->vlane, vlane);
-	PSMX2_SET_TAG(psm2_tag, tag, tag32);
+	PSMX2_SET_TAG_LAST32(psm2_tag, tag32);
 
-	err = psm2_mq_send2(ep_priv->domain->psm2_mq, psm2_epaddr, 0,
+	err = psm2_mq_send2(ep_priv->trx_ctxt->psm2_mq, psm2_epaddr, 0,
 			    &psm2_tag, buf, len);
 
 	if (err != PSM2_OK)
@@ -867,22 +988,28 @@ psmx2_tagged_inject_no_flag_av_table(struct fid_ep *ep, const void *buf,
 	int err;
 	size_t idx;
 
-	if (len > PSMX2_INJECT_SIZE)
+	if (OFI_UNLIKELY(len > psmx2_env.inject_size))
 		return -FI_EMSGSIZE;
 
 	ep_priv = container_of(ep, struct psmx2_fid_ep, ep);
 
 	av = ep_priv->av;
-	idx = (size_t)dest_addr;
-	if (idx >= av->last)
-		return -FI_EINVAL;
+	assert(av != NULL);
+	if (PSMX2_SEP_ADDR_TEST(dest_addr)) {
+		psm2_epaddr = psmx2_av_translate_sep(av, ep_priv->trx_ctxt, dest_addr);
+		vlane = 0;
+	} else {
+		idx = (size_t)dest_addr;
+		if (idx >= av->last)
+			return -FI_EINVAL;
 
-	psm2_epaddr = av->epaddrs[idx];
-	vlane = av->vlanes[idx];
+		psm2_epaddr = av->epaddrs[idx];
+		vlane = av->vlanes[idx];
+	}
 	tag32 = PSMX2_TAG32(0, ep_priv->vlane, vlane);
 	PSMX2_SET_TAG(psm2_tag, tag, tag32);
 
-	err = psm2_mq_send2(ep_priv->domain->psm2_mq, psm2_epaddr, 0,
+	err = psm2_mq_send2(ep_priv->trx_ctxt->psm2_mq, psm2_epaddr, 0,
 			    &psm2_tag, buf, len);
 
 	if (err != PSM2_OK)
@@ -992,7 +1119,10 @@ ssize_t psmx2_tagged_sendv_generic(struct fid_ep *ep,
 	}
 
 	av = ep_priv->av;
-	if (av && av->type == FI_AV_TABLE) {
+	if (av && PSMX2_SEP_ADDR_TEST(dest_addr)) {
+		psm2_epaddr = psmx2_av_translate_sep(av, ep_priv->trx_ctxt, dest_addr);
+		vlane = 0;
+	} else  if (av && av->type == FI_AV_TABLE) {
 		idx = (size_t)dest_addr;
 		if (idx >= av->last) {
 			free(req);
@@ -1014,12 +1144,12 @@ ssize_t psmx2_tagged_sendv_generic(struct fid_ep *ep,
 		no_completion = 1;
 
 	if (flags & FI_INJECT) {
-		if (len > PSMX2_INJECT_SIZE) {
+		if (len > psmx2_env.inject_size) {
 			free(req);
 			return -FI_EMSGSIZE;
 		}
 
-		err = psm2_mq_send2(ep_priv->domain->psm2_mq, psm2_epaddr,
+		err = psm2_mq_send2(ep_priv->trx_ctxt->psm2_mq, psm2_epaddr,
 				    send_flag, &psm2_tag, req->buf, len);
 
 		free(req);
@@ -1057,7 +1187,7 @@ ssize_t psmx2_tagged_sendv_generic(struct fid_ep *ep,
 	PSMX2_CTXT_USER(fi_context) = req;
 	PSMX2_CTXT_EP(fi_context) = ep_priv;
 
-	err = psm2_mq_isend2(ep_priv->domain->psm2_mq, psm2_epaddr,
+	err = psm2_mq_isend2(ep_priv->trx_ctxt->psm2_mq, psm2_epaddr,
 			     send_flag, &psm2_tag, req->buf, len,
 			     (void *)fi_context, &psm2_req);
 
@@ -1078,7 +1208,7 @@ ssize_t psmx2_tagged_sendv_generic(struct fid_ep *ep,
 		PSMX2_SET_TAG(psm2_tag, tag, tag32);
 		for (i=0; i<count; i++) {
 			if (iov[i].iov_len) {
-				err = psm2_mq_isend2(ep_priv->domain->psm2_mq,
+				err = psm2_mq_isend2(ep_priv->trx_ctxt->psm2_mq,
 						     psm2_epaddr, send_flag, &psm2_tag,
 						     iov[i].iov_base, iov[i].iov_len,
 						     (void *)fi_context, &psm2_req);
@@ -1101,7 +1231,7 @@ static ssize_t psmx2_tagged_send(struct fid_ep *ep,
 	ep_priv = container_of(ep, struct psmx2_fid_ep, ep);
 
 	return psmx2_tagged_send_generic(ep, buf, len, desc, dest_addr,
-					 tag, context, ep_priv->flags, 0);
+					 tag, context, ep_priv->tx_flags, 0);
 }
 
 static ssize_t psmx2_tagged_sendmsg(struct fid_ep *ep,
@@ -1153,7 +1283,7 @@ psmx2_tagged_sendv##suffix(struct fid_ep *ep, const struct iovec *iov,	\
 		return psmx2_tagged_sendv_generic(ep, iov, desc, count, \
 						  dest_addr, tag,	\
 						  context,		\
-						  ep_priv->flags, 0);	\
+						  ep_priv->tx_flags, 0);\
 	} else if (count) {						\
 		buf = iov[0].iov_base;					\
 		len = iov[0].iov_len;					\
@@ -1182,7 +1312,7 @@ static ssize_t psmx2_tagged_inject(struct fid_ep *ep,
 
 	return psmx2_tagged_send_generic(ep, buf, len, NULL, dest_addr,
 					 tag, NULL,
-				  	 ep_priv->flags | FI_INJECT | PSMX2_NO_COMPLETION,
+				  	 ep_priv->tx_flags | FI_INJECT | PSMX2_NO_COMPLETION,
 					 0);
 }
 
