@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2014 Intel Corporation, Inc.  All rights reserved.
+ * Copyright (c) 2016 Cisco Systems, Inc.  All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -46,6 +47,7 @@
 #endif
 
 #include "prov.h"
+#include "fi_osd.h"
 
 #include "sock.h"
 #include "sock_util.h"
@@ -205,10 +207,6 @@ int sock_verify_fabric_attr(struct fi_fabric_attr *attr)
 	if (!attr)
 		return 0;
 
-	if (attr->name &&
-	    strcmp(attr->name, sock_fab_name))
-		return -FI_ENODATA;
-
 	if (attr->prov_version) {
 		if (attr->prov_version !=
 		   FI_VERSION(SOCK_MAJOR_VERSION, SOCK_MINOR_VERSION))
@@ -218,7 +216,7 @@ int sock_verify_fabric_attr(struct fi_fabric_attr *attr)
 	return 0;
 }
 
-int sock_verify_info(struct fi_info *hints)
+int sock_verify_info(uint32_t version, struct fi_info *hints)
 {
 	uint64_t caps;
 	enum fi_ep_type ep_type;
@@ -267,6 +265,7 @@ int sock_verify_info(struct fi_info *hints)
 	case FI_SOCKADDR_IN:
 		break;
 	default:
+		SOCK_LOG_DBG("Unsupported address format\n");
 		return -FI_ENODATA;
 	}
 
@@ -278,7 +277,7 @@ int sock_verify_info(struct fi_info *hints)
 			return -FI_ENODATA;
 		}
 	}
-	ret = sock_verify_domain_attr(hints->domain_attr);
+	ret = sock_verify_domain_attr(version, hints->domain_attr);
 	if (ret)
 		return ret;
 
@@ -316,7 +315,7 @@ static int sock_fabric_close(fid_t fid)
 {
 	struct sock_fabric *fab;
 	fab = container_of(fid, struct sock_fabric, fab_fid);
-	if (atomic_get(&fab->ref))
+	if (ofi_atomic_get32(&fab->ref))
 		return -FI_EBUSY;
 
 	sock_fab_remove_from_list(fab);
@@ -356,9 +355,6 @@ static int sock_fabric(struct fi_fabric_attr *attr,
 {
 	struct sock_fabric *fab;
 
-	if (strcmp(attr->name, sock_fab_name))
-		return -FI_EINVAL;
-
 	fab = calloc(1, sizeof(*fab));
 	if (!fab)
 		return -FI_ENOMEM;
@@ -373,7 +369,7 @@ static int sock_fabric(struct fi_fabric_attr *attr,
 	fab->fab_fid.fid.ops = &sock_fab_fi_ops;
 	fab->fab_fid.ops = &sock_fab_ops;
 	*fabric = &fab->fab_fid;
-	atomic_initialize(&fab->ref, 0);
+	ofi_atomic_initialize32(&fab->ref, 0);
 #if ENABLE_DEBUG
 	fab->num_send_msg = 0;
 #endif
@@ -381,65 +377,13 @@ static int sock_fabric(struct fi_fabric_attr *attr,
 	return 0;
 }
 
-static struct sock_service_entry *sock_fabric_find_service(struct sock_fabric *fab,
-							   int service)
-{
-	struct dlist_entry *entry;
-	struct sock_service_entry *service_entry;
-
-	for (entry = fab->service_list.next; entry != &fab->service_list;
-	     entry = entry->next) {
-		service_entry = container_of(entry,
-					     struct sock_service_entry, entry);
-		if (service_entry->service == service)
-			return service_entry;
-	}
-	return NULL;
-}
-
-int sock_fabric_check_service(struct sock_fabric *fab, int service)
-{
-	struct sock_service_entry *entry;
-
-	fastlock_acquire(&fab->lock);
-	entry = sock_fabric_find_service(fab, service);
-	fastlock_release(&fab->lock);
-	return (entry == NULL) ? 1 : 0;
-}
-
-void sock_fabric_add_service(struct sock_fabric *fab, int service)
-{
-	struct sock_service_entry *entry;
-
-	entry = calloc(1, sizeof(*entry));
-	if (!entry)
-		return;
-
-	entry->service = service;
-	fastlock_acquire(&fab->lock);
-	dlist_insert_tail(&entry->entry, &fab->service_list);
-	fastlock_release(&fab->lock);
-}
-
-void sock_fabric_remove_service(struct sock_fabric *fab, int service)
-{
-	struct sock_service_entry *service_entry;
-	fastlock_acquire(&fab->lock);
-	service_entry = sock_fabric_find_service(fab, service);
-	if (service_entry) {
-		dlist_remove(&service_entry->entry);
-		free(service_entry);
-	}
-	fastlock_release(&fab->lock);
-}
-
 int sock_get_src_addr(struct sockaddr_in *dest_addr,
 		      struct sockaddr_in *src_addr)
 {
 	int sock, ret;
 	socklen_t len;
 
-	sock = socket(AF_INET, SOCK_DGRAM, 0);
+	sock = ofi_socket(AF_INET, SOCK_DGRAM, 0);
 	if (sock < 0)
 		return -errno;
 
@@ -447,7 +391,6 @@ int sock_get_src_addr(struct sockaddr_in *dest_addr,
 	ret = connect(sock, (struct sockaddr *) dest_addr, len);
 	if (ret) {
 		SOCK_LOG_DBG("Failed to connect udp socket\n");
-
 		ret = sock_get_src_addr_from_hostname(src_addr, NULL);
 		goto out;
 	}
@@ -459,11 +402,25 @@ int sock_get_src_addr(struct sockaddr_in *dest_addr,
 		ret = -errno;
 	}
 out:
-	close(sock);
+	ofi_close_socket(sock);
 	return ret;
 }
 
-static int sock_ep_getinfo(const char *node, const char *service, uint64_t flags,
+static int sock_fi_checkinfo(struct fi_info *info, struct fi_info *hints)
+{
+	if (hints && hints->domain_attr && hints->domain_attr->name &&
+             strcmp(info->domain_attr->name, hints->domain_attr->name))
+		return -FI_ENODATA;
+
+	if (hints && hints->fabric_attr && hints->fabric_attr->name &&
+             strcmp(info->fabric_attr->name, hints->fabric_attr->name))
+		return -FI_ENODATA;
+
+	return 0;
+}
+
+static int sock_ep_getinfo(uint32_t version, const char *node,
+			   const char *service, uint64_t flags,
 			   struct fi_info *hints, enum fi_ep_type ep_type,
 			   struct fi_info **info)
 {
@@ -486,7 +443,6 @@ static int sock_ep_getinfo(const char *node, const char *service, uint64_t flags
 			return -FI_ENODATA;
 		}
 		src_addr = (struct sockaddr_in *) rai->ai_addr;
-
 		if (hints && hints->dest_addr)
 			dest_addr = hints->dest_addr;
 	} else {
@@ -497,7 +453,7 @@ static int sock_ep_getinfo(const char *node, const char *service, uint64_t flags
 				return -FI_ENODATA;
 			}
 			dest_addr = (struct sockaddr_in *) rai->ai_addr;
-		} else {
+		} else if (hints) {
 			dest_addr = hints->dest_addr;
 		}
 
@@ -518,13 +474,13 @@ static int sock_ep_getinfo(const char *node, const char *service, uint64_t flags
 
 	switch (ep_type) {
 	case FI_EP_MSG:
-		ret = sock_msg_fi_info(src_addr, dest_addr, hints, info);
+		ret = sock_msg_fi_info(version, src_addr, dest_addr, hints, info);
 		break;
 	case FI_EP_DGRAM:
-		ret = sock_dgram_fi_info(src_addr, dest_addr, hints, info);
+		ret = sock_dgram_fi_info(version, src_addr, dest_addr, hints, info);
 		break;
 	case FI_EP_RDM:
-		ret = sock_rdm_fi_info(src_addr, dest_addr, hints, info);
+		ret = sock_rdm_fi_info(version, src_addr, dest_addr, hints, info);
 		break;
 	default:
 		ret = -FI_ENODATA;
@@ -533,93 +489,101 @@ static int sock_ep_getinfo(const char *node, const char *service, uint64_t flags
 
 	if (rai)
 		freeaddrinfo(rai);
+
+	if (ret == 0)
+		return sock_fi_checkinfo(*info, hints);
+
 	return ret;
 }
 
-void sock_getnodename(char *buf, int buflen)
+void sock_insert_loopback_addr(struct slist *addr_list)
 {
-	int ret;
-	struct addrinfo ai, *rai = NULL;
-	struct ifaddrs *ifaddrs, *ifa;
+	struct sock_host_list_entry *addr_entry;
 
-	ret = gethostname(buf, buflen);
-	if (ret == 0) {
-		memset(&ai, 0, sizeof(ai));
-		ai.ai_family = AF_INET;
-		ret = getaddrinfo(buf, NULL, &ai, &rai);
-		if (!ret) {
-			freeaddrinfo(rai);
-			return;
-		}
-	}
-
-#if HAVE_GETIFADDRS
-	ret = getifaddrs(&ifaddrs);
-	for(ifa = ifaddrs; ifa != NULL; ifa = ifa->ifa_next) {
-		if (ifa->ifa_addr == NULL || (ifa->ifa_addr->sa_family != AF_INET))
-			continue;
-
-		ret = getnameinfo(ifa->ifa_addr, sizeof(struct sockaddr_in),
-				  buf, buflen, NULL, 0, NI_NUMERICHOST);
-		if (ret == 0) {
-			freeifaddrs(ifaddrs);
-			return;
-		}
-	}
-	freeifaddrs(ifaddrs);
-#endif
-	/* no reasonable address found, try loopback */
-	strncpy(buf, "127.0.0.1", buflen);
+	addr_entry = calloc(1, sizeof(struct sock_host_list_entry));
+	strncpy(addr_entry->hostname, "127.0.0.1", sizeof(addr_entry->hostname));
+	slist_insert_tail(&addr_entry->entry, addr_list);
 }
 
-static int sock_getinfo(uint32_t version, const char *node, const char *service,
-			uint64_t flags, struct fi_info *hints, struct fi_info **info)
+#if HAVE_GETIFADDRS
+void sock_get_list_of_addr(struct slist *addr_list)
 {
-	struct fi_info *cur, *tail;
-	enum fi_ep_type ep_type;
-	char hostname[HOST_NAME_MAX];
 	int ret;
+	struct sock_host_list_entry *addr_entry;
+	struct ifaddrs *ifaddrs, *ifa;
 
-	if (!(flags & FI_SOURCE) && hints && hints->src_addr &&
-	    (hints->src_addrlen != sizeof(struct sockaddr_in)))
-		return -FI_ENODATA;
+	ret = ofi_getifaddrs(&ifaddrs);
+	if (!ret) {
+		for (ifa = ifaddrs; ifa != NULL; ifa = ifa->ifa_next) {
+			if (ifa->ifa_addr == NULL || !(ifa->ifa_flags & IFF_UP) ||
+			     (ifa->ifa_addr->sa_family != AF_INET) ||
+			     !strcmp(ifa->ifa_name, "lo"))
+				continue;
 
-	if (((!node && !service) || (flags & FI_SOURCE)) &&
-	    hints && hints->dest_addr &&
-	    (hints->dest_addrlen != sizeof(struct sockaddr_in)))
-		return -FI_ENODATA;
-
-	ret = sock_verify_info(hints);
-	if (ret)
-		return ret;
-
-	if (!node && !service && !hints) {
-		flags |= FI_SOURCE;
-		sock_getnodename(hostname, sizeof(hostname));
-		node = hostname;
+			addr_entry = calloc(1, sizeof(struct sock_host_list_entry));
+			ret = getnameinfo(ifa->ifa_addr, sizeof(struct sockaddr_in),
+					  addr_entry->hostname, sizeof(addr_entry->hostname),
+					  NULL, 0, NI_NUMERICHOST);
+			if (ret) {
+				SOCK_LOG_DBG("getnameinfo failed: %d\n", ret);
+				free(addr_entry);
+				continue;
+			}
+			slist_insert_tail(&addr_entry->entry, addr_list);
+		}
+		freeifaddrs(ifaddrs);
 	}
+	// Always add loopback address at the end
+	sock_insert_loopback_addr(addr_list);
+}
+#elif defined HAVE_MIB_IPADDRTABLE
+void sock_get_list_of_addr(struct slist *addr_list)
+{
+	sock_get_ip_addr_table(addr_list);
+	sock_insert_loopback_addr(addr_list);
+}
+#else
+void sock_get_list_of_addr(struct slist *addr_list)
+{
+	sock_insert_loopback_addr(addr_list);
+}
+#endif
 
-	if (!node && !service && !(flags & FI_SOURCE) && !hints->dest_addr) {
-		sock_getnodename(hostname, sizeof(hostname));
-		node = hostname;
-	}
+int sock_node_getinfo(uint32_t version, const char *node, const char *service,
+		      uint64_t flags, struct fi_info *hints, struct fi_info **info,
+		      struct fi_info **tail)
+{
+	enum fi_ep_type ep_type;
+	struct fi_info *cur;
+	int ret;
 
 	if (hints && hints->ep_attr) {
 		switch (hints->ep_attr->type) {
 		case FI_EP_RDM:
 		case FI_EP_DGRAM:
 		case FI_EP_MSG:
-			return sock_ep_getinfo(node, service, flags, hints,
-						hints->ep_attr->type, info);
+			ret = sock_ep_getinfo(version, node, service, flags,
+					      hints, hints->ep_attr->type, &cur);
+			if (ret) {
+				if (ret == -FI_ENODATA)
+					return ret;
+				goto err;
+			}
+
+			if (!*info)
+				*info = cur;
+			else
+				(*tail)->next = cur;
+			for (*tail = cur; (*tail)->next; *tail = (*tail)->next)
+				;
+			return 0;
 		default:
 			break;
 		}
 	}
-
-	*info = tail = NULL;
 	for (ep_type = FI_EP_MSG; ep_type <= FI_EP_RDM; ep_type++) {
-		ret = sock_ep_getinfo(node, service, flags,
-					hints, ep_type, &cur);
+		ret = sock_ep_getinfo(version, node, service, flags, hints,
+				      ep_type, &cur);
 		if (ret) {
 			if (ret == -FI_ENODATA)
 				continue;
@@ -629,8 +593,8 @@ static int sock_getinfo(uint32_t version, const char *node, const char *service,
 		if (!*info)
 			*info = cur;
 		else
-			tail->next = cur;
-		for (tail = cur; tail->next; tail = tail->next)
+			(*tail)->next = cur;
+		for (*tail = cur; (*tail)->next; *tail = (*tail)->next)
 			;
 	}
 	if (!*info) {
@@ -646,6 +610,127 @@ err_no_free:
 	return ret;
 }
 
+static int sock_match_src_addr(struct slist_entry *entry, const void *src_addr)
+{
+	struct sock_host_list_entry *host_entry;
+	host_entry = container_of(entry, struct sock_host_list_entry, entry);
+
+        return (strcmp(host_entry->hostname, (char *) src_addr) == 0);
+}
+
+static int sock_addr_matches_interface(struct slist *addr_list, struct sockaddr_in *src_addr)
+{
+	struct slist_entry *entry;
+
+	/* Always match if it's localhost */
+	if (ofi_is_loopback_addr((struct sockaddr *)src_addr))
+		return 1;
+
+	entry = slist_find_first_match(addr_list, sock_match_src_addr,
+					inet_ntoa(src_addr->sin_addr));
+
+	return entry ? 1 : 0;
+}
+
+static int sock_node_matches_interface(struct slist *addr_list, const char *node)
+{
+	struct addrinfo ai, *rai = NULL;
+	struct sockaddr_in addr;
+	int ret;
+
+	memset(&ai, 0, sizeof(ai));
+	ai.ai_family = AF_INET;
+	ai.ai_socktype = SOCK_STREAM;
+
+	ret = getaddrinfo(node, 0, &ai, &rai);
+	if (ret) {
+		SOCK_LOG_DBG("getaddrinfo failed!\n");
+		return -FI_EINVAL;
+	}
+	addr = *(struct sockaddr_in *)rai->ai_addr;
+	freeaddrinfo(rai);
+
+	return sock_addr_matches_interface(addr_list, &addr);
+}
+
+static void sock_free_addr_list(struct slist *addr_list)
+{
+	struct slist_entry *entry;
+	struct sock_host_list_entry *host_entry;
+
+	while (!slist_empty(addr_list)) {
+		entry = slist_remove_head(addr_list);
+		host_entry = container_of(entry, struct sock_host_list_entry,
+					   entry);
+		free(host_entry);
+	}
+}
+
+static int sock_getinfo(uint32_t version, const char *node, const char *service,
+			uint64_t flags, struct fi_info *hints,
+			struct fi_info **info)
+{
+	int ret = 0;
+	struct slist addr_list;
+	struct slist_entry *entry;
+	struct sock_host_list_entry *host_entry;
+	struct fi_info *tail;
+
+	if (!(flags & FI_SOURCE) && hints && hints->src_addr &&
+	    (hints->src_addrlen != sizeof(struct sockaddr_in)))
+		return -FI_ENODATA;
+
+	if (((!node && !service) || (flags & FI_SOURCE)) &&
+	    hints && hints->dest_addr &&
+	    (hints->dest_addrlen != sizeof(struct sockaddr_in)))
+		return -FI_ENODATA;
+
+	ret = sock_verify_info(version, hints);
+	if (ret)
+		return ret;
+
+	slist_init(&addr_list);
+	/* Returns loopback address if no other interfaces are available */
+	sock_get_list_of_addr(&addr_list);
+
+	ret = 1;
+	if (flags & FI_SOURCE) {
+		if (node)
+			ret = sock_node_matches_interface(&addr_list, node);
+	} else if (hints && hints->src_addr) {
+		ret = sock_addr_matches_interface(&addr_list, (struct sockaddr_in *)hints->src_addr);
+	}
+	if (!ret) {
+		SOCK_LOG_ERROR("Couldn't find a match with local interfaces\n");
+		return -FI_ENODATA;
+	}
+
+	*info = tail = NULL;
+	if (node ||
+	     (!(flags & FI_SOURCE) && hints && hints->src_addr) ||
+	     (!(flags & FI_SOURCE) && hints && hints->dest_addr)) {
+		sock_free_addr_list(&addr_list);
+		return sock_node_getinfo(version, node, service, flags, hints, info, &tail);
+	}
+
+	while (!slist_empty(&addr_list)) {
+		entry = slist_remove_head(&addr_list);
+		host_entry = container_of(entry, struct sock_host_list_entry, entry);
+		node = host_entry->hostname;
+		flags |= FI_SOURCE;
+		ret = sock_node_getinfo(version, node, service, flags, hints, info, &tail);
+		free(host_entry);
+		if (ret) {
+			if (ret == -FI_ENODATA)
+				continue;
+			sock_free_addr_list(&addr_list);
+			return ret;
+		}
+	}
+
+	return (!*info) ? ret : 0;
+}
+
 static void fi_sockets_fini(void)
 {
 	fastlock_destroy(&sock_list_lock);
@@ -654,7 +739,7 @@ static void fi_sockets_fini(void)
 struct fi_provider sock_prov = {
 	.name = sock_prov_name,
 	.version = FI_VERSION(SOCK_MAJOR_VERSION, SOCK_MINOR_VERSION),
-	.fi_version = FI_VERSION(1, 3),
+	.fi_version = FI_VERSION(1, 5),
 	.getinfo = sock_getinfo,
 	.fabric = sock_fabric,
 	.cleanup = fi_sockets_fini
