@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2016, Cisco Systems, Inc. All rights reserved.
+ * Copyright (c) 2014-2017, Cisco Systems, Inc. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -63,15 +63,103 @@
 
 #include "usdf.h"
 #include "usdf_av.h"
+#include "usdf_cm.h"
 #include "usdf_timer.h"
+#include "usdf_rdm.h"
 
 #include "fi_ext_usnic.h"
 
+static int usdf_av_alloc_dest(struct usdf_dest **dest_o)
+{
+	struct usdf_dest *dest;
+
+	dest = calloc(1, sizeof(**dest_o));
+	if (dest == NULL)
+		return -errno;
+
+	SLIST_INIT(&dest->ds_rdm_rdc_list);
+
+	*dest_o = dest;
+	return 0;
+}
+
+static void usdf_av_free_dest(struct usdf_dest *dest)
+{
+	struct usdf_rdm_connection *rdc = NULL;
+
+	LIST_REMOVE(dest, ds_addresses_entry);
+
+	while (!SLIST_EMPTY(&dest->ds_rdm_rdc_list)) {
+		rdc = SLIST_FIRST(&dest->ds_rdm_rdc_list);
+		rdc->dc_dest = NULL;
+
+		SLIST_REMOVE(&dest->ds_rdm_rdc_list, rdc, usdf_rdm_connection,
+			     dc_addr_link);
+		if (rdc)
+			rdc->dc_dest = NULL;
+	}
+
+	free(dest);
+}
+
+static int usdf_av_close_(struct usdf_av *av)
+{
+	struct usdf_dest *entry;
+
+	USDF_TRACE_SYS(AV, "\n");
+
+	pthread_spin_lock(&av->av_lock);
+
+	if (av->av_eq)
+		ofi_atomic_dec32(&av->av_eq->eq_refcnt);
+
+	ofi_atomic_dec32(&av->av_domain->dom_refcnt);
+
+	while (!LIST_EMPTY(&av->av_addresses)) {
+		entry = LIST_FIRST(&av->av_addresses);
+		usdf_av_free_dest(entry);
+	}
+
+	pthread_spin_destroy(&av->av_lock);
+	free(av);
+
+	USDF_DBG_SYS(AV, "AV successfully destroyed\n");
+
+	return 0;
+}
+
+static int usdf_av_close(struct fid *fid)
+{
+	struct usdf_av *av;
+	int pending;
+
+	USDF_TRACE_SYS(AV, "\n");
+
+	av = container_of(fid, struct usdf_av, av_fid.fid);
+	if (ofi_atomic_get32(&av->av_refcnt) > 0)
+		return -FI_EBUSY;
+
+	pending = ofi_atomic_get32(&av->av_active_inserts);
+	assert(pending >= 0);
+
+	if (pending) {
+		USDF_DBG_SYS(AV, "%d pending inserts, defer closing\n",
+			     pending);
+		ofi_atomic_set32(&av->av_closing, 1);
+	} else {
+		usdf_av_close_(av);
+	}
+
+	return 0;
+}
+
 static void
 usdf_av_insert_async_complete(struct usdf_av_insert *insert)
 {
 	struct fi_eq_entry entry;
 	struct usdf_av *av;
+	int pending;
+	int closing;
 
 	av = insert->avi_av;
 
@@ -81,17 +169,16 @@ usdf_av_insert_async_complete(struct usdf_av_insert *insert)
 	usdf_eq_write_internal(av->av_eq,
 		FI_AV_COMPLETE, &entry, sizeof(entry), 0);
 
-	pthread_spin_lock(&av->av_lock);
-
 	usdf_timer_free(av->av_domain->dom_fabric, insert->avi_timer);
 
-	atomic_dec(&av->av_active_inserts);
-	if (atomic_get(&av->av_active_inserts) == 0 && av->av_closing) {
-		pthread_spin_destroy(&av->av_lock);
-		free(av);
-	} else {
-		pthread_spin_unlock(&av->av_lock);
-	}
+	pending = ofi_atomic_dec32(&av->av_active_inserts);
+	USDF_DBG_SYS(AV, "new active insert value: %d\n", pending);
+	assert(pending >= 0);
+
+	closing = ofi_atomic_get32(&av->av_closing);
+
+	if (!pending && closing)
+		usdf_av_close_(av);
 
 	free(insert);
 }
@@ -123,22 +210,6 @@ usdf_post_insert_request_error(struct usdf_av_insert *insert,
 		USDF_EVENT_FLAG_ERROR);
 }
 
-static int
-usdf_av_alloc_dest(struct usdf_dest **dest_o)
-{
-	struct usdf_dest *dest;
-
-	dest = calloc(1, sizeof(**dest_o));
-	if (dest == NULL) {
-		return -errno;
-	}
-	SLIST_INIT(&dest->ds_rdm_rdc_list);
-
-	*dest_o = dest;
-	return 0;
-}
-
-
 /*
  * Called by progression thread to look for AV completions on this domain
  */
@@ -174,6 +245,9 @@ usdf_av_insert_progress(void *v)
 			if (ret == 0) {
 				++insert->avi_successes;
 				*(struct usdf_dest **)req->avr_fi_addr = dest;
+
+				LIST_INSERT_HEAD(&insert->avi_av->av_addresses,
+						 dest, ds_addresses_entry);
 			} else {
 				usdf_post_insert_request_error(insert, req);
 			}
@@ -227,33 +301,36 @@ usdf_am_insert_async(struct fid_av *fav, const void *addr, size_t count,
 			  fi_addr_t *fi_addr, uint64_t flags, void *context)
 {
 	const struct sockaddr_in *sin;
+	const char **addr_str;
+	struct sockaddr_in *cur_sin;
 	struct usd_device_attrs *dap;
 	struct usdf_av_insert *insert;
 	struct usdf_av_req *req;
 	struct usdf_av *av;
 	struct usdf_fabric *fp;
 	struct usd_dest *u_dest;
+	struct fi_info *info;
 	int ret;
 	size_t i;
+	bool addr_format_str;
 
 	USDF_TRACE_SYS(AV, "\n");
 
-	if ((flags & ~FI_MORE) != 0) {
+	if ((flags & ~(FI_MORE)) != 0)
 		return -FI_EBADFLAGS;
-	}
 
 	av = av_ftou(fav);
 	fp = av->av_domain->dom_fabric;
 	dap = fp->fab_dev_attrs;
+	info = av->av_domain->dom_info;
+	addr_format_str = (info->addr_format == FI_ADDR_STR);
 
-	if (av->av_flags & FI_READ) {
-		return -FI_EACCES;
-	}
 	if (av->av_eq == NULL) {
 		return -FI_ENOEQ;
 	}
 
 	sin = addr;
+	addr_str = (const char **)addr;
 
 	/* allocate an insert record and N requests */
 	insert = calloc(1, sizeof(*insert) + count * sizeof(*req));
@@ -270,19 +347,25 @@ usdf_am_insert_async(struct fid_av *fav, const void *addr, size_t count,
 	TAILQ_INIT(&insert->avi_req_list);
 	insert->avi_arps_left = USDF_AV_MAX_ARPS;
 
+	ret = ofi_atomic_inc32(&av->av_active_inserts);
+	USDF_DBG_SYS(AV, "new active insert value: %d\n", ret);
+
 	/* If no addresses, complete now */
 	if (count == 0) {
 		usdf_av_insert_async_complete(insert);
 		return 0;
 	}
 
-	atomic_inc(&av->av_active_inserts);
-
 	req = (struct usdf_av_req *)(insert + 1);
 
 	for (i = 0; i < count; i++) {
 		req->avr_fi_addr = &fi_addr[i];
 
+		if (addr_format_str) {
+			usdf_str_toaddr(addr_str[i], &cur_sin);
+			sin = cur_sin;
+		}
+
 		/* find the address we actually need to look up */
 		ret = usnic_nl_rt_lookup(dap->uda_ipaddr_be,
 				sin->sin_addr.s_addr, dap->uda_ifindex,
@@ -314,7 +397,13 @@ usdf_am_insert_async(struct fid_av *fav, const void *addr, size_t count,
 			TAILQ_INSERT_TAIL(&insert->avi_req_list, req, avr_link);
 		}
 
-		++sin;
+		if (addr_format_str) {
+			free(cur_sin);
+			cur_sin = NULL;
+		} else {
+			++sin;
+		}
+
 		++req;
 	}
 
@@ -338,26 +427,51 @@ usdf_am_insert_sync(struct fid_av *fav, const void *addr, size_t count,
 			  fi_addr_t *fi_addr, uint64_t flags, void *context)
 {
 	const struct sockaddr_in *sin;
+	const char **addr_str;
+	struct sockaddr_in *cur_sin;
 	struct usdf_av *av;
 	struct usd_dest *u_dest;
 	struct usdf_dest *dest;
+	struct fi_info *info;
 	int ret_count;
 	int ret;
+	int *errors;
+	uint32_t api_version;
 	size_t i;
+	bool addr_format_str;
 
 	USDF_TRACE_SYS(AV, "\n");
 
-	if ((flags & ~FI_MORE) != 0) {
+	ret_count = 0;
+	av = av_ftou(fav);
+	api_version = av->av_domain->dom_fabric->fab_attr.fabric->api_version;
+	info = av->av_domain->dom_info;
+	addr_format_str = (info->addr_format == FI_ADDR_STR);
+	errors = context;
+
+	/* Screen out unsupported flags. */
+	if ((flags & ~(FI_MORE|FI_SYNC_ERR)) != 0)
 		return -FI_EBADFLAGS;
+
+	/* If user set FI_SYNC_ERR, we have to report back to user's buffer. */
+	if (flags & FI_SYNC_ERR) {
+		if (FI_VERSION_LT(api_version, FI_VERSION(1, 5)))
+			return -FI_EBADFLAGS;
+
+		memset(errors, 0, sizeof(int) * count);
 	}
 
-	av = av_ftou(fav);
-
-	ret_count = 0;
 	sin = addr;
+	addr_str = (const char **)addr;
 
 	/* XXX parallelize, this will also eliminate u_dest silliness */
 	for (i = 0; i < count; i++) {
+
+		if (addr_format_str) {
+			usdf_str_toaddr(addr_str[i], &cur_sin);
+			sin = cur_sin;
+		}
+
 		dest = NULL;
 		u_dest = NULL;
 		ret = usdf_av_alloc_dest(&dest);
@@ -373,13 +487,24 @@ usdf_am_insert_sync(struct fid_av *fav, const void *addr, size_t count,
 				htons(IP_DF);
 			dest->ds_dest = *u_dest;
 			fi_addr[i] = (fi_addr_t)dest;
+			LIST_INSERT_HEAD(&av->av_addresses, dest,
+					 ds_addresses_entry);
 			++ret_count;
 		} else {
+			if (flags & FI_SYNC_ERR)
+				errors[i] = -ret;
+
 			fi_addr[i] = FI_ADDR_NOTAVAIL;
 			free(dest);
 		}
 		free(u_dest);
-		++sin;
+
+		if (addr_format_str) {
+			free(cur_sin);
+			cur_sin = NULL;
+		} else {
+			++sin;
+		}
 	}
 
 	return ret_count;
@@ -417,28 +542,36 @@ static int usdf_av_insertsvc(struct fid_av *fav, const char *node,
 		const char *service, fi_addr_t *fi_addr, uint64_t flags,
 		void *context)
 {
-	struct sockaddr_in *addr;
+	struct sockaddr_in addr;
+	struct usdf_av *av;
+	struct fi_info *info;
 	int ret;
+	bool addr_format_str;
 
 	USDF_TRACE_SYS(AV, "\n");
 
+	av = av_ftou(fav);
+	info = av->av_domain->dom_info;
+	addr_format_str = (info->addr_format == FI_ADDR_STR);
+
 	if (!fav)
 		return -FI_EINVAL;
 
-	addr = calloc(1, sizeof(*addr));
-	if (!addr) {
-		USDF_DBG("address allocation failed\n");
-		return -FI_ENOMEM;
+	if (addr_format_str) {
+		/* string format should not come with service param. */
+		if (service)
+			return -FI_EINVAL;
+
+		ret = fav->ops->insert(fav, &node, 1, fi_addr, flags, context);
+	} else {
+		ret = usdf_resolve_addr(node, service, &addr);
+		if (ret)
+			goto fail;
+
+		ret = fav->ops->insert(fav, &addr, 1, fi_addr, flags, context);
 	}
 
-	ret = usdf_resolve_addr(node, service, addr);
-	if (ret)
-		goto fail;
-
-	ret = fav->ops->insert(fav, addr, 1, fi_addr, flags, context);
-
 fail:
-	free(addr);
 	return ret;
 }
 
@@ -447,21 +580,17 @@ usdf_am_remove(struct fid_av *fav, fi_addr_t *fi_addr, size_t count,
 			  uint64_t flags)
 {
 	struct usdf_dest *dest;
-	struct usdf_av *av;
 	size_t i;
 
 	USDF_TRACE_SYS(AV, "\n");
 
-	av = av_ftou(fav);
-
-	if (av->av_flags & FI_READ) {
-		return -FI_EACCES;
-	}
-
 	for (i = 0; i < count; ++i) {
 		if (fi_addr[i] != FI_ADDR_NOTAVAIL) {
 			dest = (struct usdf_dest *)(uintptr_t)fi_addr[i];
-			free(dest);
+			usdf_av_free_dest(dest);
+
+			/* Mark invalid by setting to FI_ADDR_NOTAVAIL*/
+			fi_addr[i] = FI_ADDR_NOTAVAIL;
 		}
 	}
 
@@ -469,17 +598,28 @@ usdf_am_remove(struct fid_av *fav, fi_addr_t *fi_addr, size_t count,
 }
 
 static int
-usdf_am_lookup(struct fid_av *av, fi_addr_t fi_addr, void *addr,
+usdf_am_lookup(struct fid_av *fav, fi_addr_t fi_addr, void *addr,
 			  size_t *addrlen)
 {
 	struct usdf_dest *dest;
+	struct usdf_av *av;
+	struct fi_info *info;
 	struct sockaddr_in sin = { 0 };
 	size_t copylen;
+	bool addr_format_str;
 
 	USDF_TRACE_SYS(AV, "\n");
 
-	dest = (struct usdf_dest *)(uintptr_t)fi_addr;
+	av = av_ftou(fav);
+	info = av->av_domain->dom_info;
+	addr_format_str = (info->addr_format == FI_ADDR_STR);
 
+	if (fi_addr == FI_ADDR_NOTAVAIL) {
+		USDF_WARN_SYS(AV, "invalid address, can't lookup\n");
+		return -FI_EINVAL;
+	}
+
+	dest = (struct usdf_dest *)(uintptr_t)fi_addr;
 	if (*addrlen < sizeof(sin)) {
 		copylen = *addrlen;
 	} else {
@@ -488,26 +628,30 @@ usdf_am_lookup(struct fid_av *av, fi_addr_t fi_addr, void *addr,
 
 	sin.sin_family = AF_INET;
 	usd_expand_dest(&dest->ds_dest, &sin.sin_addr.s_addr, &sin.sin_port);
-	memcpy(addr, &sin, copylen);
 
-	*addrlen = sizeof(sin);
+	if (addr_format_str)
+		usdf_addr_tostr(&sin, addr, addrlen);
+	else {
+		memcpy(addr, &sin, copylen);
+		*addrlen = sizeof(sin);
+	}
 	return 0;
 }
 
 static const char *
-usdf_av_straddr(struct fid_av *av, const void *addr,
+usdf_av_straddr(struct fid_av *fav, const void *addr,
 				    char *buf, size_t *len)
 {
-	const struct sockaddr_in *sin;
-	char straddr[24];
-	int size;
+	struct fi_info *info;
+	struct usdf_av *av;
 
-	sin = addr;
-	size = snprintf(straddr, sizeof straddr, "%s:%d",
-			inet_ntoa(sin->sin_addr), sin->sin_port);
-	snprintf(buf, *len, "%s", straddr);
-	*len = size + 1;
-	return buf;
+	if (!len || !addr || !buf)
+		return NULL;
+
+	av = av_fidtou(fav);
+	info = av->av_domain->dom_info;
+
+	return ofi_straddr(buf, len, info->addr_format, addr);
 }
 
 static int
@@ -525,7 +669,7 @@ usdf_av_bind(struct fid *fid, struct fid *bfid, uint64_t flags)
 			return -FI_EINVAL;
 		}
 		av->av_eq = eq_fidtou(bfid);
-		atomic_inc(&av->av_eq->eq_refcnt);
+		ofi_atomic_inc32(&av->av_eq->eq_refcnt);
 		break;
 	default:
 		return -FI_EINVAL;
@@ -534,35 +678,6 @@ usdf_av_bind(struct fid *fid, struct fid *bfid, uint64_t flags)
 	return 0;
 }
 
-static int
-usdf_av_close(struct fid *fid)
-{
-	struct usdf_av *av;
-
-	USDF_TRACE_SYS(AV, "\n");
-
-	av = container_of(fid, struct usdf_av, av_fid.fid);
-	if (atomic_get(&av->av_refcnt) > 0) {
-		return -FI_EBUSY;
-	}
-
-	pthread_spin_lock(&av->av_lock);
-
-	if (av->av_eq != NULL) {
-		atomic_dec(&av->av_eq->eq_refcnt);
-	}
-	atomic_dec(&av->av_domain->dom_refcnt);
-
-	if (atomic_get(&av->av_active_inserts) > 0) {
-		av->av_closing = 1;
-		pthread_spin_unlock(&av->av_lock);
-	} else {
-		pthread_spin_destroy(&av->av_lock);
-		free(av);
-	}
-	return 0;
-}
-
 static struct fi_ops usdf_av_fi_ops = {
 	.size = sizeof(struct fi_ops),
 	.close = usdf_av_close,
@@ -591,24 +706,68 @@ static struct fi_ops_av usdf_am_ops_sync = {
 	.straddr = usdf_av_straddr
 };
 
+static int usdf_av_process_attr(struct fi_av_attr *attr)
+{
+	USDF_TRACE_SYS(AV, "\n");
+
+	if (attr == NULL) {
+		USDF_WARN_SYS(AV, "NULL AV attribute structure is invalid\n");
+		return -FI_EINVAL;
+	}
+
+	if (attr->name || attr->map_addr || (attr->flags & FI_READ)) {
+		USDF_WARN_SYS(AV, "named AVs are not supported\n");
+		return -FI_ENOSYS;
+	}
+
+	if (attr->flags & ~FI_EVENT) {
+		USDF_WARN_SYS(AV, "invalid flag, only FI_EVENT is supported\n");
+		return -FI_EINVAL;
+	}
+
+	if (attr->rx_ctx_bits) {
+		USDF_WARN_SYS(AV, "scalable endpoints not supported\n");
+		return -FI_EINVAL;
+	}
+
+	if (attr->ep_per_node > 1)
+		USDF_WARN_SYS(AV, "ep_per_node not supported, ignoring\n");
+
+	switch (attr->type) {
+	case FI_AV_UNSPEC:
+		USDF_DBG_SYS(AV, "no AV type specified, using FI_AV_MAP\n");
+	case FI_AV_MAP:
+		break;
+	case FI_AV_TABLE:
+		USDF_DBG_SYS(AV, "FI_AV_TABLE is unsupported\n");
+		return -FI_ENOSYS;
+	default:
+		USDF_WARN_SYS(AV, "unknown AV type %d, not supported",
+			      attr->type);
+		return -FI_EINVAL;
+	}
+
+	return FI_SUCCESS;
+}
+
 int
 usdf_av_open(struct fid_domain *domain, struct fi_av_attr *attr,
 		 struct fid_av **av_o, void *context)
 {
 	struct usdf_domain *udp;
 	struct usdf_av *av;
+	int ret;
 
 	USDF_TRACE_SYS(AV, "\n");
 
-	if (attr == NULL || av_o == NULL)
+	if (!av_o) {
+		USDF_WARN_SYS(AV, "provided AV pointer can not be NULL\n");
 		return -FI_EINVAL;
-
-	if ((attr->flags & ~(FI_EVENT | FI_READ)) != 0) {
-		return -FI_ENOSYS;
 	}
 
-	if (attr->name)
-		return -FI_ENOSYS;
+	ret = usdf_av_process_attr(attr);
+	if (ret)
+		return ret;
 
 	udp = dom_ftou(domain);
 
@@ -622,18 +781,112 @@ usdf_av_open(struct fid_domain *domain, struct fi_av_attr *attr,
 	} else {
 		av->av_fid.ops = &usdf_am_ops_sync;
 	}
+
+	LIST_INIT(&av->av_addresses);
+
 	av->av_fid.fid.fclass = FI_CLASS_AV;
 	av->av_fid.fid.context = context;
 	av->av_fid.fid.ops = &usdf_av_fi_ops;
 	av->av_flags = attr->flags;
 
 	pthread_spin_init(&av->av_lock, PTHREAD_PROCESS_PRIVATE);
-	atomic_initialize(&av->av_active_inserts, 0);
+	ofi_atomic_initialize32(&av->av_active_inserts, 0);
+	ofi_atomic_initialize32(&av->av_closing, 0);
 
-	atomic_initialize(&av->av_refcnt, 0);
-	atomic_inc(&udp->dom_refcnt);
+	ofi_atomic_initialize32(&av->av_refcnt, 0);
+	ofi_atomic_inc32(&udp->dom_refcnt);
 	av->av_domain = udp;
 
 	*av_o = av_utof(av);
 	return 0;
 }
+
+/* Look up if the sin address has been already inserted.
+ * if match, return the address of the dest pointer. otherwise,
+ * returns FI_ADDR_NOTAVAIL.
+ */
+fi_addr_t usdf_av_lookup_addr(struct usdf_av *av,
+			      const struct sockaddr_in *sin)
+{
+	struct usdf_dest *cur;
+	struct usd_udp_hdr u_hdr;
+
+	for (cur = av->av_addresses.lh_first; cur;
+	     cur = cur->ds_addresses_entry.le_next) {
+		u_hdr = cur->ds_dest.ds_dest.ds_udp.u_hdr;
+		if (sin->sin_addr.s_addr == u_hdr.uh_ip.daddr &&
+		    sin->sin_port == u_hdr.uh_udp.dest)
+			return (fi_addr_t)(uintptr_t)cur;
+	}
+	return FI_ADDR_NOTAVAIL;
+}
+
+/* Return sockaddr_in pointer. Must be used with usdf_free_sin_if_needed()
+ * to cleanup properly.
+ */
+struct sockaddr_in *usdf_format_to_sin(struct fi_info *info, const void *addr)
+{
+	struct sockaddr_in *sin;
+
+	if (!info)
+		return (struct sockaddr_in *)addr;
+
+	switch (info->addr_format) {
+	case FI_FORMAT_UNSPEC:
+	case FI_SOCKADDR:
+	case FI_SOCKADDR_IN:
+		return (struct sockaddr_in *)addr;
+	case FI_ADDR_STR:
+		usdf_str_toaddr(addr, &sin);
+		return sin;
+	default:
+		return NULL;
+	}
+}
+
+/* Utility function to free the sockaddr_in allocated from usdf_format_to_sin()
+ */
+void usdf_free_sin_if_needed(struct fi_info *info, struct sockaddr_in *sin)
+{
+	if (info && info->addr_format == FI_ADDR_STR)
+		free(sin);
+}
+
+/* Convert sockaddr_in pointer to appropriate format.
+ * If conversion happens, destroy the origin. (to minimize cleaning up code)
+ */
+void *usdf_sin_to_format(struct fi_info *info, void *addr, size_t *len)
+{
+	size_t addr_strlen;
+	char *addrstr;
+
+	if (!info)
+		return addr;
+
+	switch (info->addr_format) {
+	case FI_FORMAT_UNSPEC:
+	case FI_SOCKADDR:
+	case FI_SOCKADDR_IN:
+		if (len)
+			*len = sizeof(struct sockaddr_in);
+		return addr;
+	case FI_ADDR_STR:
+		addrstr = calloc(1, USDF_ADDR_STR_LEN);
+		if (addrstr == NULL) {
+			USDF_DBG_SYS(AV, "memory allocation failed\n");
+			return NULL;
+		}
+
+		addr_strlen = USDF_ADDR_STR_LEN;
+		usdf_addr_tostr(addr, addrstr, &addr_strlen);
+
+		if (len)
+			*len = addr_strlen;
+
+		free(addr);
+		return addrstr;
+	default:
+		return NULL;
+	}
+
+}
