@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2016, Cisco Systems, Inc. All rights reserved.
+ * Copyright (c) 2014-2017, Cisco Systems, Inc. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -51,7 +51,6 @@
 #include <rdma/fabric.h>
 #include <rdma/fi_cm.h>
 #include <rdma/fi_domain.h>
-#include <rdma/fi_prov.h>
 #include <rdma/fi_endpoint.h>
 #include <rdma/fi_rma.h>
 #include <rdma/fi_errno.h>
@@ -89,10 +88,12 @@ usdf_cq_readerr(struct fid_cq *fcq, struct fi_cq_err_entry *entry,
 	        uint64_t flags)
 {
 	struct usdf_cq *cq;
+	uint32_t api_version;
 
 	USDF_TRACE_SYS(CQ, "\n");
 
 	cq = container_of(fcq, struct usdf_cq, cq_fid);
+	api_version = cq->cq_domain->dom_fabric->fab_attr.fabric->api_version;
 
 	// The return values are analogous to sockets cq_readerr
 	if (cq->cq_comp.uc_status == 0) {
@@ -123,6 +124,10 @@ usdf_cq_readerr(struct fid_cq *fcq, struct fi_cq_err_entry *entry,
 
 	cq->cq_comp.uc_status = 0;
 
+	/* We don't have err_data to give back to the user. */
+	if (FI_VERSION_GE(api_version, FI_VERSION(1, 5)))
+		entry->err_data_size = 0;
+
 	return 1;
 }
 
@@ -295,7 +300,7 @@ usdf_cq_read_common(struct fid_cq *fcq, void *buf, size_t count,
 
 	if (cq->cq_waiting) {
 		cq->cq_waiting = false;
-		atomic_dec(&fab->num_blocked_waiting);
+		ofi_atomic_dec32(&fab->num_blocked_waiting);
 	}
 
 	return copied > 0 ? copied : -FI_EAGAIN;
@@ -367,11 +372,7 @@ usdf_cq_readfrom_context(struct fid_cq *fcq, void *buf, size_t count,
 			sin.sin_addr.s_addr = hdr->uh_ip.saddr;
 			sin.sin_port = hdr->uh_udp.source;
 
-			ret = fi_av_insert(av_utof(ep->e.dg.ep_av), &sin, 1,
-					src_addr, 0, NULL);
-			if (ret != 1) {
-				*src_addr = FI_ADDR_NOTAVAIL;
-			}
+			*src_addr = usdf_av_lookup_addr(ep->e.dg.ep_av, &sin);
 			++src_addr;
 		}
 
@@ -438,7 +439,7 @@ usdf_progress_hard_cq(struct usdf_cq_hard *hcq)
 
 void
 usdf_cq_post_soft(struct usdf_cq_hard *hcq, void *context, size_t len,
-		int prov_errno)
+		int prov_errno, uint64_t flags)
 {
 	int ret;
 	struct usdf_cq_soft_entry *entry;
@@ -461,6 +462,7 @@ usdf_cq_post_soft(struct usdf_cq_hard *hcq, void *context, size_t len,
 	entry->cse_context = context;
 	entry->cse_len = len;
 	entry->cse_prov_errno = prov_errno;
+	entry->cse_flags = flags;
 
 	/* update with wrap */
 	entry++;
@@ -576,7 +578,7 @@ static ssize_t usdf_cq_sread_fd(struct fid_cq *fcq, void *buf, size_t count,
 
 	ret = usdf_cq_trywait(&fcq->fid);
 	if (ret == FI_SUCCESS) {
-		atomic_inc(&fabric->num_blocked_waiting);
+		ofi_atomic_inc32(&fabric->num_blocked_waiting);
 
 		ret = usdf_fabric_wake_thread(fabric);
 		if (ret) {
@@ -594,7 +596,7 @@ static ssize_t usdf_cq_sread_fd(struct fid_cq *fcq, void *buf, size_t count,
 			goto err;
 		}
 
-		atomic_dec(&fabric->num_blocked_waiting);
+		ofi_atomic_dec32(&fabric->num_blocked_waiting);
 	} else if ((ret < 0) && (ret != -FI_EAGAIN)) {
 		return ret;
 	}
@@ -602,7 +604,7 @@ static ssize_t usdf_cq_sread_fd(struct fid_cq *fcq, void *buf, size_t count,
 	return fi_cq_read(fcq, buf, count);
 
 err:
-	atomic_dec(&fabric->num_blocked_waiting);
+	ofi_atomic_dec32(&fabric->num_blocked_waiting);
 	return ret;
 }
 
@@ -752,7 +754,7 @@ static int usdf_cq_unbind_wait(struct usdf_cq *cq)
 
 	fid_list_remove(&wait_priv->list, &wait_priv->lock, &cq->cq_fid.fid);
 
-	atomic_dec(&wait_priv->wait_refcnt);
+	ofi_atomic_dec32(&wait_priv->wait_refcnt);
 
 	USDF_DBG_SYS(CQ,
 			"dissasociated CQ FD %d from epoll FD %d using FID: %p\n",
@@ -774,7 +776,7 @@ usdf_cq_close(fid_t fid)
 	cq = container_of(fid, struct usdf_cq, cq_fid.fid);
 	fab = cq->cq_domain->dom_fabric;
 
-	if (atomic_get(&cq->cq_refcnt) > 0) {
+	if (ofi_atomic_get32(&cq->cq_refcnt) > 0) {
 		return -FI_EBUSY;
 	}
 
@@ -787,7 +789,7 @@ usdf_cq_close(fid_t fid)
 	if (cq->cq_is_soft) {
 		while (!TAILQ_EMPTY(&cq->c.soft.cq_list)) {
 			hcq = TAILQ_FIRST(&cq->c.soft.cq_list);
-			if (atomic_get(&hcq->cqh_refcnt) > 0) {
+			if (ofi_atomic_get32(&hcq->cqh_refcnt) > 0) {
 				return -FI_EBUSY;
 			}
 			TAILQ_REMOVE(&cq->c.soft.cq_list, hcq, cqh_link);
@@ -811,7 +813,7 @@ usdf_cq_close(fid_t fid)
 	}
 
 	if (cq->cq_waiting)
-		atomic_dec(&fab->num_blocked_waiting);
+		ofi_atomic_dec32(&fab->num_blocked_waiting);
 
 	free(cq);
 	return 0;
@@ -987,8 +989,8 @@ usdf_cq_make_soft(struct usdf_cq *cq)
 			hcq->cqh_ucq = ucq;
 			hcq->cqh_progress = usdf_progress_hard_cq;
 
-			atomic_initialize(&hcq->cqh_refcnt,
-					atomic_get(&cq->cq_refcnt));
+			ofi_atomic_initialize32(&hcq->cqh_refcnt,
+					ofi_atomic_get32(&cq->cq_refcnt));
 			TAILQ_INSERT_HEAD(&cq->c.soft.cq_list, hcq, cqh_link);
 		}
 
@@ -1139,11 +1141,11 @@ int usdf_cq_trywait(struct fid *fcq)
 	}
 
 	cq->cq_waiting = true;
-	atomic_inc(&fab->num_blocked_waiting);
+	ofi_atomic_inc32(&fab->num_blocked_waiting);
 	ret = usdf_fabric_wake_thread(fab);
 	if (ret) {
 		USDF_DBG_SYS(FABRIC, "error while waking progress thread\n");
-		atomic_dec(&fab->num_blocked_waiting);
+		ofi_atomic_dec32(&fab->num_blocked_waiting);
 	}
 
 	if (cq->cq_is_soft) {
@@ -1296,7 +1298,7 @@ usdf_cq_open(struct fid_domain *domain, struct fi_cq_attr *attr,
 	 */
 	if (attr->wait_obj == FI_WAIT_SET) {
 		wait_priv = wait_ftou(attr->wait_set);
-		atomic_inc(&wait_priv->wait_refcnt);
+		ofi_atomic_inc32(&wait_priv->wait_refcnt);
 	}
 
 	cq->object.fd = -1;
@@ -1304,7 +1306,7 @@ usdf_cq_open(struct fid_domain *domain, struct fi_cq_attr *attr,
 	cq->cq_fid.fid.fclass = FI_CLASS_CQ;
 	cq->cq_fid.fid.context = context;
 	cq->cq_fid.fid.ops = &usdf_cq_fi_ops;
-	atomic_initialize(&cq->cq_refcnt, 0);
+	ofi_atomic_initialize32(&cq->cq_refcnt, 0);
 
 	switch (attr->format) {
 	case FI_CQ_FORMAT_CONTEXT:
