@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2016, Cisco Systems, Inc. All rights reserved.
+ * Copyright (c) 2014-2017, Cisco Systems, Inc. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -51,15 +51,16 @@
 #include <rdma/fabric.h>
 #include <rdma/fi_cm.h>
 #include <rdma/fi_domain.h>
-#include <rdma/fi_prov.h>
 #include <rdma/fi_endpoint.h>
 #include <rdma/fi_rma.h>
 #include <rdma/fi_errno.h>
 #include "fi.h"
 #include "fi_enosys.h"
+#include "fi_util.h"
 
 #include "usd.h"
 #include "usdf.h"
+#include "usnic_direct.h"
 #include "usdf_endpoint.h"
 #include "fi_ext_usnic.h"
 #include "usdf_rudp.h"
@@ -114,11 +115,32 @@ static const struct fi_ep_attr rdm_dflt_ep_attr = {
 };
 
 static const struct fi_domain_attr rdm_dflt_domain_attr = {
+	.caps = USDF_DOM_CAPS,
 	.threading = FI_THREAD_ENDPOINT,
 	.control_progress = FI_PROGRESS_AUTO,
 	.data_progress = FI_PROGRESS_MANUAL,
 	.resource_mgmt = FI_RM_DISABLED,
-	.mr_mode = FI_MR_BASIC
+	.mr_mode = OFI_MR_BASIC_MAP | FI_MR_LOCAL,
+	.cntr_cnt = USDF_RDM_CNTR_CNT,
+	.mr_iov_limit = USDF_RDM_MR_IOV_LIMIT,
+	.mr_cnt = USDF_RDM_MR_CNT,
+};
+
+static struct fi_ops_atomic usdf_rdm_atomic_ops = {
+	.size = sizeof(struct fi_ops_atomic),
+	.write = fi_no_atomic_write,
+	.writev = fi_no_atomic_writev,
+	.writemsg = fi_no_atomic_writemsg,
+	.inject = fi_no_atomic_inject,
+	.readwrite = fi_no_atomic_readwrite,
+	.readwritev = fi_no_atomic_readwritev,
+	.readwritemsg = fi_no_atomic_readwritemsg,
+	.compwrite = fi_no_atomic_compwrite,
+	.compwritev = fi_no_atomic_compwritev,
+	.compwritemsg = fi_no_atomic_compwritemsg,
+	.writevalid = fi_no_atomic_writevalid,
+	.readwritevalid = fi_no_atomic_readwritevalid,
+	.compwritevalid = fi_no_atomic_compwritevalid,
 };
 
 /*******************************************************************************
@@ -167,14 +189,19 @@ out:
 
 }
 
-int usdf_rdm_fill_dom_attr(struct fi_info *hints, struct fi_info *fi)
+int usdf_rdm_fill_dom_attr(uint32_t version, struct fi_info *hints,
+			   struct fi_info *fi, struct usd_device_attrs *dap)
 {
+	int ret;
 	struct fi_domain_attr defaults;
 
 	defaults = rdm_dflt_domain_attr;
+	ret = usdf_domain_getname(version, dap, &defaults.name);
+	if (ret < 0)
+		return -FI_ENODATA;
 
 	if (!hints || !hints->domain_attr)
-		goto out;
+		goto catch;
 
 	/* how to handle fi_thread_fid, fi_thread_completion, etc?
 	 */
@@ -212,39 +239,55 @@ int usdf_rdm_fill_dom_attr(struct fi_info *hints, struct fi_info *fi)
 		return -FI_ENODATA;
 	}
 
-	switch (hints->domain_attr->mr_mode) {
-	case FI_MR_UNSPEC:
-	case FI_MR_BASIC:
+	switch (hints->domain_attr->caps) {
+	case 0:
+	case FI_REMOTE_COMM:
 		break;
 	default:
+		USDF_WARN_SYS(DOMAIN,
+			"invalid domain capabilities\n");
 		return -FI_ENODATA;
 	}
 
-out:
+	if (usdf_check_mr_mode(version, hints, defaults.mr_mode))
+		return -FI_ENODATA;
+
+	if (hints->domain_attr->mr_cnt <= USDF_RDM_MR_CNT) {
+		defaults.mr_cnt = hints->domain_attr->mr_cnt;
+	} else {
+		USDF_DBG_SYS(DOMAIN, "mr_count exceeded provider limit\n");
+		return -FI_ENODATA;
+	}
+
+catch:
+	/* catch the version changes here. */
+	ret = usdf_catch_dom_attr(version, hints, &defaults);
+	if (ret)
+		return ret;
+
 	*fi->domain_attr = defaults;
 
 	return FI_SUCCESS;
 }
 
-int usdf_rdm_fill_tx_attr(struct fi_info *hints, struct fi_info *fi)
+int usdf_rdm_fill_tx_attr(uint32_t version, struct fi_info *hints,
+			  struct fi_info *fi)
 {
+	int ret;
 	struct fi_tx_attr defaults;
 
 	defaults = rdm_dflt_tx_attr;
 
 	if (!hints || !hints->tx_attr)
-		goto out;
+		goto catch;
 
 	/* make sure we can support the caps that are requested*/
 	if (hints->tx_attr->caps & ~USDF_RDM_CAPS)
 		return -FI_ENODATA;
 
 	/* clear the mode bits the app doesn't support */
-	defaults.mode &= (hints->mode | hints->tx_attr->mode);
-
-	/* make sure the app supports our required mode bits */
-	if ((defaults.mode & USDF_RDM_REQ_MODE) != USDF_RDM_REQ_MODE)
-		return -FI_ENODATA;
+	if (hints->mode || hints->tx_attr->mode)
+		defaults.mode &= (hints->mode | hints->tx_attr->mode);
 
 	defaults.op_flags |= hints->tx_attr->op_flags;
 
@@ -268,31 +311,35 @@ int usdf_rdm_fill_tx_attr(struct fi_info *hints, struct fi_info *fi)
 	if (hints->tx_attr->size > defaults.size)
 		return -FI_ENODATA;
 
-out:
+catch:
+	/* catch version changes here. */
+	ret = usdf_catch_tx_attr(version, &defaults);
+	if (ret)
+		return ret;
+
 	*fi->tx_attr = defaults;
 
 	return FI_SUCCESS;
 }
 
-int usdf_rdm_fill_rx_attr(struct fi_info *hints, struct fi_info *fi)
+int usdf_rdm_fill_rx_attr(uint32_t version, struct fi_info *hints,
+			  struct fi_info *fi)
 {
+	int ret;
 	struct fi_rx_attr defaults;
 
 	defaults = rdm_dflt_rx_attr;
 
 	if (!hints || !hints->rx_attr)
-		goto out;
+		goto catch;
 
 	/* make sure we can support the capabilities that are requested */
 	if (hints->rx_attr->caps & ~USDF_RDM_CAPS)
 		return -FI_ENODATA;
 
 	/* clear the mode bits the app doesn't support */
-	defaults.mode &= (hints->mode | hints->rx_attr->mode);
-
-	/* make sure the app supports our required mode bits */
-	if ((defaults.mode & USDF_RDM_REQ_MODE) != USDF_RDM_REQ_MODE)
-		return -FI_ENODATA;
+	if (hints->mode || hints->rx_attr->mode)
+		defaults.mode &= (hints->mode | hints->rx_attr->mode);
 
 	defaults.op_flags |= hints->rx_attr->op_flags;
 
@@ -313,7 +360,12 @@ int usdf_rdm_fill_rx_attr(struct fi_info *hints, struct fi_info *fi)
 	if (hints->rx_attr->size > defaults.size)
 		return -FI_ENODATA;
 
-out:
+catch:
+	/* catch version changes here. */
+	ret = usdf_catch_rx_attr(version, &defaults);
+	if (ret)
+		return ret;
+
 	*fi->rx_attr = defaults;
 
 	return FI_SUCCESS;
@@ -538,48 +590,17 @@ fail:
 
 static int
 usdf_ep_rdm_enable(struct fid_ep *fep)
-{
-	return usdf_ep_rdm_get_queues(ep_ftou(fep));
-}
-
-static int
-usdf_ep_rdm_getopt(fid_t fid, int level, int optname,
-		  void *optval, size_t *optlen)
 {
 	struct usdf_ep *ep;
+	int ret;
 
-	USDF_TRACE_SYS(EP_CTRL, "\n");
+	ep = ep_ftou(fep);
 
-	ep = ep_fidtou(fid);
-	(void)ep;
+	ret = usdf_ep_rdm_get_queues(ep);
+	if (ret == FI_SUCCESS)
+		ep->flags |= USDF_EP_ENABLED;
 
-	switch (level) {
-	case FI_OPT_ENDPOINT:
-		return -FI_ENOPROTOOPT;
-	default:
-		return -FI_ENOPROTOOPT;
-	}
-	return 0;
-}
-
-static int
-usdf_ep_rdm_setopt(fid_t fid, int level, int optname,
-		  const void *optval, size_t optlen)
-{
-	struct usdf_ep *ep;
-
-	USDF_TRACE_SYS(EP_CTRL, "\n");
-
-	ep = ep_fidtou(fid);
-	(void)ep;
-
-	switch (level) {
-	case FI_OPT_ENDPOINT:
-		return -FI_ENOPROTOOPT;
-	default:
-		return -FI_ENOPROTOOPT;
-	}
-	return 0;
+	return ret;
 }
 
 static ssize_t
@@ -665,7 +686,7 @@ usdf_ep_rdm_bind_cq(struct usdf_ep *ep, struct usdf_cq *cq, uint64_t flags)
 			goto fail;
 
 		hcq->cqh_cq = cq;
-		atomic_initialize(&hcq->cqh_refcnt, 0);
+		ofi_atomic_initialize32(&hcq->cqh_refcnt, 0);
 		hcq->cqh_progress = usdf_rdm_hcq_progress;
 		hcq->cqh_post = usdf_cq_post_soft;
 		TAILQ_INSERT_TAIL(&cq->c.soft.cq_list, hcq, cqh_link);
@@ -674,8 +695,8 @@ usdf_ep_rdm_bind_cq(struct usdf_ep *ep, struct usdf_cq *cq, uint64_t flags)
 		TAILQ_INSERT_TAIL(&ep->ep_domain->dom_hcq_list,
 				hcq, cqh_dom_link);
 	}
-	atomic_inc(&hcq->cqh_refcnt);
-	atomic_inc(&cq->cq_refcnt);
+	ofi_atomic_inc32(&hcq->cqh_refcnt);
+	ofi_atomic_inc32(&cq->cq_refcnt);
 	*hcqp = hcq;
 	return 0;
 
@@ -689,11 +710,18 @@ fail:
 static int
 usdf_ep_rdm_bind(struct fid *fid, struct fid *bfid, uint64_t flags)
 {
+	int ret;
 	struct usdf_ep *ep;
 	struct usdf_cq *cq;
+	struct usdf_av *av;
 
 	USDF_TRACE_SYS(EP_CTRL, "\n");
 
+	/* Check if the binding flags are valid. */
+	ret = ofi_ep_bind_valid(&usdf_ops, bfid, flags);
+	if (ret)
+		return ret;
+
 	ep = ep_fidtou(fid);
 
 	switch (bfid->fclass) {
@@ -702,9 +730,11 @@ usdf_ep_rdm_bind(struct fid *fid, struct fid *bfid, uint64_t flags)
 		if (ep->e.rdm.ep_av != NULL) {
 			return -FI_EINVAL;
 		}
-		ep->e.rdm.ep_av = av_fidtou(bfid);
-		break;
 
+		av = av_fidtou(bfid);
+		ep->e.rdm.ep_av = av;
+		ofi_atomic_inc32(&av->av_refcnt);
+		break;
 
 	case FI_CLASS_CQ:
 		if (flags & FI_SEND) {
@@ -731,7 +761,7 @@ usdf_ep_rdm_bind(struct fid *fid, struct fid *bfid, uint64_t flags)
 			return -FI_EINVAL;
 		}
 		ep->ep_eq = eq_fidtou(bfid);
-		atomic_inc(&ep->ep_eq->eq_refcnt);
+		ofi_atomic_inc32(&ep->ep_eq->eq_refcnt);
 		break;
 	default:
 		return -FI_EINVAL;
@@ -753,14 +783,14 @@ usdf_rdm_rx_ctx_close(fid_t fid)
 
 	rx = rx_fidtou(fid);
 
-	if (atomic_get(&rx->rx_refcnt) > 0) {
+	if (ofi_atomic_get32(&rx->rx_refcnt) > 0) {
 		return -FI_EBUSY;
 	}
 
 	hcq = rx->r.rdm.rx_hcq;
 	if (hcq != NULL) {
-		atomic_dec(&hcq->cqh_refcnt);
-		atomic_dec(&hcq->cqh_cq->cq_refcnt);
+		ofi_atomic_dec32(&hcq->cqh_refcnt);
+		ofi_atomic_dec32(&hcq->cqh_cq->cq_refcnt);
 	}
 	if (rx->r.rdm.rx_sock != -1) {
 		close(rx->r.rdm.rx_sock);
@@ -771,7 +801,7 @@ usdf_rdm_rx_ctx_close(fid_t fid)
 		free(rx->r.rdm.rx_rqe_buf);
 		usd_destroy_qp(rx->rx_qp);
 	}
-	atomic_dec(&rx->rx_domain->dom_refcnt);
+	ofi_atomic_dec32(&rx->rx_domain->dom_refcnt);
 
 	free(rx);
 
@@ -791,14 +821,14 @@ usdf_rdm_tx_ctx_close(fid_t fid)
 
 	tx = tx_fidtou(fid);
 
-	if (atomic_get(&tx->tx_refcnt) > 0) {
+	if (ofi_atomic_get32(&tx->tx_refcnt) > 0) {
 		return -FI_EBUSY;
 	}
 
 	hcq = tx->t.rdm.tx_hcq;
 	if (hcq != NULL) {
-		atomic_dec(&hcq->cqh_refcnt);
-		atomic_dec(&hcq->cqh_cq->cq_refcnt);
+		ofi_atomic_dec32(&hcq->cqh_refcnt);
+		ofi_atomic_dec32(&hcq->cqh_cq->cq_refcnt);
 	}
 
 	if (tx->tx_qp != NULL) {
@@ -806,7 +836,7 @@ usdf_rdm_tx_ctx_close(fid_t fid)
 		free(tx->t.rdm.tx_wqe_buf);
 		usd_destroy_qp(tx->tx_qp);
 	}
-	atomic_dec(&tx->tx_domain->dom_refcnt);
+	ofi_atomic_dec32(&tx->tx_domain->dom_refcnt);
 
 	free(tx);
 
@@ -822,11 +852,15 @@ usdf_rx_rdm_port_bind(struct usdf_rx *rx, struct fi_info *info)
 	int ret;
 
 	if (info->src_addr != NULL) {
-		if (info->addr_format != FI_SOCKADDR &&
-		    info->addr_format != FI_SOCKADDR_IN) {
+		switch (info->addr_format) {
+		case FI_SOCKADDR:
+		case FI_SOCKADDR_IN:
+		case FI_ADDR_STR:
+			sin = usdf_format_to_sin(info, info->src_addr);
+			break;
+		default:
 			return -FI_EINVAL;
 		}
-		sin = (struct sockaddr_in *)info->src_addr;
 	} else {
 		memset(&src, 0, sizeof(src));
 		sin = &src;
@@ -834,7 +868,7 @@ usdf_rx_rdm_port_bind(struct usdf_rx *rx, struct fi_info *info)
 		sin->sin_addr.s_addr =
 			rx->rx_domain->dom_fabric->fab_dev_attrs->uda_ipaddr_be;
 	}
-		
+
 	rx->r.rdm.rx_sock = socket(AF_INET, SOCK_DGRAM, 0);
 	if (rx->r.rdm.rx_sock == -1) {
 		return -errno;
@@ -850,6 +884,14 @@ usdf_rx_rdm_port_bind(struct usdf_rx *rx, struct fi_info *info)
 		 return -errno;
 	}
 
+	/* This has to be here because usdf_sin_to_format will allocate
+	 * new piece of memory if the string conversion happens.
+	 */
+	if (info->addr_format == FI_ADDR_STR)
+		free(info->src_addr);
+
+	info->src_addr = usdf_sin_to_format(info, sin, &info->src_addrlen);
+
 	return 0;
 }
 
@@ -862,29 +904,32 @@ usdf_ep_rdm_close(fid_t fid)
 
 	ep = ep_fidtou(fid);
 
-	if (atomic_get(&ep->ep_refcnt) > 0) {
+	if (ofi_atomic_get32(&ep->ep_refcnt) > 0) {
 		return -FI_EBUSY;
 	}
 
 	if (ep->ep_rx != NULL) {
-		atomic_dec(&ep->ep_rx->rx_refcnt);
+		ofi_atomic_dec32(&ep->ep_rx->rx_refcnt);
 		if (rx_utofid(ep->ep_rx)->fclass  == FI_CLASS_RX_CTX) {
 			(void) usdf_rdm_rx_ctx_close(rx_utofid(ep->ep_rx));
 		}
 	}
 
 	if (ep->ep_tx != NULL) {
-		atomic_dec(&ep->ep_tx->tx_refcnt);
+		ofi_atomic_dec32(&ep->ep_tx->tx_refcnt);
 		if (tx_utofid(ep->ep_tx)->fclass  == FI_CLASS_TX_CTX) {
 			(void) usdf_rdm_tx_ctx_close(tx_utofid(ep->ep_tx));
 		}
 	}
 
-	atomic_dec(&ep->ep_domain->dom_refcnt);
+	ofi_atomic_dec32(&ep->ep_domain->dom_refcnt);
 	if (ep->ep_eq != NULL) {
-		atomic_dec(&ep->ep_eq->eq_refcnt);
+		ofi_atomic_dec32(&ep->ep_eq->eq_refcnt);
 	}
-	
+
+	if (ep->e.rdm.ep_av)
+		ofi_atomic_dec32(&ep->e.rdm.ep_av->av_refcnt);
+
 	free(ep);
 	return 0;
 }
@@ -892,8 +937,8 @@ usdf_ep_rdm_close(fid_t fid)
 static struct fi_ops_ep usdf_base_rdm_ops = {
 	.size = sizeof(struct fi_ops_ep),
 	.cancel = usdf_ep_rdm_cancel,
-	.getopt = usdf_ep_rdm_getopt,
-	.setopt = usdf_ep_rdm_setopt,
+	.getopt = usdf_ep_getopt_unconnected,
+	.setopt = usdf_ep_setopt,
 	.tx_ctx = fi_no_tx_ctx,
 	.rx_ctx = fi_no_rx_ctx,
 	.rx_size_left = usdf_rdm_rx_size_left,
@@ -910,6 +955,7 @@ static struct fi_ops_cm usdf_cm_rdm_ops = {
 	.accept = fi_no_accept,
 	.reject = fi_no_reject,
 	.shutdown = fi_no_shutdown,
+	.join = fi_no_join,
 };
 
 static struct fi_ops_msg usdf_rdm_ops = {
@@ -964,6 +1010,7 @@ usdf_ep_rdm_open(struct fid_domain *domain, struct fi_info *info,
 	struct usdf_rx *rx;
 	struct usdf_ep *ep;
 	int ret;
+	uint32_t api_version;
 
 	USDF_TRACE_SYS(EP_CTRL, "\n");
 
@@ -975,6 +1022,7 @@ usdf_ep_rdm_open(struct fid_domain *domain, struct fi_info *info,
 	}
 
 	udp = dom_ftou(domain);
+	api_version = udp->dom_fabric->fab_attr.fabric->api_version;
 
 	/* allocate peer table if not done */
 	if (udp->dom_peer_tab == NULL) {
@@ -997,6 +1045,7 @@ usdf_ep_rdm_open(struct fid_domain *domain, struct fi_info *info,
 	ep->ep_fid.ops = &usdf_base_rdm_ops;
 	ep->ep_fid.cm = &usdf_cm_rdm_ops;
 	ep->ep_fid.msg = &usdf_rdm_ops;
+	ep->ep_fid.atomic = &usdf_rdm_atomic_ops;
 	ep->ep_domain = udp;
 	ep->ep_caps = info->caps;
 	ep->ep_mode = info->mode;
@@ -1012,14 +1061,14 @@ usdf_ep_rdm_open(struct fid_domain *domain, struct fi_info *info,
 			goto fail;
 		}
 		tx->tx_fid.fid.fclass = FI_CLASS_TX_CTX;
-		atomic_initialize(&tx->tx_refcnt, 0);
+		ofi_atomic_initialize32(&tx->tx_refcnt, 0);
 		tx->tx_domain = udp;
 		tx->tx_progress = usdf_rdm_tx_progress;
-		atomic_initialize(&tx->t.rdm.tx_next_msg_id, 1);
-		atomic_inc(&udp->dom_refcnt);
+		ofi_atomic_initialize32(&tx->t.rdm.tx_next_msg_id, 1);
+		ofi_atomic_inc32(&udp->dom_refcnt);
 
 		/* info is both hints and output */
-		ret = usdf_rdm_fill_tx_attr(info, info);
+		ret = usdf_rdm_fill_tx_attr(api_version, info, info);
 		if (ret)
 			goto fail;
 		tx->tx_attr = *info->tx_attr;
@@ -1029,7 +1078,7 @@ usdf_ep_rdm_open(struct fid_domain *domain, struct fi_info *info,
 		TAILQ_INIT(&tx->t.rdm.tx_rdc_have_acks);
 
 		ep->ep_tx = tx;
-		atomic_inc(&tx->tx_refcnt);
+		ofi_atomic_inc32(&tx->tx_refcnt);
 	}
 
 	/* implicitly create RX context if not to be shared */
@@ -1042,11 +1091,11 @@ usdf_ep_rdm_open(struct fid_domain *domain, struct fi_info *info,
 		}
 
 		rx->rx_fid.fid.fclass = FI_CLASS_RX_CTX;
-		atomic_initialize(&rx->rx_refcnt, 0);
+		ofi_atomic_initialize32(&rx->rx_refcnt, 0);
 		rx->rx_domain = udp;
 		rx->r.rdm.rx_tx = tx;
 		rx->r.rdm.rx_sock = -1;
-		atomic_inc(&udp->dom_refcnt);
+		ofi_atomic_inc32(&udp->dom_refcnt);
 
 		ret = usdf_rx_rdm_port_bind(rx, info);
 		if (ret) {
@@ -1054,7 +1103,7 @@ usdf_ep_rdm_open(struct fid_domain *domain, struct fi_info *info,
 		}
 
 		/* info is both hints and output */
-		ret = usdf_rdm_fill_rx_attr(info, info);
+		ret = usdf_rdm_fill_rx_attr(api_version, info, info);
 		if (ret) {
 			goto fail;
 		}
@@ -1064,11 +1113,11 @@ usdf_ep_rdm_open(struct fid_domain *domain, struct fi_info *info,
 		TAILQ_INIT(&rx->r.rdm.rx_posted_rqe);
 
 		ep->ep_rx = rx;
-		atomic_inc(&rx->rx_refcnt);
+		ofi_atomic_inc32(&rx->rx_refcnt);
 	}
 
-	atomic_initialize(&ep->ep_refcnt, 0);
-	atomic_inc(&udp->dom_refcnt);
+	ofi_atomic_initialize32(&ep->ep_refcnt, 0);
+	ofi_atomic_inc32(&udp->dom_refcnt);
 
 	*ep_o = ep_utof(ep);
 	return 0;
@@ -1078,11 +1127,11 @@ fail:
 			close(rx->r.rdm.rx_sock);
 		}
 		free(rx);
-		atomic_dec(&udp->dom_refcnt);
+		ofi_atomic_dec32(&udp->dom_refcnt);
 	}
 	if (tx != NULL) {
 		free(tx);
-		atomic_dec(&udp->dom_refcnt);
+		ofi_atomic_dec32(&udp->dom_refcnt);
 	}
 	if (ep != NULL) {
 		free(ep);
