@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2016, Cisco Systems, Inc. All rights reserved.
+ * Copyright (c) 2014-2017, Cisco Systems, Inc. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -52,17 +52,18 @@
 #include <rdma/fabric.h>
 #include <rdma/fi_cm.h>
 #include <rdma/fi_domain.h>
-#include <rdma/fi_prov.h>
 #include <rdma/fi_endpoint.h>
 #include <rdma/fi_rma.h>
 #include <rdma/fi_errno.h>
 #include "fi.h"
 #include "fi_enosys.h"
+#include "fi_file.h"
 
 #include "fi_ext_usnic.h"
 #include "usnic_direct.h"
 #include "usd.h"
 #include "usdf.h"
+#include "usdf_endpoint.h"
 #include "usdf_cm.h"
 #include "usdf_msg.h"
 
@@ -82,7 +83,7 @@ usdf_pep_bind(fid_t fid, fid_t bfid, uint64_t flags)
 			return -FI_EINVAL;
 		}
 		pep->pep_eq = eq_fidtou(bfid);
-		atomic_inc(&pep->pep_eq->eq_refcnt);
+		ofi_atomic_inc32(&pep->pep_eq->eq_refcnt);
 		break;
 
 	default:
@@ -119,7 +120,9 @@ usdf_pep_conn_info(struct usdf_connreq *crp)
 	sin->sin_addr.s_addr = reqp->creq_ipaddr;
 	sin->sin_port = reqp->creq_port;
 
-	ip->dest_addr = sin;
+	ip->dest_addr = usdf_sin_to_format(pep->pep_info, sin,
+					   &ip->dest_addrlen);
+
 	ip->handle = (fid_t) crp;
 	return ip;
 fail:
@@ -168,8 +171,8 @@ usdf_pep_read_connreq(void *v)
 
 	n = read(crp->cr_sockfd, crp->cr_ptr, crp->cr_resid);
 	if (n == -1) {
-		usdf_cm_msg_connreq_failed(crp, -errno);
-		return 0;
+		ret = -errno;
+		goto report_failure_skip_data;
 	}
 
 	crp->cr_ptr += n;
@@ -185,37 +188,41 @@ usdf_pep_read_connreq(void *v)
 	/* if resid is 0 now, completely done */
 	if (crp->cr_resid == 0) {
 		ret = usdf_pep_creq_epoll_del(crp);
-		if (ret != 0) {
-			usdf_cm_msg_connreq_failed(crp, ret);
-			return 0;
-		}
+		if (ret != 0)
+			goto report_failure_skip_data;
 
 		/* create CONNREQ EQ entry */
 		entry_len = sizeof(*entry) + reqp->creq_datalen;
 		entry = malloc(entry_len);
 		if (entry == NULL) {
-			usdf_cm_msg_connreq_failed(crp, -errno);
-			return 0;
+			ret = -errno;
+			goto report_failure_skip_data;
 		}
 
 		entry->fid = &pep->pep_fid.fid;
 		entry->info = usdf_pep_conn_info(crp);
 		if (entry->info == NULL) {
-			free(entry);
-			usdf_cm_msg_connreq_failed(crp, -FI_ENOMEM);
-			return 0;
+			ret = -FI_ENOMEM;
+			goto free_entry_and_report_failure;
 		}
+
 		memcpy(entry->data, reqp->creq_data, reqp->creq_datalen);
 		ret = usdf_eq_write_internal(pep->pep_eq, FI_CONNREQ, entry,
 				entry_len, 0);
+
+		if (ret != (int)entry_len)
+			goto free_entry_and_report_failure;
+
 		free(entry);
-		if (ret != (int)entry_len) {
-			usdf_cm_msg_connreq_failed(crp, ret);
-			return 0;
-		}
 	}
 
 	return 0;
+
+free_entry_and_report_failure:
+	free(entry);
+report_failure_skip_data:
+	usdf_cm_report_failure(crp, ret, false);
+	return 0;
 }
 
 static int
@@ -266,8 +273,7 @@ usdf_pep_listen_cb(void *v)
 	ret = epoll_ctl(pep->pep_fabric->fab_epollfd, EPOLL_CTL_ADD,
 			crp->cr_sockfd, &ev);
 	if (ret == -1) {
-		crp->cr_pollitem.pi_rtn = NULL;
-		usdf_cm_msg_connreq_failed(crp, -errno);
+		usdf_cm_report_failure(crp, -errno, false);
 		return 0;
 	}
 
@@ -282,12 +288,17 @@ usdf_pep_listen(struct fid_pep *fpep)
 	struct usdf_pep *pep;
 	struct epoll_event ev;
 	struct usdf_fabric *fp;
+	struct sockaddr_in *sin;
+	socklen_t socklen;
 	int ret;
+	bool addr_format_str;
 
 	USDF_TRACE_SYS(EP_CTRL, "\n");
 
 	pep = pep_ftou(fpep);
 	fp = pep->pep_fabric;
+	addr_format_str = (pep->pep_info->addr_format == FI_ADDR_STR);
+	sin = NULL;
 
 	switch (pep->pep_state) {
 	case USDF_PEP_UNBOUND:
@@ -301,7 +312,8 @@ usdf_pep_listen(struct fid_pep *fpep)
 			"PEP already consumed, you may only fi_close() now\n");
 		return -FI_EOPBADSTATE;
 	default:
-		USDF_WARN_SYS(EP_CTRL, "unhandled case!\n");
+		USDF_WARN_SYS(EP_CTRL, "unhandled case! (%d)\n",
+			      pep->pep_state);
 		abort();
 	}
 
@@ -309,17 +321,40 @@ usdf_pep_listen(struct fid_pep *fpep)
 	 * already did the bind in a previous call to usdf_pep_listen() and the
 	 * listen(2) call failed */
 	if (pep->pep_state == USDF_PEP_UNBOUND) {
-		ret = bind(pep->pep_sock, (struct sockaddr *)&pep->pep_src_addr,
-				sizeof(struct sockaddr_in));
+		sin = usdf_format_to_sin(pep->pep_info, &pep->pep_src_addr);
+		if (sin == NULL)
+			goto fail;
+
+		ret = bind(pep->pep_sock, sin, sizeof(struct sockaddr_in));
 		if (ret == -1) {
-			return -errno;
+			goto fail;
 		}
+
+		/* Get the actual port (since we may have requested
+		 * port 0)
+		 */
+		socklen = sizeof(*sin);
+		ret = getsockname(pep->pep_sock, sin,
+				&socklen);
+		if (ret == -1)
+			goto fail;
+
+		/* If it's FI_ADDR_STR, we have to update the string
+		 * with this method. (FI_SOCKADDR_IN got taken care of, above)
+		*/
+		if (addr_format_str) {
+			pep->pep_info->src_addrlen = USDF_ADDR_STR_LEN;
+			usdf_addr_tostr(sin, pep->pep_src_addr.addr_str,
+					&pep->pep_info->src_addrlen);
+		}
+
+		/* Update the state to bound. */
 		pep->pep_state = USDF_PEP_BOUND;
 	}
 
 	ret = listen(pep->pep_sock, pep->pep_backlog);
 	if (ret != 0) {
-		return -errno;
+		goto fail;
 	}
 	pep->pep_state = USDF_PEP_LISTENING;
 
@@ -329,10 +364,108 @@ usdf_pep_listen(struct fid_pep *fpep)
 	ev.data.ptr = &pep->pep_pollitem;
 	ret = epoll_ctl(fp->fab_epollfd, EPOLL_CTL_ADD, pep->pep_sock, &ev);
 	if (ret == -1) {
-		return -errno;
+		goto fail;
 	}
 
 	return 0;
+
+fail:
+	usdf_free_sin_if_needed(pep->pep_info, sin);
+
+	return -errno;
+}
+
+/* Register as a callback triggered by the socket becoming writeable. Write as
+ * much data as can be written in a single write, and keep track of how much
+ * data is left. If the data is not fully written, it will finish getting
+ * written in another iteration of the progression.
+ */
+static int usdf_pep_reject_async(void *vreq)
+{
+	struct usdf_connreq *crp;
+	int ret;
+
+	crp = vreq;
+
+	do {
+		ret = write(crp->cr_sockfd, crp->cr_ptr, crp->cr_resid);
+	} while ((ret < 0) && (errno == EINTR));
+
+	if ((ret <= 0) && (errno != EAGAIN)) {
+		USDF_DBG_SYS(EP_CTRL, "write failed: %s\n",
+				strerror(errno));
+		usdf_cm_report_failure(crp, -errno, false);
+		return -errno;
+	}
+
+	crp->cr_resid -= ret;
+	crp->cr_ptr += ret;
+
+	if (crp->cr_resid == 0)
+		usdf_cm_msg_connreq_cleanup(crp);
+
+	return FI_SUCCESS;
+}
+
+static int usdf_pep_reject(struct fid_pep *fpep, fid_t handle, const void *param,
+		size_t paramlen)
+{
+	struct usdf_pep *pep;
+	struct usdf_connreq *crp;
+	struct epoll_event event;
+	struct usdf_connreq_msg *reqp;
+	int ret;
+
+	if (paramlen > USDF_MAX_CONN_DATA) {
+		USDF_WARN_SYS(EP_CTRL,
+				"reject payload size %zu exceeds max %zu\n",
+				paramlen, USDF_MAX_CONN_DATA);
+		return -FI_EINVAL;
+	}
+
+	if (!fpep || !handle) {
+		USDF_WARN_SYS(EP_CTRL,
+				"handle and passive ep needed for reject\n");
+		return -FI_EINVAL;
+	}
+
+	if (!param && paramlen > 0) {
+		USDF_WARN_SYS(EP_CTRL,
+				"NULL data pointer with non-zero data length\n");
+		return -FI_EINVAL;
+	}
+
+	/* usdf_pep_conn_info stashed the pep pointer into the handle field of
+	 * the info struct previously returned
+	 */
+	crp = (struct usdf_connreq *) handle;
+	pep = pep_ftou(fpep);
+
+	crp->cr_ptr = crp->cr_data;
+	crp->cr_resid = sizeof(*reqp) + paramlen;
+
+	reqp = (struct usdf_connreq_msg *) crp->cr_data;
+
+	/* The result field is used on the remote end to detect whether the
+	 * connection succeeded or failed.
+	 */
+	reqp->creq_result = htonl(-FI_ECONNREFUSED);
+	reqp->creq_datalen = htonl(paramlen);
+	memcpy(reqp->creq_data, param, paramlen);
+
+	crp->cr_pollitem.pi_rtn = usdf_pep_reject_async;
+	crp->cr_pollitem.pi_context = crp;
+
+	event.events = EPOLLOUT;
+	event.data.ptr = &crp->cr_pollitem;
+
+	ret = epoll_ctl(pep->pep_fabric->fab_epollfd, EPOLL_CTL_ADD,
+			crp->cr_sockfd, &event);
+
+	if (ret)
+		return -errno;
+
+	return FI_SUCCESS;
 }
 
 static void
@@ -383,7 +516,7 @@ usdf_pep_close(fid_t fid)
 	USDF_TRACE_SYS(EP_CTRL, "\n");
 
 	pep = pep_fidtou(fid);
-	if (atomic_get(&pep->pep_refcnt) > 0) {
+	if (ofi_atomic_get32(&pep->pep_refcnt) > 0) {
 		return -FI_EBUSY;
 	}
 
@@ -391,9 +524,10 @@ usdf_pep_close(fid_t fid)
 	close(pep->pep_sock);
 	pep->pep_sock = -1;
 	if (pep->pep_eq != NULL) {
-		atomic_dec(&pep->pep_eq->eq_refcnt);
+		ofi_atomic_dec32(&pep->pep_eq->eq_refcnt);
 	}
-	atomic_dec(&pep->pep_fabric->fab_refcnt);
+	ofi_atomic_dec32(&pep->pep_fabric->fab_refcnt);
+	fi_freeinfo(pep->pep_info);
 	free(pep);
 
 	return 0;
@@ -403,14 +537,16 @@ static int usdf_pep_getname(fid_t fid, void *addr, size_t *addrlen)
 {
 	int ret;
 	struct usdf_pep *pep;
+	struct fi_info *info;
 	size_t copylen;
 
 	USDF_TRACE_SYS(EP_CTRL, "\n");
 
 	ret = FI_SUCCESS;
 	pep = pep_fidtou(fid);
+	info = pep->pep_info;
 
-	copylen = sizeof(pep->pep_src_addr);
+	copylen = info->src_addrlen;
 	memcpy(addr, &pep->pep_src_addr, MIN(copylen, *addrlen));
 
 	if (*addrlen < copylen) {
@@ -426,32 +562,50 @@ static int usdf_pep_setname(fid_t fid, void *addr, size_t addrlen)
 {
 	int ret;
 	struct usdf_pep *pep;
+	struct fi_info *info;
+	struct sockaddr_in *sin;
 	uint32_t req_addr_be;
 	socklen_t socklen;
 	char namebuf[INET_ADDRSTRLEN];
 	char servbuf[INET_ADDRSTRLEN];
+	bool addr_format_str;
 
 	USDF_TRACE_SYS(EP_CTRL, "\n");
 
 	pep = pep_fidtou(fid);
+	info = pep->pep_info;
+	addr_format_str = (info->addr_format == FI_ADDR_STR);
+	sin = NULL;
+
 	if (pep->pep_state != USDF_PEP_UNBOUND) {
 		USDF_WARN_SYS(EP_CTRL, "PEP cannot be bound\n");
 		return -FI_EOPBADSTATE;
 	}
-	if (((struct sockaddr *)addr)->sa_family != AF_INET) {
-		USDF_WARN_SYS(EP_CTRL, "non-AF_INET address given\n");
-		return -FI_EINVAL;
-	}
-	if (addrlen != sizeof(struct sockaddr_in)) {
-		USDF_WARN_SYS(EP_CTRL, "unexpected src_addrlen\n");
+
+	switch (info->addr_format) {
+	case FI_SOCKADDR:
+	case FI_SOCKADDR_IN:
+		/* It is possible for passive endpoint to not have src_addr. */
+		if (info->src_addr) {
+			ret = usdf_cm_addr_is_valid_sin(info->src_addr,
+							info->src_addrlen,
+							info->addr_format);
+			if (!ret)
+				return -FI_EINVAL;
+		}
+		break;
+	case FI_ADDR_STR:
+		break;
+	default:
 		return -FI_EINVAL;
 	}
 
-	req_addr_be = ((struct sockaddr_in *)addr)->sin_addr.s_addr;
+	sin = usdf_format_to_sin(info, addr);
+	req_addr_be = sin->sin_addr.s_addr;
 
 	namebuf[0] = '\0';
 	servbuf[0] = '\0';
-	ret = getnameinfo((struct sockaddr*)addr, addrlen,
+	ret = getnameinfo((struct sockaddr *)sin, sizeof(struct sockaddr_in),
 		namebuf, sizeof(namebuf),
 		servbuf, sizeof(servbuf),
 		NI_NUMERICHOST|NI_NUMERICSERV);
@@ -465,16 +619,15 @@ static int usdf_pep_setname(fid_t fid, void *addr, size_t addrlen)
 		return -FI_EADDRNOTAVAIL;
 	}
 
-	ret = bind(pep->pep_sock, (struct sockaddr *)addr,
-		sizeof(struct sockaddr_in));
+	ret = bind(pep->pep_sock, sin, sizeof(*sin));
 	if (ret == -1) {
 		return -errno;
 	}
 	pep->pep_state = USDF_PEP_BOUND;
 
 	/* store the resulting port so that can implement getname() properly */
-	socklen = sizeof(pep->pep_src_addr);
-	ret = getsockname(pep->pep_sock, &pep->pep_src_addr, &socklen);
+	socklen = sizeof(*sin);
+	ret = getsockname(pep->pep_sock, sin, &socklen);
 	if (ret == -1) {
 		ret = -errno;
 		USDF_WARN_SYS(EP_CTRL, "getsockname failed %d (%s), PEP may be in bad state\n",
@@ -482,6 +635,18 @@ static int usdf_pep_setname(fid_t fid, void *addr, size_t addrlen)
 		return ret;
 	}
 
+	if (addr_format_str) {
+		/* We have to reset src_addrlen here and
+		 * the conversion will update it to the correct len.
+		 */
+		info->src_addrlen = USDF_ADDR_STR_LEN;
+		usdf_addr_tostr(sin, pep->pep_src_addr.addr_str,
+				&info->src_addrlen);
+		free(sin);
+	} else {
+		memcpy(&pep->pep_src_addr, sin, sizeof(*sin));
+	}
+
 	return 0;
 }
 
@@ -496,8 +661,8 @@ struct fi_ops usdf_pep_ops = {
 static struct fi_ops_ep usdf_pep_base_ops = {
 	.size = sizeof(struct fi_ops_ep),
 	.cancel = fi_no_cancel,
-	.getopt = fi_no_getopt,
-	.setopt = fi_no_setopt,
+	.getopt = usdf_ep_getopt_connected,
+	.setopt = usdf_ep_setopt,
 	.tx_ctx = fi_no_tx_ctx,
 	.rx_ctx = fi_no_rx_ctx,
 	.rx_size_left = fi_no_rx_size_left,
@@ -512,8 +677,9 @@ static struct fi_ops_cm usdf_pep_cm_ops = {
 	.connect = fi_no_connect,
 	.listen = usdf_pep_listen,
 	.accept = fi_no_accept,
-	.reject = fi_no_reject,
+	.reject = usdf_pep_reject,
 	.shutdown = fi_no_shutdown,
+	.join = fi_no_join,
 };
 
 int
@@ -543,24 +709,23 @@ usdf_pep_open(struct fid_fabric *fabric, struct fi_info *info,
 
 	switch (info->addr_format) {
 	case FI_SOCKADDR:
-		if (((struct sockaddr *)info->src_addr)->sa_family != AF_INET) {
-			USDF_WARN_SYS(EP_CTRL, "non-AF_INET src_addr specified\n");
-			return -FI_EINVAL;
+	case FI_SOCKADDR_IN:
+		/* It is possible for passive endpoint to not have src_addr. */
+		if (info->src_addr) {
+			ret = usdf_cm_addr_is_valid_sin(info->src_addr,
+							info->src_addrlen,
+							info->addr_format);
+			if (!ret)
+				return -FI_EINVAL;
 		}
 		break;
-	case FI_SOCKADDR_IN:
+	case FI_ADDR_STR:
 		break;
 	default:
 		USDF_WARN_SYS(EP_CTRL, "unknown/unsupported addr_format\n");
 		return -FI_EINVAL;
 	}
 
-	if (info->src_addrlen &&
-			info->src_addrlen != sizeof(struct sockaddr_in)) {
-		USDF_WARN_SYS(EP_CTRL, "unexpected src_addrlen\n");
-		return -FI_EINVAL;
-	}
-
 	fp = fab_ftou(fabric);
 
 	pep = calloc(1, sizeof(*pep));
@@ -581,16 +746,11 @@ usdf_pep_open(struct fid_fabric *fabric, struct fi_info *info,
 		ret = -errno;
 		goto fail;
 	}
-        ret = fcntl(pep->pep_sock, F_GETFL, 0);
-        if (ret == -1) {
-                ret = -errno;
-                goto fail;
-        }
-        ret = fcntl(pep->pep_sock, F_SETFL, ret | O_NONBLOCK);
-        if (ret == -1) {
-                ret = -errno;
-                goto fail;
-        }
+	ret = fi_fd_nonblock(pep->pep_sock);
+	if (ret) {
+		ret = -errno;
+		goto fail;
+	}
 
 	/* set SO_REUSEADDR to prevent annoying "Address already in use" errors
 	 * on successive runs of programs listening on a well known port */
@@ -622,11 +782,14 @@ usdf_pep_open(struct fid_fabric *fabric, struct fi_info *info,
 
 		sin->sin_family = AF_INET;
 		sin->sin_addr.s_addr = fp->fab_dev_attrs->uda_ipaddr_be;
-		pep->pep_info->src_addr = sin;
+
+		pep->pep_info->src_addr =
+			usdf_sin_to_format(pep->pep_info,
+					   sin, &pep->pep_info->src_addrlen);
 	}
 
 	memcpy(&pep->pep_src_addr, pep->pep_info->src_addr,
-			pep->pep_info->src_addrlen);
+	       pep->pep_info->src_addrlen);
 
 	/* initialize connreq freelist */
 	ret = pthread_spin_init(&pep->pep_cr_lock, PTHREAD_PROCESS_PRIVATE);
@@ -637,13 +800,15 @@ usdf_pep_open(struct fid_fabric *fabric, struct fi_info *info,
 	TAILQ_INIT(&pep->pep_cr_free);
 	TAILQ_INIT(&pep->pep_cr_pending);
 	pep->pep_backlog = 10;
+	pep->pep_cr_max_data = USDF_MAX_CONN_DATA;
+
 	ret = usdf_pep_grow_backlog(pep);
 	if (ret != 0) {
 		goto fail;
 	}
 
-	atomic_initialize(&pep->pep_refcnt, 0);
-	atomic_inc(&fp->fab_refcnt);
+	ofi_atomic_initialize32(&pep->pep_refcnt, 0);
+	ofi_atomic_inc32(&fp->fab_refcnt);
 
 	*pep_o = pep_utof(pep);
 	return 0;
